//===- llvm/ADT/UnReservedVector.h - vectors with no space to grow --*- C++ -*-===//
// These vectors take up 4 bytes less space that a regular vector but
// are much less efficient to construct


#ifndef LLVM_ADT_UNRESERVEDVECTOR_H
#define LLVM_ADT_UNRESERVEDVECTOR_H

#include "llvm/ADT/iterator"
#include <algorithm>
#include <memory>
#include <assert.h>

namespace llvm
{

/// UnReservedVectorImpl - This class consists of code factored out of
/// The UnReservedVector class because it was factored out of the SmallVector
/// class for a good reason.  Not so much here.
template <typename T> class UnReservedVectorImpl
{
protected:
	T *Begin;
	unsigned Size;

public:
	// Default ctor - Initialize to empty.
	UnReservedVectorImpl() :
		Begin(NULL), Size(0)
	{
	}

	~UnReservedVectorImpl()
	{
		// Destroy the constructed elements in the vector.
		delete[] Begin;
	}

	typedef size_t size_type;
	typedef T* iterator;
	typedef const T* const_iterator;

	typedef std::reverse_iterator<const_iterator> const_reverse_iterator;
	typedef std::reverse_iterator<iterator> reverse_iterator;

	typedef T& reference;
	typedef const T& const_reference;

	bool empty() const
	{
		return Size == 0;
	}
	size_type size() const
	{
		return Size;
	}

	// forward iterator creation methods.
	iterator begin()
	{
		return Begin;
	}
	const_iterator begin() const
	{
		return Begin;
	}
	iterator end()
	{
		return Begin+Size;
	}
	const_iterator end() const
	{
		return Begin+Size;
	}

	// reverse iterator creation methods.
	reverse_iterator rbegin()
	{
		return reverse_iterator(end());
	}
	const_reverse_iterator rbegin() const
	{
		return const_reverse_iterator(end());
	}
	reverse_iterator rend()
	{
		return reverse_iterator(begin());
	}
	const_reverse_iterator rend() const
	{
		return const_reverse_iterator(begin());
	}

	reference operator[](unsigned idx)
	{
		return Begin[idx];
	}
	const_reference operator[](unsigned idx) const
	{
		return Begin[idx];
	}

	reference front()
	{
		return begin()[0];
	}
	const_reference front() const
	{
		return begin()[0];
	}

	reference back()
	{
		return end()[-1];
	}
	const_reference back() const
	{
		return end()[-1];
	}

	void push_back(const_reference Elt)
	{
		unsigned oldSize = Size;
		resize(Size+1);
		Begin[oldSize] = Elt;
	} 

	void clear()
	{
		if (Begin != NULL)
			delete [] Begin;
		Begin = NULL;
		Size = 0;
	}

	void resize(unsigned N)
	{
		if(Size == N)
			;
		else if (N == 0)
		{
			clear();
		}
		else if (Size == 0)
		{
			Begin = new T[N];
			Size = N;
		}
		else
		{
			//allocate mem without calling constructors
			T *NewElts = reinterpret_cast<T*>(new char[N*sizeof(T)]);
			// Copy the elements over.
			if (N < Size)
			{
				T *End = Begin+N;
				std::uninitialized_copy(Begin, End, NewElts);
			}
			else
			{
				T *End = Begin+Size;
				std::uninitialized_copy(Begin, End, NewElts);
				construct_range(NewElts+Size, NewElts+N);
			}
			delete [] Begin;
			Begin = NewElts;
			assert(Begin != NULL);
			Size = N;
		}
	}

	void resize(unsigned N, const T &NV)
	{
		if (N < size())
		{
			resize(N);
		}
		else if (N > size())
		{
			//allocate mem without calling constructors
			T *NewElts = reinterpret_cast<T*>(new char[N*sizeof(T)]);
			// Copy the elements over.
			T *End = Begin+Size;
			std::uninitialized_copy(Begin, End, NewElts);
			construct_range(NewElts+Size, NewElts+N, NV);
			
			delete [] Begin;
			Begin = NewElts;
			Size = N;
			assert(Begin != NULL);
		}
	}



	void swap(UnReservedVectorImpl &RHS);

	/// append - Add the specified range to the end of the UnReservedVector.
	///
	template<typename in_iter> void append(in_iter in_start, in_iter in_end)
	{
		unsigned NumInputs = std::distance(in_start, in_end);
		unsigned oldSize = Size;
		// Grow allocated space 
		resize(Size+NumInputs);		

		// Copy the new elements over.
		std::uninitialized_copy(in_start, in_end, Begin+oldSize);
	}

	void assign(unsigned NumElts, const T &Elt)
	{
		resize(NumElts);
		destroy_range(Begin, Begin+Size);
		construct_range(Begin, Begin+Size, Elt);
	}


	const UnReservedVectorImpl &operator=(const UnReservedVectorImpl &RHS);

	bool operator==(const UnReservedVectorImpl &RHS) const
	{
		if(RHS.Size != Size)
			return false;
		
		for(unsigned i = 0, n = Size; i < n; i++)
		{
			if(Begin[i] != RHS.Begin[i])
				return false;
		}
		return true;
	}
	
private:


	void construct_range(T *S, T *E)
	{
		for (; S != E; ++S)
			new (S) T();
	}

	void construct_range(T *S, T *E, const T &Elt)
	{
		for (; S != E; ++S)
			new (S) T(Elt);
	}

	void destroy_range(T *S, T *E)
	{
		while (S != E)
		{
			--E;
			E->~T();
		}
	}
};


template <typename T> void UnReservedVectorImpl<T>::swap(
		UnReservedVectorImpl<T> &RHS)
{
	if (this == &RHS)
		return;

	// We can only avoid copying elements if neither vector is small.

	std::swap(Begin, RHS.Begin);
	std::swap(Size, RHS.Size);
	return;
}

template <typename T> const UnReservedVectorImpl<T> & UnReservedVectorImpl<T>::operator=(const UnReservedVectorImpl<T> &RHS)
{
	// Avoid self-assignment.
	if (this == &RHS) return *this;

	clear();
	
	//allocate mem without calling constructors
	Begin = reinterpret_cast<T*>(new char[RHS.Size*sizeof(T)]);
	Size = RHS.Size;
	// Copy the elements over.
	std::copy(RHS.Begin, RHS.Begin+RHS.Size, Begin);
	
	return *this;
}

/// UnReservedVector - This is a 'vector' (really, a variable-sized array), 
/// optimized to be space efficient by sacraficing construction efficiency.
///
template <typename T>
class UnReservedVector : public UnReservedVectorImpl<T>
{
	
public:
	UnReservedVector() : UnReservedVectorImpl<T>()
	{
	}

	explicit UnReservedVector(unsigned Size, const T &Value = T())
	: UnReservedVectorImpl<T>()
	{
		this->assign(Size, Value);		
	}

	template<typename ItTy>
	UnReservedVector(ItTy S, ItTy E) : UnReservedVectorImpl<T>()
	{
		append(S, E);
	}

	UnReservedVector(const UnReservedVector &RHS) : UnReservedVectorImpl<T>()
	{
		if (!RHS.empty())
		operator=(RHS);
	}

	bool operator==(const UnReservedVector &RHS) const
	{
		return UnReservedVectorImpl<T>::operator==(RHS);
	}
	
	const UnReservedVector &operator=(const UnReservedVector &RHS)
	{
		UnReservedVectorImpl<T>::operator=(RHS);
		return *this;
	}
};

} // End llvm namespace

namespace std
{
/// Implement std::swap in terms of UnReservedVector swap.
template<typename T> inline void swap(llvm::UnReservedVectorImpl<T> &LHS,
		llvm::UnReservedVectorImpl<T> &RHS)
{
	LHS.swap(RHS);
}

/// Implement std::swap in terms of UnReservedVector swap.
template<typename T> inline void swap(
		llvm::UnReservedVector<T> &LHS, llvm::UnReservedVector<T> &RHS)
{
	LHS.swap(RHS);
}
}

#endif
