//===- llvm/Analysis/ProfileInfo.h - Profile Info Interface -----*- C++ -*-===//
//
//                     The LLVM Compiler Infrastructure
//
// This file was developed by the LLVM research group and is distributed under
// the University of Illinois Open Source License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
//
// This file defines the generic ProfileInfo interface, which is used as the
// common interface used by all clients of profiling information, and
// implemented either by making static guestimations, or by actually reading in
// profiling information gathered by running the program.
//
// Note that to be useful, all profile-based optimizations should preserve
// ProfileInfo, which requires that they notify it when changes to the CFG are
// made.
//
//===----------------------------------------------------------------------===//

#ifndef LLVM_ANALYSIS_PROFILEINFO_H
#define LLVM_ANALYSIS_PROFILEINFO_H

#include <string>
#include <map>
#include "llvm/ADT/hash_map"
#include "llvm/Pass.h"
#include "llvm/BasicBlock.h"
#include "llvm/Function.h"

namespace llvm {
  class BasicBlock;
  class Pass;

  /// ProfileInfo Class - This class holds and maintains edge profiling
  /// information for some unit of code.
  class ProfileInfo {
  protected:
    // EdgeCounts - Count the number of times a transition between two blocks is
    // executed.  As a special case, we also hold an edge from the null
    // BasicBlock to the entry block to indicate how many times the function was
    // entered.
	  //dkoes - make this textual based instead of pointer value based
	  typedef std::pair<std::string, std::string> edgeType;
	  typedef std::map< edgeType, unsigned> edgeCountType;
	  std::map<std::string, edgeCountType > EdgeCounts;
  public:
    static char ID; // Class identification, replacement for typeinfo
    virtual ~ProfileInfo();  // We want to be subclassed

    //===------------------------------------------------------------------===//
    /// Profile Information Queries
    ///
    virtual unsigned getExecutionCount(const BasicBlock *BB);

    unsigned getEdgeWeight(const BasicBlock *Src, const BasicBlock *Dest) {
    	std::string fname = Src->getParent()->getName();
    	if(EdgeCounts.count(fname))
    	{
    		std::string sname = Src->getName();
    		std::string dname = Dest->getName();
    		edgeType edge(sname, dname);
    		edgeCountType& counts = EdgeCounts[fname];
    		if(counts.count(edge))
    		{
    			return counts[edge];
    		}
    	}
    	return 0;
    }

    /// Return true if have meaningful profile data
    virtual bool isValid() { return EdgeCounts.size() != 0; }

    //===------------------------------------------------------------------===//
    /// Analysis Update Methods
    ///

  };

  /// createProfileLoaderPass - This function returns a Pass that loads the
  /// profiling information for the module from the specified filename, making
  /// it available to the optimizers.
  Pass *createProfileLoaderPass();
  ImmutablePass *createNoProfileInfoPass();

} // End llvm namespace

#endif
