#ifndef DATAFLOWLIVEVARIABLES_H_
#define DATAFLOWLIVEVARIABLES_H_


#include "llvm/CodeGen/LiveVariables.h"
#include "llvm/Function.h"
#include "llvm/CodeGen/MachineFunction.h"
#include "llvm/CodeGen/MachineFunctionPass.h"
#include "llvm/CodeGen/MachineInstr.h"
#include "llvm/CodeGen/Passes.h"
#include "llvm/Target/TargetInstrInfo.h"
#include "llvm/Target/TargetMachine.h"
#include "llvm/ADT/Statistic.h"
#include "llvm/ADT/STLExtras.h"
#include "llvm/ADT/hash_set"
#include "llvm/ADT/hash_map"
#include "llvm/Support/Debug.h"
#include "llvm/Support/Compiler.h"
#include <deque>
#include "llvm/Support/CFG.h"
#include "llvm/CodeGen/SSARegMap.h"

/*  Compute the live-in and live-out sets of each block as well as marking
 * the liveness properties of every machine operand.  Uses iterative dataflow
 * so does not require SSA form.
 */

using namespace llvm;

namespace llvm
{
class DataFlowLiveVariables : public MachineFunctionPass
{
	MachineFunction *MF;

	hash_map<const MachineBasicBlock *, BitVector> liveins;
	hash_map<const MachineBasicBlock *, hash_set<unsigned> > temp_liveins; //for exception registers
	hash_map<const MachineBasicBlock*, BitVector> liveouts;

	hash_map<const MachineBasicBlock*, BitVector> block_gen;
	hash_map<const MachineBasicBlock*, BitVector> block_killc; //complement of kill

	void compute_block_gen_kill(const MachineBasicBlock*, BitVector& gen, BitVector& kill);
	void compute_dataflow(std::deque<const MachineBasicBlock*>& );
	void update_liveness_info();
	unsigned nVars;

	//recomputes gen/kill of block and updates liveness information
	void update(MachineBasicBlock *block);
	void update(const std::vector<MachineBasicBlock*>& block);

	//set reg in bitmap, properly handles subregs of hard registers
	void setReg(BitVector& bits, unsigned r);
	//unset reg in bitmap
	void unsetReg(BitVector& bits, unsigned r);
public:
	static char ID;
	DataFlowLiveVariables() :
		MachineFunctionPass((intptr_t)&ID)
	{
	}

	virtual ~DataFlowLiveVariables() {}

	virtual const char* getPassName() const
	{
		return "DataFlow Live Variable Analysis";
	}

	virtual void getAnalysisUsage(AnalysisUsage &AU) const
	{
		AU.setPreservesAll();
		MachineFunctionPass::getAnalysisUsage(AU);
	}

	/// runOnMachineFunction - compute liveness
	bool runOnMachineFunction(MachineFunction&);

	//heavyweight update of all blocks
	void update_full();

	const BitVector& getLiveOut(MachineBasicBlock *MBB)
	{
		assert(liveouts.count(MBB) > 0);
		return liveouts[MBB];
	}

	const BitVector& getLiveIn(MachineBasicBlock *MBB)
	{
		assert(liveins.count(MBB) > 0);
		return liveins[MBB];
	}

	//return true if r (a physical register in MBB's "livein" list
	//is truly live in and used by the block
	const bool isTrulyLiveIn(MachineBasicBlock *MBB, unsigned r)
	{
		if(temp_liveins[MBB].count(r) > 0 || liveins[MBB].test(r))
			return true;
		return false;
	}

	const void setTrulyLiveIn(MachineBasicBlock *MBB, std::vector<unsigned>& ret)
	{
		ret.clear();
		BitVector& live = liveins[MBB];
		for(int r = live.find_first(); r >= 0; r = live.find_next(r))
		{
			ret.push_back(r);
		}
		hash_set<unsigned>& tlive = temp_liveins[MBB];
		for(hash_set<unsigned>::iterator itr = tlive.begin(), e = tlive.end(); itr != e; itr++)
		{
			ret.push_back(*itr);
		}
	}
};

}

#endif /*DATAFLOWLIVEVARIABLES_H_*/
