//===-- llvm/CodeGen/SelectionDAGISel.h - Common Base Class------*- C++ -*-===//
//
//                     The LLVM Compiler Infrastructure
//
// This file was developed by the LLVM research group and is distributed under
// the University of Illinois Open Source License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
//
// This file implements the SelectionDAGISel class, which is used as the common
// base class for SelectionDAG-based instruction selectors.
//
//===----------------------------------------------------------------------===//

#ifndef LLVM_CODEGEN_SELECTIONDAG_ISEL_H
#define LLVM_CODEGEN_SELECTIONDAG_ISEL_H

#include "llvm/Pass.h"
#include "llvm/Constant.h"
#include "llvm/CodeGen/SelectionDAG.h"
#include "llvm/CodeGen/SelectionDAGNodes.h"
#include "llvm/ADT/hash_set"
#include "llvm/ADT/hash_map"
#include "llvm/ADT/STLExtras.h"
#include <list>
#include <algorithm>
#include <queue>
#include <iostream> 


namespace llvm
{
class SelectionDAGLowering;
class SDOperand;
class SSARegMap;
class MachineBasicBlock;
class MachineFunction;
class MachineInstr;
class TargetLowering;
class FunctionLoweringInfo;
class HazardRecognizer;
class SelectionDAGISel;
class Tile;

/// Sorting functions for the selection queue.
struct isel_sort : public std::binary_function<SDNode*, SDNode*, bool> {
  bool operator()(const SDNode* left, const SDNode* right) const {
    return (left->getNodeId() > right->getNodeId());
  }
};
 
struct TileCost
{
	Tile *tile;
	int cost;
	
	TileCost(Tile *t, int c) :tile(t), cost(c) {}
	TileCost(): tile(NULL), cost(0) {}
};

}
namespace llvm {

 
/// SelectionDAGISel - This is the common base class used for SelectionDAG-based
/// pattern-matching instruction selectors.
class SelectionDAGISel : public FunctionPass
{
public:

  TargetLowering &TLI;
  SSARegMap *RegMap;
  SelectionDAG *CurDAG;
  MachineBasicBlock *BB;
  AliasAnalysis *AA;
  std::vector<SDNode*> TopOrder;
  unsigned DAGSize;
  static char ID;

  explicit SelectionDAGISel(TargetLowering &tli) : 
      FunctionPass((intptr_t)&ID), TLI(tli), DAGSize(0),
      deletedNodes(NULL), ISelQueued(NULL), ISelSelected(NULL) {}  

	TargetLowering &getTargetLowering()
	{
		return TLI;
	}

	virtual void getAnalysisUsage(AnalysisUsage &AU) const;

	virtual bool runOnFunction(Function &Fn);

	unsigned MakeReg(MVT::ValueType VT);

	virtual void EmitFunctionEntryCode(Function &Fn, MachineFunction &MF)
	{
	}
	virtual void InstructionSelectBasicBlock(SelectionDAG &SD) = 0;
	virtual void SelectRootInit()
	{
		DAGSize = CurDAG->AssignTopologicalOrder(TopOrder);
	}

	/// SelectInlineAsmMemoryOperand - Select the specified address as a target
	/// addressing mode, according to the specified constraint code.  If this does
	/// not match or is not implemented, return true.  The resultant operands
	/// (which will appear in the machine instruction) should be added to the
	/// OutOps vector.
	virtual bool SelectInlineAsmMemoryOperand(const SDOperand &Op,
			char ConstraintCode, std::vector<SDOperand> &OutOps,
			SelectionDAG &DAG)
	{
		return true;
	}	
	
	// register node N as being covered by tile T
	void setNodeTile(SDNode *N, Tile *T)
	{
		nodeTiles[N].insert(T);
	}
	
	int countNodeTile(SDNode *N)
	{
		return nodeTiles[N].size();
	}
	
	void unsetNodeTile(SDNode *N, Tile *T)
	{
		nodeTiles[N].erase(T);
	}
	
	void clearNodeTiles()
	{
		nodeTiles.clear();
	}
	
	void setSharedNodes(SelectionDAG &DAG);
	bool isSharedNode(SDNode *N) { return sharedNodes.count(N) != 0; }
	
  /// CanBeFoldedBy - Returns true if the specific operand node N of U can be
  /// folded during instruction selection that starts at Root?
  virtual bool CanBeFoldedBy(SDNode *N, SDNode *U, SDNode *Root) const {
    return true;
  }
  
  /// CreateTargetHazardRecognizer - Return a newly allocated hazard recognizer
  /// to use for this target when scheduling the DAG.
  virtual HazardRecognizer *CreateTargetHazardRecognizer();
  
  /// CaseBlock - This structure is used to communicate between SDLowering and
  /// SDISel for the code generation of additional basic blocks needed by multi-
  /// case switch statements.
  struct CaseBlock {
    CaseBlock(ISD::CondCode cc, Value *cmplhs, Value *cmprhs, Value *cmpmiddle,
              MachineBasicBlock *truebb, MachineBasicBlock *falsebb,
              MachineBasicBlock *me)
      : CC(cc), CmpLHS(cmplhs), CmpMHS(cmpmiddle), CmpRHS(cmprhs),
        TrueBB(truebb), FalseBB(falsebb), ThisBB(me) {}
    // CC - the condition code to use for the case block's setcc node
    ISD::CondCode CC;
    // CmpLHS/CmpRHS/CmpMHS - The LHS/MHS/RHS of the comparison to emit.
    // Emit by default LHS op RHS. MHS is used for range comparisons:
    // If MHS is not null: (LHS <= MHS) and (MHS <= RHS).
    Value *CmpLHS, *CmpMHS, *CmpRHS;
    // TrueBB/FalseBB - the block to branch to if the setcc is true/false.
    MachineBasicBlock *TrueBB, *FalseBB;
    // ThisBB - the block into which to emit the code for the setcc and branches
    MachineBasicBlock *ThisBB;
  };
  struct JumpTable {
    JumpTable(unsigned R, unsigned J, MachineBasicBlock *M,
              MachineBasicBlock *D): Reg(R), JTI(J), MBB(M), Default(D) {}
    
    /// Reg - the virtual register containing the index of the jump table entry
    //. to jump to.
    unsigned Reg;
    /// JTI - the JumpTableIndex for this jump table in the function.
    unsigned JTI;
    /// MBB - the MBB into which to emit the code for the indirect jump.
    MachineBasicBlock *MBB;
    /// Default - the MBB of the default bb, which is a successor of the range
    /// check MBB.  This is when updating PHI nodes in successors.
    MachineBasicBlock *Default;
  };
  struct JumpTableHeader {
    JumpTableHeader(uint64_t F, uint64_t L, Value* SV, MachineBasicBlock* H,
                    bool E = false):
      First(F), Last(L), SValue(SV), HeaderBB(H), Emitted(E) {}
    uint64_t First;
    uint64_t Last;
    Value *SValue;
    MachineBasicBlock *HeaderBB;
    bool Emitted;
  };
  typedef std::pair<JumpTableHeader, JumpTable> JumpTableBlock;

  struct BitTestCase {
    BitTestCase(uint64_t M, MachineBasicBlock* T, MachineBasicBlock* Tr):
      Mask(M), ThisBB(T), TargetBB(Tr) { }
    uint64_t Mask;
    MachineBasicBlock* ThisBB;
    MachineBasicBlock* TargetBB;
  };
  
  typedef SmallVector<BitTestCase, 3> BitTestInfo;

  struct BitTestBlock {
    BitTestBlock(uint64_t F, uint64_t R, Value* SV,
                 unsigned Rg, bool E,
                 MachineBasicBlock* P, MachineBasicBlock* D,
                 const BitTestInfo& C):
      First(F), Range(R), SValue(SV), Reg(Rg), Emitted(E),
      Parent(P), Default(D), Cases(C) { }
    uint64_t First;
    uint64_t Range;
    Value  *SValue;
    unsigned Reg;
    bool Emitted;
    MachineBasicBlock *Parent;
    MachineBasicBlock *Default;
    BitTestInfo Cases;
  };
	
	  //return tre if N only has one use; for tiling use precomputed
	  //shared information to avoid getting confused by the junk nodes in the DAG
	  bool hasOneUse(SDOperand N)
	  {
		  if(sharedNodes.empty())
			  return N.hasOneUse();
		  else
		  {
			  return !isSharedNode(N.Val); //considers only data edges, ok?
		  }
	  }
protected:
	hash_set<SDNode*> *deletedNodes; //nodes that were deleted while tiling
	hash_set<SDNode*> sharedNodes; //DAG nodes
	hash_set<SDNode*> fixedNodes; //nodes which are fixed to produce output
	hash_map<SDNode*, std::set<Tile*> > nodeTiles; //tiles covering each node
	hash_map<SDNode*, TileCost> bestTileForNode; //optimal selection info
	
	// Instruction selector priority queue:
	std::vector<SDNode*> ISelQueue;
	/// Keep track of nodes which have already been added to queue.
	unsigned char *ISelQueued;
	/// Keep track of nodes which have already been selected.
	unsigned char *ISelSelected;
	/// Dummy parameter to ReplaceAllUsesOfValueWith().
	std::vector<SDNode*> ISelKilled;
	
	//duplication aware replacement
	void ReplaceUses(SDNode *F, SDNode *T);
	void ReplaceUses(SDOperand F, SDOperand T);
	
	  //return true if this is a call to a constant address 
	virtual bool isDirectCall(SDNode *N) 
	{ 
		assert(false && "need to implement directcall on target");
	}
	  
	//selection helpers
	inline void setQueued(int Id) {
	  ISelQueued[Id / 8] |= 1 << (Id % 8);
	}
	inline bool isQueued(int Id) {
	  return ISelQueued[Id / 8] & (1 << (Id % 8));
	}
	inline void setSelected(int Id) {
	  if(ISelSelected) ISelSelected[Id / 8] |= 1 << (Id % 8);
	}
	inline bool isSelected(int Id) {
	  return ISelSelected != NULL && ISelSelected[Id / 8] & (1 << (Id % 8));
	}

	void AddToISelQueue(SDOperand N)  {
	  int Id = N.Val->getNodeId();
	  if (Id != -1 && !isQueued(Id)) {
	    ISelQueue.push_back(N.Val);
	    std::push_heap(ISelQueue.begin(), ISelQueue.end(), isel_sort());
	    setQueued(Id);
	  }
	}

	inline void RemoveKilled() {
	  unsigned NumKilled = ISelKilled.size();
	  if (NumKilled) {
	    for (unsigned i = 0; i != NumKilled; ++i) {
	      SDNode *Temp = ISelKilled[i];
	      ISelQueue.erase(std::remove(ISelQueue.begin(), ISelQueue.end(), Temp), ISelQueue.end());
	    };
	    std::make_heap(ISelQueue.begin(), ISelQueue.end(), isel_sort());
	    if(deletedNodes) deletedNodes->insert(ISelKilled.begin(), ISelKilled.end());
	    ISelKilled.clear();
	  }
	}


	virtual SDOperand SelectRoot(SDOperand Root) = 0;
	virtual SDNode *Select(SDOperand Op) = 0;
	virtual void Match(SDOperand Op, std::vector<Tile*>& Tiles)
	{
		assert(0 && "Not implemented yet");
	}

	/// Pick a safe ordering and emit instructions for each target node in the
	/// graph.
	void ScheduleAndEmitDAG(SelectionDAG &DAG);

	/// SelectInlineAsmMemoryOperands - Calls to this are automatically generated
	/// by tblgen.  Others should not call it.
	void SelectInlineAsmMemoryOperands(std::vector<SDOperand> &Ops,
			SelectionDAG &DAG);

	// Calls to these predicates are generated by tblgen.
	bool CheckAndMask(SDOperand LHS, ConstantSDNode *RHS, int64_t DesiredMaskS) const;
	bool CheckOrMask(SDOperand LHS, ConstantSDNode *RHS, int64_t DesiredMaskS) const;
	void SelectDAG(SelectionDAG &DAG); //actual perform selection


private:
	static int debug_blocknum;

	void SelectBasicBlock(BasicBlock *BB, MachineFunction &MF,
		FunctionLoweringInfo &FuncInfo);
	void BuildSelectionDAG(SelectionDAG &DAG, BasicBlock *LLVMBB,
		std::vector<std::pair<MachineInstr*, unsigned> > &PHINodesToUpdate,
		FunctionLoweringInfo &FuncInfo);
	void CodeGenAndEmitDAG(SelectionDAG &DAG);
	void LowerArguments(BasicBlock *BB, SelectionDAGLowering &SDL,
		std::vector<SDOperand> &UnorderedChains);

	bool InvalidTile(Tile *tile);
	void AcceptableTiles(SDNode *N, std::vector<Tile *>& tiles);
	int PerformSelection(std::vector<Tile*>& matchedTiles);
	void SetBestTile(SDNode *N);
	void MaxMunchTreeSelect(SelectionDAG &DAG);
	void ImproveCSEDecisions(SelectionDAG &DAG);
	void BottomUpDynamicProgramming(std::vector<SDNode*>& topo);
	void TopDownSelectTiles(SelectionDAG &DAG,				
				std::vector<Tile*>& matchedTiles);	
	void exploreAreaSubNodes(SDNode *N, const std::set<SDNode*>& enodes, 
			const std::set<Tile*> &tiles, std::queue<Tile*>& q, 
			hash_set<SDNode*> seen = hash_set<SDNode*>());
	int tileCutCost(Tile *tile, SDNode *N);
	int getPartialTileCost(Tile *tile, SDNode *N);
	int getOverlapAreaCost(const std::set<Tile*> &tiles, SDNode *N);
	void OptimalTreeSelect(SelectionDAG &DAG);
	void WriteILP(SelectionDAG &DAG, std::ostream &o);
 
	/// SwitchCases - Vector of CaseBlock structures used to communicate
	/// SwitchInst code generation information.
	std::vector<CaseBlock> SwitchCases;


  /// JTCases - Vector of JumpTable structures which holds necessary information
  /// for emitting a jump tables during SwitchInst code generation.
  std::vector<JumpTableBlock> JTCases;
 
  std::vector<BitTestBlock> BitTestCases;

	//merges token factors; used by replace uses
	void mergeTokenFactors(SDNode *TF);
	
};

//A Tile matches LLVM DAG nodes to target nodes
class Tile
{
	static int currID;
	int id;
public:
	virtual void doMatch() = 0; //perform the transformation represented by this tile
	virtual ~Tile()
	{
	}
	Tile(SelectionDAGISel *isel) :
		ResNode(NULL), CurDAG(isel->CurDAG),
				ISel(isel)
	{
		id = currID++;
	}

	int getID()
	{
		return id;
	}
	
	int getCost()
	{
		return cost;
	}
	
	int getSize()
	{
		return size;
	}
	
	bool hasInternalDagNodes()
	{
		return shared_nodes.size() > 0;
	}

	const std::vector<SDOperand>& getNodes()
	{
		return nodes;
	}
	
	const std::set<SDNode*>& getInternalNodes()
	{
		return internal_nodes;
	}
	

	virtual const std::vector<SDOperand>& getEdgeNodes()
	{
		return edge_nodes;
	}
	
	const std::vector<SDNode*>& getSharedNodes()
	{
		return shared_nodes;
	}

	SDNode *getRootNode()
	{
		if(nodes.size() == 0)
			return NULL;
		else
			return nodes[0].Val;
	}
	
	const SDNode *getResNode()
	{
		return ResNode;
	}

	const std::vector<SDNode *>& getOtherNodes()
	{
		return OtherNodes;
	}

	const std::vector<SDNode *>& getCopiesBefore()
	{
		return CopiesBefore;
	}

	const std::vector<SDNode *>& getCopiesAfter()
	{
		return CopiesAfter;
	}

	//registers each node of the tile as being covered by this tile
	//must be called before mutating the DAG
	void apply()
	{
		if(nodes.size() > 0)
			apply_r(nodes[0].Val);
	}
	//de-registers each node of the tile as being covered by this tile
	void unapply()
	{
		for(unsigned i = 0; i < nodes.size(); i++)
		{
			ISel->unsetNodeTile(nodes[i].Val, this);		
		}
		for(unsigned i = 0; i < edge_nodes.size(); i++)
		{
			ISel->unsetNodeTile(edge_nodes[i].Val, this);		
		}		
	}

private:
	//return true if vec contains N
	bool vectorContains(const std::vector<SDOperand>& vec, SDNode *N)
	{
		unsigned i;
		unsigned n = vec.size();
		for(i = 0; i < n; i++)
		{
			if(vec[i].Val == N)
				return true;
		}
		return false;
	}

	//recursive helper for apply
	void apply_r(SDNode *N)
	{		
		if (vectorContains(edge_nodes, N)) //an edge node
		{		
			return;
		}		
		ISel->setNodeTile(N, this);	
		if(!vectorContains(nodes, N))
			nodes.push_back(SDOperand(N,0)); //keep track of all nodes for unapply
		for (unsigned i = 0; i < N->getNumOperands(); i++)
		{
			apply_r(N->getOperand(i).Val);
		}
	}
	
	//recursive helper for determining how many nodes are in this tile
	//also has side-effect of setting internalDAGNodes and internal_nodes
	int countNodes(SDNode *N)
	{
		int ret = 0;
		if (vectorContains(edge_nodes, N)) //an edge node
		{		
			return ret;
		}

		internal_nodes.insert(N);
		if(N != nodes[0].Val && ISel->isSharedNode(N)) //it's okay for root to have uses
			shared_nodes.push_back(N);
		ret++; //this node
		for (unsigned i = 0; i < N->getNumOperands(); i++)
		{
			ret += countNodes(N->getOperand(i).Val);
		}

		return ret;
	}

protected:
	void setSizeAndCost() //also sets internalDAGnodes and internalNodes
	{
		//get cost
		cost = 0; 
		if (ResNode != NULL)
			cost += CurDAG->SizeNode(ISel->BB, ResNode);
		for (unsigned i = 0; i < OtherNodes.size(); i++)
			cost += CurDAG->SizeNode(ISel->BB, OtherNodes[i]);
		for (unsigned i = 0; i < CopiesBefore.size(); i++)
			cost += CurDAG->SizeNode(ISel->BB, CopiesBefore[i]);
		for (unsigned i = 0; i < CopiesAfter.size(); i++)
			cost += CurDAG->SizeNode(ISel->BB, CopiesAfter[i]);

		//compute the size by counting all nodes reached from the
		//root node stopping at the edge nodes; include edges nodes
		//which are also nodes, these are presumably operands which
		//get fully included into the insn
		if(nodes.size() > 0)
			size = countNodes(nodes[0].Val);
		else
			size = 0;
		
		
	}
	
	std::vector<SDOperand> nodes; //nodes covered by tile, first is root
	std::vector<SDOperand> edge_nodes; //children of this tile
	std::vector<SDNode*> shared_nodes; //internal DAG nodes
	std::set<SDNode*> internal_nodes; //internal, non edge, nodes
	SDNode *ResNode; //resulting target node, unique to this tile
	std::vector<SDNode *> OtherNodes; //additional insns
	//any register copies associated with this tile
	std::vector<SDNode *> CopiesBefore; //performed before ResNode
	std::vector<SDNode *> CopiesAfter; //performed after ResNode
	int cost;
	int size;
	SelectionDAG *CurDAG;
	SelectionDAGISel *ISel;

};

}

#endif /* LLVM_CODEGEN_SELECTIONDAG_ISEL_H */
