#ifndef ALLOCATIONCLASSINFO_H_
#define ALLOCATIONCLASSINFO_H_

#include <string>
#include <vector>
#include "llvm/Support/Compiler.h"
#include "llvm/Support/CommandLine.h"
#include "llvm/Target/TargetMachine.h"
#include "llvm/Target/MRegisterInfo.h"
#include "llvm/ADT/hash_map"
namespace llvm
{

enum MetricToUse {
	Size,
  Speed
};
extern cl::opt<MetricToUse> WhatMetric;

class RAATInfo;
/* Provides information about the allocation classes provided by an architecture.
 * The target backend must derive a subclass for every set of allocation classes
 * it wishes to use.
 *
 * This is mostly to easy porting the code from gcc.  Eventually it would be
 * a good idea to auto generate this functionality from tables.
 * */
class AllocationClassInfo
{
public:
	/*
	   Register is a normal register.
	   Memory requires antivariable to model stores
	   Constant is for const_equiv variable evictions
	   AntiOnly is for the side network of all memory */
	enum class_type {None,Register, Memory, Constant, AntiOnly};

	AllocationClassInfo(const TargetMachine &tm): TM(tm), MRI(*tm.getRegisterInfo()) {}
	virtual ~AllocationClassInfo() {}

	virtual void initialize() {} //because virtuals inside constructors don't work

	virtual bool hasMultiRegClasses() const { return false; }
	virtual unsigned numClasses() const { return allocclasses.size(); }
	virtual unsigned numCalleeSaves() const { return 0; }
	virtual unsigned numCallerSaves() const { return 0; }
	/* Return type (reg,mem,const,anti) of class c */
	virtual class_type getType(int c) const;
	/* Return (first) register for this class, 0 if not a reg */
	virtual unsigned getRegister(int c) const;
	/* Return all canonical registers for a register class */
	virtual const std::vector<unsigned>& getRegisters(int c) const;
	/* cost of move form class c1 to c2 */
	virtual unsigned getMoveCost(int c1, int c2) const;
	virtual unsigned getSelfMoveCost(int c) const { return 0; } //only relevant for multireg class
	/* any additional cost of allocating the operand at op_index to class c of MI*/
	virtual int getInsnCost(MachineInstr *MI, int op_index, int c) const { return 0; }
	virtual int getInsnRemovalCost(MachineInstr *MI, int isize) const { if(WhatMetric == Size) return isize; else return 1; }
	/* printable unique name of class */
	virtual const std::string& getName(int c) const;
	/* TargetRegisterClass for a register allocation class (null if not reg) */
	virtual const TargetRegisterClass* getTargetClass(int c) const;

	/* true if allocation class index is compatible with RC */
	virtual bool isCompatible(const MachineFunction &MF, const TargetRegisterClass* RC, int index) const;
	/* index of register allocation class given regno */
	virtual int getRegisterIndex(unsigned r) const;
	/* Return corresponding antionly class of a memory class;
	 * returns -1 if invalid input */
	virtual int getAntiClass(int c) const;
	/* Return store cost */
	virtual int getStoreCost(int c) const
	{
		return allocclasses[c].move_from_cost[getAntiClass(c)];
	}
	/* Return the id of a class */
	virtual int getClassId(int c) const;
	/* Return capacity of class c */
	virtual int getCapacity(int c) const;
	/* Return vector of memory classes */
	const std::vector<int>& getMemoryClasses() const { return memory_classes; }
	/* Return vector of constant classes */
	const std::vector<int>& getConstantClasses() const { return constant_classes;}
	/* Return vector of register classes */
	const std::vector<int>& getRegisterClasses() const { return register_classes;}
	//no special restrictions on using class across block boundaries
	virtual bool usableAcrossBlocks(int c) const { return true; }

	bool isSubClassOf(int c, const AllocationClassInfo* simple, int sc) const;
protected:
	const TargetMachine &TM;
	const MRegisterInfo &MRI;

	class AllocationClass
	{
	public:
		/* a memory class has the index of the
						       side eviction network of the antivariables  in capcity*/
		short capacity;    /* capacity of the class, negative if uncapcitated */
		short id; /* special id for class, usualy unused */
		enum class_type type;
		const TargetRegisterClass *TC; // if reg, the most restrictive llvm register class it's assigned
		std::vector<short> move_from_cost; /* indexed by other classes, cost to move from this class to that class */
		std::string name; //printable name
		std::vector<unsigned> regNums; //canonical registers of a register class

		AllocationClass(): capacity(0), id(0), type(None), TC(NULL)
		{
		}
	};

	std::vector<AllocationClass> allocclasses;
	std::vector<int> register_classes;
	std::vector<int> memory_classes;
	std::vector<int> constant_classes;

	  /* compute cost to move between allocatin classes, this should
	   * be overloaded */
	  virtual int computeMoveCost(unsigned src, unsigned dst);
		/* create allocation classes, these methods should be caleld by
		 * deriving classes.  They return the new allocation class number */
		int addRegisterClass(unsigned reg, const TargetRegisterClass *trc);
		int addRegisterClass(const TargetRegisterClass *trc,  const std::vector<unsigned>& regs, std::string name);
		int addMemoryClass(int capacity, short id);
		int addConstantClass(short id);

		/* derived classes MUST call this after adding allocation classes */
		void setupCosts();

};

}

#endif /*ALLOCATIONCLASSINFO_H_*/
