#ifndef GLOBALMCNF_H_
#define GLOBALMCNF_H_

#include "llvm/ADT/BitVector.h"
#include "llvm/ADT/hash_map"
#include "llvm/ADT/hash_set"
#include "llvm/ADT/SmallSet.h"
#include "llvm/ADT/UnReservedVector.h"
#include "llvm/Analysis/LoopInfo.h"
#include "llvm/Analysis/ProfileInfo.h"
#include "llvm/CodeGen/LiveIntervalAnalysis.h"
#include "llvm/CodeGen/MachineFunction.h"
#include "llvm/CodeGen/MachineFrameInfo.h"
#include "llvm/CodeGen/DataFlowLiveVariables.h"
#include "llvm/CodeGen/SSARegMap.h"
#include "llvm/PrincipledBackend/RAATInfo.h"
#include "llvm/Support/Debug.h"
#include "llvm/Support/Compiler.h"
#include "llvm/Support/CommandLine.h"
#include "llvm/Target/TargetMachine.h"
#include "llvm/Target/TargetInstrInfo.h"
#include <vector>
#include <map>
#include <functional>
#include <queue>
#include <iostream>

#include "llvm/PrincipledBackend/ProgCommon.h"
#include "llvm/PrincipledBackend/BitSet.h"

#define MAX_K	2 /* maximum number of k shortest paths */
#define MAX_CANDIDATES 1  /* maximum number of candidates for global paths */
#define UNREACHABLE_INT ((int)0x80000000)

/* A Global MCNF data structure for register allocation.
 To ease the porting from C code this is just a container
 structure with publicly accessibly data.
 */
namespace llvm
{

struct mcnf_edge
{
	mcnf_node *src;
	mcnf_node *dst;

//	SmallSet<commod_type, 3> commod_map;
	BitSet commod_map;
	UnReservedVector<commod_type> commodities; /*commodities that can use this edge */
	int cost; /* cost of this edge for any commodity */

	mcnf_edge() :
		src(NULL), dst(NULL), cost(0)
	{
	}

	~mcnf_edge()
	{
	}

	int getCost(int q);
	double getPrice(int q);
	std::string getName(int q);
	bool commodity_valid(int q);

	//make this edge a part of the solution for commodity q
	void set_solution(int q);
	//return true if q is part of solution on this edge
	bool is_set_solution(int q);
	//is part of a feasible solution
	void mark_feasible_edge(int q);
	void unmark_feasible_edge(int q);
};

int mcnf_edge_cost(mcnf_edge *edge, int q);
double mcnf_edge_price(mcnf_edge *edge, int q);
bool mcnf_boundary_node_valid(mcnf_node *node, int q);
double mcnf_edge_price_fixed_boundary_exits(mcnf_edge *edge, int q);

/* Per commodity data needed for shortest paths computations */
struct shortest_info
{
	double value; /* value of shortest path to this node */
	int cost;
	mcnf_edge *prev_edge; /* pointer to previous edge along _a_ shortest path */
//	mcnf_edge *next_edge; /* pointer to next edge along shortest path, set by fsf */
//	index_type index; /* for kth shortest, which shortest path to prev */

	shortest_info(double val = 0.0) :
		value(val), cost(0), prev_edge(NULL) //, next_edge(NULL), index(0)
	{
	}
};

/* Per commodity data needed for balanced shortest paths computations */
struct balanced_info
{
	double ratio; /* number between 0 and 1, what percent of shortest paths go through here */
	double num_shortest; /* number of shortest paths */
	unsigned prevs; /* bitmask of all shortest paths (index based on preds array */

	balanced_info() :
		ratio(0), num_shortest(0), prevs(0)
	{
	}
};

struct feasible_solution_info
{
	double priced_cost;
	mcnf_edge *prev_edge; /* prev edge in this allocation */
	mcnf_edge *next_edge;

	feasible_solution_info() :
		priced_cost(0), prev_edge(NULL), next_edge(NULL)
	{
	}
};

struct mcnf_candidates
{
	/* holds candidate paths for k shortest, each candidate
	 has a back pointer to the previous node and index as
	 well as the length of the candidate path */
	std::vector<double> length; /* length of this candidate for each prev*/
	/* this is an indexed array implementation, the index
	 of the array is the index of the previous node (-1 if no path) */
	std::vector<index_type> kindex; /* which shortest path version to use from previous path */
	commod_type c; /* what commodities this represents */

	mcnf_candidates() :
		c(0)
	{
	}
};

/* Data structure for mapping a quantity number to
 an index in a node or edge.  Don't use stl hash_map
 since it has a minimum size of 53 buckets. Instead implement
 an open address hash table. */
class commodity_index
{
	struct hash_entry
	{
		commod_type key;
		index_type value;

		hash_entry() :
			key(0), value(0)
		{
		}
	};

	UnReservedVector<hash_entry> table;
	unsigned size; //number of elements

	/* look for the correct spot in an open address hash table */
	int get_index_helper(commod_type k, int index)
	{
		int i;
		int cap = table.size();

		/* look for key */
		for (i = 0; i < cap; i++)
		{
			int table_key = table[index].key;
			if (table_key == k) /* found it */
			{
				return index;
			}
			else if (table_key == 0)
			{
				return index; /* first empty spot, elements never removed so this is where it should be */
			}

			index = (index+1)%cap;
		}

		abort(); /* full? */
		return -1;
	}

	/* return the index for q in cindex
	 make this as small as possible to ensure inlining
	 */
	int get_index(commod_type k)
	{
		unsigned index;
		unsigned cap = table.size();
		int key;

		index = ((unsigned)k) % cap;

		key = table[index].key;

		if (key == k)
			return index;
		else if (key == 0)
			return index;
		else /* collision */
		{
			return get_index_helper(k, index);
		}
	}

	//double (roughly) the size of the hash table
	void resize()
	{
		UnReservedVector<hash_entry> table2(table.size()*2+1);
		//put current contents into table2, increase size of table
		std::swap(table2, table);
		size = 0; //no elements in empty table
		for (unsigned i = 0, n = table2.size(); i < n; i++)
		{
			if (table2[i].key != 0)
				add(table2[i].key, table2[i].value);
		}
	}
public:
	commodity_index() :
		table(11), size(0)
	{
	}
	index_type get(commod_type q)
	{
		int spot = get_index(q);
		if(table[spot].key == 0)
		    return -1;
		return table[spot].value;

	}
	void add(commod_type q, index_type index)
	{
		int spot = get_index(q);
		if (table[spot].key == 0) /* have to add it */
		{
			if (2*size+1 > table.size()) //max load factor of .5
			{
				resize();
				spot = get_index(q);

				if (table[spot].key != 0)
					abort();
			}
			table[spot].key = q;
			table[spot].value = index;
			size++;
		}
		else if(table[spot].key != q || table[spot].value != index)
			abort();
	}
};

struct usable_info
{
	bool set:1; //this quantity is blocking this node
	bool unusable:1; //number of other quantities using this node that conflict with this quantity
	int allocClass:14; /* class this is set to */

	usable_info() :
		set(false), unusable(false), allocClass(0)
	{
	}
};

enum mcnf_node_type
{	nError, nDef, nSink, nXBar, nInsn, nEntry, nExit, nLast};
extern const char *node_type_prefix[];

enum visited_type
{
	VISITED_CLEAR, /* not seen by any algorithm */
	VISITED_SHORTEST_SEEN, /* seen by shortest algorithm */
	VISITED_SHORTEST_QUEUE, /* in the queue of shortest algo */
	VISITED_SHORTEST_DONE, /* seen and no longer in queue of shortest algo */
	VISITED_BALANCED_SEEN, /* seen by balancing algorithm */
	VISITED_BALANCED_QUEUE, /* in queue of balancing algorithm */
	VISITED_BALANCED_DONE, /* out of queue of balancing algorithm */
	VISITED_OVERFLOW_SEEN, /* seen by shortest-overflow */
	VISITED_OVERFLOW_QUEUE, /*in queue */
	VISITED_OVERFLOW_DONE, /* finalized */
	VISITED_PRICE_INIT, /* we initialized the capacity price on this node */
	VISITED_SHORTEST_BACK_SEEN, /* seen by shortest algorithm */
	VISITED_SHORTEST_BACK_QUEUE, /* in the queue of shortest algo */
	VISITED_SHORTEST_BACK_DONE /* seen and no longer in queue of shortest algo */
};
/* Arrays referring to commodity info in a node are indexed
 by a node specific index.
 A base node contains everything shortest path algorithms need.
 Some attempt is made to optimize for spatial locality in the field layout.
 */
struct mcnf_node
{
	double price; /* lagrangian price */

	UnReservedVector<commod_type> commodities;
	UnReservedVector<int> nonanti_costs; /* cost of a non-anti commodity being allocated to this node */

	double *dual_prices; /* for memory nodes, the price for a commodity pair is in the positive position*/

	/* boundary constraint prices for entry/exit nodes
	 indexed first be pred/succ, then by commodity */
	double **boundary_prices;

	short capacity; /* how many commodities are allowed through this node */
	short flowcnt; /* number commodities currently allocated to this node (does not include antis) */
	short constrained_flowcnt; /* number of commodities _fixed_ to this node */

	enum mcnf_node_type type : 4;
	enum visited_type visited : 4;

	UnReservedVector<mcnf_edge *> succs; /* array of out edges */
	UnReservedVector<mcnf_edge *> preds; /* array of predecessor edges */

	shortest_info* sinfo[MAX_K];
	balanced_info* binfo;

	int block_num; /* number of the mcnf_block this node is in */
	int node_order; /* topological position in block (within this block, unique to this node) */
	int position; /* insn # for xbar, insn, commodity for def/sink */
	short subindex; /* level of xbar, which def/sink for a given commodity */
	short allocClass; /* for insn/xbar */

	int vecpos; /* position in block node vector TODO: make this obsolete */
	mcnf_node *mirror_node; /* mem/anti node pointer */

	//gradients
	double gradient; /* lagrangian gradient */
	double *dual_gradients;
	double **boundary_gradients;

	double *primal_solution;
	usable_info *used_classes; /* indexed by quantity, set if unusable has been set for that quantity */

	commodity_index cindex; /* maps from quantity to index */

	/* when constructing several candidate allocations for a single quantity,
	 need to keep track of them all */
//	std::vector<mcnf_edge> candidate_prev[MAX_CANDIDATES]; /* path for a candidate */
//	usable_info candidate_used_class[MAX_CANDIDATES];

	feasible_solution_info *fsinfo; /* indexed by  quantity; info for mark */
	std::vector<bool> feas_marked; //bitvector indexed by quantity, set if part of feasible solutio
	mcnf_node() :
		price(0), dual_prices(NULL), boundary_prices(NULL), capacity(0),
				flowcnt(0), constrained_flowcnt(0), type(nError),
				visited(VISITED_CLEAR), binfo(NULL), block_num(0),
				node_order(0), position(0), subindex(0), allocClass(0),
				vecpos(0), mirror_node(NULL), gradient(0),
				dual_gradients(NULL), boundary_gradients(NULL),
				primal_solution(NULL), used_classes(NULL), fsinfo(NULL)
	{
		for (int i = 0; i < MAX_K; i++)
			sinfo[i] = NULL;
	}

	~mcnf_node();

	std::string getName();

	//return index of edge in preds
	int pred_index(mcnf_edge *edge);

	bool has_class(); //return true if class field is meaningful

	/* Return the total amount of flow in a node, assumes
	 balanced paths have been computed.
	 */
	double node_flow();

	bool dual_constraint_matters(int c);
};

void initialize_node_for_feasible_solution_finder(mcnf_node *node);

/* Registers represented by this quantity
 Sources and sinks throught the mcnf for this quantity
 (vector does not have ownership)
 */
struct quantity
{
	std::vector<unsigned> regs;

	std::vector<mcnf_node*> defs;
	std::vector<mcnf_node*> sinks;
	std::vector<std::vector<mcnf_node*> > antidefs;
	std::vector<std::vector<mcnf_node*> > antisinks;

	/* shortest path info */
	double priced_cost;
	double anti_priced_cost;
	int unpriced_cost; /* cost of unconstraid path */
	int anti_unpriced_cost;
	float weight; //llvm spill cost heuristic
	short prefClass; //from llvm preference register
	short lastAlloced; //what we chose last time
	bool fixed; /* true if fixed */
	bool spilled; //true if spilled at some point

	quantity(int numAntis) :
		antidefs(numAntis), antisinks(numAntis), priced_cost(0),
				anti_priced_cost(0), unpriced_cost(0), anti_unpriced_cost(0),
				weight(0), prefClass(-1), lastAlloced(-1),  fixed(false), spilled(false)
	{
	}
};

class RQMap
{
	hash_map<unsigned, commod_type> regmap;
	//a map from registers to quantities that returns a zero quantity
	//if the register isn't in the map
public:
	RQMap()
	{
	}
	virtual ~RQMap()
	{
	}

	commod_type get(unsigned r)
	{
		if (regmap.find(r) != regmap.end())
			return regmap[r];
		return 0;
	}

	void put(unsigned r, commod_type q)
	{
		regmap[r] = q;
	}

	void clear()
	{
		regmap.clear();
	}
};


/* Information on each RAAT with weighted TiedConstraints */
struct TC_RAATInfo
{
	int srcq;
	int dstq;
	int cost; //weighted cost

	std::vector<mcnf_node*> inodes; //insn nodes

	TC_RAATInfo(): srcq(0), dstq(0), cost(0) {}

	TC_RAATInfo(int s, int d, int c): srcq(s), dstq(d), cost(c) {}
};

enum entry_exit
{	Entry, Exit};
typedef double (*edge_weight_fn)(mcnf_edge *edge, int q);
class GlobalMCNF;

typedef std::map<int, unsigned> RegMap;

int compute_interval_level(int position, int subindex, bool isXBar);
extern int offset_to_xbar_middle;
int getNodeIntervalLevel(mcnf_node *node);
int getIntervalLevelPosition(int level);
int getIntervalLevelSubindex(int level);
enum mcnf_node_type getIntervalLevelType(int level);


/* A mcnf representation of a basic block */
struct mcnf_block
{
	GlobalMCNF *cfg;
	std::vector<mcnf_node*> defs; /* includes antidefs */
	std::vector<mcnf_node*> sinks; /* includes antisinks */
	std::vector<mcnf_node*> entries;
	std::vector<mcnf_node*> exits;
	std::vector<mcnf_node*> body; /* insn and xbars, intermingled */

	//the next two vectors "own" the memory
	std::vector<mcnf_node*> allnodes;
	std::vector<mcnf_edge*> alledges; //local edges

	MachineBasicBlock *b;
	int index; /* index in the mcnf_cfg array */

	std::vector<RAATInfo> RAATs; // persistent insn info plus alloc info
	std::vector<TC_RAATInfo> TC_RAATs; //info of RAATs with tied constraints
	hash_map<int, int> TCRAATMap; //from level to index in TC_RAATs

	std::vector<std::vector<int> > quantity_next_use; /* not particularly efficient, indexed by current insn level
	 and quantity, the number of the next insn (or exit) that
	 uses that quantity */
	/*  indexed by quantity of price of path in this block
	 if quantity isn't used, price is left at zero */
	hash_map<int, double> quantity_prices;

	std::vector<mcnf_block*> succs; /* successor blocks */
	std::vector<mcnf_block*> preds; /* predecessor blocks */

	hash_map<int, hash_set<int> > quantities_used_in_class; //for multireg quantities, quantity is using this class in a connected block
	struct constrained_block
	{
		mcnf_block *block;
		enum entry_exit type;

		constrained_block(mcnf_block *b, enum entry_exit t) :
			block(b), type(t)
		{
		}
	};

	//all blocks that are constrained by this block's entry allocations, includes the block itself
	std::vector<constrained_block> blocks_constrained_by_entry;
	//similarly, by the exit
	std::vector<constrained_block> blocks_constrained_by_exit;
	//true if the entry of this block constrains the exit
	bool entry_exit_constrained; /* true if has succ/pred equal to self */
	bool is_self_loop; //true if has itself as pred/succ

	std::vector<unsigned> liveins_vec; //regs livein
	std::vector<unsigned> hard_liveins; //physical regs that are live into the block

	hash_map<int, mcnf_block*> def_dominator_in;
	hash_map<int, mcnf_block*> def_dominator_out;

	mcnf_block(GlobalMCNF *gmcnf) :
		cfg(gmcnf), b(NULL), index(-1), entry_exit_constrained(0), is_self_loop(false), lastXBarPos(-1)
	{
	}

	~mcnf_block()
	{
		for (unsigned i = 0, n = allnodes.size(); i < n; i++)
			delete allnodes[i];
		for (unsigned i = 0, n = alledges.size(); i < n; i++)
			delete alledges[i];
	}

	template<class NodeProcessor> void traverse_nodes(NodeProcessor& process);
	//trace shortest path back
	template <class EdgeMarker> bool trace_path(mcnf_node *sink, int q,
			EdgeMarker& func);
	template <class EdgeWeight, class EdgeCost> double
			shortest_paths(EdgeWeight& edge_weight, EdgeCost& edge_cost,
					bool calcbinfo = false);
	template <class EdgeWeight, class EdgeCost, class TieBreaker> void
			shortest_paths_backwards(EdgeWeight& edge_weight,
					EdgeCost& edge_cost, TieBreaker& this_edge_better_if_tied2);

	void initialize_path_info();
	void set_balanced_flows();
	void print_quantity(std::ostream& out, int q);

	bool is_entry_level(mcnf_node **level);
	bool is_exit_level(mcnf_node **level);
	bool is_insn_level(mcnf_node **level);
	void get_exit_level(mcnf_node **level);
	int get_level(int start, mcnf_node **nextnodes, mcnf_node **insnnodes);
	void set_to_previous_level(mcnf_node **level);

	void compute_entry_exit_constraints();
	void check_feasible_links();

	bool edge_insertable(); //return true if can insert edges between this block and successors

	//return amount shortest paths solution should be adjusted by to
	//reflecti mpact of weighted tied constraints; also adjust shortest paths accordingly
	double shortest_tied_constraint_adjust();

	//keep track of register assignments within multireg classes

	//return the register being sued by q at node
	//will only return a valid register if node is a transition node
	//(the assignment is changing)
	//datastructures for multireg nodes
	int lastXBarPos;

	hash_map<int, RegMap> quantityRegMap; //quantity, position, reg
	hash_map<int, unsigned> quantityRegMapOut;

	typedef std::map<int, int> UsageMap;
	hash_map<unsigned, UsageMap> regUsageMap; //for each reg, what quantity it is mapped to
	hash_map<unsigned, int> regUsageMapOut;
	hash_map<int, hash_set<unsigned> > regUnavailableToQuantities;

	typedef std::map<int, bool> HardUsageMap; //true means in use
	hash_map<unsigned, HardUsageMap> hardRegUsageMap; //for each reg, map from position to whether the hard reg is in use (unavailable)

	int lastAvailablePosForQuantityInReg(int q, unsigned r, int start, int end, bool isExit);
	unsigned getRegisterForQuantity(int q, mcnf_node *node, int offset=0);
	unsigned getRegisterForQuantityAtPos(int q, int level);
	unsigned getRegisterForQuantityIn(int q); //register into block
	unsigned getRegisterForQuantityOut(int q); //register out of block
	bool regAvailableForQuantityAtInterval(unsigned reg, int q, int start, int end, bool isExit);
	bool regAvailableForQuantityAtNode(unsigned reg, int q, mcnf_node *node);
	bool regAvailableForQuantityAtOut(unsigned reg, int q);
	void setQuantityRegAtPos(int q, unsigned r, int pos);
	void setQuantityReg(int q, unsigned r, mcnf_node *node, int offset = 0);
	void setQuantityRegIn(int q, unsigned r);
	void setQuantityRegOut(int q, unsigned r);
	void setRegUnavailableForQuantityOut(int q, unsigned r);
	void setRegAvailableForQuantityOut(int q, unsigned r);
	int getQuantityForReg(unsigned reg, mcnf_node *node);
	int getQuantityForRegAtPos(unsigned reg, int pos);
	void propagateQuantityRegOut(int q);
	void clearQuantityRegs();
	void printMaps();

	bool getHardRegUnavailable(unsigned r, int position, int subindex, bool isXBar);
	bool getHardRegUnavailableAtNode(unsigned r, mcnf_node* node);
	void setHardRegUnavailability(int position, int subindex, const std::set<unsigned>& unavailable_regs, bool isXBar);

private:

	mcnf_block(const mcnf_block& rhs)
	{
		assert(0);
	}

	void debug_print_quantity_shortest(int q);
	void compute_entry_exit_constraint_helper(mcnf_block *block,
			hash_map<mcnf_block*, char>& visited, enum entry_exit type,
			bool started_at_entry);

};

//set the flow count of a feasible solution
class set_flow_count : public std::binary_function<mcnf_edge *, int, bool>
{
public:

	bool operator()(mcnf_edge *edge, int q)
	{
		mcnf_node *node = edge->src;
		node->flowcnt++;
		return true;
	}
};

//used to accumulate the cost of a path
class tally_path_cost : public std::binary_function<mcnf_edge *, int, bool>
{
	int cost;
public:
	tally_path_cost() :
		cost(0)
	{
	}

	bool operator()(mcnf_edge *edge, int q)
	{
		cost += edge->getCost(q);
		return true;
	}

	int getCost()
	{
		return cost;
	}
};


//used to accumulate the price of a path
class tally_path_price : public std::binary_function<mcnf_edge *, int, bool>
{
	double price;
public:
	tally_path_price() :
		price(0)
	{
	}

	bool operator()(mcnf_edge *edge, int q)
	{
		price += edge->getPrice(q);
//		double justprice = edge->getPrice(q) - edge->getCost(q);
//std::cerr << "TPP " << edge->dst->getName() << " " << justprice << "\n";
		return true;
	}

	double getPrice()
	{
		return price;
	}
};

class GlobalMCNF
{
	//avoid fragmentation by maintaining memory pools for path info
	//these are allocated with malloc
	char *binfo_pool;
	char *sinfo_pool[MAX_K];
	char *fsinfo_pool;
public:
	//sub data structures

	std::vector<mcnf_block*> blocks;
	std::vector<int> block_order; /*topo order of blocks */
	std::vector<int> block_order_index; /* indexed by block_num, gives index in block_order */

	hash_map<unsigned, BitVector> conflicts; //asymetric conflict map
	hash_set<MachineInstr*> coalesceableMoves; //safe to coalesce
	MachineFunction *MF;
	const TargetMachine* TM;
	DataFlowLiveVariables* LV;
	LiveIntervals *LI;
	LoopInfo *Loops;
	std::vector<quantity> quantities; //quantities
	RQMap reg_quantity; //current quantity mapped to reg
	hash_map<int, int> savedRegHardRegs; //which quantities are just saving callee save registers
	hash_set<unsigned> clobberedClasses; //hard registers which are clobbered in this function

	hash_map<mcnf_edge*, std::vector<int> > blockedEdges; //when constraining a full model based off simple model solution for 2phase solving
public:
	GlobalMCNF() :
		binfo_pool(NULL), fsinfo_pool(NULL),
		quantities(1, quantity(0))
	{
		for(unsigned i = 0; i < MAX_K; i++)
			sinfo_pool[i] = NULL;
		clear();
	} //create empty model
	virtual ~GlobalMCNF()
	{
		clear();
	}

	void build(MachineFunction *mf, DataFlowLiveVariables *LV, LiveIntervals *LI, LoopInfo *Loops);
	void setSavedReg(unsigned r, unsigned hard_reg);

	void clear()
	{
		if(binfo_pool != NULL)
			free(binfo_pool);
		binfo_pool = NULL;

		for(unsigned i = 0; i < MAX_K; i++)
		{
			if(sinfo_pool[i] != NULL)
				free(sinfo_pool[i]);
			sinfo_pool[i] = NULL;
		}

		if(fsinfo_pool != NULL)
			free(fsinfo_pool);
		fsinfo_pool = NULL;

		for (unsigned i = 0, n = blocks.size(); i < n; i++)
			delete blocks[i];

		blocks.resize(0);
		block_order.resize(0);
		block_order_index.resize(0);
		MF = NULL;
		TM = NULL;
		coalesceableMoves.clear();
		conflicts.clear();
		quantities.clear();
		reg_quantity.clear();
		savedRegHardRegs.clear();
		clobberedClasses.clear();
		quantities.push_back(quantity(0));
		num_edges = 0;
		num_nodes = 0;
		num_RAATs = 0;
		num_ignoreable = 0;
		maxLive = 0;
		maxLiveAcrossCalls = 0;
		maxIntLive = 0;
		maxFPLive = 0;
		max_block_cost = 0;
	}

	/* Traverse the whole mcnf cfg and apply process to
	 each node */
	template<class NodeProcessor> void traverse_nodes(NodeProcessor& process);
	template <class EdgeMarker>
	//trace feasible paths
	void trace_all_feasible_paths(int q, EdgeMarker& func);
	template <class EdgeMarker> bool trace_feasible_path(mcnf_node *sink,
			int q, EdgeMarker& func);

	int num_edges;
	int num_nodes;
	int num_alloc_classes;
	int num_RAATs;
	int num_ignoreable;
	unsigned maxLive;
	unsigned maxLiveAcrossCalls;
	unsigned maxIntLive;
	unsigned maxFPLive;
	unsigned max_block_cost;

	int numEmptyBlocks()
	{
		int ret = 0;
		for(unsigned i = 0, n = blocks.size(); i < n; i++)
		{
			mcnf_block *block = blocks[i];
			if(block->b->getFirstTerminator() == block->b->begin())
				ret++;
		}
		return ret;
	}

	GlobalMCNF(const GlobalMCNF& rhs)
	{
		assert(0);
	}

	void copyPrices(GlobalMCNF& simple, const AllocationClassInfo *simple_alloc_classes);
	void constrainModel(GlobalMCNF& simple, const AllocationClassInfo *simple_alloc_classes);

	void dfo_blockorder_r(mcnf_block *block, int *curr,
			std::vector<char>& visited);
	void dfo_blockorder();

	int getNumTiedconstraints();
	/*** Building the MCNF */
	void mcnf_cfg_init(); //basic initialization

	int create_reg_quantity(int var);
	void add_reg_quantity(int var, int q);
	void setup_livesets();
	void setup_conflicts();
	void setup_def_dominators();
	void setup_reg_prefs();
	bool hasConflict(unsigned r1, unsigned r2);
	void setup_quantities();
	void generateRAATs();

	void initialize_block_available_regs(mcnf_block *block,
			std::vector<char>& available_regs, std::set<unsigned>& unavailable_regs);
	void
			initialize_usable_classes(
					hash_map<unsigned, BitVector>& usable_classes);
	void update_block_usable_classes(mcnf_block* b,
			hash_map<unsigned, BitVector>& usable_classes);

	short const_load_cost(int ac, const BitVector& classes);
	short memory_store_cost(int c);
	short memory_load_cost(int c);
	int cost_block_multiplier(mcnf_block *block);
	int after_control_flow_cost(mcnf_block *b, int movecost);
	int class_bandwidth(RAATInfo *ii, int c);

	mcnf_node* mcnf_node_allocate(mcnf_block *block);
	void mcnf_initialize_node_for_class(mcnf_node *node, int c, std::vector<char>& available_regs);
	void mcnf_node_add_quantity(mcnf_node *node, commod_type q);
	mcnf_edge* mcnf_edge_allocate(mcnf_block *block);
	mcnf_edge* mcnf_edge_find_or_allocate(mcnf_block *block, mcnf_node* src, mcnf_node* dst);
	void mcnf_edge_add_commodity(mcnf_edge *edge, commod_type q);
	void mcnf_initialize_edge_node_commodities(mcnf_node *src, mcnf_node *dest,
			mcnf_edge *edge);
	void mcnf_connect_nodes(mcnf_node *src, mcnf_node *dest, mcnf_edge *edge);

	mcnf_node* create_single_commodity_node(mcnf_block *block,
			mcnf_node_type type, commod_type q, int position, short subindex,
			short aclass);
	void create_nodes_from_bitmap(mcnf_block *block, mcnf_node_type type,
			int position, short subindex, std::vector<char>& available_regs,
			hash_map<unsigned, BitVector>& usable_classes,
			const BitVector& live, mcnf_node **nodes);

	void create_def(mcnf_block *block, commod_type q);
	void create_sink(mcnf_block *block, commod_type q);
	void connect_def(commod_type q, const BitVector& classes,
			mcnf_block *block, mcnf_node **nodes, int class_equiv);
	void connect_sink(commod_type q, const BitVector& classes,
			mcnf_block *block, mcnf_node **nodes);
	void get_exception_landingpads(mcnf_block *block, entry_exit type,
			std::vector<mcnf_block*>& blocks, hash_set<int>& visited);
	void create_entry_nodes(mcnf_block *block,
			std::vector<char>& available_regs,
			std::set<unsigned>& unavailable_regs,
			hash_map<unsigned, BitVector>& usable_classes,
			const BitVector& live, mcnf_node **nodes);
	void create_exit_nodes(mcnf_block *block, std::vector<char>& available_regs,
			std::set<unsigned>& unavailable_regs,
			hash_map<unsigned, BitVector>& usable_classes,
			const BitVector& live, mcnf_node **nodes, int position);
	void create_xbar(mcnf_block *block, RAATInfo *ii, RAATInfo *nextii,
			int position, std::vector<char>& available_regs_top,
			std::vector<char>& available_regs_bottom,
			std::set<unsigned>& unavailable_regs_top,
			std::set<unsigned>& unavailable_regs_bottom,
			hash_map<unsigned, BitVector>& usable_classes, BitVector& livein,
			BitVector& liveout, mcnf_node **top, mcnf_node **bottom,
			int issimple);
	void
			link_straight(mcnf_block *block, mcnf_node **first,
					mcnf_node **second);
	void update_insn_info_before(RAATInfo *ii,
			std::vector<char>& available_regs_top,
			std::vector<char>& available_regs_bottom,
			std::set<unsigned>& unavailable_regs_top,
			std::set<unsigned>& unavailable_regs_bottom,
			hash_map<unsigned, BitVector>& usable_classes, BitVector& top,
			BitVector& bottom);
	void update_insn_info_after(RAATInfo *ii,
			std::vector<char>& available_regs_top,
			std::vector<char>& available_regs_bottom,
			std::set<unsigned>& unavailable_regs_top,
			std::set<unsigned>& unavailable_regs_bottom,
			hash_map<unsigned, BitVector>& usable_classes, BitVector& top,
			BitVector& bottom);
	void create_insn(mcnf_block *block, int insn_number, RAATInfo *ii,
			std::vector<char>& available_regs_top,
			std::vector<char>& available_regs_bottom,
			std::set<unsigned>& unavailable_regs_top,
			std::set<unsigned>& unavailable_regs_bottom,
			hash_map<unsigned, BitVector>& usable_classes, BitVector& top,
			BitVector& bottom, mcnf_node **nodes);
	void link_insn(mcnf_block *block, RAATInfo *ii, mcnf_node **before,
			mcnf_node **insn, mcnf_node **after);
	bool can_ignore_insn(RAATInfo *ii);
	bool no_pressure_between_insns(RAATInfo *ii, RAATInfo *nextii,
			BitVector& live, std::vector<char>& available_regs,
			hash_map<unsigned, BitVector>& usable_classes);
	void mcnf_block_order_nodes(mcnf_block *block);
	void initialize_quantity_next_use(mcnf_block *block);
	void mcnf_link_blocks(void);

	void dump_info(void);
	void dump_live_quantities_map(const BitVector& live);

	//return amoutn solution cost should be adjusted by to reflect
	//impact of weighted tied constraints
	int feas_sol_tied_constraint_adjust();

	/* solver helpers */
	void initialize_path_info();
	void initialize_solver_data();
	void clear_solver_data();
	int sanity_check_feasible_solution();
	double shortest_balanced_paths();

	/* return the sum of the product of each bundle constraint with
	   the corresponding price */
	double price_of_constraints();

	void clear_prices();

	void compute_max_block_cost();
};

/* CPLEX routines */
enum CPLEXUseSolution {IgnoreSolution, UseInitSolution, ConstrainToSolution };

void mcnf_cplex_output(GlobalMCNF& mcnf_cfg, std::string ext);
int mcnf_cplex_solve(GlobalMCNF& mcnf_cfg, bool useInitSolution);
void cplex_output_coalesce_model(GlobalMCNF& mcnf_cfg);
void mcnf_cplex_output_mst(GlobalMCNF& mcnf_cfg);

/* Template functions */

/* Traces the  path back from sink for quantity q.  Calls func on each edge.
 Returns 1 if the trace was successfully executed all the way to a def.
 */
template <class EdgeMarker> bool mcnf_block::trace_path(mcnf_node *sink, int q,
		EdgeMarker& func)
{
	mcnf_node *node;
	mcnf_edge *prev_edge;
	bool keepgoing = true;

	node = sink;

	if (sink->type != nSink)
		abort();

	do
	{
		/* previous node on shortest path */
		int index = node->cindex.get(q);

		prev_edge = node->sinfo[0][index].prev_edge;

		if (prev_edge == NULL)
		{
			if (node->type == nDef)
				return true;
			else
				return false;
		}

		/* call func on this edge */
		keepgoing = func(prev_edge, q);

		node = prev_edge->src;

	} while (keepgoing);

	return keepgoing;
}

//trace along every feasible path for q, calling func
//on each edge of the path
template <class EdgeMarker> void GlobalMCNF::trace_all_feasible_paths(int q,
		EdgeMarker& func)
{
	unsigned i, s, n, ns;
	assert(q != 0);
	quantity& qinfo = quantities[abs(q)];

	if (q> 0)
	{
		for (s = 0, ns = qinfo.sinks.size(); s < ns; s++)
		{
			mcnf_node *sink = qinfo.sinks[s];
			if (!trace_feasible_path(sink, q, func))
				abort();
		}
	}
	else
	{ //antivar
		for (i = 0, n = qinfo.antisinks.size(); i < n; i++)
		{
			for (s = 0, ns = qinfo.antisinks[i].size(); s < ns; s++)
			{
				mcnf_node *sink = qinfo.antisinks[i][s];
				if (!trace_feasible_path(sink, q, func))
					abort();
			}
		}
	}
}

//trace the feasible path from sink calling func on each edge,
//checks for consistency
template <class EdgeMarker> bool GlobalMCNF::trace_feasible_path(
		mcnf_node *sink, int q, EdgeMarker& func)
{
	mcnf_edge *prev_edge;
	bool keepgoing = true;

	mcnf_node *node = sink;

	if (sink->type != nSink)
		abort();

	do
	{
		/* previous node on shortest path */
		int index = node->cindex.get(q);
		if (index < 0)
			abort();
		prev_edge = node->fsinfo[index].prev_edge;

		if (prev_edge == NULL)
		{
			if (node->type == nDef)
				return true;
			else
				return false;
		}

		/* call func on this edge */
		keepgoing = func(prev_edge, q);

		node = prev_edge->src;

	} while (keepgoing);

	return keepgoing;
}
int print_path(mcnf_edge* edge, int q);
typedef bool (*node_compare_fn)(mcnf_node *n1, mcnf_node *n2);
bool node_order_compare(mcnf_node *n1, mcnf_node *n2);
bool reverse_node_order_compare(mcnf_node *n1, mcnf_node *n2);

// Compute all the shortest paths in this block using edge_weight for
// cost information.  Assumes visited has been cleared, ignores bundle
// constraints; if calcbinfo is true, compute forward balanced path info
template <class EdgeWeight, class EdgeCost> double mcnf_block::shortest_paths(
		EdgeWeight& edge_weight, EdgeCost& edge_cost, bool calcbinfo /* = false*/)
{
	int d, dn, s, sn;
	double newcost;
	double prevcost;
	double npaths = 0;
	double total = 0;

	/* put all the defs into the priority queue */
	dn = defs.size();

	std::priority_queue<mcnf_node*, std::vector<mcnf_node*>, node_compare_fn>
			nodeq(reverse_node_order_compare);

	for (d = 0; d < dn; d++)
	{
		int i;
		int nc;
		mcnf_node *def = defs[d];
		nc = def->commodities.size();

		memset(def->sinfo[0], 0, sizeof(shortest_info)*nc);
		if (calcbinfo)
		{
			memset(def->binfo, 0, sizeof(balanced_info)*nc);
			for (i = 0; i < nc; i++)
				def->binfo[i].num_shortest = 1.0;
		}

		def->visited = VISITED_SHORTEST_QUEUE;
		nodeq.push(def);
	}

	/* Now repeatedly iterate over the nodes/edges updating
	 the distances for each commodity. */
	while (!nodeq.empty())
	{
		int j, nsuccs;
		mcnf_node *node = nodeq.top();
		nodeq.pop();
		if (node->visited != VISITED_SHORTEST_QUEUE)
			abort();
		node->visited = VISITED_SHORTEST_DONE;

		/* for each edge */
		nsuccs = node->succs.size();
		for (j = 0; j < nsuccs; j++)
		{
			int k;
			int nc;
			mcnf_edge *edge = node->succs[j];
			mcnf_node *dest = edge->dst;

			/* stop at block boundary */
			if (dest->block_num != index)
				continue;
			if (node->type == nExit && dest->type == nEntry)
				continue; /* catch loops */

			if (dest->visited == VISITED_CLEAR)
			{
				int nc;
				nc = dest->commodities.size();
			    memset(dest->sinfo[0], 0, sizeof(shortest_info)*nc);

				if (calcbinfo)
					memset(dest->binfo, 0, sizeof(balanced_info)*nc);
				dest->visited = VISITED_SHORTEST_SEEN;
				 for(int i = 0; i < nc; i++)
					 dest->sinfo[0][i].value = UNREACHABLE;
			}
			else if (dest->visited == VISITED_SHORTEST_DONE)
				abort(); /* nodes not being processed topologically! */

			/* consider every commodity over this edge */
			nc = edge->commodities.size();
			for (k = 0; k < nc; k++)
			{
				/* commodity */
				int c = edge->commodities[k];
				/* index in nodes */
				int sindex = node->cindex.get(c);
				int dindex = dest->cindex.get(c);

				if (sindex < 0 || dindex < 0)
					continue;

				newcost = node->sinfo[0][sindex].value + edge_weight(edge, c);

				if (newcost == UNREACHABLE)
					continue;

				prevcost = dest->sinfo[0][dindex].value;

				if (calcbinfo)
				{
					npaths = node->binfo[sindex].num_shortest;

					if (!finite(npaths))
						abort();
				}

				if (newcost < prevcost || (newcost == prevcost
						&& node->allocClass == dest->allocClass))
				{
					/* update balanced path info */
					if (calcbinfo)
					{
						int predindex = dest->pred_index(edge);
						if (dest->preds.size() < 32)
							dest->binfo[dindex].prevs = 1<<predindex;
						else
							dest->binfo[dindex].prevs = 0; /* have to use prev_edge */
						dest->binfo[dindex].num_shortest = npaths;
					}

					/* update path to this node */
					dest->sinfo[0][dindex].prev_edge = edge;
					dest->sinfo[0][dindex].value = newcost;
					dest->sinfo[0][dindex].cost = node->sinfo[0][sindex].cost
							+ edge_cost(edge, c);

					/* if it's not already in the queue, add it */
					if (dest->visited != VISITED_SHORTEST_QUEUE)
					{
						if (dest->visited == VISITED_SHORTEST_DONE)
							abort();
						dest->visited = VISITED_SHORTEST_QUEUE;
						nodeq.push(dest);
					}
				}
				else if (newcost == prevcost && calcbinfo)
				{
					/* just update balanced path info */
					double newnpaths;
					int predindex = dest->pred_index(edge);

					if (predindex >= 32)
						abort(); /* this should only happen on def'ing node.. */

					/* very rarely we may have more paths than fit in a double,
					 in this case just ignore additional paths */
					newnpaths = dest->binfo[dindex].num_shortest + npaths;
					if (finite(newnpaths))
					{
						dest->binfo[dindex].prevs |= 1<<predindex;
						dest->binfo[dindex].num_shortest += npaths;
					}
				}
			}
		}
	}

	/* now go over all the sinks and sum up the path values */
	sn = sinks.size();

	if (DebugPathsPRA)
	{
		std::cerr << "Block " << index << "\n";
	}
	for (s = 0; s < sn; s++)
	{
		int nc, c;
		double pcost;
		mcnf_node *node = sinks[s];

		nc = node->commodities.size();

		if (nc != 1)
			abort();

		if (node->visited == VISITED_CLEAR || !finite(node->sinfo[0][0].value))
		{
			debug_print_quantity_shortest(node->position);
			abort();
		}

		pcost = node->sinfo[0][0].value;
		total += pcost;

		c = node->commodities[0];

		tally_path_cost tally;
		trace_path(node, c, tally);
		int cost = tally.getCost();

		if (DebugPathsPRA)
		{
			std::cerr << c << ": " << pcost << "(" << cost << ")\n";
			std::cerr << "\t";
			trace_path(node, c, print_path);
			std::cerr << "\n";
		}

		if (c> 0)
		{
			cfg->quantities[c].priced_cost += pcost;
			cfg->quantities[c].unpriced_cost += cost;
			quantity_prices[c] += pcost;
		}
		else
		{
			c = -c;
			cfg->quantities[c].anti_priced_cost += pcost;
			cfg->quantities[c].anti_unpriced_cost += cost;
			quantity_prices[c] += pcost;
		}
	}

	return total;
}

extern void reinitialize_node_for_fsf_trace(mcnf_node *node);
template <class EdgeWeight, class EdgeCost, class TieBreaker> void mcnf_block::shortest_paths_backwards(
		EdgeWeight& edge_weight, EdgeCost& edge_cost,
		TieBreaker& this_edge_better_if_tied2)
{
	int s, sn;
	double newcost;
	double prevcost;

	if (MAX_K < 2)
		abort();
	/* put all the sinks into the priority queue */
	sn = sinks.size();
	std::priority_queue<mcnf_node*, std::vector<mcnf_node*>,
	node_compare_fn> nodeq(node_order_compare);

	for (s = 0; s < sn; s++)
	{
		int nc;
		mcnf_node *node = sinks[s];

		nc = node->commodities.size();

		memset(node->sinfo[1], 0, sizeof(shortest_info)*nc);
		node->visited = VISITED_SHORTEST_BACK_QUEUE;
		nodeq.push(node);
	}

	/* Now repeatedly iterate over the nodes/edges updating
	 the distances for each commodity. */
	while (!nodeq.empty())
	{
		int j, npreds;
		mcnf_node *node = nodeq.top();
		nodeq.pop();

		if (node->visited != VISITED_SHORTEST_BACK_QUEUE)
			abort();
		node->visited = VISITED_SHORTEST_BACK_DONE;

		/* for each edge */
		npreds = node->preds.size();
		for (j = 0; j < npreds; j++)
		{
			int k;
			int nc;
			mcnf_edge *edge = node->preds[j];
			mcnf_node *src = edge->src;

			/* stop at block boundary */
			if (src->block_num != index)
				continue;
			if (node->type == nEntry && src->type == nExit)
				continue; /* catch loops */

			if (src->visited < VISITED_SHORTEST_BACK_SEEN)
			{
				int nc;
				nc = src->commodities.size();

				memset(src->sinfo[1], 0, sizeof(shortest_info)*nc);
				for(int i = 0; i < nc; i++)
					src->sinfo[1][i].value = UNREACHABLE;
				src->visited = VISITED_SHORTEST_BACK_SEEN;
			}
			else if (src->visited == VISITED_SHORTEST_BACK_DONE)
				abort(); /* nodes not being processed topologically! */

			/* consider every commodity over this edge */
			nc = edge->commodities.size();
			for (k = 0; k < nc; k++)
			{
				/* commodity */
				int c = edge->commodities[k];
				/* index in nodes */
				int sindex = node->cindex.get(c);
				int dindex = src->cindex.get(c);

				if (sindex < 0 || dindex < 0)
					continue;

				newcost = node->sinfo[1][sindex].value + edge_weight(edge, c);

				if (newcost == UNREACHABLE)
					continue;

				prevcost = src->sinfo[1][dindex].value;

				if (newcost < prevcost || (newcost == prevcost
						&& this_edge_better_if_tied2(edge,
								src->sinfo[1][dindex].prev_edge, c)))
				{
					/* update path to this node */
					src->sinfo[1][dindex].prev_edge = edge;
					src->sinfo[1][dindex].value = newcost;
					src->sinfo[1][dindex].cost = node->sinfo[1][sindex].cost
							+ edge_cost(edge, c);

					/* if it's not already in the queue, add it */
					if (src->visited != VISITED_SHORTEST_BACK_QUEUE)
					{
						src->visited = VISITED_SHORTEST_BACK_QUEUE;
						nodeq.push(src);
					}
				}

			}
		}
	}
	//sanity check, make sure this is valid path to every def
	for (int d = 0, dn = defs.size(); d < dn; d++)
	{
		int nc;
		mcnf_node *def = defs[d];
		nc = def->commodities.size();
		if (nc != 1)
			abort();

		if(DebugPRA > 5)
		{
			std::cerr << "BackPathPrice " << def->position << ": " <<def->sinfo[1][0].value << "\n";
		}
		if (def->visited == VISITED_CLEAR || !finite(def->sinfo[1][0].value))
		{
			static bool hitonce = false;
			if(!hitonce)
			{
				hitonce = true;
				std::cerr << "backwards path blocked for " << def->position << "\n";
				traverse_nodes(reinitialize_node_for_fsf_trace);
				shortest_paths_backwards(edge_weight, edge_cost, this_edge_better_if_tied2);
			}
			else
				abort();
		}
	}

}

template<class NodeProcessor> void mcnf_block::traverse_nodes(
		NodeProcessor& process)
{
	int n;
	int nn = allnodes.size();
	for (n = 0; n < nn; n++)
	{
		mcnf_node *node = allnodes[n];
		process(node);
	}
}

/* Traverse the whole mcnf cfg and apply process to
 each node */
template<class NodeProcessor> void GlobalMCNF::traverse_nodes(
		NodeProcessor& process)
{
	int b, nb;

	nb = blocks.size();

	for (b = 0; b < nb; b++)
	{
		mcnf_block *block = blocks[b];
		block->traverse_nodes(process);
	}
}

}
#endif /*GLOBALMCNF_H_*/
