#ifndef RAATINFO_H_
#define RAATINFO_H_

#include <vector>
#include "llvm/PrincipledBackend/ProgCommon.h"
#include "llvm/PrincipledBackend/AllocationClassInfo.h"
#include "llvm/PrincipledBackend/AllocClassMap.h"
#include "llvm/ADT/BitVector.h"
#include "llvm/ADT/hash_map"
#include "llvm/CodeGen/ValueTypes.h"
#include "llvm/Target/MRegisterInfo.h"
#include "llvm/Support/Compiler.h"

namespace llvm
{

/* Information on a single register-allocation aware tile.  This implementation
 * is highly biased towards assuming a single MachineInsruction equals a tile.
 * It may get completely rewritten when we rewrite the instruction selector.
 *
 * Store information about the usage of variables and the constraints placed
 * upon them.  Support an expressive constraint and cost framework:
 * -class capacity constraints
 * -individual per op class costs
 * -tied constraints (op1 alloc == op2 alloc, op1 lastuse, op2 def)
 * -weighted tied constraint (op1 alloc != op2 alloc, additional cost)
 * NOT YET IMPLEMENTED:
 * -conditional constrait (if op1 alloc == X then {} else {} )
 * --these can nest (they contain additional RAATInfos)
 */
class RAATInfo
{
public:
	// information about a variable/hardreg used by this RAAT
	struct OperandInfo
	{
		unsigned regNum; //hard reg or virtual
		bool isHard; //is a hard register, should be redundant w/regnum
		bool isUse;
		bool isDef;
		bool isLastUse;
		bool isExtend; //lives past insn, then dead, e.g., unused reg
		bool isClobber; //available before and after, but not during, hard reg only
		bool isImplicit; //not directly addressed

		int equivalentClass; //this defined op is equivalent to mem/const

		BitVector allowableClasses; //allocation classes for this var

		OperandInfo() :
			regNum(0), isHard(false), isUse(false), isDef(false),
					isLastUse(false), isExtend(false), isClobber(false), isImplicit(true),
					equivalentClass(-1)
		{
		}
	};

	class TiedConstraint
	{
		unsigned inReg; //live in virtual reg
		unsigned outReg; //live out virtual reg
		int costModifier; //change in cost if equal (probably negative)
		BitVector allowableClasses; //constraint only applies to these constraints
		bool unmergeable; //if true, must be a weighted constraint, not hard

public:
		TiedConstraint(unsigned in, unsigned out, int cost, const BitVector& ac, bool um) :
			inReg(in), outReg(out), costModifier(cost), allowableClasses(ac), unmergeable(um)
		{
		}

		unsigned getInReg() const
		{
			return inReg;
		}
		unsigned getOutReg() const
		{
			return outReg;
		}
		int getCost() const
		{
			return costModifier;
		}

		bool goodAllocClass(int c) const
		{
			return allowableClasses.test(c);
		}

		bool isMergeable() const
		{
			return !unmergeable;
		}
	};

	RAATInfo(): reg2index(1)
	{
		assert(0 && "Must initialize RAAT with insn for now");
	}
	RAATInfo(MachineInstr *MI): reg2index(1)
	{
		initializeFromInsn(MI);
	}

	virtual ~RAATInfo()
	{
	}

	//return an iterator pointing to the begining of this RAAT
	MachineInstr * begin()
	{
		return insn;
	}

	bool hasSavedReg()
	{
		return insn->hasSavedReg();
	}

	unsigned getNumDefs()
	{
		int ret = 0;
		for(unsigned i = 0, n = ops.size(); i < n; i++)
		{
			if(ops[i].isDef)
				ret++;
		}
		for(unsigned i = 0, n = hardops.size(); i < n; i++)
		{
			if(hardops[i].isDef)
				ret++;
		}
		return ret;
	}

	unsigned getNumOps()
	{
		return ops.size();
	}

	unsigned getNumHardOps()
	{
		return hardops.size();
	}

	unsigned getNumTiedConstraints()
	{
		return tiedConstraints.size();
	}

	const OperandInfo& getOpFromIndex(unsigned i)
	{
		return ops[i];
	}

	const OperandInfo& getHardOpFromIndex(unsigned i)
	{
		return hardops[i];
	}

	//return index for a register (-1 if not present)
	const unsigned getOpIndex(unsigned r)
	{
		if (reg2index.count(r) > 0)
			return reg2index[r];
		else
			return -1;
	}

	const TiedConstraint& getTiedConstraint(unsigned i)
	{
		return tiedConstraints[i];
	}

	//remove tied constraint at index i
	void removeTiedConstraint(unsigned i)
	{
		tiedConstraints.erase(tiedConstraints.begin()+i);
	}

	int getClassCapacity(unsigned i)
	{
		if (classCapacityConstraints.count(i) > 0)
			return classCapacityConstraints[i];
		else
			return -1;
	}

	//return additional cost of allocating the op at opindex to class c
	virtual int getClassCost(unsigned opindex, unsigned c)
	{
		assert(c < opClassCosts[opindex].size());
		return opClassCosts[opindex][c];
	}

	//return register preference of op at opindex (presumably cheaper)
	//-1 if none
	virtual int getClassPref(unsigned opindex)
	{
		if(classPref.count(opindex))
		{
			return classPref[opindex];
		}
		else
		{
			return -1;
		}
	}

	virtual void update_regs(AllocClassMap& map); //modify insn to use mappings

	int getCost() const
	{
		return baseCost;
	}
protected:
	int baseCost; //cost of RAAT regardless of allocation
	std::vector<OperandInfo> ops; //virtual reg operands
	std::vector<OperandInfo> hardops; //physical registers
	hash_map<unsigned, unsigned> reg2index; //map from virt reg to op index

	hash_map<unsigned, int> classCapacityConstraints;
	//indexed by op, a map of class costs, cost zero if absent
	std::vector<std::vector<short> > opClassCosts;
	//indexed by opindex, class preference if present
	hash_map<unsigned, int> classPref;
	std::vector<TiedConstraint> tiedConstraints;

	MachineInstr *insn;

	virtual void initializeFromInsn(MachineInstr *MI);
	virtual int computeBaseCost(MachineInstr *MI);

	/*** PROGRESSIVE ALLOCATOR FUNCTIONALITY ***/
	/* All the below is for use by the allocator; we may want to factor
	 * it out at some point */
public:
	// initialize allocation specifiec structures
	void initializeForAllocator();
	//set the quantity corresponding to op in this insn
	void setOpQuantity(unsigned op, int q)
	{
		if(op >= opQuantities.size())
			opQuantities.resize(op+1,0);
		assert(opQuantities[op] == 0 || opQuantities[op] == q);
		opQuantities[op] = q;
	}

	int getOpQuantity(unsigned op)
	{
		assert(op < opQuantities.size());
		return opQuantities[op];
	}

	bool isControlFlow();

	struct QuantityInfo
	{
		//merges op info for a quantity
		int q;
		std::vector<unsigned> opindex;
		BitVector classes;
		int equivalentClass;
		bool use;
		bool def;
		bool lastuse;
		bool extend;

		bool livein;
		bool liveout;

		QuantityInfo() :
			q(-1), equivalentClass(-1), use(false), def(false), lastuse(false),
					extend(false), livein(false), liveout(false)
		{
		}
	};

	std::vector<QuantityInfo>& getQuantityInfo()
	{
		if(info.size() == 0)
			analyze_RAAT_quantities();
		return info;
	}
	/* keep track of how much of each class
	   is free in and out of an insn both in
	   terms of the  unallocated network and
	   a partially allocated network */
	struct Freeness
	{
	    int free_in:8;
	    int free_out:8;
	    int curr_free_in:8;
	    int curr_free_out:8;

	    Freeness(): free_in(0), free_out(0), curr_free_in(0), curr_free_out(0) {}
	};

	Freeness& getClassFreeness(int c) { return frees[c]; }
	/* Return the maximum number of memory ops allowed in an
	   instruction. This should be target specific. */
	virtual int total_memory_bandwidth() { return 1; }

	void dump();
	void print(std::ostream& out);

protected:
	std::vector<int> opQuantities;
	std::vector<QuantityInfo> info;
	std::vector<Freeness> frees;
	void analyze_RAAT_quantities();
};

}

#endif /*RAATINFO_H_*/
