#ifndef SOLUTION_H_
#define SOLUTION_H_

/* Implementation of the Solution class, which stores the tranformations
 * dictated by an MCNF solution */

#include "llvm/ADT/hash_map"
#include "llvm/ADT/hash_set"
#include "llvm/PrincipledBackend/AllocationClassInfo.h"
#include "llvm/PrincipledBackend/AllocClassMap.h"
#include "llvm/PrincipledBackend/ProgCommon.h"
#include <vector>

namespace llvm
{

/* type of transformation, a Store is when the antivar
 is evicted; there  are also corresponding Moves for the var */
enum trans_type
{	Error, Birth, Death, Move, Store};

struct transformation
{
	enum trans_type type;
	int position; /* insn number, happens after insn */
	int level; /* additional sorting criterion */
	int order; //for sorting non model moves
	int cost; /* for debugging */
	index_type src_class;
	unsigned srcr; //actual reg nums
	index_type dst_class;
	unsigned dstr;
	commod_type q;

	transformation() :
		type(Error), position(0), level(0), cost(0), src_class(0),
				dst_class(0), q(0)
	{
	}

	transformation(trans_type t, int p, int l, int c, index_type sc,
			unsigned sr, index_type dc, unsigned dr, commod_type quant) :
		type(t), position(p), level(l), order(0), cost(c), src_class(sc), srcr(
				sr), dst_class(dc), dstr(dr), q(quant)
	{
		if (sc >= 0 && alloc_classes->getType(sc)
				== AllocationClassInfo::Register
				&& alloc_classes->getRegisterIndex(sr) != sc)
			abort();
		if (dc >= 0 && alloc_classes->getType(dc)
				== AllocationClassInfo::Register
				&& alloc_classes->getRegisterIndex(dr) != dc)
			abort();
	}
	transformation(trans_type t, int p, int l, int o, int c, index_type sc,
			unsigned sr, index_type dc, unsigned dr, commod_type quant) :
		type(t), position(p), level(l), order(o), cost(c), src_class(sc), srcr(
				sr), dst_class(dc), dstr(dr), q(quant)
	{
		if (sc >= 0 && alloc_classes->getType(sc)
				== AllocationClassInfo::Register
				&& alloc_classes->getRegisterIndex(sr) != sc)
			abort();
		if (dc >= 0 && alloc_classes->getType(dc)
				== AllocationClassInfo::Register
				&& alloc_classes->getRegisterIndex(dr) != dc)
			abort();
		if(sr == dr && sr != 0)
			abort();
	}

	bool operator<(const transformation& r) const;
	bool operator==(const transformation& r) const;

	//apply the transformation prior to RI, updating map
	void apply(RAATInfo& RI, AllocClassMap& map);
	//apply the transformation at the end of block, updating map
	void applyEnd(mcnf_block *block, AllocClassMap& map);

	void print(std::ostream& out);
private:
	int compare(const transformation& r) const;
	void apply(MachineBasicBlock *MBB, MachineBasicBlock::iterator MI, AllocClassMap& map);

};

struct block_solution
{
	mcnf_block *block;
	hash_map<commod_type, index_type> q_in_alloc; /* class allocation of quantities at */
	hash_map<commod_type, index_type> q_out_alloc; /* block boundaries (zero if unalloced) */

	//live range mappings
	hash_map<int, RegMap> quantityRegMap; //quantity, position, reg

	std::vector<transformation> transforms;

	block_solution() :
		block(NULL)
	{
	}

	void record_solution(std::vector<transformation>& shufflecode);

	void setup_block_in(AllocClassMap& map);
	void print_solution(std::ostream& out);

	unsigned getRegisterForQuantityIn(int q, int c); //register into block
	unsigned getRegisterForQuantity(int q, mcnf_node *node, int offset=0);

};

struct Solution
{
	std::vector<block_solution> block_sols;

	Solution() {}
	//record the solution present in the GMCNF model
	//overwrites any existing solution
	void record_solution(GlobalMCNF& mcnf_cfg,  std::vector< std::vector<transformation> >& shufflecode);
	void record_solution(GlobalMCNF& mcnf_cfg);
	void print_solution(std::ostream& out);
	void clear() { block_sols.clear(); }
};

}
#endif /*SOLUTION_H_*/
