#include "llvm/Pass.h"
#include "llvm/Function.h"
#include "llvm/Module.h"
#include "llvm/BasicBlock.h"
#include "llvm/Instructions.h"
#include "llvm/Type.h"
#include "llvm/ADT/hash_map"
#include "llvm/ADT/hash_set"
#include "llvm/Support/CFG.h"

#include <vector>
#include <deque>
#include <set>
#include <algorithm>
#include <ostream>
#include <fstream>
#include <iostream>

using namespace llvm;

namespace
{

class LivenessHistogram : public ModulePass
{
	std::vector<unsigned> counts;

	//in the counts histogram, increment category c
	void increment(unsigned c)
	{
		if (counts.size() <= c)
			counts.resize(c+1, 0);
		counts[c]++;
	}

	//return the number of instructions in F
	unsigned countInsns(Function &F)
	{
		unsigned ret = 0;
		for (Function::iterator i = F.begin(), e = F.end(); i != e; ++i)
			ret += i->size();
		return ret;
	}
	
	//output the histogram to a file
	void printHistogram(Module& M)
	{
		std::cerr << "Printing histogram\n";
		std::string name = M.getModuleIdentifier() + ".lhist";
	    std::ofstream file(name.c_str());
	    for(unsigned i = 0, n = counts.size(); i < n; i++)
	    {
	    	file << i << ": " << counts[i] << "\n";
	    }
	}
	
public:
	static char ID;

	LivenessHistogram() :
		ModulePass((intptr_t)&ID)
	{
	}

	~LivenessHistogram()
	{
	}

	virtual bool runOnFunction(Function &F)
	{
		hash_map<BasicBlock*, std::set<Instruction*> > block_kill;
		hash_map<BasicBlock*, std::set<Instruction*> > block_gen;
		hash_map<BasicBlock*, std::set<Instruction*> > block_livein;
		hash_map<BasicBlock*, std::set<Instruction*> > block_liveout;
		hash_map<BasicBlock*, std::set<Instruction*> > block_ssa_liveout;

		//initialize the gen and kill sets of each block as well
		//as the ssa liveout
		for (Function::iterator BB = F.begin(), E = F.end(); BB != E; ++BB)
		{
			std::set<Instruction*>& kill = block_kill[BB];
			std::set<Instruction*>& gen = block_gen[BB];

			for (BasicBlock::InstListType::reverse_iterator I =
					BB->getInstList().rbegin(), EI = BB->getInstList().rend(); I != EI; I++)
			{
				Instruction *insn = &*I;
				//each instruction is a definer, and so will kill any uses
				kill.insert(insn);
				gen.erase(insn);
				if (PHINode *PN = dyn_cast<PHINode>(insn))
				{
					//phi nodes are special, the def is killed, but the uses
					//are only live out of their corresponding blocks

					for (unsigned i = 0, e = PN->getNumIncomingValues(); i != e; ++i)
					{
						if (Instruction *Op
								= dyn_cast<Instruction>(PN->getIncomingValue(i)))
						{
							BasicBlock *PredBB = PN->getIncomingBlock(i);
							block_ssa_liveout[PredBB].insert(Op);
						}
					}
				}
				else
				{
					//every insn op is a use an is now live
					for (User::op_iterator OI = insn->op_begin(), OE =
							insn->op_end(); OI != OE; ++OI)
					{
						if (Instruction *Used = dyn_cast<Instruction>(*OI))
						{
							gen.insert(Used);
							kill.erase(Used); //this actually shouldn't be necessary for SSA
						}

					}
				}
			}

			block_livein[BB] = gen;
		}

		//now compute the livein and liveout sets
		std::deque<BasicBlock*> worklist;
		hash_set<BasicBlock*> onlist;

		for (Function::iterator BB = F.begin(), E = F.end(); BB != E; ++BB)
		{
			BasicBlock *bb = BB;
			worklist.push_front(bb);
			onlist.insert(bb);
		}

		while (worklist.size() > 0)
		{
			BasicBlock *bb = worklist.front();
			worklist.pop_front();
			onlist.erase(bb);

			std::set<Instruction*> newliveout, newlivein;
			//compute the union of all successors blocks
			for (succ_iterator SI = succ_begin(bb), E = succ_end(bb); SI != E; ++SI)
			{
				BasicBlock *Succ = *SI;
				newliveout.insert(block_livein[Succ].begin(),
						block_livein[Succ].end());
			}
			//include the ssa liveout values
			newliveout.insert(block_ssa_liveout[bb].begin(),
					block_ssa_liveout[bb].end());

			//remove the insns killed by this block
			std::set_difference(newliveout.begin(), newliveout.end(),
					block_kill[bb].begin(), block_kill[bb].end(), inserter(
							newlivein, newlivein.begin()));
			//add the insns generated
			newlivein.insert(block_gen[bb].begin(), block_gen[bb].end());

			block_liveout[bb] = newliveout;
			if (newlivein != block_livein[bb])
			{
				block_livein[bb] = newlivein;
				//add predecessors to the worklist if they aren't in it already
				for (pred_iterator PI = pred_begin(bb), E = pred_end(bb); PI
						!= E; ++PI)
				{
					BasicBlock *Pred = *PI;
					if (onlist.count(Pred) == 0)
					{
						onlist.insert(Pred);
						worklist.push_back(Pred);
					}
				}
			}
		}

		//next compute the actual histogram information
		for (Function::iterator BB = F.begin(), E = F.end(); BB != E; ++BB)
		{
			BasicBlock *bb = BB;
			std::set<Instruction*> live = block_liveout[bb];
			increment(live.size());
			for (BasicBlock::iterator I = bb->end(); I != bb->begin();)
			{
				--I;
				Instruction *insn = I;

				//update liveness info
				live.erase(insn);

				if (!isa<PHINode>(insn))
				{
					//every insn op is a use an is now live
					for (User::op_iterator OI = insn->op_begin(), OE =
							insn->op_end(); OI != OE; ++OI)
					{
						if (Instruction *Used = dyn_cast<Instruction>(*OI))
						{
							live.insert(Used);
						}
					}
					increment(live.size());
				}
			}
		}
		return false;
	}

	virtual bool runOnModule(Module& M)
	{
		for (Module::iterator MI = M.begin(), ME = M.end(); MI != ME; ++MI)
		{
			runOnFunction(*MI);
		}
		printHistogram(M);
		return false;
	}
};

char LivenessHistogram::ID = 0;
RegisterPass<LivenessHistogram> X("liveness-hist", "Liveness Histogram");
}
