/*
 * ComputeBlockFrequencies.h
 *
 *  Created on: Jul 9, 2008
 *      Author: dkoes
 */


#include "llvm/ADT/hash_map"
#include "llvm/Pass.h"
#include "llvm/BasicBlock.h"
#include "llvm/Function.h"
#include "llvm/Analysis/ProfileInfo.h"
#include "llvm/Analysis/LoopInfo.h"
#include "llvm/Support/Debug.h"
#include "llvm/Support/Compiler.h"
#include "llvm/Support/CommandLine.h"

namespace llvm
{

cl::opt<bool> UseProfile("use-profile", cl::desc(
		"Use profile information to calculate block frequencies"));

class ComputeBlockFrequencies: public FunctionPass
{
	LoopInfo* LI;
	ProfileInfo *PI;
public:

	static char ID;
	ComputeBlockFrequencies() :
		FunctionPass((intptr_t) & ID), LI(NULL), PI(NULL)
	{
	}

	virtual ~ComputeBlockFrequencies()
	{
	}

	virtual const char* getPassName() const
	{
		return "Compute Block Frequencies";
	}

	virtual void getAnalysisUsage(AnalysisUsage &AU) const
	{
		AU.addRequired<LoopInfo>();
		if (UseProfile)
			AU.addRequired<ProfileInfo> ();
		FunctionPass::getAnalysisUsage(AU);
	}

	/// runOnFunction - annotate basic blocks with frequency info
	bool runOnFunction(Function& F)
	{
		LI = &getAnalysis<LoopInfo>();
		if(UseProfile)
		{
			PI = &getAnalysis<ProfileInfo>();
			if(!PI->isValid())
				PI = NULL;
		}

		if (PI != NULL) //use profile info
		{
			for (Function::iterator BB = F.begin(), E = F.end(); BB != E; ++BB)
			{
				BasicBlock *block = BB;
				unsigned cnt = PI->getExecutionCount(block);
				BB->setFrequency(cnt);
			//	cerr << BB->getName() << "\t" << cnt << "\n";
			}
		}
		else
		{
			//very simple static estimation
			for (Function::iterator BB = F.begin(), E = F.end(); BB != E; ++BB)
			{
				unsigned depth = LI->getLoopDepth(BB);
				int mult = 1;
				while (depth > 0 && mult < 10000)
				{
					mult *= 10;
					depth--;
				}
				BB->setFrequency(mult);
			}
		}

		return false;
	}

};

char ComputeBlockFrequencies::ID = 0;
RegisterPass<ComputeBlockFrequencies> X("compute-block-freqs",
                          "Basic block frequency computation");

Pass *createComputeBlockFrequenciesPass() { return new ComputeBlockFrequencies(); }

}

