//===- ProfileInfoLoaderPass.cpp - LLVM Pass to load profile info ---------===//
//
//                     The LLVM Compiler Infrastructure
//
// This file was developed by the LLVM research group and is distributed under
// the University of Illinois Open Source License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
//
// This file implements a concrete implementation of profiling information that
// loads the information from a profile dump file.
//
//===----------------------------------------------------------------------===//

#define DEBUG_TYPE "profileloader"

#include "llvm/Function.h"
#include "llvm/BasicBlock.h"
#include "llvm/InstrTypes.h"
#include "llvm/Pass.h"
#include "llvm/Analysis/Passes.h"
#include "llvm/Analysis/ProfileInfo.h"
#include "llvm/Analysis/ProfileInfoLoader.h"
#include "llvm/Support/CommandLine.h"
#include "llvm/Support/Compiler.h"
#include "llvm/Support/Debug.h"

#include "llvm/Support/Streams.h"
#include <iostream>
using namespace llvm;

namespace {
  cl::opt<std::string>
  ProfileInfoFilename("profile-info-file", cl::init("llvmprof.out"),
                      cl::value_desc("filename"),
                      cl::desc("Profile file loaded by -profile-loader"));

  class VISIBILITY_HIDDEN LoaderPass : public FunctionPass, public ProfileInfo {
    std::string Filename;
    bool haveLoaded;
    hash_map<std::string, unsigned> BlockCounts;
  public:
    static char ID; // Class identification, replacement for typeinfo
    explicit LoaderPass(const std::string &filename = "")
      : FunctionPass((intptr_t)&ID), Filename(filename), haveLoaded(false) {
      if (filename.empty()) Filename = ProfileInfoFilename;
    }

    virtual void getAnalysisUsage(AnalysisUsage &AU) const {
      AU.setPreservesAll();
    }

    virtual const char *getPassName() const {
      return "Profiling information loader";
    }

    virtual bool runOnFunction(Function &F)
	{
		if (!haveLoaded)
		{
			runOnModule(*F.getParent());
			haveLoaded = true;
		}

		return false;
	}

    /// run - Load the profile information from the specified file.
    virtual bool runOnModule(Module &M);

    virtual bool isValid() { return BlockCounts.size() != 0; }

    virtual unsigned getExecutionCount(const BasicBlock *BB)
    {
  	  std::string fname = BB->getParent()->getName();
  	  std::string bname = BB->getName();
  	  std::string name = fname+bname;
    	if(BlockCounts.count(name))
    		return BlockCounts[name];
    	else
    		return 0;
    }

  };

  char LoaderPass::ID = 0;
  RegisterPass<LoaderPass>
  X("profile-loader", "Load profile information from llvmprof.out");

  RegisterAnalysisGroup<ProfileInfo, true> Y(X);
}  // End of anonymous namespace

Pass *llvm::createProfileLoaderPass() { return new LoaderPass(); }


bool LoaderPass::runOnModule(Module &M) {

	//be silent if file doesn't exist
	FILE *F = fopen(Filename.c_str(), "r");
	if(F == NULL)
		return false;
	else
		fclose(F);
  ProfileInfoLoader PIL("profile-loader", Filename, M);
  EdgeCounts.clear();

  std::vector<std::pair<BasicBlock*, unsigned> > BCs;
  PIL.getBlockCounts(BCs);

  for (unsigned i = 0, e = BCs.size(); i != e; ++i)
  {
	  BasicBlock *BB = BCs[i].first;
	  std::string fname = BB->getParent()->getName();
	  std::string bname = BB->getName();
	  BlockCounts[fname+bname] = BCs[i].second;

	  DOUT << fname << " " << bname << "\t" << BCs[i].second << "\n";
  }

  return false;
}
