#include "llvm/CodeGen/DataFlowLiveVariables.h"

#include <iostream>

using namespace llvm;

namespace llvm
{
char DataFlowLiveVariables::ID = 0;
static RegisterPass<DataFlowLiveVariables> X("dataflow-livevars",
		"Dataflow Live Variables");

//set reg in bitmap, properly handles subregs of hard registers
void DataFlowLiveVariables::setReg(BitVector& bits, unsigned r)
{
	bits.set(r);

	//set all aliasing regs as well
	if(MRegisterInfo::isPhysicalRegister(r))
	{
		const MRegisterInfo *RegInfo = MF->getTarget().getRegisterInfo();
		const TargetRegisterDesc& TD = RegInfo->get(r);
		if(TD.AliasSet)
		{
			const unsigned *subs = TD.AliasSet;
			while(*subs)
			{
				bits.set(*subs);
				subs++;
			}
		}
	}
}

//unset reg in bitmap
void DataFlowLiveVariables::unsetReg(BitVector& bits, unsigned r)
{
	bits.reset(r);
	//unset all sub regs as well
	if (MRegisterInfo::isPhysicalRegister(r))
	{
		const MRegisterInfo *RegInfo = MF->getTarget().getRegisterInfo();
		const TargetRegisterDesc& TD = RegInfo->get(r);
		if (TD.AliasSet)
		{
			const unsigned *subs = TD.AliasSet;
			while (*subs)
			{
				bits.reset(*subs);
				subs++;
			}
		}
	}
}

/* Compute gen and kill sets for a basic block */
void DataFlowLiveVariables::compute_block_gen_kill(
		const MachineBasicBlock* MBB, BitVector& gen, BitVector& killc)
{
	MachineBasicBlock::const_iterator I = MBB->end();
	//clear sets
	int size = gen.size();
	gen.clear();
	gen.resize(size);
	killc.clear();
	killc.resize(size);

	while (I != MBB->begin())
	{
		I--;
		const MachineInstr *MI = I;
		for (unsigned op = 0, n = MI->getNumOperands(); op < n; ++op)
		{
			//register operand?
			const MachineOperand &MO = MI->getOperand(op);
			unsigned reg;
			if (MO.isReg() && (reg = MO.getReg()))
			{
				if (MO.isUse())
					setReg(gen, reg);
				if (MO.isDef())
				{
					unsetReg(gen, reg);
					setReg(killc, reg);
				}
			}
		}
	}

	temp_liveins[MBB].clear();
	//remove livein regs since they should not be propagated
	for (MachineBasicBlock::const_livein_iterator LI = MBB->livein_begin(), LE =
			MBB->livein_end(); LI != LE; ++LI)
	{
		setReg(killc, *LI);
		//keep track of which "livein" (ie, exception) registers are actually
		//used and live (since sometimes they are ignored)
		if(gen.test(*LI))
			temp_liveins[MBB].insert(*LI);
		unsetReg(gen, *LI);
	}

	killc.flip();

}

/* changedblocks may have different gen/kill sets, so rerun the iterative
 * data flow using these blocks.  Empties worklist. */
void DataFlowLiveVariables::compute_dataflow(
		std::deque<const MachineBasicBlock*>& worklist)
{
	hash_set<const MachineBasicBlock*> inQ(worklist.begin(), worklist.end());

	while (!worklist.empty())
	{
		const MachineBasicBlock *block = worklist.front();
		worklist.pop_front();
		inQ.erase(block);
		BitVector& livein = liveins[block];
		BitVector& liveout = liveouts[block];
		unsigned livein_cnt = livein.count();
		//compute liveout from union of successors live in
		for (MachineBasicBlock::const_succ_iterator SI = block->succ_begin(),
				E = block->succ_end(); SI != E; ++SI)
		{
			const MachineBasicBlock *succ = *SI;
			liveout |= liveins[succ];
		}

		//compute livein using block gen and kill
		livein = (liveout & block_killc[block]) | block_gen[block];
		if (livein.count() != livein_cnt)
		{
			//changed, need to make sure predecessors are updated
			for (MachineBasicBlock::const_pred_iterator PI =
					block->pred_begin(), E = block->pred_end(); PI != E; ++PI)
			{
				const MachineBasicBlock *pred = *PI;
				if (inQ.count(pred) == 0)
				{
					inQ.insert(pred);
					worklist.push_back(pred);
				}
			}
		}
	}

}

/* Recompute isKill and isDead info on machine operands */
void DataFlowLiveVariables::update_liveness_info()
{
	for (MachineFunction::iterator MBB = MF->begin(), MBBe = MF->end(); MBB
			!= MBBe; MBB++)
	{
		MachineBasicBlock *block = MBB;
		MachineBasicBlock::iterator I = MBB->end();
		BitVector live = liveouts[block];
		while (I != MBB->begin())
		{
			I--;
			MachineInstr *MI = I;
			BitVector liveout = live; //liveout of instruction
			//do all defs, then all uses
			//usually operands are in the right order, but sometimes the
			//implicit operand list gets changed dynamically and has interleaved
			//defs and uses
			for (unsigned op = 0, n = MI->getNumOperands(); op < n; ++op)
			{
				//register operand?
				MachineOperand &MO = MI->getOperand(op);
				unsigned reg;
				if (MO.isReg() && (reg = MO.getReg()))
				{
					if (MO.isDef())
					{
						if (!liveout.test(reg))
							MO.setIsDead();
						else
							MO.unsetIsDead();
						unsetReg(live, reg);

					}
					if (MO.isUse())
					{
						if (!liveout.test(reg))
						{
							MO.setIsKill();
						}
						else
						{
							MO.unsetIsKill();
						}
						setReg(live, reg);
					}
				}
			}
		}
	}
}

/* Recompute gen and kill and update liveness info
 * */
void DataFlowLiveVariables::update(const std::vector<MachineBasicBlock*>& blocks)
{
	assert(nVars == MF->getSSARegMap()->getLastVirtReg()+1);

	//update gen and kill
	for (unsigned i = 0, n = blocks.size(); i < n; i++)
	{
		const MachineBasicBlock *block = blocks[i];
		compute_block_gen_kill(block, block_gen[block], block_killc[block]);
	}
	//data flow
	std::deque<const MachineBasicBlock*> wlist(blocks.begin(), blocks.end());
	compute_dataflow(wlist);
	update_liveness_info(); //this could be more efficient..
}

void DataFlowLiveVariables::update(MachineBasicBlock* block)
{
	assert(nVars == MF->getSSARegMap()->getLastVirtReg()+1);
	std::vector<MachineBasicBlock*> vec(1, block);
	update(vec);
}

//update all blocks
void DataFlowLiveVariables::update_full()
{
	//redo everything, in order to support more incremental updates
	//we need to be smarter about the order we process blocks to avoid
	//using stale liveness info in the dataflow
	runOnMachineFunction(*MF);

}

// Remove simple dead code.  For now just remove obviously useless moves.
bool DataFlowLiveVariables::runOnMachineFunction(MachineFunction& mf)
{
	MF = &mf;
	nVars = MF->getSSARegMap()->getLastVirtReg()+1;
	liveins.clear();
	liveouts.clear();
	block_gen.clear();
	block_killc.clear();
	//initialize livein liveout sets and kill/gen sets
	for (MachineFunction::iterator MBB = MF->begin(), MBBe = MF->end(); MBB
			!= MBBe; MBB++)
	{
		MachineBasicBlock *block = MBB;
		liveins[block].resize(nVars);
		liveouts[block].resize(nVars);
		block_gen[block].resize(nVars);
		block_killc[block].resize(nVars);
	}

	//compute gen and kill for each block
	for (MachineFunction::iterator MBB = MF->begin(), MBBe = MF->end(); MBB
			!= MBBe; MBB++)
	{
		const MachineBasicBlock *block = MBB;
		compute_block_gen_kill(block, block_gen[block], block_killc[block]);
	}

	//now perform iterative dataflow
	std::deque<const MachineBasicBlock*> worklist;
	for (MachineFunction::iterator MBB = MF->begin(), MBBe = MF->end(); MBB
			!= MBBe; MBB++)
	{
		const MachineBasicBlock *block = MBB;
		worklist.push_front(block);
	}

	compute_dataflow(worklist);

	//now go through and set the liveness information on each operand
	update_liveness_info();

	return false;
}

} //namespace llvm
