#include "llvm/PrincipledBackend/AllocationClassInfo.h"
#include "llvm/ADT/StringExtras.h"

namespace llvm
{

cl::opt<MetricToUse> WhatMetric(cl::Optional,
  cl::desc("What metric to use for the progressive allocator."),
  cl::init(Size),
  cl::values(
    clEnumValN(Size, "size-metric",  "Size metric(default)"),
    clEnumValN(Speed,  "speed-metric",   "Speed metric"),
    clEnumValEnd
  )
);

/* Return type (reg,mem,const,anti) of class c */
AllocationClassInfo::class_type AllocationClassInfo::getType(int c) const
{
	assert(c < (int)allocclasses.size() && c >= 0);
	return allocclasses[c].type;
}

/* Return capacity of class c */
int AllocationClassInfo::getCapacity(int c) const
{
	assert(c < (int)allocclasses.size() && c >= 0);
	return allocclasses[c].capacity;
}

/* Return register for this class, 0 if not a reg */
unsigned AllocationClassInfo::getRegister(int c) const
{
	assert(c < (int)allocclasses.size() && c >= 0);
	if(allocclasses[c].type == Register)
		return allocclasses[c].regNums[0];
	else
		return 0;
}

/* Return canonical registers for this class  */
const std::vector<unsigned>& AllocationClassInfo::getRegisters(int c) const
{
	assert(c < (int)allocclasses.size() && c >= 0);
	return allocclasses[c].regNums;
}

/* cost of move form class c1 to c2 */
unsigned AllocationClassInfo::getMoveCost(int c1, int c2) const
{
	assert(c1 < (int)allocclasses.size() >= 0);
	assert(c2 < (int)allocclasses.size() >= 0);

	return allocclasses[c1].move_from_cost[c2];
}

/* printable unique name of class */
const std::string& AllocationClassInfo::getName(int c) const
{
	assert(c < (int)allocclasses.size() >= 0);
	return allocclasses[c].name;
}

/* TargetRegisterClass for a register allocation class (null if not reg) */
const TargetRegisterClass* AllocationClassInfo::getTargetClass(int c) const
{
	assert(c < (int)allocclasses.size() >= 0);
	return allocclasses[c].TC;
}

/* true if allocation class index is compatible with RC */
bool AllocationClassInfo::isCompatible(const MachineFunction &MF, const TargetRegisterClass* RC, int index) const
{
	const TargetRegisterClass *RC2 = getTargetClass(index);
	if(RC2 == NULL)
		return false;
	unsigned reg = getRegister(index);
	//is reg a member of RC?
	if(reg > 0)
	{
		if(RC->allocableContains(MF, reg))
			return true;
		//check subregs of reg
		const unsigned *subs = MRI.get(reg).SubRegs;
		while(*subs)
		{
			unsigned sub = *subs;
			if(RC->allocableContains(MF, sub))
				return true;
			subs++;
		}
	}
	return false;
}

/* index of register allocation class given regno
 * */
int AllocationClassInfo::getRegisterIndex(unsigned r) const
{
	//return first class that contains r
	if(r == 0)
		return -1;
	for(unsigned i = 0, n = allocclasses.size(); i < n; i++)
	{
		if(getType(i) == Register)
		{
			const std::vector<unsigned>& regs = getRegisters(i);
			for(unsigned j = 0, nr = regs.size(); j < nr; j++)
			{
				if(MRI.regsOverlap(r, regs[j]))
					return i;
			}
		}
	}
	return -1;
}


/* create allocation classes, these methods should be caleld by
 * deriving classes.  They return the new allocation class number */
int AllocationClassInfo::addRegisterClass(unsigned reg, const TargetRegisterClass *trc)
{
	int index = allocclasses.size();
	AllocationClass c;
	c.regNums.push_back(reg);
	c.capacity = 1;
	c.type = Register;
	c.TC = trc;
	c.name = MRI.getName(reg);
	allocclasses.push_back(c);
	register_classes.push_back(index);
	return index;
}

/* create a register class that contains all regs in trc */
int AllocationClassInfo::addRegisterClass(const TargetRegisterClass *trc, const std::vector<unsigned>& regs, std::string name)
{
	int index = allocclasses.size();
	AllocationClass c;
	c.regNums = regs;
	c.capacity = regs.size();
	c.type = Register;
	c.TC = trc;
	c.name = name;
	allocclasses.push_back(c);
	register_classes.push_back(index);
	return index;
}

/* takes care of setting up anti class */
int AllocationClassInfo::addMemoryClass(int capacity, short id)
{
	int index = allocclasses.size();
	int aindex = index+1;
	assert(index > 0);
	AllocationClass c;
	AllocationClass ac; //anti
	c.capacity = -aindex;
	ac.capacity = -index;
	c.type = Memory;
	ac.type = AntiOnly;
	c.id = id;
	ac.id = id;
	c.TC = NULL;
	ac.TC = NULL;
	c.name =  "memory_" + itostr(id);
	ac.name = "anti_" + itostr(id);
	allocclasses.push_back(c);
	allocclasses.push_back(ac);

	memory_classes.push_back(index);
	return index;
}

 int AllocationClassInfo::getAntiClass(int c) const
 {
	 assert(c < (int)allocclasses.size());
	 if(allocclasses[c].type != Memory && allocclasses[c].type != AntiOnly)
		 return -1;
	 else
		 return -allocclasses[c].capacity;
 }

int AllocationClassInfo::addConstantClass(short id)
{
	int index = allocclasses.size();
	AllocationClass c;
	c.capacity = -1;
	c.type = Constant;
	c.id = id;
	c.name = "constant_"+itostr(id);
	allocclasses.push_back(c);
	register_classes.push_back(index);
	return index;
}

int AllocationClassInfo::getClassId(int c) const
{
	assert(c < (int)allocclasses.size());
	return allocclasses[c].id;
}

/* compute cost to move between allocatin classes, this should
   * be overloaded otherwise moves will be inhibited. */
int AllocationClassInfo::computeMoveCost(unsigned src, unsigned dst)
{
	int type1 = getType(src);
	int type2 = getType(dst);

	switch(type1)
	{
	case Register:
		switch(type2)
		{
		case Register: //reg reg move
			if(src == dst)
				return 0;
			else
				return -1;

		case Memory:
			return 0; //store costs are paid by antivariables
		case Constant:
			return 0; //no actual code need to "store" a constant
		case AntiOnly:
			return -1; //not possible
		}
		break;
	case Memory:
		switch(type2)
		{
		case Register: //load cost
			return -1;
		case Memory:
			return 0;
		case Constant:
			return 0; // no code needed
		case AntiOnly:
			return -1;
		}

	case Constant:
		switch(type2)
		{
		case Register: //load cost
			return -1;
		case Memory:
			return -1; //store cost paid by anti
		case Constant:
			return 0;
		case AntiOnly:
			return -1;
		}
	case AntiOnly:
		if(type2 == Memory || type2 == AntiOnly)
			return 0;
		else
			return -1;
	default:
		assert(0 && "unkown allocation class type");
	}
	return -1;
}

void AllocationClassInfo::setupCosts()
{
	for(int i = 0, n = allocclasses.size(); i < n; i++)
	{
		allocclasses[i].move_from_cost.resize(n,-1);
		for(int j = 0; j < n; j++)
		{
			allocclasses[i].move_from_cost[j] =
					computeMoveCost(i,j);
		}
	}
}

/* Return true if c is a more specific class of sc in the provided simple ac */
bool AllocationClassInfo::isSubClassOf(int c, const AllocationClassInfo *simple, int sc) const
{
	if(getType(c) != simple->getType(sc))
	{
		return false;
	}

	if(getType(c) == Register)
	{
		//compare register classes
		unsigned r = getRegister(c);
		if(simple->getRegisterIndex(r) == sc)
			return true;
		else
			return false;
	}
	else
		return true; //don't subclass mem etc
}


}
