#include "BlockTraces.h"
#include <algorithm>

/* A trace of blocks is a linear order of successive blocks that a solver
 * can essentially treat as a single block.
 * */
namespace llvm
{

cl::opt<TraceOrderings> TraceOrdering(cl::Optional,
  cl::desc("What trace ordering to use for the progressive allocator."),
  cl::init(DefaultTrace),
  cl::values(
	clEnumValN(DefaultTrace, "default-traces",  "Default trace ordering"),
    clEnumValN(GreedyTrace, "greedy-traces",  "Greedy trace ordering"),
    clEnumValN(LLVMTrace,  "llvm-traces",   "LLVM numbering trace ordering"),
    clEnumValN(LoopTrace,  "loop-traces",   "Loop priority trace ordering"),
    clEnumValN(LongTrace,  "long-traces",   "Long as possible trace ordering"),
    clEnumValEnd
  )
);


/* compute length of longest good trace from block; this is exponential */
int BlockTraces::compute_trace_length(mcnf_block *block)
{
	int bindex = block->index;

	if(blocks_trace(block) >= 0)
		return 0;

	trace_lengths[bindex] = -2; //sentinal saying we're exploring this block
	int max = trace_lengths[bindex];
	if(max < 0) max = 0;

	for (unsigned i = 0, n = block->succs.size(); i < n; i++)
	{
		mcnf_block *next = block->succs[i];
		if(trace_lengths[next->index] != -2 &&
				blocks_trace(next) < 0 && good_for_trace(block, next))
		{
			int len = compute_trace_length(next);
			if(len > max)
			{
				max = len;
				next_succ[bindex] = i;
			}
			else if(len == max)
			{
				//break ties by prefering a successor with only this block
				//as a successor
				if(next->preds.size() == 1)
				{
					next_succ[bindex] = i;
				}
			}
		}
	}
	trace_lengths[bindex] = max+1;

	if(DebugPRA > 8)
	{
		std::cerr << "Block " << bindex << " trace length " << max+1 << "\n";
	}
	return trace_lengths[bindex];
}

//construct traces in llvm order, all traces are sequential lists of block numbers
void BlockTraces::construct_llvm_traces()
{
	add_trace(NULL);
	for (unsigned i = 0, n = mcnf_cfg->blocks.size(); i < n; i++)
	{
		mcnf_block *block = mcnf_cfg->blocks[i];
		add_to_current_trace(block);
		//if none of block's successors are block+1, create a new trace
		unsigned s;
		unsigned ns = block->succs.size();
		for (s = 0; s < ns; s++)
		{
			mcnf_block *next = block->succs[s];
			if(next->index == block->index+1)
				break;
		}
		if(s == ns && i < n-1)
		{
			//need to create new trace, find prev block with highest trace
			mcnf_block *next = mcnf_cfg->blocks[i+1];
			mcnf_block *bestpred = NULL;
			int bestT = -1;
			for(unsigned p = 0, np = next->preds.size(); p < np; p++)
			{
				mcnf_block *pred = next->preds[p];
				if(block_traces[pred->index] > bestT)
				{
					bestT = block_traces[pred->index];
					bestpred = pred;
				}
			}
			add_trace(bestpred);
		}
	}
}

//used to sort blocks by loop depth
class block_loop_compare
{
	LoopInfo *LI;

public:
	block_loop_compare(LoopInfo *li): LI(li)
	{

	}

	//return true if lhs < rhs
	bool operator()(mcnf_block *lhs, mcnf_block *rhs)
	{
		int ldepth = LI->getLoopDepth(lhs->b->getBasicBlock());
		int rdepth = LI->getLoopDepth(rhs->b->getBasicBlock());

		if(ldepth == rdepth) //sort by num
			return lhs->index < rhs->index;
		else
			return rdepth < ldepth;
	}

};


//construct traces from block, but only of successors with teh same
//loop depth
void BlockTraces::construct_loop_traces_r(mcnf_block *block)
{
	add_to_current_trace(block);
	int bdepth = LI->getLoopDepth(block->b->getBasicBlock());
	//pick successor block with lowest number and equal loop depth
	mcnf_block *next_block_for_trace = NULL;
	int min = INT_MAX;
	for (unsigned i = 0, n = block->succs.size(); i < n; i++)
	{
		mcnf_block *next = block->succs[i];
		int d = LI->getLoopDepth(next->b->getBasicBlock());
		if (block_traces[next->index] < 0 && d == bdepth)
		{
			if(next->index < min)
			{
				next_block_for_trace = next;
				min = next->index;
			}
		}
	}

	if(next_block_for_trace != NULL)
		construct_loop_traces_r(next_block_for_trace);
	//now take care of other successors
	for (unsigned i = 0, n = block->succs.size(); i < n; i++)
	{
		mcnf_block *next = block->succs[i];
		int d = LI->getLoopDepth(next->b->getBasicBlock());
		if (block_traces[next->index] < 0 && d == bdepth)
		{
			add_trace(block);
			construct_loop_traces_r(next);
		}
	}
}

//do innermost loops first
void BlockTraces::construct_loop_traces()
{
	block_loop_compare lcmp(LI);
	std::vector<mcnf_block*> blocks = mcnf_cfg->blocks;
	std::sort(blocks.begin(), blocks.end(), lcmp);

	for(int b = 0, nb = blocks.size(); b < nb; b++)
	{
		mcnf_block *block = blocks[b];
		if(block_traces[block->index] < 0)
		{
			//create a trace for this block
			//look for a valid predecessor
			mcnf_block *bestpred = NULL;
			int bestT = -1;
			for(unsigned p = 0, np = block->preds.size(); p < np; p++)
			{
				mcnf_block *pred = block->preds[p];
				if(block_traces[pred->index] > bestT)
				{
					bestT = block_traces[pred->index];
					bestpred = pred;
				}
			}
			add_trace(bestpred);

			//now construct all the traces in this loop
			construct_loop_traces_r(block);
		}
	}

}

void BlockTraces::construct_traces()
{
	if(TraceOrdering == LLVMTrace)
	{
		construct_llvm_traces();
	}
	else if(TraceOrdering == LoopTrace)
	{
		construct_loop_traces();
	}
	else
	{
		hash_set<mcnf_block*> visited_blocks;
		add_trace(NULL);

		construct_traces_r(mcnf_cfg->blocks[0], visited_blocks);
	}

	/* sanity check, every block should have a trace */
	for (unsigned i = 0, n = mcnf_cfg->blocks.size(); i < n; i++)
	{
		mcnf_block *block = mcnf_cfg->blocks[i];
		int whicht = blocks_trace(block);
		if (whicht < 0)
			abort();
	}

	if(DebugPRA > 1)
	{
		std::cerr << "Traces:\n";
		for(unsigned t = 0, nt = traces.size(); t < nt; t++)
		{
			std::cerr << t << ":";
			for(unsigned b = 0, nb = traces[t].size(); b < nb; b++)
			{
				std::cerr << " " << traces[t][b]->index;
			}
			std::cerr << "\n";
		}
	}
}

void BlockTraces::construct_traces_r(mcnf_block *block,
		hash_set<mcnf_block*>& visited)
{
	visited.insert(block);
//	compute_trace_length(block);
	add_to_current_trace(block);

	if(DebugPRA > 8)
	{
		std::cerr << "Constructing trace at block " << block->index << "\n";
	}

	//pick successor block with lowest number
	mcnf_block *next_block_for_trace = NULL;
	int min = INT_MAX;
	for (unsigned i = 0, n = block->succs.size(); i < n; i++)
	{
		mcnf_block *next = block->succs[i];
		if (visited.count(next) == 0 && good_for_trace(block, next))
		{
			if(next->index < min)
			{
				next_block_for_trace = next;
				min = next->index;
			}
		}
	}

	if(next_block_for_trace != NULL)
		construct_traces_r(next_block_for_trace, visited);
	//now take care of other successors
	for (unsigned i = 0, n = block->succs.size(); i < n; i++)
	{
		mcnf_block *next = block->succs[i];
		if (visited.count(next) == 0 && good_for_trace(block, next))
		{
			add_trace(block);
			construct_traces_r(next, visited);
		}
	}
	/* now anyone we missed */
	for (unsigned i = 0, n = block->succs.size(); i < n; i++)
	{
		mcnf_block *next = block->succs[i];
		if (visited.count(next) == 0)
		{
			add_trace(block);
			construct_traces_r(next, visited);
		}
	}
}

void BlockTraces::add_trace(mcnf_block *from) //create a new (blank) trace
{
	if(DebugPRA>8)
	{
		std::cerr << "New trace " << traces.size() << "\n";
	}
	traces.push_back(std::vector<mcnf_block*>());
	ancestors.push_back(from);
	trace_lengths.assign(mcnf_cfg->blocks.size(), -1);
	next_succ.assign(mcnf_cfg->blocks.size(), -1);
}

void BlockTraces::add_to_current_trace(mcnf_block *block)
{
	int trace = traces.size() -1;
	int index = traces.back().size();
	unsigned b = block->index;

	if(DebugPRA > 8)
	{
		std::cerr << "Adding to current trace block " << b << "\n";
	}
	traces.back().push_back(block);
	assert(b < block_traces.size() && block_traces[b] < 0);
	block_traces[b] = trace;
	trace_indexes[b] = index;
}

bool BlockTraces::good_for_trace(mcnf_block *block, mcnf_block *next)
{
	//stop traces at loop headers
	//this forces a trace recompute inside of the loop
	if(TraceOrdering != LongTrace && LI->isLoopHeader(next->b->getBasicBlock()))
		return false;

	return true;
}

}


