#ifndef BLOCKTRACES_H_
#define BLOCKTRACES_H_
#include "llvm/PrincipledBackend/GlobalMCNF.h"
#include "llvm/PrincipledBackend/AllocationClassInfo.h"
#include <iostream>
#include <stdio.h>
#include "llvm/Function.h"
#include "llvm/Module.h"
#include "llvm/ADT/StringExtras.h"
#include "llvm/Analysis/LoopInfo.h"

/* A trace of blocks is a linear order of successive blocks that a solver
 * can essentially treat as a single block.
 * */
namespace llvm
{

enum TraceOrderings
{
	DefaultTrace, //can override
	GreedyTrace,
	LLVMTrace,
	LoopTrace,
	LongTrace
};

extern cl::opt<TraceOrderings> TraceOrdering;

/* Use traces of blocks instead of individual blocks.  Each trace
 * acts effectively as a single block. */
class BlockTraces
{
	std::vector<std::vector<mcnf_block*> > traces;
	std::vector<int> block_traces; /* what trace a block is assigned, index by block */
	std::vector<int> trace_indexes; /* the position of a block within its trace */
	std::vector<mcnf_block*> ancestors; /* where this trace is branching from */
	std::vector<int> trace_lengths; /* used internally to find longest traces */
	std::vector<int> next_succ; //next successor in longest trace
	GlobalMCNF *mcnf_cfg; //model derived from
	LoopInfo *LI;

public:
	BlockTraces(GlobalMCNF *mcfg, LoopInfo *li) :
		mcnf_cfg(mcfg), LI(li)
	{
		if (TraceOrdering == DefaultTrace && WhatMetric == Speed)
		{
			TraceOrdering = LongTrace; //works better
		}

		block_traces.assign(mcnf_cfg->blocks.size(), -1);
		trace_indexes.assign(mcnf_cfg->blocks.size(), -1);
		trace_lengths.assign(mcnf_cfg->blocks.size(), -1);
		next_succ.assign(mcnf_cfg->blocks.size(), -1);
		construct_traces(); //construct traces from model
	}

	void changeModel(GlobalMCNF *newcfg)
	{
		mcnf_cfg = newcfg;
		traces.clear();
		ancestors.clear();
		block_traces.assign(mcnf_cfg->blocks.size(), -1);
		trace_indexes.assign(mcnf_cfg->blocks.size(), -1);
		trace_lengths.assign(mcnf_cfg->blocks.size(), -1);
		next_succ.assign(mcnf_cfg->blocks.size(), -1);
		construct_traces(); //construct traces from model
	}

	int num_traces()
	{
		return traces.size();
	}
	const std::vector<mcnf_block*>& get_trace(unsigned which) const
	{
		assert(which < traces.size());
		return traces[which];
	}

	int blocks_trace(mcnf_block *b) //what trace a block is assigned
	{
		return block_traces[b->index];
	}
	int blocks_trace(int index) //what trace a block is assigned
	{
		return block_traces[index];
	}
	int block_trace_index(mcnf_block *b) //blocks position in its trace
	{
		return trace_indexes[b->index];
	}
	int block_trace_index(int index) //blocks position in its trace
	{
		return trace_indexes[index];
	}
	/* Returns the ancestor block number of trace whicht */
	int trace_ancestor(int whicht)
	{
		if (ancestors[whicht] == NULL)
			return -1;
		return ancestors[whicht]->index;
	}

private:
	//trace construction routines
	void construct_traces(); //construct the traces from the model
	void construct_traces_r(mcnf_block *block, hash_set<mcnf_block*>& visited); //recursive helper
	void construct_llvm_traces();
	void construct_loop_traces();
	void construct_loop_traces_r(mcnf_block *block); //recursive helper
	void add_trace(mcnf_block *from); //create a new (blank) trace
	void add_to_current_trace(mcnf_block *block);
	bool good_for_trace(mcnf_block *block, mcnf_block *next);
	int compute_trace_length(mcnf_block *block);
};
}

#endif /*BLOCKTRACES_H_*/

