#include "llvm/PrincipledBackend/GlobalMCNF.h"
#include <iostream>
#include <stdio.h>
#include "llvm/Function.h"
#include "llvm/Module.h"
#include "llvm/ADT/StringExtras.h"

#ifdef CPLEX_AVAILABLE

#define IL_STD

#include <ilcplex/ilocplex.h>
ILOSTLBEGIN

/* Generate an ILP of the coalescing problem
 * */
namespace llvm
{

struct move_info
{
	RAATInfo *ii;
	unsigned src;
	unsigned dst;
	unsigned varIndex;
	
	move_info(): ii(NULL), src(0), dst(0), varIndex(0) {}
	move_info(RAATInfo *i, unsigned s, unsigned d, unsigned index):
		ii(i), src(s), dst(d), varIndex(index) {}
};
/* Use the CPLEX callable library to construct the aggressive coalescing problem 
 * Each var corresponds to a model integer var
 * conflicting vars cannot have the same value
 * coalescable vars improve the objective function if they have the same value
 * 
 * Make this a 0-1 programming problem where we model the trivial graph
 * coloing problem (n colors for n nodes) with the additional coalescing constraints
 * 
 * for each node v_i, sum v_i_k = 1  // each node gets some color k
 * for all (v_i,v_j) in Conflict, for all k, v_i_k + v_j_k <= 1 //interfering nodes must have different colors
 * 
 * for all k and m(v_i,v_j) in Coalescable  C_m <= 1 - v_i_k + v_j_k
 * 							 	and   C_m <= 1 + v_i_k - v_j_k
 * 
 * max sum C_m
 * */
void cplex_build_coalesce_model(GlobalMCNF& mcnf_cfg, IloModel& model)
{
	MachineFunction *MF = mcnf_cfg.MF;
	IloEnv env = model.getEnv();
	try
	{
		IloConstraintArray con(env); //constraints
		hash_map<unsigned, unsigned> varIndex;
		IloNumVarArray var(env);

		//create a variable for each var
		unsigned numRegs = MF->getSSARegMap()->getLastVirtReg() - MRegisterInfo::FirstVirtualRegister+1; 
		for (unsigned r = MRegisterInfo::FirstVirtualRegister, n =
				MF->getSSARegMap()->getLastVirtReg(); r <= n; r++)
		{
			unsigned index = var.getSize();
			varIndex[r] = index;
			//create a variable for each color
			for(unsigned c = 0; c < numRegs; c++)
			{
				var.add(IloNumVar(env, 0, 1.0, ILOBOOL, ("r"+utostr(r) + "_" + utostr(c)).c_str()));
			}
			//require that the var get a color
			IloRange color(env, 1.0, 1.0);
			for(unsigned c = 0; c < numRegs; c++)
			{
				color.setLinearCoef(var[index+c], 1.0);
			}
			con.add(color);
		}
		
		// add the conflict constraints
		for (unsigned r = MRegisterInfo::FirstVirtualRegister, n =
				MF->getSSARegMap()->getLastVirtReg(); r <= n; r++)
		{		
			const BitVector& live = mcnf_cfg.conflicts[r];
			
			for (int r2 = live.find_first(); r2 >= 0; r2 = live.find_next(r2))
			{
				if(MRegisterInfo::isVirtualRegister(r2))
				{
					unsigned i1 = varIndex[r];
					unsigned i2 = varIndex[r2];
					for(unsigned c = 0; c < numRegs; c++)
					{
						//v_i_k + v_j_k <= 1
						con.add(var[i1+c] + var[i2+c] <= 1);
					}
				}
			}
		}
		
		//find all the coalescable moves
		unsigned move_start = var.getSize();
		std::vector<move_info> moves;
		for(unsigned b = 0, nb = mcnf_cfg.blocks.size(); b < nb; b++)
		{
			mcnf_block *block = mcnf_cfg.blocks[b];
			// loop over each instruction
			for (unsigned r = 0, nr = block->RAATs.size(); r < nr; r++)
			{
				RAATInfo *ii = &block->RAATs[r];
				if(ii->getNumTiedConstraints() == 1)
				{
					//assume this is a move				
					const RAATInfo::TiedConstraint c = ii->getTiedConstraint(0);
					unsigned src = c.getInReg();
					unsigned dst = c.getOutReg();
					
					moves.push_back(move_info(ii, src, dst, var.getSize()));
					var.add(IloNumVar(env, 0.0, 1.0, ILOBOOL, ("m"+utostr(moves.size()-1)).c_str()));
				}
			}
		}
		
		//create objective function
		IloObjective obj = IloMaximize(env);
		for(unsigned m = move_start, nm = var.getSize(); m < nm; m++)
		{
			obj.setLinearCoef(var[m], 1.0);
		}
		
		//create coalescing conditions
		for(unsigned i = 0, n = moves.size(); i < n; i++)
		{
			unsigned sindex = varIndex[moves[i].src];
			unsigned dindex = varIndex[moves[i].dst];
			IloNumVar& m = var[moves[i].varIndex];

			for(unsigned c = 0; c < numRegs; c++)
			{
				IloNumVar& src = var[sindex+c];
				IloNumVar& dst = var[dindex+c];
				// m <= 1 - v_i_k + v_j_k
				con.add(m <= 1.0 - src + dst);				
				// m <= 1 + v_i_k - v_j_k
				con.add(m <= 1.0 + src - dst);
			}
		}
		
		model.add(obj);
		model.add(con);
	}
	catch (IloException& e)
	{
		cerr << "Concert exception caught: " << e << endl;
	}
	catch (...)
	{
		cerr << "Unknown exception caught" << endl;
	}

}

/* Output the coalesce model */
void cplex_output_coalesce_model(GlobalMCNF& mcnf_cfg)
{
	IloEnv env;
	MachineFunction *MF = mcnf_cfg.MF;
	std::string fname = MF->getFunction()->getParent()->getModuleIdentifier() + "_" + MF->getFunction()->getName() + "_C.lp";

	try
	{
		IloModel model(env, fname.c_str());

		cplex_build_coalesce_model(mcnf_cfg, model);
		IloCplex cplex(model);

		cplex.exportModel(fname.c_str());
	}
	catch (IloException& e)
	{
		cerr << "Concert exception caught: " << e << endl;
	}
	catch (...)
	{
		cerr << "Unknown exception caught" << endl;
	}

	env.end();
}

} //namespace

#else
namespace llvm
{
void cplex_output_coalesce_model(GlobalMCNF& mcnf_cfg) { abort(); }
}
#endif //CPLEX_AVAILABLE
