#include "llvm/PrincipledBackend/GlobalMCNF.h"
#include <iostream>
#include <stdio.h>
#include "llvm/Function.h"
#include "llvm/Module.h"
#include "llvm/ADT/StringExtras.h"

#ifdef CPLEX_AVAILABLE

#define IL_STD

#include <ilcplex/ilocplex.h>
ILOSTLBEGIN

/* Routines for interacting with cplex.  In particular, can write the problem
 * to a file or solve it if the cplex library is installed.
 * */
namespace llvm
{

/*
 * read any existing solution off of mcnf_cfg and make it a strting solution
 */
void cplex_start_sol(GlobalMCNF& mcnf_cfg, IloCplex& cplex, IloNumVarArray& var, hash_map<mcnf_edge*, hash_map<int, unsigned> >& edgepos)
{
	IloEnv env = var.getEnv();
	IloNumArray x(env, var.getSize());

	unsigned b;
	const unsigned nb = mcnf_cfg.blocks.size();
	for ( b = 0; b < nb; b++)
	{
		int n;
		mcnf_block *block = mcnf_cfg.blocks[b];
		int nn = block->allnodes.size();

		for (n = 0; n < nn; n++)
		{
			int e, en;
			mcnf_node *node = block->allnodes[n];
			en = node->succs.size();

			for (e = 0; e < en; e++)
			{
				int c, cn;
				mcnf_edge *edge = node->succs[e];
				cn = edge->commodities.size();

				for (c = 0; c < cn; c++)
				{
					int q = edge->commodities[c];
					mcnf_node *src = edge->src;
					mcnf_node *dst = edge->dst;
					int sindex = src->cindex.get(q);
					int dindex = dst->cindex.get(q);
					if(sindex >= 0 && dindex >= 0 && src->feas_marked[sindex] && dst->feas_marked[dindex])
					{
						x[edgepos[edge][q]] = 1;
					}
					else
					{
						x[edgepos[edge][q]] = 0;
					}
				}
			}
		}
	}
	cplex.setVectors(x, 0, var, 0, 0, 0);
}


/* Use the CPLEX callable library to solve the global mcnf problem */
void cplex_build_model(GlobalMCNF& mcnf_cfg, IloModel& model,
		IloNumVarArray& var, int &TCstart, int &TCend,
		hash_map<mcnf_edge*, hash_map<int, unsigned> >& edgepos)
{
	int ncnt = 0, bcnt = 0, dcnt = 0, ecnt = 0, bdcnt = 0;
	IloEnv env = model.getEnv();
	try
	{
		IloRangeArray con(env); //constraints

		//setup mapping of variables (edges) filling in edgpos and var
		unsigned b;
		const unsigned nb = mcnf_cfg.blocks.size();
		for (b = 0; b < nb; b++)
		{
			int n;
			mcnf_block *block = mcnf_cfg.blocks[b];
			int nn = block->allnodes.size();

			for (n = 0; n < nn; n++)
			{
				int e, en;
				mcnf_node *node = block->allnodes[n];
				en = node->succs.size();

				for (e = 0; e < en; e++)
				{
					int c, cn;
					mcnf_edge *edge = node->succs[e];
					cn = edge->commodities.size();

					for (c = 0; c < cn; c++)
					{
						int q = edge->commodities[c];
						int cost = edge->getCost(q);
						if (cost != UNREACHABLE_INT)
						{
							edgepos[edge][q] = var.getSize();
							var.add(IloNumVar(env, 0, 1.0, ILOBOOL,
											edge->getName(q).c_str()));
						}
					}
				}
			}
		}

		//create objective function
		IloObjective obj = IloMinimize(env);
		for ( b = 0; b < nb; b++)
		{
			int n;
			mcnf_block *block = mcnf_cfg.blocks[b];
			int nn = block->allnodes.size();

			for (n = 0; n < nn; n++)
			{
				int e, en;
				mcnf_node *node = block->allnodes[n];
				en = node->succs.size();

				for (e = 0; e < en; e++)
				{
					int c, cn;
					mcnf_edge *edge = node->succs[e];
					cn = edge->commodities.size();

					for (c = 0; c < cn; c++)
					{
						int q = edge->commodities[c];
						int cost = edge->getCost(q);
						if (cost != 0 && cost != UNREACHABLE_INT)
						{
							obj.setLinearCoef(var[edgepos[edge][q]], (double)cost);
						}
					}
				}
			}
		}

		TCstart = var.getSize();
		if(TiedConstraints)
		{
			//setup tied constraints
			//for each node, tc_node <= src edge; tc_node <= dst_edge; and
			//cost*tc_ndoe in the objective function
			for (b = 0; b < nb; b++)
			{
				mcnf_block *block = mcnf_cfg.blocks[b];
				for(unsigned t = 0, nt = block->TC_RAATs.size(); t < nt; t++)
				{
					TC_RAATInfo& TC = block->TC_RAATs[t];
					int srcq = TC.srcq;
					int dstq = TC.dstq;
					for(unsigned i = 0, n = TC.inodes.size(); i < n; i++)
					{
						mcnf_node *node = TC.inodes[i];
						int sindex = node->cindex.get(srcq);
						int dindex = node->cindex.get(dstq);
						if(sindex >= 0 && dindex >= 0)
						{
							unsigned tc_node = var.getSize();
							string name = "tc_" + utostr(block->index) + "_" + utostr(t) + "_" + utostr(node->allocClass) + "_" + utostr(srcq) + "_" +utostr(dstq);
							//set cost
							var.add(IloNumVar(env, 0, 1.0, ILOBOOL, name.c_str()));
							obj.setLinearCoef(var[tc_node], (double)TC.cost);

							//find src and dst edges, should be just one
							for(int p = 0, np = node->preds.size(); p < np; p++)
							{
								mcnf_edge *pred = node->preds[p];
								if(pred->commodity_valid(srcq))
								{
									//add constraint
									con.add(var[tc_node] - var[edgepos[pred][srcq]] <= 0);
								}
							}
							for(int s = 0, ns = node->succs.size(); s < ns; s++)
							{
								mcnf_edge *succ = node->succs[s];
								if(succ->commodity_valid(dstq))
								{
									//add constraint
									con.add(var[tc_node] - var[edgepos[succ][dstq]] <= 0);
								}
							}
						}
					}
				}
			}
		}
		TCend = var.getSize();

		model.add(obj);

		//now add constraints
		/* first the network constraints */

		/* the inflow of a node is equal to its outflow
		 srcs/sinks generate their own inflow/outflow
		 entry nodes multiply their outflow (ie, merge the inflow)
		 exit nodes multiply the inflow (ie, split the outflow)
		 */
		for (b = 0; b < nb; b++)
		{
			int n;
			mcnf_block *block = mcnf_cfg.blocks[b];
			int nn = block->allnodes.size();

			for (n = 0; n < nn; n++)
			{
				int c, cn;
				mcnf_node *node = block->allnodes[n];
				double inmult = 1;
				double outmult = -1;
				double inadd = 0;
				double outadd = 0;

				if (node->succs.size() == 0 && node->preds.size() == 0)
				continue;

				if (node->type == nDef)
				inadd = 1;
				else if (node->type == nSink)
				outadd = -1;

				/* for each commodity going through this node */
				cn = node->commodities.size();
				for (c = 0; c < cn; c++)
				{
					int e, en;
					int q = node->commodities[c];

					/* calculate number of pred/succ for this commodity for entry/exit nods */
					if (node->type == nEntry)
					{
						/* for each pred, add the constraint in == out */
						//have to handle quantity renaming here
						if (node->succs.size() != 1)
						abort();

						mcnf_edge *outedge = node->succs[0];
						if (outedge->commodity_valid(q))
						{
							hash_set<mcnf_edge*> seen;
							int en = node->preds.size();
							for (e = 0; e < en; e++)
							{
								mcnf_edge *edge = node->preds[e];
								//avoid double couting special crossing antiedges
								if(seen.count(edge) > 0)
									continue;
								//may be a crossblock edge, so specify which side
								//of the edge the quantity is on
								if (edge->commodity_valid(q))
								{
									std::string cname = "bound_" + utostr(bdcnt++);
									//look to see if this is a memory node with
									//an additional edge coming from an anti
									//edge formulation is differnt then node used by lagrangian...
									mcnf_edge *edge2 = NULL;
									if(edge->src->type == nExit)
									{
										for(int e2 = e+1; e2 < en; e2++)
										{
											mcnf_edge *tmp = node->preds[e2];
											if(tmp->src->type == nExit &&
													tmp->src->block_num == edge->src->block_num)
											{
												assert(edge2 == NULL); //should only be one
												edge2 = tmp;
											}
										}
									}
									if(edge2 != NULL)
									{
										seen.insert(edge2);
										con.add(var[edgepos[edge2][q]] +
												var[edgepos[edge][q]]
												- var[edgepos[outedge][q]] == 0);
									}
									else
									{
										con.add(
											var[edgepos[edge][q]]
											- var[edgepos[outedge][q]] == 0);
									}
									con[con.getSize()-1].setName(cname.c_str());
								}
							}
						}
					}
					else if (node->type == nExit)
					{
						/* for each succ, add the constraint in == out */
						if (node->preds.size() != 1)
						abort();
						mcnf_edge *inedge = node->preds[0];
						if (inedge->commodity_valid(q))
						{
							hash_set<mcnf_edge*> seen;
							en = node->succs.size();
							for (e = 0; e < en; e++)
							{
								mcnf_edge *edge = node->succs[e];
								if(seen.count(edge) > 0)
									continue;
								if (edge->commodity_valid(q))
								{
									std::string cname = "bound_" + utostr(bdcnt++);
									//look to see if this is a memory node with
									//an additional edge coming from an anti
									mcnf_edge *edge2 = NULL;
									if(edge->dst->type == nEntry)
									{
										for(int e2 = e+1; e2 < en; e2++)
										{
											mcnf_edge *tmp = node->succs[e2];
											if(tmp->dst->type == nEntry &&
													tmp->dst->block_num == edge->dst->block_num)
											{
												assert(edge2 == NULL); //should only be one
												edge2 = tmp;
											}
										}
									}
									if(edge2 != NULL)
									{
										seen.insert(edge2);
										con.add(IloRange(env, 0.0,
												var[edgepos[inedge][q]]
												    - (var[edgepos[edge][q]] + var[edgepos[edge2][q]]), 0.0,
												    cname.c_str()));
									}
									else
									{
										con.add(IloRange(env, 0.0,
													var[edgepos[inedge][q]]
													- var[edgepos[edge][q]], 0.0,
													cname.c_str()));
									}
								}
							}
						}
					}
					else
					{
						inmult = 1;
						outmult = -1;

						double flow = -outadd-inadd;
						std::string cname = "net_" + utostr(ncnt++);
						IloRange c(env, flow, flow, cname.c_str());

						/* output summation of inflow */
						en = node->preds.size();
						for (e = 0; e < en; e++)
						{
							mcnf_edge *edge = node->preds[e];

							if (edge->commodity_valid(q))
							{
								c.setLinearCoef(var[edgepos[edge][q]], inmult);
							}
						}

						/*output summation of outflow */

						en = node->succs.size();
						for (e = 0; e < en; e++)
						{
							mcnf_edge *edge = node->succs[e];

							if (edge->commodity_valid(q))
							{
								c.setLinearCoef(var[edgepos[edge][q]], outmult);
							}
						}

						con.add(c);
					}
				}

				/* now output bundle constraints */
				if (cn> 1)
				{
					if (node->dual_prices != NULL) /* memory node, just bundle vars with antivars */
					{
						/* bundle every pair of var/antivar together */
						/* make this a node constraint, not edge constraint
						 since otherwise a var can go in and out of memory in the xbar */
						for (c = 0; c < cn; c++)
						{
							int q = node->commodities[c];

							if (q> 0)
							{
								int e, en;
								en = node->preds.size();
								std::string cname = "dual_" + utostr(dcnt++);
								IloRange c(env, 0.0, 1.0, cname.c_str());
								for (e = 0; e < en; e++)
								{
									mcnf_edge *edge = node->preds[e];

									if (edge->commodity_valid(q))
									{
										c.setLinearCoef(var[edgepos[edge][q]], 1.0);
									}
									if (edge->commodity_valid(-q))
									{
										c.setLinearCoef(var[edgepos[edge][-q]], 1.0);
									}
								}
								con.add(c);

								en = node->succs.size();
								cname = "dual_" + utostr(dcnt++);
								IloRange d(env, 0.0, 1.0, cname.c_str());
								for (e = 0; e < en; e++)
								{
									mcnf_edge *edge = node->succs[e];

									if (edge->commodity_valid(q))
									{
										d.setLinearCoef(var[edgepos[edge][q]], 1.0);
									}
									if (edge->commodity_valid(-q))
									{
										d.setLinearCoef(var[edgepos[edge][-q]], 1.0);
									}
								}
								con.add(d);
							}
						}

					}
					else if (node->capacity >= 0)
					{
						if (node->type != nXBar && node->type != nInsn)
						abort(); /* nothing else should be capacitated */

						/* non dual capacitates only affect one type of commodity */
						if (node->preds.size()> 0)
						{
							/* sum of incoming is <= capacity */
							std::string cname = "bin_" + utostr(bcnt++);
							IloRange bin(env, 0.0, (double)node->capacity,cname.c_str());
							int cnt = 0;
							for (c = 0; c < cn; c++)
							{
								int e, en;
								int q = node->commodities[c];

								if (alloc_classes->getType(node->allocClass)
										== AllocationClassInfo::AntiOnly)
								{
									if (q> 0)
									continue;
								}
								else
								{
									if (q < 0)
									continue;
								}

								en = node->preds.size();
								for (e = 0; e < en; e++)
								{
									mcnf_edge *edge = node->preds[e];
									if (edge->commodity_valid(q))
									{
										bin.setLinearCoef(var[edgepos[edge][q]], 1.0);
										cnt++;
									}
								}
							}
							if(cnt> 1)
							con.add(bin);
						}

						if (node->succs.size()> 0)
						{
							/* sum of outgoing is <= capacity */
							std::string cname = "bout_" + utostr(bcnt++);
							IloRange bout(env, 0.0, (double)node->capacity, cname.c_str());
							int cnt = 0;

							for (c = 0; c < cn; c++)
							{
								int e, en;
								int q = node->commodities[c];

								if (alloc_classes->getType(node->allocClass)
										== AllocationClassInfo::AntiOnly)
								{
									if (q> 0)
									continue;
								}
								else
								{
									if (q < 0)
									continue;
								}

								en = node->succs.size();
								for (e = 0; e < en; e++)
								{
									mcnf_edge *edge = node->succs[e];
									if (edge->commodity_valid(q))
									{
										bout.setLinearCoef(var[edgepos[edge][q]], 1.0);
										cnt++;
									}
								}
							}
							con.add(bout);
						}
					}

					/* require split and merge flows to be equal */
					if (false && node->type == nEntry && node->preds.size()> 1)
					{
						for (c = 0; c < cn; c++)
						{
							int e, en;
							int q = node->commodities[c];

							en = node->preds.size();
							for (e = 1; e < en; e++)
							{
								//find two successive edges with q valid
								mcnf_edge *edge1 = node->preds[e-1];
								while(!edge1->commodity_valid(q) && e < en)
								{
									e++;
									edge1 = node->preds[e-1];
								}
								if(e == en)
								break;

								mcnf_edge *edge2 = node->preds[e];
								while(!edge2->commodity_valid(q) && e < en-1)
								{
									e++;
									edge2 = node->preds[e];
								}
								if(!edge2->commodity_valid(q))
								break;

								std::string cname = "equ_" + utostr(ecnt++);
								con.add(IloRange(env, 0.0,
												var[edgepos[edge1][q]]
												-var[edgepos[edge2][q]],
												0.0, cname.c_str()));
							}
						}
					}
					if (false && node->type == nExit && node->succs.size()> 1)
					{
						for (c = 0; c < cn; c++)
						{
							int e, en;
							int q = node->commodities[c];

							en = node->succs.size();
							for (e = 1; e < en; e++)
							{
								//find two successive edges with q valid
								mcnf_edge *edge1 = node->succs[e-1];
								while(!edge1->commodity_valid(q) && e < en)
								{
									e++;
									edge1 = node->succs[e-1];
								}
								if(e == en)
								break;

								mcnf_edge *edge2 = node->succs[e];
								while(!edge2->commodity_valid(q) && e < en-1)
								{
									e++;
									edge2 = node->succs[e];
								}
								if(!edge2->commodity_valid(q))
								break;

								std::string cname = "equ_" + utostr(ecnt++);
								assert(edgepos[edge1].count(q)> 0);
								con.add(IloRange(env, 0.0,
												var[edgepos[edge1][q]]
												-var[edgepos[edge2][q]],
												0.0, cname.c_str()));
							}
						}
					}
				}
			}
		}

		//now add any blocked edges
		for(hash_map<mcnf_edge*, std::vector<int> >::iterator  itr = mcnf_cfg.blockedEdges.begin(),
				eitr = mcnf_cfg.blockedEdges.end(); itr != eitr; itr++)
		{
			mcnf_edge *edge = itr->first;
			const std::vector<int>& qs = itr->second;
			for(int i = 0, n = qs.size(); i < n; i++)
			{
				int q = qs[i];
				con.add(IloRange(env, 0.0, var[edgepos[edge][q]], 0.0));
			}
		}

		model.add(con);
	}
	catch (IloException& e)
	{
		cerr << "Concert exception caught: " << e << endl;
	}
	catch (...)
	{
		cerr << "Unknown exception caught" << endl;
	}

}

/* Given a model with a feasible solution in it, output a corresonding mst file */
void mcnf_cplex_output_mst(GlobalMCNF& mcnf_cfg)
{
	MachineFunction *MF = mcnf_cfg.MF;
	std::string fname = MF->getFunction()->getParent()->getModuleIdentifier() + "_" + MF->getFunction()->getName() + ".mst";
	ofstream file(fname.c_str());

	file << "<?xml version = \"1.0\" standalone=\"yes\"?>\n";
	file << "<?xml-stylesheet href=\"http://www.ilog.com/products/cplex/xmlv1.0/solution.xsl\"\n";
	file << "type=\"text/xsl\"?>\n";
	file << "<CPLEXSolution version=\"1.0\">\n\n";
	file << " <variables>\n";

	unsigned b;
	const unsigned nb = mcnf_cfg.blocks.size();
	for (b = 0; b < nb; b++)
	{
		int n;
		mcnf_block *block = mcnf_cfg.blocks[b];
		int nn = block->allnodes.size();

		for (n = 0; n < nn; n++)
		{
			int e, en;
			mcnf_node *node = block->allnodes[n];
			en = node->succs.size();

			for (e = 0; e < en; e++)
			{
				int c, cn;
				mcnf_edge *edge = node->succs[e];
				cn = edge->commodities.size();

				for (c = 0; c < cn; c++)
				{
					int q = edge->commodities[c];
					int cost = edge->getCost(q);
					if (cost != UNREACHABLE_INT)
					{
						file << "<variable name=\"";
						std::string name = edge->getName(q);
						if(name[0] == 'e')
							file << '_';
						file << edge->getName(q);
						file << "\" value=\"";
						if(edge->is_set_solution(q))
							file << "1\"/>\n";
						else
							file << "0\"/>\n";
					}
				}
			}
		}
	}
	file << " </variables>\n";
	file << "</CPLEXSolution>\n";

}

/* Output the current mcnf_cfg as cplex lp format */
void mcnf_cplex_output(GlobalMCNF& mcnf_cfg, std::string ext)
{
	IloEnv env;
	MachineFunction *MF = mcnf_cfg.MF;
	std::string fname = MF->getFunction()->getParent()->getModuleIdentifier() + "_" + MF->getFunction()->getName() + ext + ".lp";

	try
	{
		int TCstart, TCend;
		IloNumVarArray var(env);
		hash_map<mcnf_edge*, hash_map<int, unsigned> > edgepos;
		IloModel model(env, fname.c_str());

		cplex_build_model(mcnf_cfg, model, var, TCstart, TCend, edgepos );
		IloCplex cplex(model);

		//branch first on coalescing constraints
		for(int i = TCstart; i < TCend; i++)
		{
			cplex.setPriority(var[i], 1.0);
		}

		cplex.exportModel(fname.c_str());
	}
	catch (IloException& e)
	{
		cerr << "Concert exception caught: " << e << endl;
	}
	catch (...)
	{
		cerr << "Unknown exception caught" << endl;
	}

	env.end();
}

/* Use cplex to find the solution to the global MCNF problem
 * and put the solution into mcnf_cfg. */
int mcnf_cplex_solve(GlobalMCNF& mcnf_cfg, bool useInitSolution)
{
	IloEnv env;
	int ret = 0;
	MachineFunction *MF = mcnf_cfg.MF;
	std::string fname = MF->getFunction()->getParent()->getModuleIdentifier() + "_" + MF->getFunction()->getName() + ".lp";

	try
	{
		int TCstart, TCend;
		IloModel model(env, fname.c_str());
		IloNumVarArray var(env);
		hash_map<mcnf_edge*, hash_map<int, unsigned> > edgepos;

		cplex_build_model(mcnf_cfg, model, var, TCstart, TCend, edgepos);
		IloCplex cplex(model);
		if(useInitSolution)
		{
			cplex_start_sol(mcnf_cfg, cplex, var, edgepos);
		}

		// Optimize the problem and obtain solution.
		if ( !cplex.solve() )
		{
			env.error() << "Failed to optimize LP" << endl;
			throw(-1);
		}

		std::cerr << "Solution status = " << cplex.getStatus() << "\n";
		double solval = cplex.getObjValue();
		std::cerr << "Solution value  = " << solval << "\n";
		ret = (int)solval;

		mcnf_cfg.traverse_nodes(initialize_node_for_feasible_solution_finder);

		IloNumArray vals(env);
		cplex.getValues(var, vals);

		for(hash_map<mcnf_edge*, hash_map<int, unsigned> > ::iterator itr =
				edgepos.begin(); itr != edgepos.end(); itr++)
		{
			mcnf_edge *edge = itr->first;
			hash_map<int, unsigned>& cmap = itr->second;
			for(hash_map<int, unsigned>::iterator citr = cmap.begin();
					citr != cmap.end(); citr++)
			{
				int q = citr->first;
				unsigned pos = citr->second;
				if(vals[pos] > .1)
					edge->set_solution(q);
			}
		}

		//int checkval = mcnf_cfg.sanity_check_feasible_solution();
		//hmm.. rounding can make this fail..assert(checkval == solval && "objective values don't match");

	}
	catch (IloException& e)
	{
		std::cerr << "Concert exception caught: " << e << endl;
	}
	catch (...)
	{
		std::cerr << "Unknown exception caught" << endl;
	}

	env.end();
	return ret;
}

} //namespace

#else //CPLEX_AVAILABLE

namespace llvm
{
void mcnf_cplex_output(GlobalMCNF& mcnf_cfg, std::string ext) { abort();}
int mcnf_cplex_solve(GlobalMCNF& mcnf_cfg, bool useInitSolution) { abort(); return 0; }
void mcnf_cplex_output_mst(GlobalMCNF& mcnf_cfg) { abort(); }
}
#endif
