#include "llvm/PrincipledBackend/GlobalMCNF.h"
#include <iostream>
#include <stdio.h>
#include "llvm/Function.h"
#include "llvm/Module.h"
#include "llvm/ADT/StringExtras.h"
#include "ColumnGenSolver.h"
#include <sys/time.h>
#include <time.h>

#ifdef CPLEX_AVAILABLE



/* Routines for interacting with cplex.  In particular, can write the problem
 * to a file or solve it if the cplex library is installed.
 * */
namespace llvm
{

int ColumnGenSolver::compute_feas_cost(mcnf_node *sink)
{
	tally_path_cost tally;
	mcnf_cfg.trace_feasible_path(sink, sink->commodities[0], tally);
	return tally.getCost();
}

//allocate paths for q going through node to the appropriate predecessors
void ColumnGenSolver::update_paths_node(mcnf_node *node, int q,
		const std::vector<BitVector>& curr, std::vector<BitVector>& next,
		std::vector<int>& pcosts)
{
	BlockPathInfo& bpi = bpInfo[node->block_num];
	std::vector<QPathInfo> ranges;
	bpi.qMap.getRanges(q, ranges);
	int index = node->cindex.get(q);

	for (int r = 0, nr = ranges.size(); r < nr; r++)
	{
		int first = ranges[r].start;
		int end = ranges[r].end;

		//how many paths are currently going through this node representing q?
		int livepaths = 0;
		if (node->has_class())
		{
			for (int i = first; i < end; i++)
			{
				if (curr[node->allocClass].test(i))
					livepaths++;
			}
		}
		else
		{

			if (node != ranges[r].sink)
				continue; //skip
			livepaths = node->binfo[index].num_shortest+1; //sink, include feas
			if (!node->feas_marked[index])
				livepaths--;
		}
		assert(livepaths <= end-first);

		if (livepaths == 0)
			continue; //nothing to propagate

		//first take care of feasible path
		if (node->type == nSink || curr[node->allocClass].test(first))
		{
			assert(node->feas_marked[index]);
			mcnf_edge *pedge = node->fsinfo[index].prev_edge;
			mcnf_node *prev = pedge->src;
			pcosts[first] += pedge->getCost(q);
			if (prev->has_class())
			{
				next[prev->allocClass].set(first);

				int pindex = prev->cindex.get(q);
				mcnf_edge *dedge = prev->fsinfo[pindex].prev_edge;
				if (dedge->src->type == nDef) //include def cost
				{
					pcosts[first] += dedge->getCost(q);
				}
			}
			path_edges[bpi.path_index_offset+first].push_back(pedge);
			livepaths--;
		}
		first++; //ignore feasible path from now on

		if (livepaths == 0)
			return; //nothing to propagate
		//num shortest is the number of paths TO this node, not the number
		//through the node (which what livepaths is), calculate the scaling factor
		assert(livepaths >= node->binfo[index].num_shortest);
		int scale = livepaths/node->binfo[index].num_shortest;

		//process all predecessors
		unsigned predmask = node->binfo[index].prevs;
		assert(predmask);
		//should never encounter entry node here

		for (int p = 0, np = node->preds.size(); p < np; p++)
		{
			if (predmask & (1<<p))
			{
				int cnt = 0;
				mcnf_edge *pedge = node->preds[p];
				mcnf_node *pred = pedge->src;
				assert(pred->has_class());
				int pindex = pred->cindex.get(q);
				if (pindex >= 0)
				{
					int npaths = pred->binfo[pindex].num_shortest*scale;
					int cost = pedge->getCost(q);
					//allocate npaths of the paths currently going through node to
					//pred's alloc class
					if (node->has_class())
					{
						if (pred->sinfo[0][pindex].prev_edge->src->type == nDef)
						{
							cost
									+= pred->sinfo[0][pindex].prev_edge->getCost(q); //include def cost
						}
						//get first npaths of paths currently allocated node's alloc class
						//and assign them to pred's alloc class
						int currc = node->allocClass;
						if (!curr[currc].test(first))
							first = curr[currc].find_next(first);
						for (; cnt < npaths; cnt++)
						{
							assert(first < end && first >= 0);
							next[pred->allocClass].set(first);
							pcosts[first] += cost;
							path_edges[bpi.path_index_offset+first].push_back(pedge);
							first = curr[currc].find_next(first);
						}
					}
					else //better be a sink
					{
						assert(node->type == nSink);
						for (; cnt < npaths; first++, cnt++)
						{
							pcosts[first] += cost;
							//always push first edge
							path_edges[bpi.path_index_offset+first].push_back(pedge);
							next[pred->allocClass].set(first);
						}
						assert(first <= end);
					}
				}
			}
		}
	}
}

//update path bitvectors to reflect what paths are where at level
//curr contains the path allocation for the successive xbar level,
//insn for the successive insn level
//put the result in next
//accumulate the cost of each path in pcosts
void ColumnGenSolver::update_paths(mcnf_node **level, const std::vector<BitVector>& curr, 
		const std::vector<BitVector>& insn, std::vector<BitVector>& next,
		std::vector<int>& pcosts)
{
	int numClasses = alloc_classes->numClasses();
	hash_map<mcnf_node*, hash_set<int> > processed;
	
	for(int i = 0, n = next.size(); i < n; i++)
	{
		next[i].reset();
	}
	
	for(int c = 0; c < numClasses; c++)
	{
		mcnf_node *node = level[c];
		if(node != NULL)
		{
			//process every one of node's in block successors
			//only process once per a commodity though
			for(int s = 0, ns = node->succs.size(); s < ns; s++)
			{
				mcnf_edge *edge = node->succs[s];
				mcnf_node *succ = edge->dst;
				if(succ->type == nEntry)
					continue;
				for(int i = 0, n = edge->commodities.size(); i < n; i++)
				{
					int q = edge->commodities[i];
					if(processed.count(succ))
					{
						if(processed[succ].count(q))
							continue;
					}
					processed[succ].insert(q);;
					
					if(succ->type == nInsn)
					{
						update_paths_node(succ, q, insn, next, pcosts);
					}
					else 
					{
						update_paths_node(succ, q, curr, next, pcosts);
					}
				}
				
			}
		}
	}
}

//add capacity and dual constraint coefficients to rows for node
void ColumnGenSolver::set_constraints(mcnf_node *node,
		IloNumVarArray& bpaths, const BitVector& paths, const BlockPathInfo& bpi)
{
	if(node == NULL)
		return;

	if (node->capacity > 0)
	{
		assert(priceMap.count(node));
		//set coefficient for all (nonanti) paths in this class
		int row = priceMap[node];
		for (int p = paths.find_first(); p >= 0; p = paths.find_next(p))
		{
			int q = pathMap.find(p+bpi.path_index_offset)->second;
			if(q > 0)
			{
				rows[row].setLinearCoef(bpaths[p], 1.0);
			}
		}
	}
	else if (node->dual_prices != NULL)
	{
		assert(priceMap.count(node));
		int row = priceMap[node];
		int nc = node->commodities.size();
		/* for each positive commodity */
		for (int i = 0; i < nc; i++)
		{
			int q = node->commodities[i];
			if (q > 0 && node->dual_constraint_matters(q))
			{
				/* get the index for the corresponding anti commodity */
				int ai = node->cindex.get(-q);
				if (ai >= 0)
				{
					//all q paths plus all q paths
					std::vector<QPathInfo> ranges;
					bpi.qMap.getRanges(q, ranges);
					for (int r = 0, nr = ranges.size(); r < nr; r++)
					{
						int first = ranges[r].start;
						int second = ranges[r].end;
						for (int p = first; p < second; p++)
						{
							if (paths.test(p))
							{
								rows[row].setLinearCoef(bpaths[p], 1.0);
							}
						}
					}
					
					bpi.qMap.getRanges(-q, ranges);
					for (int r = 0, nr = ranges.size(); r < nr; r++)
					{
						int first = ranges[r].start;
						int second = ranges[r].end;
						for (int p = first; p < second; p++)
						{
							if (paths.test(p))
							{
								rows[row].setLinearCoef(bpaths[p], 1.0);
							}
						}
					}
					
					row++;
				}
			}
		}
	}
}


//Construct a path formulation of the problem using all the equal cost shortest
//paths and the feasible path, solve it, and then extract the prices
void ColumnGenSolver::column_generate_prices()
{
	//create the model
	int numClasses = alloc_classes->numClasses();
	mcnf_node **curr_nodes = (mcnf_node**)alloca(numClasses
				*sizeof(mcnf_node*));
		
	//first do the block local constraints
	for(int b = 0, nb = mcnf_cfg.blocks.size(); b < nb; b++)
	{
		mcnf_block *block = mcnf_cfg.blocks[b];
		BlockPathInfo& bpi = bpInfo[b];
		int npaths = 0;;
		bpi.qMap.clear();
		bpi.path_index_offset = path_offset;
		
		//figure out the number of paths in this block and
		//add single path constraints
		for (int s = 0, sn = block->sinks.size(); s < sn; s++)
		{
			mcnf_node *sink = block->sinks[s];
			int q = sink->commodities[0];
			int np = sink->binfo[0].num_shortest;
			np += sink->feas_marked[0];//include feas path
			pathMap.addRange(bpi.path_index_offset+npaths, bpi.path_index_offset+npaths+np, q);
			bpi.qMap.addRange(sink, npaths, npaths+np);
			npaths += np;					
		}
		
		path_edges.resize(path_edges.size()+npaths);
		std::vector<int> pcosts(npaths, 0); //cost of each path
		//create path variables
		IloNumVarArray bpaths(env, npaths, 0, 1);
		
		if(DebugPRA)
		{
			std::cerr << "Block " << b << " npaths " << npaths << "\n";
			int nps = 0;
			for (int s = 0, sn = block->sinks.size(); s < sn; s++)
			{
				mcnf_node *sink = block->sinks[s];
				int q = sink->commodities[0];
				int np = sink->binfo[0].num_shortest + sink->feas_marked[0]; //include feas path
				for(int i = 0; i < np; i++)
				{
					std::string name = (q < 0 ? "pm" : "p") + itostr(abs(q)) + "_" + itostr(b) + "_" + itostr(i) +"_" + itostr(path_offset+nps+i);
					bpaths[nps+i].setName(name.c_str());
				}
				nps += np;					
			}
		}
		paths.add(bpaths);

		//add single path per quantity constraints
		for (int s = 0, sn = block->sinks.size(); s < sn; s++)
		{
			mcnf_node *sink = block->sinks[s];
			int row = priceMap[sink];
			int start, end;
			bpi.qMap.getRange(sink, start, end);
			for(int pnum = start; pnum < end; pnum++)
			{
				rows[row].setLinearCoef(bpaths[pnum], 1.0);
			}				
		}
		
		//now go through the block level by level maintaing for each alloc class
		//what paths are active in that class and filling out the constraint rows
		//for the capacity and dual constraints
		//keep track of insn allocations 
		std::vector<BitVector> currp(numClasses, BitVector(npaths));		
		std::vector<BitVector> insnp(numClasses, BitVector(npaths));		
		std::vector<BitVector> nextp(numClasses, BitVector(npaths));		

		//go through the block backwards a level at a time
		block->get_exit_level(curr_nodes);
		update_paths(curr_nodes, currp, insnp, nextp, pcosts);
		currp.swap(nextp);
		bpi.exit_paths = currp; //save what paths are were at exit
		
		while(!block->is_entry_level(curr_nodes))
		{
			block->set_to_previous_level(curr_nodes);
			//for the current level update what paths are in each node (curr)
			update_paths(curr_nodes, currp, insnp, nextp, pcosts);
			
			//capacity/dual constraints
			for(int c = 0; c < numClasses; c++)
			{
				mcnf_node *node = curr_nodes[c];
				if(node == NULL || node_has_no_paths(node))
					continue;
				
				set_constraints(node, bpaths, nextp[c], bpi);
			}
			
			if(block->is_insn_level(curr_nodes))
			{
				//set insnp, don't worry about stale data in curr since
				//it won't get looked at
				insnp.swap(nextp);
			}
			else
			{
				//update curr
				currp.swap(nextp);
			}
		}
		
		//setup objective
		//first path is the feasible path
		for(int pnum = 0; pnum < npaths; pnum++)
		{
			obj.setLinearCoef(bpaths[pnum], pcosts[pnum]);
		}
		//save entry paths
		bpi.entry_paths = currp;
		
		path_offset += npaths;
	}	
				
	//now fill out the boundary constraints
	for (int b = 0, nb = mcnf_cfg.blocks.size(); b < nb; b++)
	{
		mcnf_block *block = mcnf_cfg.blocks[b];
		BlockPathInfo& bpi = bpInfo[b];

		//entries
		for (int i = 0, n = block->entries.size(); i < n; i++)
		{
			mcnf_node *entry = block->entries[i];

			int i, nc;
			int row = priceMap[entry];
			nc = entry->commodities.size();
			/* for each commodity */
			for (i = 0; i < nc; i++)
			{
				int p, np;
				int c = entry->commodities[i];
				/* look at all the predecessors and enforce
				 the equality constraint  */
				np = entry->preds.size();
				for (p = 0; p < np; p++)
				{
					mcnf_node *pred = entry->preds[p]->src;
					int pindex = pred->cindex.get(c);
					if (pindex >= 0 && pred->type == nExit && pred->allocClass
							== entry->allocClass)
					{
						//support crossblock antis
						if (CrossBlockAntis && c < 0)
						{
							// these are <= constraints
							if (alloc_classes->getType(entry->allocClass)
									== AllocationClassInfo::AntiOnly)
							{
								rows[row].setLB(-IloInfinity);
							}
							else //inverse, mem node
							{
								row++;
								continue; //no actual constraint
							}
						}

						//sum of this quanties paths in pred must equal paths in node
						BitVector& node_paths =
								bpi.entry_paths[entry->allocClass];
						
						std::vector<QPathInfo> ranges;
						bpi.qMap.getRanges(c, ranges);
						hash_set<int> seenpaths;
						for (int r = 0, nr = ranges.size(); r < nr; r++)
						{
							int first = ranges[r].start;
							int end = ranges[r].end;

							for (int path = first; path < end; path++)
							{
								if (node_paths.test(path))
								{
									rows[row].setLinearCoef(
											paths[bpi.path_index_offset+path],
											1.0);
									seenpaths.insert(bpi.path_index_offset+path);
								}
							}
						}
						
						BlockPathInfo& pbpi = bpInfo[pred->block_num];
						BitVector& pred_paths =	pbpi.exit_paths[pred->allocClass];
						
						pbpi.qMap.getRanges(c, ranges);
						for (int r = 0, nr = ranges.size(); r < nr; r++)
						{
							int pfirst = ranges[r].start;
							int pend = ranges[r].end;

							for (int path = pfirst; path < pend; path++)
							{
								if (pred_paths.test(path))
								{
									if(seenpaths.count(pbpi.path_index_offset+path))
									{
										//apparently can't get/modify the coefficent,
										//so set to zero (1-1)
										rows[row].setLinearCoef(
												paths[pbpi.path_index_offset+path],
												0.0);
									}
									else
									{
										rows[row].setLinearCoef(
											paths[pbpi.path_index_offset+path],
											-1.0);
									}
								}
							}
						}
						row++;
					}
				}
			}
		}
		
		//exits
		for(int i = 0, n = block->exits.size(); i < n; i++)
		{
			mcnf_node *exit = block->exits[i];
			int i, nc;
			nc = exit->commodities.size();
			int row = priceMap[exit];

			/* for each commodity */
			for (i = 0; i < nc; i++)
			{
				int s, ns;
				int c = exit->commodities[i];

				/* look at all the successors and enforce
						the equality constraint  */
				ns = exit->succs.size();

				for (s = 0; s < ns; s++)
				{
					mcnf_node *succ = exit->succs[s]->dst;
					int sindex = succ->cindex.get(c);
					if (sindex >= 0 && succ->type == nEntry && succ->allocClass == exit->allocClass)
					{					
						//support crossblock antis
						if(CrossBlockAntis && c < 0)
						{									
							// these are <= constraints
							if(alloc_classes->getType(exit->allocClass) == AllocationClassInfo::Memory)
							{
								rows[row].setLB(-IloInfinity);
							}
							else  //inverse, anti node
							{
								row++;
								continue; //no actual constraint
							}
						}
								
						std::vector<QPathInfo> ranges;
						bpi.qMap.getRanges(c, ranges);
						hash_set<int> seenpaths;
						for (int r = 0, nr = ranges.size(); r < nr; r++)
						{
							int first = ranges[r].start;
							int end = ranges[r].end;

							//sum of this quanties paths in succ must equal paths in node
							BitVector& node_paths =
									bpi.exit_paths[exit->allocClass];
							for (int path = first; path < end; path++)
							{
								if (node_paths.test(path))
								{
									rows[row].setLinearCoef(
											paths[bpi.path_index_offset+path],
											1.0);
									seenpaths.insert(bpi.path_index_offset+path);
								}
							}
						}
						
						BlockPathInfo& sbpi = bpInfo[succ->block_num];
						BitVector& succ_paths = sbpi.entry_paths[succ->allocClass];
						
						sbpi.qMap.getRanges(c, ranges);
						for (int r = 0, nr = ranges.size(); r < nr; r++)
						{
							int sfirst = ranges[r].start;
							int send = ranges[r].end;
							for (int path = sfirst; path < send; path++)
							{
								if (succ_paths.test(path))
								{
									if(seenpaths.count(sbpi.path_index_offset+path))
									{
										rows[row].setLinearCoef(
												paths[sbpi.path_index_offset+path],
												0.0);
									}
									else
									{
										rows[row].setLinearCoef(
											paths[sbpi.path_index_offset+path],
											-1.0);
									}
								}
							}
						}
						row++;
					}				
				}
			}			
		}
		
	}
	
	//solve
	model.add(obj);
	model.add(rows);
	cplex.extract(model);
	
	if(DebugPRA > 14)
	{
		std::string name = fname + "_path.lp";
		cplex.exportModel(name.c_str());
	}

	// Optimize the problem and obtain solution.
	if ( !cplex.solve() )
	{
		env.error() << "Failed to optimize" << endl;
		abort();
	}

	std::cerr << "Num Rows: " << rows.getSize() << "  Num Cols: " << paths.getSize() << "\n";
	std::cerr << "Solution status = " << cplex.getStatus() << "\n";
	double solval = cplex.getObjValue();
	std::cerr << "Solution value  = " << solval << "\n";
	if(DebugPRA)
	{
		std::cerr << "Solution Time = " << cplex.getTime() << "\n";
		
		if(DebugPRA > 3)
		{
			std::string name = fname + "_path.sol";
			cplex.writeSolution(name.c_str());
		}
	}
		
	
	//extract price information
	extract_prices(cplex, rows);
	//get primal values
	extract_primals(cplex, paths);
	//clear out memory, eh, be nice to save what's useful
	cplex.end();
//	model = IloModel(env);
	cplex = IloCplex(env);
}

static void allocate_primals(mcnf_node *node)
{
	int nc = node->commodities.size();
	node->primal_solution = new double[nc];
	memset(node->primal_solution, 0, nc*sizeof(double));	
}

static void initialize_primals(mcnf_node *node)
{
	int nc = node->commodities.size();
	memset(node->primal_solution, 0, nc*sizeof(double));	
}

// use column generation to solve for prices to guide heuristic solver
// include all same priced paths as columns plus the feasible path
// We're basically doing Dantzig-Wolfe decomposition with no column removal here
int ColumnGenSolver::solve()
{
try{	
	double value;
	int k;
	int max_iter= MAX_ITER;
    struct timeval start, end;    

	double best_lower = -HUGE_VAL;
	int best_solution = INT_MAX;

	if (flag_prog_ra_iter >= 0)
		max_iter = flag_prog_ra_iter;

	best_solution = solver.construct_feasible_solution(true);
	solution.record_solution(mcnf_cfg);
	
	mcnf_cfg.traverse_nodes(allocate_primals);
	setupRows();
    if(DebugPRA) gettimeofday(&start,NULL);

	for (k = 1; k <= max_iter; k++)
	{
		int solution_value;

		/* find the value of all the shortest paths
		ignore bundle and boundary constraints */
		value = mcnf_cfg.shortest_balanced_paths();

		if (DebugPRA)
			fprintf(stderr, "Shortest value = %f\n", value);

		/* subtract the bundleconstraints multiplied by the price */
		double p_of_c = mcnf_cfg.price_of_constraints();
		value -= p_of_c;
		if(DebugPRA)
			fprintf(stderr, "Price of constraints = %f\n", p_of_c);
		if (value> best_lower)
			best_lower = value;
		
		prune_shortest_with_commodity_costs();
		//heavyweight price computation
		column_generate_prices();

		/* find a feasible solution, using prices as a guide */
		solution_value = solver.construct_feasible_solution();
		//mcnf_cfg.sanity_check_feasible_solution();

		if (solution_value < best_solution)
		{
			best_solution = solution_value;
			solution.record_solution(mcnf_cfg);
		}

		if (!finite(value))
			abort();


		if (DebugPRA)
		{
			fprintf(stderr, "****ITERATION %d *******\n", k);
			fprintf(stderr, "L(w) = %.20f\n", value);
			fprintf(stderr, "Lower bound = %.20f\n", best_lower);
			fprintf(stderr, "Feasible solution = %d\n", solution_value);
			fprintf(stderr, "Best solution = %d\n", best_solution);
		}

		if (!flag_prog_ra_max_iter && best_lower > (double)(best_solution - 1.0))
			break;
	}
	
    if(DebugPRA)
    {
        gettimeofday(&end,NULL);
        if(k == max_iter)
        	k--;

        double time = (double)(end.tv_sec - start.tv_sec) + (end.tv_usec-start.tv_usec)/1000000.0;
        std::cerr << "TIME = " << time << " per iteration " << time/k << "\n";
    }
    
	return best_solution;
}
catch (IloException& e)
{
	cerr << "Concert exception caught: " << e << endl;
	abort();
}	
	return 0;
}

//return true if no paths currently go through node
bool ColumnGenSolver::node_has_no_paths(mcnf_node *node)
{
	int nc = node->commodities.size();
	
	for(int i = 0; i < nc; i++)
	{
		if(node->feas_marked[i])
			return false;
		if(node->binfo[i].ratio > 0)
			return false;
	}
	return true;
}

//every capacity, dual, and boundary constraint corresponds to a row and has a price
//only consider nodes that have shortest or feasible paths through them
void ColumnGenSolver::setupRows()
{
	priceMap.clear();
	for(int b = 0, nb = mcnf_cfg.blocks.size(); b < nb; b++)
	{
		mcnf_block *block = mcnf_cfg.blocks[b];
		for(int n = 0, nn = block->allnodes.size(); n < nn; n++)
		{
			mcnf_node *node = block->allnodes[n];
	
//			if(node_has_no_paths(node))
//				continue;
			
			//capacity (price)
			if(node->capacity > 0)
			{
				int index = rows.getSize();
				priceMap[node] = index;
				rows.add(IloRange(env, 0, node->capacity, NULL));
				if(DebugPRA > 14)
				{
					std::string name = node->getName() + "_cap";
					rows[index].setName(name.c_str());
				}
			}
			
			//anti constraints (dual_price)
			/* handle dual constraints */
			if (node->dual_prices != NULL)
			{
				int i, nc;

				if (node->capacity >= 0)
					abort();

				priceMap[node] = rows.getSize();

				nc = node->commodities.size();
				/* for each positive commodity */
				for (i = 0; i < nc; i++)
				{
					int c = node->commodities[i];
					if (c > 0 && node->dual_constraint_matters(c))
					{
						/* get the index for the corresponding anti commodity */
						int ai = node->cindex.get(-c);
						if (ai >= 0)
						{
							rows.add(IloRange(env, 0, 1.0, NULL));
							
							if(DebugPRA > 14)
							{
								std::string name = node->getName() + "_" + itostr(c) + "_dual";
								rows[rows.getSize()-1].setName(name.c_str());
							}
						}						
					}
				}
			}
			
			if (node->boundary_prices != NULL)
			{
				if (node->type == nEntry)
				{
					int i, nc;
					priceMap[node] = rows.getSize();

					nc = node->commodities.size();
					// for each commodity 
					for (i = 0; i < nc; i++)
					{
						int p, np;
						int c = node->commodities[i];
						/* look at all the predecessors and enforce
						the equality constraint with the price */
						np = node->preds.size();
						for (p = 0; p < np; p++)
						{
							mcnf_node *pred = node->preds[p]->src;
							int pindex = pred->cindex.get(c);
							if (pindex >= 0 && pred->type == nExit && pred->allocClass == node->allocClass)
							{
								rows.add(IloRange(env, 0, 0, NULL));
								if(DebugPRA > 14)
								{
									std::string name = node->getName() + "_" + itostr(c) + "_bound_"+itostr(p);
									rows[rows.getSize()-1].setName(name.c_str());
								}
							}
						}
					}
				}
				else if (node->type == nExit)
				{
					int i, nc;
					priceMap[node] = rows.getSize();

					nc = node->commodities.size();
					/* for each commodity */
					for (i = 0; i < nc; i++)
					{
						int s, ns;
						int c = node->commodities[i];
						/* look at all the successors and enforce
						the equality constraint with the price */
						ns = node->succs.size();
						for (s = 0; s < ns; s++)
						{
							mcnf_node *succ = node->succs[s]->dst;
							int sindex = succ->cindex.get(c);
							if (sindex >= 0 && succ->type == nEntry && succ->allocClass == node->allocClass)
							{
								rows.add(IloRange(env, 0, 0, NULL));
								if(DebugPRA > 14)
								{
									std::string name = node->getName() + "_" + itostr(c) + "_bound_"+itostr(s);
									rows[rows.getSize()-1].setName(name.c_str());
								}
							}
						}
					}
				}
			}			
			
			if(node->type == nSink) //there are with the single path constraints go
			{
				assert(priceMap.count(node) == 0);
				priceMap[node] = rows.getSize();
				rows.add(IloRange(env, 1, 1, NULL));
				if(DebugPRA > 14)
				{
					std::string name = node->getName() + "_single";
					rows[rows.getSize()-1].setName(name.c_str());
				}
			}
		}	//nodes	
	}//blocks
	
}

//extract prices from solution and put them in the mcnf model
void ColumnGenSolver::extract_prices(IloCplex& cplex, IloRangeArray& rows)
{	
	for (int b = 0, nb = mcnf_cfg.blocks.size(); b < nb; b++)
	{
		mcnf_block *block = mcnf_cfg.blocks[b];
		for (int n = 0, nn = block->allnodes.size(); n < nn; n++)
		{
			mcnf_node *node = block->allnodes[n];

			//capacity (price)
			if (node->capacity > 0)
			{
				int row = priceMap[node];
				node->price = -cplex.getDual(rows[row]);
				
				if(DebugPRA > 1 && node->price != 0) std::cerr << node->getName() << "_cap " << node->price << "\n";

				if(node->price < 0)
					node->price = 0;
			}

			if(node->type == nSink) //reduced cost of commodities
			{
				int row = priceMap[node];
				double ccost = cplex.getDual(rows[row]);
				
				commodity_costs[node->position] = ccost;
				if(DebugPRA > 1)
				{
					std::string name = node->getName() + "_single ";
					std::cerr << name << cplex.getDual(rows[row]) << "\n";
				}
			}
			//anti constraints (dual_price)
			/* handle dual constraints */
			if (node->dual_prices != NULL)
			{
				int i, nc;

				if (node->capacity >= 0)
					abort();

				int row = priceMap[node];

				nc = node->commodities.size();
				/* for each positive commodity */
				for (i = 0; i < nc; i++)
				{
					int c = node->commodities[i];
					if (c > 0 && node->dual_constraint_matters(c))
					{
						/* get the index for the corresponding anti commodity */
						int ai = node->cindex.get(-c);
						if (ai >= 0)
						{
							node->dual_prices[i] = -cplex.getDual(rows[row]);							
							if(DebugPRA > 1 && node->dual_prices[i] != 0) std::cerr << node->getName() << "_" << c << " " << node->dual_prices[i] << "\n";
							if(node->dual_prices[i] < 0)
								node->dual_prices[i] = 0;
							row++;
						}
					}
				}
			}

			if (node->boundary_prices != NULL)
			{
				if (node->type == nEntry)
				{
					int i, nc;
					int row = priceMap[node];

					nc = node->commodities.size();
					// for each commodity 
					for (i = 0; i < nc; i++)
					{
						int p, np;
						int c = node->commodities[i];
						/* look at all the predecessors and enforce
						 the equality constraint with the price */
						np = node->preds.size();
						for (p = 0; p < np; p++)
						{
							mcnf_node *pred = node->preds[p]->src;
							int pindex = pred->cindex.get(c);
							if (pindex >= 0 && pred->type == nExit
									&& pred->allocClass == node->allocClass)
							{
								node->boundary_prices[p][i] = cplex.getDual(rows[row]);
								if(DebugPRA && node->boundary_prices[p][i] != 0)
									std::cerr << node->getName() << "_bound_" << p << "_" << c << " " << node->boundary_prices[p][i] <<"\n";
								row++;
							}
						}
					}
				}
				else if (node->type == nExit)
				{
					int i, nc;
					int row = priceMap[node];

					nc = node->commodities.size();
					/* for each commodity */
					for (i = 0; i < nc; i++)
					{
						int s, ns;
						int c = node->commodities[i];
						/* look at all the successors and enforce
						 the equality constraint with the price */
						ns = node->succs.size();
						for (s = 0; s < ns; s++)
						{
							mcnf_node *succ = node->succs[s]->dst;
							int sindex = succ->cindex.get(c);
							if (sindex >= 0 && succ->type == nEntry
									&& succ->allocClass == node->allocClass)
							{
								node->boundary_prices[s][i] = cplex.getDual(rows[row]);
								if(DebugPRA && node->boundary_prices[s][i] != 0)
									std::cerr << node->getName() << "_bound_" << s << "_" << c << " " << node->boundary_prices[s][i] <<"\n";

								row++;
							}
						}
					}
				}
			}
		} //nodes	
	}//blocks
}

//set the primal value of path p int bpi's block to be val
void ColumnGenSolver::set_path_primal(int q, int p, double val)
{
	std::list<mcnf_edge*>& edges = path_edges[p];

	if(DebugPRA > 8)
	{
		std::cerr << "Setting primal for path " << p << " for quantity " << q << " with value " << val << "\n";
	}
	
	for(std::list<mcnf_edge*>::iterator e = edges.begin(), end = edges.end(); e != end; e++)
	{
		mcnf_node *node = (*e)->dst;
		int index = node->cindex.get(q);		
		if(DebugPRA > 10)
		{
			std::cerr << "  " << node->getName();
		}
		node->primal_solution[index] += val;

	}
						
	if (DebugPRA > 10)
	{
		std::cerr << "\n";
	}
	
}

/* Follow all the paths in the feasible solution and mark them in model */
void ColumnGenSolver::extract_primals(IloCplex& cplex, IloNumVarArray& paths)
{
	IloNumArray vals(env);

	cplex.getValues(paths, vals);
	mcnf_cfg.traverse_nodes(initialize_primals);

	for (PathMap::iterator itr = pathMap.begin(), end = pathMap.end(); itr
			!= end; itr++)
	{
		int p = itr->first;
		int q = itr->second;
		double val = vals[p];
		if (val != 0)
		{
			set_path_primal(q, p, val);
		}
	}
}

//unmark shortest paths which are >= to the previous commodity cost of that
//quantity as they will not contribute to the basis
void ColumnGenSolver::prune_shortest_with_commodity_costs()
{
	int num_pruned = 0;
	if(commodity_costs.size() == 0)
		return;
	for (int b = 0, nb = mcnf_cfg.blocks.size(); b < nb; b++)
	{
		mcnf_block *block = mcnf_cfg.blocks[b];
		for(int s = 0, ns = block->sinks.size(); s < ns; s++)
		{
			mcnf_node *sink = block->sinks[s];
			int q = sink->position;
			assert(commodity_costs.count(q));
			double ccost = commodity_costs[q];			
			if(sink->sinfo[0][0].value >= ccost)
			{
				//pretend these paths don't exist
				sink->binfo[0].num_shortest = 0;
				sink->binfo[0].prevs = 0;
				sink->feas_marked[0] = false;
				num_pruned++;
			}
			else if(DebugPRA > 1)
			{
				std::cerr << "NotPruned " << q << " " << sink->sinfo[0][0].value << " " << ccost << "\n";
			}
		}
	}
	
	if(DebugPRA)
	{
		std::cerr << "NumPruned " << num_pruned << "\n";
	}
}

} //namespace

#else //CPLEX_AVAILABLE

namespace llvm
{

int ColumnGenSolver::solve()
{
	abort();
}


}
#endif
