#ifndef COLUMNGENSOLVER_H_
#define COLUMNGENSOLVER_H_

#include "llvm/PrincipledBackend/GlobalMCNF.h"
#include "llvm/PrincipledBackend/Solution.h"
#include <iostream>
#include <stdio.h>
#include <vector>
#include <list>
#include "llvm/Function.h"
#include "llvm/Module.h"
#include "llvm/ADT/StringExtras.h"
#include "llvm/Analysis/LoopInfo.h"
#include "BlockTraces.h"
#include "HeuristicSolver.h"

#ifdef CPLEX_AVAILABLE

#define IL_STD

#include <ilcplex/ilocplex.h>
ILOSTLBEGIN

#endif

/* Use column generation to compute prices in each iteration.
 *
 */
namespace llvm
{

class PathMap : public hash_map<int, int>
{
public:

	//add mappings [m..n)->q
	//at some point use a more efficient custom data structure
	void addRange(int m, int n, int q)
	{
		for(int i = m; i < n; i++)
			(*this)[i] = q;
	}
};

/* information on paths from a sink */
struct QPathInfo
{
	mcnf_node *sink;
	int start;
	int end;

	QPathInfo(mcnf_node *n, int s, int e): sink(n), start(s), end(e) {}
};
//mapping from q to paths of q
class Q2PathMap : public hash_map<int, std::vector<QPathInfo> >
{
public:
	void addRange(mcnf_node *sink, int m, int n)
	{
		int q = sink->position;
		(*this)[q].push_back(QPathInfo(sink,m,n));
	}

	void getRanges(int q, std::vector<QPathInfo>& ranges) const
	{
		assert(count(q));
		ranges = find(q)->second;
	}

	void getRange(mcnf_node *sink, int& m, int& n)
	{
		int q = sink->position;
		assert(count(q));
		const std::vector<QPathInfo>& ranges = find(q)->second;
		for(int r = 0, nr = ranges.size(); r < nr; r++)
		{
			if(ranges[r].sink == sink)
			{
				m = ranges[r].start;
				n = ranges[r].end;
				return;
			}
		}
		abort();
	}
};

struct BlockPathInfo
{
	int path_index_offset; //add this to each path for globally uniq id
	Q2PathMap qMap; //map q to path index range
	std::vector<BitVector> entry_paths;
	std::vector<BitVector> exit_paths;

};

class ColumnGenSolver
{
	GlobalMCNF& mcnf_cfg;
	HeuristicSolver solver;
	Solution& solution;
	std::string fname; //name of function
	std::vector< std::list<mcnf_edge*> > path_edges; //only store edges different from shortest
	PathMap pathMap; //map path index to q

	hash_map<int, double> commodity_costs;
#ifdef CPLEX_AVAILABLE
	IloEnv env;
	IloRangeArray rows;
	IloNumVarArray paths;
	IloObjective obj;
	IloModel model;
	IloCplex cplex;
	int path_offset;
#endif

	hash_map<mcnf_node*, int> priceMap; //a mapping from node to row index,
	//index first constraint relating to node (boundary constraints have multiple)

	std::vector<BlockPathInfo> bpInfo; //path info for each block, need to persist for boundary constraints
public:

	ColumnGenSolver(GlobalMCNF& mcfg, LoopInfo *li, Solution& sol): mcnf_cfg(mcfg), solver(&mcfg, li), solution(sol)
#ifdef CPLEX_AVAILABLE
	, rows(env), paths(env), obj(IloMinimize(env)), model(env), cplex(env), path_offset(0)
#endif
	{
		MachineFunction *MF = mcnf_cfg.MF;
		fname = MF->getFunction()->getParent()->getModuleIdentifier() + "_" + MF->getFunction()->getName();
		//setup BlockPathInfo
		bpInfo.resize(mcnf_cfg.blocks.size());
		solver.initialize_dual_prices();
	}

	virtual ~ColumnGenSolver()
	{
#ifdef CPLEX_AVAILABLE
		model.end();
		cplex.end();
#endif
	}

	//return cost of solution found
	int solve();

private:
#ifdef CPLEX_AVAILABLE

	//compute prices from reduced lp path formulation
	void column_generate_prices();

	void prune_shortest_with_commodity_costs();
	int compute_feas_cost(mcnf_node *sink);

	void update_paths(mcnf_node **level, const std::vector<BitVector>& curr, const std::vector<BitVector>& insn, std::vector<BitVector>& next, std::vector<int>& pcosts);
	void update_paths_node(mcnf_node *node, int q, const std::vector<BitVector>& curr, std::vector<BitVector>& next, std::vector<int>& pcosts);

	void set_constraints(mcnf_node *node, IloNumVarArray& bpaths, const BitVector& paths, const BlockPathInfo& bpi);
	void setupRows();
	void extract_prices(IloCplex& cplex, IloRangeArray& rows);
	void extract_primals(IloCplex& cplex, IloNumVarArray& paths);
	void set_path_primal(int q, int p, double val);

	bool node_has_no_paths(mcnf_node *node);
#endif
};

} //namespace
#endif /*COLUMNGENSOLVER_H_*/
