#include "llvm/PrincipledBackend/GlobalMCNF.h"
#include "llvm/ADT/hash_set"
#include "llvm/ADT/StringExtras.h"
#include "llvm/PrincipledBackend/AllocationClassInfo.h"
#include "llvm/CodeGen/LiveVariables.h"
#include "llvm/BasicBlock.h"
#include "llvm/Module.h"

#include <iostream>
#include <queue>

namespace llvm
{

enum AllowedMoves
{
	Full, LimitMoves, NoMoves
};

static cl::opt<AllowedMoves> MoveModel(cl::Optional, cl::desc(
		"What fidelity of model to use for the progressive allocator."), cl::init(
				LimitMoves), cl::values(clEnumValN(Full, "full-model", "Full model"),
		clEnumValN(LimitMoves,  "limit-moves",   "Limit moves to entry/exit of block"),clEnumValN (NoMoves,"no-moves", "Disallow all move insertion"),
		clEnumValEnd
		)
		);

cl::opt<bool> CPLEXCoalesce("cplex-coalesce", cl::Hidden, cl::desc(
		"Output coalescing problem as cplex file"));

cl::opt<bool> DisableMerging("disable-merging", cl::Hidden, cl::desc(
		"don't merge coalescable moves prior to allocation"));

#ifndef NDEBUG
cl::opt<int> DebugPRA("debugpra", cl::init(0), cl::Hidden, cl::desc(
		"Output progressive register allocator debug info to stderr"));
#else
const bool DebugPRA = 0;
#endif

#ifndef NDEBUG
cl::opt<bool> DebugPathsPRA("debug-paths-pra", cl::Hidden, cl::desc(
		"Output progressive register allocator path debug info to stderr"));
#else
const bool DebugPRA = 0;
#endif

const char *node_type_prefix[] =
{ "ERROR", /* nError */
"def", /* nDef */
"snk", /* nSink */
"xbr", /* nXBar */
"ins", /* nInsn */
"ent", /* nEntry */
"ext", /* nExit */
"ERROR" };

#define XBAR_TOP_LEVEL 0
#define XBAR_BOTTOM_LEVEL 4

bool node_order_compare(mcnf_node *n1, mcnf_node *n2)
{
	if (n1->node_order < n2->node_order)
		return true;
	return false;
}

bool reverse_node_order_compare(mcnf_node *n1, mcnf_node *n2)
{
	if (n1->node_order <= n2->node_order)
		return false;
	return true;
}
//cost of using q, which is valid at loc, along this edge
int mcnf_edge::getCost(int q)
{
	int ret = cost;
	if (!commodity_valid(q))
		return UNREACHABLE_INT;

	if (q > 0 && dst->nonanti_costs.size() > 0)
	{
		int pos = dst->cindex.get(q);
		ret += dst->nonanti_costs[pos];
	}

	return ret;

}

/* Return the price of commodity q using edge
 Returns UNREACHABLE if q can't use edge
 include price of destination node
 */
double mcnf_edge::getPrice(int q)
{
	mcnf_node *node;
	double ret;

	if (!commodity_valid(q))
		return UNREACHABLE;

	ret = getCost(q);
	node = dst;

	if (node->dual_prices != NULL)
	{
		int posq = abs(q);
		int i = node->cindex.get(posq);
		assert(i >= 0);
		ret += node->dual_prices[i];
	}

	if (node->boundary_prices != NULL)
	{
		int i = node->cindex.get(q);
		if (node->type == nEntry)
		{
			int p;
			int np = node->preds.size();
			for (p = 0; p < np; p++)
				ret += node->boundary_prices[p][i];
		}
		else if (node->type == nExit)
		{
			int s;
			int ns = node->succs.size();

			for (s = 0; s < ns; s++)
				ret += node->boundary_prices[s][i];
		}
	}

	if (q > 0) //antis are unaffected by capacity constraints
		ret += node->price;

	return ret;
}

//unique name for this edge, loc is where quantity is valid
//(only meaningful for crossblock edges)
std::string mcnf_edge::getName(int q)
{
	std::string sname = src->getName();
	std::string dname = dst->getName();
	std::string qname;
	if (q < 0)
		qname = "a" + itostr(-q);
	else
		qname = itostr(q);

	assert(commodity_valid(q) && "Invalid commodity in edge name");
	return sname + ";" + dname + ";" + qname;
}

//commodity q used on this edge, location is which node quantity is valid for
//(only meaningful for crossblock edges)
bool mcnf_edge::commodity_valid(int q)
{
	return commod_map.test(q);
}

//return true if q is set along this edge in the solution
bool mcnf_edge::is_set_solution(int q)
{
	int index1, index2;

	index1 = src->cindex.get(q);
	index2 = dst->cindex.get(q);

	if (index1 < 0 || index2 < 0)
		return false;

	if (src->feas_marked[index1] && dst->feas_marked[index2])
		return true;
	return false;
}
//make this edge a part of the solution for commodity q
void mcnf_edge::set_solution(int q)
{
	int index1, index2;

	if (src->block_num != dst->block_num)
		return; /* don't set crossblock edges */
	if (src->type == nExit && dst->type == nEntry)
		return; /* self loop */

	index1 = src->cindex.get(q);
	index2 = dst->cindex.get(q);

	if (index1 < 0 || index2 < 0)
		abort();

	src->feas_marked[index1] = true;
	dst->feas_marked[index2] = true;

	src->fsinfo[index1].next_edge = this;
	dst->fsinfo[index2].prev_edge = this;
}

/* Mark edge as being along the feasible path for q */
void mcnf_edge::mark_feasible_edge(int q)
{
	int dindex = dst->cindex.get(q);
	int sindex = src->cindex.get(q);

	if (dindex < 0)
		abort();
	dst->feas_marked[dindex] = true;
	dst->fsinfo[dindex].prev_edge = this;
	src->fsinfo[sindex].next_edge = this;

	if (DebugPRA > 9)
	{
		std::cerr << "Marking feasible edge " << getName(q) << "\n";
	}
	if (q > 0)
		dst->flowcnt++;
	else if (dst->mirror_node != NULL
			&& alloc_classes->getType(dst->allocClass)
					== AllocationClassInfo::AntiOnly)
		dst->flowcnt++;

	/* mark source only if it is a def, otherwise it will get handles
	 as a dst */
	if (src->type == nDef)
	{
		if (sindex < 0)
			abort();

		src->feas_marked[sindex] = true;

		if (q > 0)
			src->flowcnt++;
	}

	/* update priced cost */
	dst->fsinfo[dindex].priced_cost = src->fsinfo[sindex].priced_cost
			+ mcnf_edge_price(this, q);

	if (dst->type == nEntry)
	{
		/* add the costs of previous blocks to priced cost */
		int p, np;
		np = dst->preds.size();
		dst->fsinfo[dindex].priced_cost = 0;
		for (p = 0; p < np; p++)
		{
			mcnf_node *pred = dst->preds[p]->src;
			if (pred->type == nExit)
			{
				int index = pred->cindex.get(q);
				if (index >= 0 && pred->feas_marked[index])
					dst->fsinfo[dindex].priced_cost
							+= pred->fsinfo[index].priced_cost;
			}
		}
	}

	/* if quantity has already been allocated at boundary,
	 obey that allocation */
	if (src->used_classes != NULL)
	{
		int cap = alloc_classes->getCapacity(src->allocClass);
		if (src->used_classes[sindex].set)
		{
			if (src->used_classes[sindex].allocClass != src->allocClass)
				abort();
		}
		else if (cap >= 0 && src->used_classes[sindex].unusable)
			abort();
	}

	//don't check dst since that would hopefully be redundant and it
	//interferes with crossblock anti eviction

	/* sanity checks */
	if (dst->capacity > 0 && dst->flowcnt > dst->capacity)
		abort();
	if (!src->feas_marked[sindex])
		abort();

}

/* Remove an edge from the feasible path of q */
void mcnf_edge::unmark_feasible_edge(int q)
{
	if (DebugPRA > 9)
	{
		std::cerr << "Unarking feasible edge " << getName(q) << "\n";
	}
	int dindex = dst->cindex.get(q);
	int sindex = src->cindex.get(q);

	if (!dst->feas_marked[dindex])
		abort(); /* suppose to be marked */
	dst->feas_marked[dindex] = false;

	if (q > 0)
		dst->flowcnt--;
	else if (dst->mirror_node != NULL
			&& alloc_classes->getType(dst->allocClass)
					== AllocationClassInfo::AntiOnly)
		dst->flowcnt--;

	if (dst->flowcnt < 0)
		abort();

	src->fsinfo[sindex].next_edge = NULL;
	dst->fsinfo[dindex].prev_edge = NULL;

	/* mark source only if it is a def, otherwise it will get handles
	 as a dst */
	if (src->type == nDef)
	{
		if (!src->feas_marked[sindex])
			abort();
		src->feas_marked[sindex] = false;

		if (q > 0)
			src->flowcnt--;
		if (src->flowcnt < 0)
			abort();
	}

	/* sanity check */
	if (src->flowcnt < 0 || dst->flowcnt < 0)
		abort();
}

/* Returns the cost of the edge.. for when we need to be
 * able to pass a function pointer */
int mcnf_edge_cost(mcnf_edge *edge, int q)
{
	return edge->getCost(q);
}

/* Return the price of commodity q using edge
 Useful when a function pointer is needed.
 */
double mcnf_edge_price(mcnf_edge *edge, int q)
{
	return edge->getPrice(q);
}

/* Return true if it's okay to allocate q to node */
bool mcnf_boundary_node_valid(mcnf_node *node, int q)
{
	if (node->used_classes != NULL)
	{
		int index = node->cindex.get(q);
		int cap = alloc_classes->getCapacity(node->allocClass);
		if (node->used_classes[index].set)
		{
			if (node->used_classes[index].allocClass != node->allocClass)
			{
				return false;
			}
		}
		else if (cap >= 0 && node->used_classes[index].unusable)
			return false;
	}
	return true;
}

/* Return the price of an edge, but constrain a commodity
 to be in the correct place at the exit of a block if
 used_classes is set. For the simultaneous allocator, we
 assume variables are allocated to the right place initially,
 and then may be possibly evicted out of their current
 entry node. */
double mcnf_edge_price_fixed_boundary_exits(mcnf_edge *edge, int q)
{
	double ret = 0;
	mcnf_node *src = edge->src;
	mcnf_node *dst = edge->dst;

	/* if quantity has already been allocated at boundary,
	 obey that allocation */
	if (src->type == nExit && !mcnf_boundary_node_valid(src, q))
	{
		return UNREACHABLE;
	}

	if (dst->type == nExit && !mcnf_boundary_node_valid(dst, q))
	{
		return UNREACHABLE;
	}

	return mcnf_edge_price(edge, q) + ret;
}

//an unique name for this node
std::string mcnf_node::getName()
{
	const char *prefix = node_type_prefix[(int) type];
	std::string ret = std::string(prefix) + "_" + itostr(block_num) + "_"
			+ itostr(position) + "_" + itostr(subindex) + "_" + itostr(
			allocClass);

	for (unsigned i = 0, n = ret.length(); i < n; i++)
	{
		if (ret[i] == '-') //minus signs aren't valid characters
			ret[i] = 'n';
	}

	return ret;
}

//return the index of edge in this node's preds
int mcnf_node::pred_index(mcnf_edge *edge)
{
	int i;
	int npreds = preds.size();

	for (i = 0; i < npreds; i++)
	{
		if (preds[i] == edge)
		{
			return i;
		}
	}
	abort();
	return 0;
}

//return true if node's class field is meaningful
bool mcnf_node::has_class()
{
	switch (type)
	{
	case nXBar:
	case nInsn:
	case nEntry:
	case nExit:
		return true;
	default:
		return false;
	}
}

/* Return the total amount of flow in a node, assumes
 balanced paths have been computed.
 */
double mcnf_node::node_flow()
{
	int i;
	int nc = commodities.size();

	double sum = 0.0;
	/* TODO: cache this info */
	for (i = 0; i < nc; i++)
	{
		if (commodities[i] > 0)
			sum += binfo[i].ratio;
	}

	return sum;
}

/*
 Return true if we should not ignore the dual constraint
 for commodity c on node.  Only constrain var/antis when
 a var can be stored.
 */
bool mcnf_node::dual_constraint_matters(int c)
{
	int p;

	if (dual_prices == NULL)
		return false;

	int np = preds.size();

	for (p = 0; p < np; p++)
	{
		mcnf_edge *edge = preds[p];

		if (edge->src->type == nXBar && alloc_classes->getType(
				edge->src->allocClass) == AllocationClassInfo::Register
				&& edge->commodity_valid(c))
			return true;
	}
	return false;
}

/* destroy memory allocated for a node */
mcnf_node::~mcnf_node()
{
	//sinfo, binfo, and fsinfo are pool allocated
	if (dual_prices != NULL)
		delete[] dual_prices;
	if (dual_gradients != NULL)
		delete[] dual_gradients;

	if (boundary_prices)
	{
		int i;
		int n = 0;
		if (type == nEntry)
			n = preds.size();
		else if (type == nExit)
			n = succs.size();
		else
			abort();

		for (i = 0; i < n; i++)
		{
			delete[] boundary_prices[i];
			delete[] boundary_gradients[i];
		}
		delete[] boundary_prices;
		delete[] boundary_gradients;
	}

	if (primal_solution != NULL)
		delete[] primal_solution;
	if (used_classes != NULL)
		delete[] used_classes;
}

//return true if can insert edges between this block and successors
//this isn't possible if the control isn't statically determined (ie jump tables)
bool mcnf_block::edge_insertable()
{
	RAATInfo *last = &RAATs.back();
	if (last->isControlFlow() && last->getNumOps() > 0)
		return false;
	return true;
}

void mcnf_block::compute_entry_exit_constraint_helper(mcnf_block *block,
		hash_map<mcnf_block*, char>& visited, enum entry_exit type,
		bool started_at_entry)
{
	const int seenEntryMask = 1;
	const int seenExitMask = 2;
	int len, i;

	if (started_at_entry && type == Exit && block == this)
		entry_exit_constrained = true;
	if (!started_at_entry && type == Entry && block == this)
		entry_exit_constrained = true;

	if (visited.count(block) == 0)
		visited[block] = 0;

	if (type == Entry)
	{
		if (visited[block] & seenEntryMask)
			return;
		visited[block] |= seenEntryMask;
	}
	else
	{
		if (visited[block] & seenExitMask)
			return;
		visited[block] |= seenExitMask;
	}
	//record this block as being constrained
	if (started_at_entry)
		blocks_constrained_by_entry.push_back(constrained_block(block, type));
	else
		blocks_constrained_by_exit.push_back(constrained_block(block, type));

	/* now look at all preds/succ */
	if (type == Entry)
	{
		len = block->preds.size();
		for (i = 0; i < len; i++)
		{
			mcnf_block *pred = block->preds[i];
			compute_entry_exit_constraint_helper(pred, visited, Exit,
					started_at_entry);
		}
	}
	else
	{
		len = block->succs.size();
		for (i = 0; i < len; i++)
		{
			mcnf_block *succ = block->succs[i];
			compute_entry_exit_constraint_helper(succ, visited, Entry,
					started_at_entry);
		}
	}
}

/* Compute and store what blocks are constrained by this blocks
 * entry/exit allocations.  Also computes is_self loop.
 */
void mcnf_block::compute_entry_exit_constraints()
{
	hash_map<mcnf_block*, char> visited;
	compute_entry_exit_constraint_helper(this, visited, Entry, true);
	visited.clear();
	compute_entry_exit_constraint_helper(this, visited, Exit, false);

	int i;
	int len = preds.size();
	for (i = 0; i < len; i++)
	{
		mcnf_block *pred = preds[i];
		if (pred == this)
			is_self_loop = true;
	}
}

//make sure that a feasible solution is correctly linking exit
//and entry nodes
void mcnf_block::check_feasible_links()
{
	unsigned i, n;
	for (i = 0, n = exits.size(); i < n; i++)
	{
		mcnf_node *node = exits[i];
		unsigned c, nc;
		for (c = 0, nc = node->commodities.size(); c < nc; c++)
		{
			int q = node->commodities[c];
			if (node->feas_marked[c])
			{
				//keep track of what successor blocks that q is live into we've
				//seen, 1 if seen, 2 if actual feasible solution
				hash_map<int, int> seen;
				//look at all successors
				unsigned s, ns;
				for (s = 0, ns = node->succs.size(); s < ns; s++)
				{
					mcnf_edge *succ = node->succs[s];
					mcnf_node *snode = succ->dst;
					if (snode->type == nEntry)
					{
						int index = snode->cindex.get(q);
						if (index >= 0)
						{ //quantity live
							if (snode->feas_marked[index])
							{
								if (seen.count(snode->block_num)
										&& seen[snode->block_num] == 2)
									abort();
								else
									seen[snode->block_num] = 2;
							}
							else if (!seen.count(snode->block_num))
							{
								seen[snode->block_num] = 1;
							}
						}
					}
				}
				for (hash_map<int, int>::iterator itr = seen.begin(), end =
						seen.end(); itr != end; itr++)
				{
					if (itr->second == 1)
					{
						std::cerr << "Lack feasible link from "
								<< node->block_num << " to " << itr->first
								<< " for quantity " << q << " in class "
								<< node->allocClass << "\n";
						abort();
					}
				}
			}
		}
	}
#if 0
	//somewhat redundant if called on all blocks, but check entry->exit
	for (i = 0, n = entries.size(); i < n; i++)
	{
		mcnf_node *node = entries[i];
		unsigned c, nc;
		for (c = 0, nc = node->commodities.size(); c < nc; c++)
		{
			int q = node->commodities[c];
			if (node->feas_marked[c])
			{
				//look at all preds
				unsigned p, np;
				for (p = 0, np = node->preds.size(); p < np; p++)
				{
					mcnf_edge *pred = node->preds[p];
					mcnf_node *pnode = pred->src;
					if (pnode->type == nExit)
					{
						int index = pnode->cindex.get(q);
						if (index >= 0 && !pnode->feas_marked[index])
						abort();
					}
				}
			}
		}
	}
#endif
}

//print quantity q (with reg names) to out
void mcnf_block::print_quantity(std::ostream& out, int q)
{
	out << itostr(q) << "(";
	if (q < 0)
		q = -q;
	for (unsigned i = 0, n = cfg->quantities[q].regs.size(); i < n; i++)
	{
		out << itostr(cfg->quantities[q].regs[i]) << " ";
	}
	out << ")";
}

/* Return true if level contains entry nodes. Returns false for a null level */
bool mcnf_block::is_entry_level(mcnf_node **level)
{
	int i;
	bool ret = false;
	int numClasses = alloc_classes->numClasses();
	for (i = 0; i < numClasses; i++)
	{
		mcnf_node *node = level[i];
		if (node != NULL && node->type != nEntry)
			return false;
		else if (node != NULL)
			ret = true;
	}

	return ret;
}

/* Return true if level contains exit nodes. Returns false for a null level */
bool mcnf_block::is_exit_level(mcnf_node **level)
{
	int i;
	int ret = 0;
	int numClasses = alloc_classes->numClasses();
	for (i = 0; i < numClasses; i++)
	{
		mcnf_node *node = level[i];
		if (node != NULL && node->type != nExit)
			return false;
		else if (node != NULL)
			ret = true;
	}

	return ret;
}

/* Return true if level contains insn nodes. Returns false for a null level */
bool mcnf_block::is_insn_level(mcnf_node **level)
{
	int i;
	int ret = 0;
	int numClasses = alloc_classes->numClasses();
	for (i = 0; i < numClasses; i++)
	{
		mcnf_node *node = level[i];
		if (node != NULL && node->type != nInsn)
			return false;
		else if (node != NULL)
			ret = true;
	}

	return ret;
}

//fill out level with exit nodes
void mcnf_block::get_exit_level(mcnf_node **level)
{
	int numClasses = alloc_classes->numClasses();
	memset(level, 0, sizeof(mcnf_node*) * numClasses);

	for (int i = 0, n = exits.size(); i < n; i++)
	{
		mcnf_node *node = exits[i];
		assert(node->allocClass >= 0 && level[node->allocClass] == NULL);
		level[node->allocClass] = node;
	}
}

/*
 * gets the level starting at position start, ignoring entry nodes
 */
int mcnf_block::get_level(int start, mcnf_node **nextnodes,
		mcnf_node **insnnodes)
{
	int i;
	std::vector<mcnf_node*>& vec = allnodes;
	int sz = vec.size();
	int position = -10;
	int subindex = -10;
	int numClasses = alloc_classes->numClasses();

	memset(insnnodes, 0, sizeof(mcnf_node*) * numClasses);
	memset(nextnodes, 0, sizeof(mcnf_node*) * numClasses);

	for (i = start; i < sz; i++)
	{
		mcnf_node *node = vec[i];
		if (node->type == nEntry)
			; /* ignored */
		else if (node->type == nInsn)
		{
			if (position == -10)
			{
				position = node->position;
				subindex = 0;
			}
			if (node->position == position && subindex == 0)
			{
				insnnodes[node->allocClass] = node;
			}
			else if (node->position > position)
				break;
		}
		else if (node->type == nXBar)
		{
			if (position == -10)
			{
				position = node->position;
				subindex = node->subindex;
			}
			if (node->position == position && node->subindex == subindex)
			{
				nextnodes[node->allocClass] = node;
			}
			else if (node->position > position || (node->position == position
					&& node->subindex > subindex))
				break;
		}
		else if (node->type == nExit)
		{
			if (position == -10)
				nextnodes[node->allocClass] = node;
			else
				break;
		}
	}

	if (i == sz)
		return -1;
	else
		return i;
}

//return the level immediately previous to level in level
void mcnf_block::set_to_previous_level(mcnf_node **level)
{
	//find first node
	int numClasses = alloc_classes->numClasses();

	mcnf_node *node = NULL;
	for (int i = 0, n = numClasses; i < n; i++)
	{
		if (level[i] != NULL)
		{
			node = level[i];
			break;
		}
	}
	assert(node != NULL);

	//clear out level
	memset(level, 0, sizeof(mcnf_node*) * numClasses);

	//the look backwards in allnodes for prev level

	int pos = node->vecpos - 1;
	int position = -2;
	int subindex = -1;
	enum mcnf_node_type type = nError;

	for (; pos >= 0; pos--)
	{
		mcnf_node *node = allnodes[pos];
		if (!node->has_class())
			continue;
		if (type == nError)
		{
			//found our level
			type = node->type;
			position = node->position;
			subindex = node->subindex;
		}
		else if (node->position != position || node->subindex != subindex
				|| node->type != type)//are we done?
		{
			break;
		}
		//have a node at appropriate level
		level[node->allocClass] = node;
	}
}

/* Do the necessary initialization for the path algorithms on all nodes
 clears the visited bit
 sets the balanced info to zero
 */
void mcnf_block::initialize_path_info()
{
	int n;
	int nn = allnodes.size();

	for (n = 0; n < nn; n++)
	{
		mcnf_node *node = allnodes[n];

		node->visited = VISITED_CLEAR;
		memset(node->binfo, 0, sizeof(balanced_info) * node->commodities.size());
	}

	quantity_prices.clear();
}

/* For debugging, go through the block and print shortest
 * path costs for quantity q
 */
void mcnf_block::debug_print_quantity_shortest(int q)
{
	int numClasses = alloc_classes->numClasses();

	mcnf_node **nodes = (mcnf_node**) alloca(numClasses *sizeof(mcnf_node*));
	mcnf_node **insn_nodes = (mcnf_node**) alloca(numClasses *sizeof(mcnf_node*));

	memset(nodes, 0, sizeof(mcnf_node*) * numClasses);
	memset(insn_nodes, 0, sizeof(mcnf_node*) * numClasses);

	std::cerr << "Shortest path info for quantity " << q;
	std::cerr << " in block " << index << "\n";
	int i;
	int n = entries.size();
	for (i = 0; i < n; i++)
	{
		mcnf_node *node = entries[i];
		nodes[node->allocClass] = node;
	}

	int levelpos = 0;
	int position = -1;
	while (!is_exit_level(nodes) && levelpos >= 0)
	{
		/*, compute next and insn nodes */
		levelpos = get_level(levelpos, nodes, insn_nodes);
		bool atlevel = false;
		for (i = 0; i < numClasses; i++)
		{
			if (nodes[i])
			{
				int pos = nodes[i]->cindex.get(q);
				if (pos >= 0)
				{
					if (!atlevel)
					{
						//first q node at this level
						atlevel = true;
						if (nodes[i]->position > position)
						{
							position = nodes[i]->position;
							if (position < (int) RAATs.size())
							{
								RAATs[position].dump();
								std::cerr << "\n";
							}
						}
						std::cerr << nodes[i]->position << ";"
								<< nodes[i]->subindex << ":";
					}
					//print out info for this quantity's shortest path at this node
					std::cerr << " (" << alloc_classes->getName(i) << ", ";
					std::cerr << nodes[i]->sinfo[0][pos].value;
					//std::cerr << ", " << nodes[i]->sinfo[0][pos].cost;
					mcnf_edge *prev = nodes[i]->sinfo[0][pos].prev_edge;
					if (prev)
						std::cerr << ", " << alloc_classes->getName(
								prev->src->allocClass);
					else
						std::cerr << ", NULL";
					std::cerr << ")";
				}
			}
		}
		if (atlevel)
			std::cerr << "\n";
	}

}

/* after the shortest path costs have been set,
 backwards propagate teh shortest path count information
 so all nodes have the right info */
void mcnf_block::set_balanced_flows()
{
	double ratio, pratio, npaths;
	int i, sn;

	sn = sinks.size();

	/* add the sinks to the queue */
	std::priority_queue<mcnf_node*, std::vector<mcnf_node*>, node_compare_fn>
			nodeq(node_order_compare);

	for (i = 0; i < sn; i++)
	{
		mcnf_node *node = sinks[i];

		if (node->visited != VISITED_SHORTEST_DONE)
			abort();
		node->visited = VISITED_BALANCED_QUEUE;
		if (node->commodities.size() != 1)
			abort(); /* it's a sink */

		node->binfo[0].ratio = 1.0; /* all shortest paths go through here */
		nodeq.push(node);
	}

	while (!nodeq.empty())
	{
		int nc;
		mcnf_node *node = nodeq.top();
		nodeq.pop();;

		node->visited = VISITED_BALANCED_DONE;
		/* upon exit from the queue, it is assumed that
		 node has the correct ratio value
		 the num_shortest value is still the old value however
		 (ie, should equal the sum of the num_shortest of hte predecessors)
		 */

		/* for every commodity */
		nc = node->commodities.size();
		for (i = 0; i < nc; i++)
		{
			int p, np;
			int c = node->commodities[i];
			int predmask = node->binfo[i].prevs;
			ratio = node->binfo[i].ratio;
			npaths = node->binfo[i].num_shortest;

			if (ratio == 0) /* not the shortest path for this commodity */
				continue;

			np = node->preds.size();

			if (node->type == nDef && fabs(node->binfo[0].ratio - 1.0)
					> THRESHOLD)
				abort();

			/* look at each predecessor node for this commodity */
			if (np >= 32)
			{
				mcnf_node *pred;
				/* handle nodes with lots of predecessors (ie, entry nodes)
				 we can only support this case if there is only one previous node though */
				if (predmask != 0)
					abort();

				pred = node->sinfo[0][i].prev_edge->src;

				if (pred != NULL)
				{
					int pindex = pred->cindex.get(c);

					if (pred->block_num != index)
						abort(); /* shortest path should stop at boundaries */

					/* compute it's contribution to the flow through this node
					 and multiply by this node's current ratio */
					pred->binfo[pindex].ratio += ratio
							* pred->binfo[pindex].num_shortest / npaths;
					pratio = pred->binfo[pindex].ratio;

					if (!finite(pratio) || pratio > (1.0 + THRESHOLD) || pratio
							< 0)
						abort();

					/* add the predecessor to the queue if it's not already there */
					if (pred->visited != VISITED_BALANCED_QUEUE)
					{
						pred->visited = VISITED_BALANCED_QUEUE;
						nodeq.push(pred);
					}
				}
			}
			else
			{
				for (p = 0; p < np; p++)
				{
					if (predmask & (1 << p))
					{
						mcnf_node *pred = node->preds[p]->src;
						int pindex = pred->cindex.get(c);

						if (pred->block_num != index)
							abort(); /* shortest path should stop at boundaries */

						/* compute it's contribution to the flow through this node
						 and multiply by this node's current ratio */
						pred->binfo[pindex].ratio += ratio
								* pred->binfo[pindex].num_shortest / npaths;
						pratio = pred->binfo[pindex].ratio;

						if (!finite(pratio) || pratio > (1.0 + THRESHOLD)
								|| pratio < 0)
							abort();

						/* add the predecessor to the queue if it's not already there */
						if (pred->visited != VISITED_BALANCED_QUEUE)
						{
							pred->visited = VISITED_BALANCED_QUEUE;
							nodeq.push(pred);
						}
					}
				}
			}
		}
	}

	if (DebugPathsPRA)
	{
		for (i = 0; i < sn; i++)
		{
			mcnf_node *node = sinks[i];
			std::cerr << "npaths " << node->commodities[0] << ": "
					<< node->binfo[0].num_shortest << "\n";
		}
	}
}

int offset_to_xbar_middle = (XBAR_BOTTOM_LEVEL+4);

int compute_interval_level(int position, int subindex, bool isXBar)
{
	int ret = position+2;
	ret *= (XBAR_BOTTOM_LEVEL+2);
	if(isXBar) ret++;
	ret += subindex;
	return ret;
}

static void compute_from_interval_level(int level, int& position, int& subindex, bool& isXBar)
{
	position = (level/(XBAR_BOTTOM_LEVEL+2))-2;
	subindex = 0;
	int rem = level%(XBAR_BOTTOM_LEVEL+2);
	if(rem == 0)
	{
		isXBar = false;
	}
	else
	{
		isXBar = true;
		subindex = rem-1;
		//slight hack to deal with gaps in zigzag
		if(subindex > 1 && subindex < XBAR_BOTTOM_LEVEL)
			subindex = XBAR_BOTTOM_LEVEL;
	}
}
//compute a level numberfor a node
int getNodeIntervalLevel(mcnf_node *node)
{
	return compute_interval_level(node->position, node->subindex, node->type == nXBar);
}

/* return the position for the given lvel */
int getIntervalLevelPosition(int level)
{
	int p, s;
	bool isX;
	compute_from_interval_level(level, p, s, isX);
	return p;
}

/* return subindex for a level */
int getIntervalLevelSubindex(int level)
{
	int p, s;
	bool isX;
	compute_from_interval_level(level, p, s, isX);
	return s;
}

//xbar or insn?
enum mcnf_node_type getIntervalLevelType(int level)
{
	int p, s;
	bool isX;
	compute_from_interval_level(level, p, s, isX);
	return isX ? nXBar : nInsn;
}

// return the last position in block where r is available for q for an interval
//starting at start with a max last position of end
//returns -1 if reg isn't available at start
int mcnf_block::lastAvailablePosForQuantityInReg(int q, unsigned reg, int start, int end, bool isExit)
{
	int lastPos = end;

	//first check hard regs
	HardUsageMap& hmap = hardRegUsageMap[reg];

	if(hmap.size() != 0)
	{
		HardUsageMap::iterator posi = hmap.upper_bound(start);
		//assume map is properly condensed

		//upper bound has to be after end
		if(posi != hmap.end() && posi->first <= lastPos)
		{
			if(!posi->second) //was unavailable at start
				return -1;
			lastPos = posi->first-1;
		}
		//and previous interval needs to be available
		if (posi != hmap.begin())
		{
			posi--;
			if (posi->second) //true means reg is in use
				return -1;
		}
	}

	if(isExit && !regAvailableForQuantityAtOut(reg, q))
	{
		//can't end allocation with this quantity
		int endPos = compute_interval_level(lastXBarPos, 1, true);
		if(endPos < start) //need to find reg available after last xbar
			return -1;
		if(endPos < lastPos)
			lastPos = endPos;
	}

	// what about other quantities?
	if(regUsageMap.count(reg))
	{
		UsageMap& map = regUsageMap[reg];
		if(map.size() == 0)
			return lastPos;
		UsageMap::iterator posi = map.upper_bound(start);
		if(posi != map.end() && posi->first <= lastPos)
		{
			if(posi->second != 0)
				lastPos = posi->first-1;
			else  //a gap, see where end is
			{
				UsageMap::iterator nexti = posi;
				nexti++;
				if(nexti != map.end() && nexti->first <= lastPos)
				{
					lastPos = nexti->first-1;
				}
			}
		}
		if(posi == map.begin())
			return lastPos;
		posi--;
		if(posi->second != 0 && posi->second != q)
		{
			if(DebugPRA > 5) std::cerr << "lastAvailable unavail reg " << reg << " for q " << q << " at " << posi->first << " " << posi->second << "\n";
			return -1; //not available at start
		}
	}

	return lastPos;
}

unsigned mcnf_block::getRegisterForQuantity(int q, mcnf_node *node, int offset /*=0*/)
{
	int level = getNodeIntervalLevel(node)+offset;
	return getRegisterForQuantityAtPos(q, level);
}
//functions for recording reg allocation of quantities in multireg classes
unsigned mcnf_block::getRegisterForQuantityAtPos(int q, int level)
{
	if(q < 0)
		return 0;
	if(quantityRegMap.count(q))
	{
		RegMap& map = quantityRegMap[q];
		if(map.size() == 0)
			return 0;
		RegMap::iterator pos = map.lower_bound(level);
		if(pos == map.end() || pos->first != level) //overshot
		{
			if(pos == map.begin())
				return 0;
			pos--;
		}
		return pos->second;
	}
	else
	{
		return 0;
	}
}

unsigned mcnf_block::getRegisterForQuantityIn(int q) //register into block
{
	if(q < 0)
		return 0;
	if(quantityRegMap.count(q))
	{
		RegMap& map = quantityRegMap[q];
		assert(map.begin() != map.end());
		return map.begin()->second;
	}
	else
	{
		return 0;
	}
}

unsigned mcnf_block::getRegisterForQuantityOut(int q) //register out of block
{
	if(q < 0)
		return 0;
	if(quantityRegMapOut.count(q))
	{
		return quantityRegMapOut[q];
	}
	else
	{
		return 0;
	}
}

void mcnf_block::printMaps()
{
	std::cerr << "quantityRegMap:\n";
	for(hash_map<int, RegMap>::iterator qitr = quantityRegMap.begin(), eitr = quantityRegMap.end(); qitr != eitr; qitr++)
	{
		int q = qitr->first;
		RegMap& map = qitr->second;
		std::cerr << q << ":";
		for(RegMap::iterator itr = map.begin(), e = map.end(); itr != e; itr++)
		{
			int pos = itr->first;
			unsigned r = itr->second;
			std::cerr << " (" << pos << "," << r << ")";
		}
		std::cerr << "\n";
	}

	std::cerr << "\nregUsageMap:\n";
	for(hash_map<unsigned, UsageMap>::iterator ritr = regUsageMap.begin(), eitr = regUsageMap.end(); ritr != eitr; ritr++)
	{
		unsigned r = ritr->first;
		UsageMap& map = ritr->second;
		std::cerr << r << ":";
		for(UsageMap::iterator itr = map.begin(), e = map.end(); itr != e; itr++)
		{
			int pos = itr->first;
			int q = itr->second;
			std::cerr << " (" << pos << "," << q << ")";
		}
		std::cerr << "\n";
	}
	std::cerr << "\n";
}

void mcnf_block::setQuantityReg(int q, unsigned r, mcnf_node *node, int offset /* = 0 */)
{
	if (q < 0)
		return;
	int pos = getNodeIntervalLevel(node) + offset;
	setQuantityRegAtPos(q, r, pos);
}
//set r to be the reg for q starting at node's position,
//does very little checking, allows insertions into the middle
void mcnf_block::setQuantityRegAtPos(int q, unsigned r, int pos)
{
	if(q < 0) return;
	RegMap& map = quantityRegMap[q];
	UsageMap& umap = regUsageMap[r];
	RegMap::iterator last = map.upper_bound(pos);

	//mark whatever q was allocated to at  pos as being free
	unsigned prevr = getRegisterForQuantityAtPos(q, pos);
	if(DebugPRA > 9) std::cerr << "prevr for q " << q << " is " << prevr << " at pos " << pos << "\n";
	if (prevr != 0 && prevr != r)
	{
		if(regUsageMap.count(prevr) == 0)
		{
			prevr = getRegisterForQuantityAtPos(q, pos);
			abort();
		}
		if(regUsageMap[prevr].count(pos) && regUsageMap[prevr][pos] != q)
		{
			//we jsut plugged the whole, the allocation is already changing so
			//no need to explicitly clear it
		}
		else
		{
			if(DebugPRA > 9)
				std::cerr << "Removign prevr " << prevr << " for q " << q << " at pos " << pos << "\n";
			regUsageMap[prevr][pos] = 0;
		}
	}

	//update usage of register r
	if (r != 0)
	{
		UsageMap::iterator posi = umap.lower_bound(pos);

		if(posi == umap.end())
		{
			//nothing recorded at this position, check to see if marking
			//would be redundant
			if(posi == umap.begin())
			{
				//nothing before, just mark
				umap[pos] = q;
			}
			else
			{
				UsageMap::iterator prev = posi;
				prev--;
				if(prev->second != q)
				{
					umap[pos] = q; //a change
				}
			}
		}
		else if(posi->first == pos) //something there
		{
			if(posi->second != 0 && posi->second != q)
			{
				std::cerr << "Register " << r << " at pos " << pos << " already has quantity " << umap[pos] << "\n";
				abort();
			}

			if (posi == umap.begin())
			{
				umap[pos] = q;
			}
			else
			{
				//check to see if this mark is necessary
				UsageMap::iterator prev = posi;
				prev--;
				if (prev->second == q)
				{
					//no need for a mark here
					umap.erase(posi);
				}
				else
				{
					posi->second = q;
				}
			}
		}
		else //overshot, nothing at pos, but something is after
		{
			UsageMap::iterator prev = posi;
			prev--;
			if (prev->second == q)
			{
				//no need for a mark at pos
			}
			else
			{
				umap[pos] = q; //insert a mark
			}

		}
	}

	if(map.size() == 0)
	{
		map[pos] = r;
	}
	else if(map.count(pos) && map.size() > 1)
	{
		//something at this position already, have to do more work to see
		//if a mark here is actually necessary
		RegMap::iterator posi = map.lower_bound(pos);
		RegMap::iterator prev = posi;
		prev--;
		if(prev->second == r)
		{
			//no need for a mark at pos
			map.erase(posi);
		}
		else
		{
			map[pos] = r;
		}
	}
	else if(map.rbegin()->second != r) //keep map sparse as possible
	{
		map[pos] = r;
	}

	if(DebugPRA > 9)
		std::cerr << "QRegSet in block " << index << " for " << q << " to " << r <<  " pos " << pos << "\n";
	if(DebugPRA > 100)
		printMaps();
}

//set regin with no consistency check
void mcnf_block::setQuantityRegIn(int q, unsigned r)
{
	if(q < 0) return;
	RegMap& map = quantityRegMap[q];
	UsageMap& umap = regUsageMap[r];

	//clear all regs that q uses in this block
	for(RegMap::iterator itr = map.begin(); itr != map.end(); itr++)
	{
		int pos = itr->first;
		unsigned reg = itr->second;
		if(reg != 0)
		{
			UsageMap& umapq = regUsageMap[reg];
			if(umapq.count(pos))
				umapq.erase(pos);
		}
	}
	map.clear();
	map[0] = r;
	umap[0] = q;

	if(DebugPRA > 9)
		std::cerr << "QRegIn in block " << index << " for " << q << " to " << r << "\n";
}

/* Return true if no quantities are currently allocated to reg at
 * node's level
 */
bool mcnf_block::regAvailableForQuantityAtNode(unsigned reg, int q, mcnf_node *node)
{
	int pos = getNodeIntervalLevel(node);
	if(reg == 0)
		return true;

	if(getHardRegUnavailableAtNode(reg, node))
		return false;

	if(regUsageMap.count(reg))
	{
		UsageMap& map = regUsageMap[reg];
		if(map.size() == 0)
			return true;
		UsageMap::iterator posi = map.upper_bound(pos);
		//pos is now past where we want to look, so adjust
		if(posi == map.begin())
			return true; //nothing set before upperbound
		int regq = 0;
		//backup one
		posi--;
		regq = posi->second;

		if(DebugPRA > 9)
			std::cerr << "Reg " << reg << " Unavailable at pos " << pos << " due to q " << posi->second << " starting at " << posi->first << "\n";
		return regq == 0 || regq == q;
	}
	else
	{
		return true;
	}
}


/* Return true if no quantities are currently allocated to reg at all the levels between
 * start and end inclusive
 */
bool mcnf_block::regAvailableForQuantityAtInterval(unsigned reg, int q, int start, int end, bool isExit)
{
	if(reg == 0)
		return true;

	//first check hard regs
	HardUsageMap& hmap = hardRegUsageMap[reg];

	if(hmap.size() != 0)
	{
		HardUsageMap::iterator posi = hmap.upper_bound(start);
		//assume map is properly condensed

		//upper bound has to be after end
		if(posi != hmap.end() && posi->first <= end)
			return false; //since map is condensed any change is bad
		//and previous interval needs to be available
		if (posi != hmap.begin())
		{
			posi--;
			if (posi->second) //true means reg is in use
				return false;
		}
	}

	//reg available at out?
	if(isExit && !regAvailableForQuantityAtOut(reg, q))
		return false;

	//hard regs okay, what about other quantities?
	if(regUsageMap.count(reg))
	{
		UsageMap& map = regUsageMap[reg];
		if(map.size() == 0)
			return true;
		UsageMap::iterator posi = map.upper_bound(start);
		if(posi != map.end() && posi->first <= end)
			return false; //again, any change bad

		//pos is now past where we want to look, so adjust
		if(posi == map.begin())
			return true; //nothing set before upperbound
		posi--;
		if(posi->second != 0 && posi->second != q)
		{
			if(DebugPRA > 3)
				std::cerr << "Interval Unavailable because of " << posi->second << " at " << posi->first << " in reg " << reg << "\n";
			return false;
		}
	}
	return true;
}

/* Return true if q can be allocated to reg out of this block (not blocked) */
bool mcnf_block::regAvailableForQuantityAtOut(unsigned r, int q)
{
	if(regUsageMapOut.count(r) == 0)
	{
		if(regUnavailableToQuantities[q].count(r))
			return false;
		return true;
	}
	if(regUsageMapOut[r] == q)
		return true;
	return false;
}

//record that q should not be allocated to r out of this block
//this is because some quantity that is not live out of this block,
//but conflicts with q at this junction is using r
void mcnf_block::setRegUnavailableForQuantityOut(int q, unsigned r)
{
	if(r == 0)
		abort();
	else
		regUnavailableToQuantities[q].insert(r);
	if(DebugPRA > 9)
		std::cerr << "Marking " << q << " unavailable with reg " << r  << " in block " << index << "\n";
}

//undo above
void mcnf_block::setRegAvailableForQuantityOut(int q, unsigned r)
{
	if(DebugPRA > 9)
		std::cerr << "Marking " << q << " Available with reg " << r  << "in block " << index << "\n";

	regUnavailableToQuantities[q].erase(r);
}

int mcnf_block::getQuantityForReg(unsigned reg, mcnf_node *node)
{
	int pos = getNodeIntervalLevel(node);
	return getQuantityForRegAtPos(reg, pos);
}

/* return quantity using reg at node's level */
int mcnf_block::getQuantityForRegAtPos(unsigned reg, int pos)
{
	if(reg == 0)
		return 0;

	if(regUsageMap.count(reg))
	{
		UsageMap& map = regUsageMap[reg];
		if(map.size() == 0)
			return 0;
		UsageMap::iterator posi = map.upper_bound(pos);
		//pos is now past where we want to look, so adjust
		if(posi == map.begin())
			return 0; //nothing set before upperbound
		int regq = 0;

		if(posi == map.end())
		{
			regq = map.rbegin()->second;
		}
		else
		{
			//backup one
			posi--;
			regq = posi->second;
		}

		return regq;
	}
	else
	{
		return 0;
	}
}

//set regout with no consistency checks
void mcnf_block::setQuantityRegOut(int q, unsigned r)
{
	if(q < 0) return;
	if(DebugPRA > 9)
		std::cerr << "QRegOut in block " << index << " for " << q << " to " << r << "\n";

	if(r == 0 && quantityRegMapOut.count(q) && quantityRegMapOut[q] != 0)
	{
		//remove previous mapping
		if(regUsageMapOut[quantityRegMapOut[q]] != q)
		{
			std::cerr << "q currently mapped to reg " << quantityRegMapOut[q] << " which maps to q " << regUsageMapOut[quantityRegMapOut[q]]  << "\n";
			abort();
		}
		regUsageMapOut.erase(quantityRegMapOut[q]);
	}
	quantityRegMapOut[q] = r;

	if(r != 0 && regUsageMapOut.count(r))
	{
		std::cerr << "OutReg for reg " << r << " already assignd to q " << regUsageMapOut[r] << " not " << q << "\n";
		abort();
	}
	regUsageMapOut[r] = q;

}

// propagate q's current allocation to regout, checking for consistency
void mcnf_block::propagateQuantityRegOut(int q)
{
	if(q < 0) return;
	assert(quantityRegMap.count(q));
	unsigned r = quantityRegMap[q].rbegin()->second;
	if(quantityRegMapOut.count(q))
	{
		if(quantityRegMapOut[q] != r)
		{
			std::cerr << "Trying to propagate q " << q << " in " << r << " to out " << quantityRegMapOut[q] << "\n";
			abort();
		}
	}
	else
	{
		quantityRegMapOut[q] = r;
	}

}

void mcnf_block::clearQuantityRegs()
{
	quantityRegMap.clear();
	quantityRegMapOut.clear();
	regUsageMap.clear();
	regUsageMapOut.clear();
	regUnavailableToQuantities.clear();
}

/* Return true if r is not available for allocation at provided position */
bool mcnf_block::getHardRegUnavailable(unsigned r, int position, int subindex, bool isXBar)
{
	int pos = compute_interval_level(position, subindex, isXBar);
	HardUsageMap& map = hardRegUsageMap[r];

	if(map.size() == 0)
		return false;

	HardUsageMap::iterator posi = map.upper_bound(pos);
	if(posi == map.begin())
		return false;
	//backup, look what the current status is
	posi--;
	return posi->second;
}

/* Return true if r is not available for allocation at provided position */
bool mcnf_block::getHardRegUnavailableAtNode(unsigned r, mcnf_node *node)
{
	return getHardRegUnavailable(r, node->position, node->subindex, node->type == nXBar);
}

/* Record where hard registers are being used and so are unavailable for allocation */
void mcnf_block::setHardRegUnavailability(int position, int subindex, const std::set<unsigned>& unavailable_regs, bool isXBar)
{
	int pos = compute_interval_level(position, subindex, isXBar);

	//first, if any reg is currently marked is unavailable but is no longer unavailable, mark it unavailable
	for(hash_map<unsigned, HardUsageMap>::iterator itr = hardRegUsageMap.begin(),
			eitr = hardRegUsageMap.end(); itr != eitr; itr++)
	{
		unsigned r = itr->first;
		HardUsageMap& map = itr->second;

		if(unavailable_regs.count(r) == 0 && getHardRegUnavailable(r, position, subindex, isXBar))
		{
			map[pos] = false;
		}

	}

	//now make sure anything that is unavailable here is marked appropriately
	for(std::set<unsigned>::iterator itr = unavailable_regs.begin(),
			eitr = unavailable_regs.end(); itr != eitr; itr++)
	{
		unsigned r = *itr;
		if(!getHardRegUnavailable(r, position, subindex, isXBar))
		{
			HardUsageMap& map = hardRegUsageMap[r];
			map[pos] = true;
		}
	}
}


/* dump all our info on the function being compiled */
void GlobalMCNF::dump_info()
{
	int b, nb;
	nb = blocks.size();
	//dump block info
	for (b = 0; b < nb; b++)
	{
		unsigned i, n;
		mcnf_block *block = blocks[b];

		std::cerr << "Block " << itostr(block->index) << "\t";
		std::cerr << block->b->getBasicBlock()->getName() << "\n";
		std::cerr << "Succs:";
		for (i = 0, n = block->succs.size(); i < n; i++)
		{
			std::cerr << " " << itostr(block->succs[i]->index);
		}
		std::cerr << "\nPreds:";
		for (i = 0, n = block->preds.size(); i < n; i++)
		{
			std::cerr << " " << itostr(block->preds[i]->index);
		}
		std::cerr << "\n";

		std::cerr << "Livein: ";

		for (i = 0, n = block->liveins_vec.size(); i < n; i++)
		{
			unsigned var = block->liveins_vec[i]; //livein var
			std::cerr << itostr(var) << "(" << reg_quantity.get(var);
			std::cerr << ") ";
		}
		std::cerr << "\n";
		//dump RAATs
		for (i = 0, n = block->RAATs.size(); i < n; i++)
		{
			std::cerr << "RAAT " << i << "\n";
			block->RAATs[i].dump();
			std::cerr << "\n";
		}
		std::cerr << "\n";
	}
}

/* setup the def dominator sets in each block.
 * For every livein/liveout quantity record the block that
 * dominates this block where a store can safely be placed (no interfening
 * defs between that block and this block).
 * */
void GlobalMCNF::setup_def_dominators()
{
	std::queue<mcnf_block*> worklist;
	hash_set<mcnf_block*> inQ;
	//first compute what quantities each block defines
	std::vector<hash_set<int> > defined(blocks.size());

	for (int b = 0, nb = blocks.size(); b < nb; b++)
	{
		mcnf_block *block = blocks[b];
		assert(block->index == b);
		for (int i = 0, n = block->RAATs.size(); i < n; i++)
		{
			RAATInfo *ii = &block->RAATs[i];
			std::vector<RAATInfo::QuantityInfo>& info = ii->getQuantityInfo();
			for (int op = 0, nq = info.size(); op < nq; op++)
			{
				const RAATInfo::QuantityInfo& opinfo = info[op];

				if (opinfo.def)
				{
					defined[b].insert(opinfo.q);
					//setup the block's out set
					block->def_dominator_out[opinfo.q] = block;
				}
			}
		}
		worklist.push(block);
		inQ.insert(block);
	}

	while (!worklist.empty())
	{
		mcnf_block *block = worklist.front();
		worklist.pop();
		inQ.erase(block);
		bool changed = false;
		//for each quantity
		for (int i = 0, nr = block->liveins_vec.size(); i < nr; i++)
		{
			int q = reg_quantity.get(block->liveins_vec[i]);
			//to compute the in set, if all previous values are identical, we propage
			//that value, otherwise the result is this block (if there are no previous
			//values, leave blank)
			for (int p = 0, np = block->preds.size(); p < np; p++)
			{
				mcnf_block *pred = block->preds[p];
				if (pred->def_dominator_out.count(q) > 0)
				{
					mcnf_block *val = pred->def_dominator_out[q];
					if (block->def_dominator_in.count(q) == 0)
					{
						block->def_dominator_in[q] = val;
					}
					else
					{
						if (block->def_dominator_in[q] != val)
						{
							//this block
							block->def_dominator_in[q] = block;
						}
						//otherwise, same as val, keep
					}
				}
			}
			//to compute the outset, and quantity defined in this block gets this block,
			//all others propagate
			if (defined[block->index].count(q))
			{
				//no need to set changed
				assert(block->def_dominator_out[q] == block);
			}
			else if (block->def_dominator_in.count(q))
			{
				if (block->def_dominator_out.count(q) == 0
						|| block->def_dominator_out[q]
								!= block->def_dominator_in[q])
				{
					changed = true;
					block->def_dominator_out[q] = block->def_dominator_in[q];
				}
			}
		}

		if (changed) //push successors on the Q
		{
			for (int s = 0, ns = block->succs.size(); s < ns; s++)
			{
				mcnf_block *succ = block->succs[s];
				if (!inQ.count(succ))
				{
					inQ.insert(succ);
					worklist.push(succ);
				}
			}
		}
	}

	if (DebugPRA > 4)
	{
		std::string fname =
				MF->getFunction()->getParent()->getModuleIdentifier() + "_"
						+ MF->getFunction()->getName();
		std::cerr << "DefDoms " << fname << "\n";
		for (int b = 0, nb = blocks.size(); b < nb; b++)
		{
			mcnf_block *block = blocks[b];
			std::cerr << "Block " << block->index << ":\n";
			for (int i = 0, nr = block->liveins_vec.size(); i < nr; i++)
			{
				int q = reg_quantity.get(block->liveins_vec[i]);
				if (block->def_dominator_in.count(q))
				{
					std::cerr << " " << q << ":"
							<< block->def_dominator_in[q]->index << ","
							<< block->def_dominator_out[q]->index;
				}
			}
			std::cerr << "\n";
		}
	}
}

/* recursive depth first traversal of blocks that
 fills out block order */
void GlobalMCNF::dfo_blockorder_r(mcnf_block *block, int *curr,
		std::vector<char>& visited)
{
	int i, n;
	if (visited[block->index])
		return;

	/* process this block */
	visited[block->index] = 1;
	block_order[*curr] = block->index;
	(*curr)++;

	/* process successors */
	for (i = 0, n = block->succs.size(); i < n; i++)
	{
		mcnf_block *next = block->succs[i];
		dfo_blockorder_r(next, curr, visited);
	}
}

/* Fill out block_order (which is assumed to be large enough)
 with a dfo order of blocks.
 */
void GlobalMCNF::dfo_blockorder()
{
	int curr = 0;
	int numb = blocks.size();
	mcnf_block *block = blocks[0];
	std::vector<char> visited(numb, 0); //index by block

	dfo_blockorder_r(block, &curr, visited);

	if (curr != numb)
	{
		int i;
		for (i = 0; i < numb; i++)
		{
			if (!visited[i])
				block_order[curr++] = i;
		}
		assert(curr == numb && "problem with dfo ordering");
	}
}

/* Return the cost to load a constant of class c into a register
 Just consider the cost of the first possible register in classes.
 */
short GlobalMCNF::const_load_cost(int ac, const BitVector& classes)
{
	int i;

	for (i = 0; i < num_alloc_classes; i++)
	{
		if (classes.test(i))
			break;
	}
	if (i == num_alloc_classes)
		i = 0; /* huh? */
	return alloc_classes->getMoveCost(ac, i);
}

/* Cost to store into memory class class.
 For now assume it's a constant (umm.. don't know what else we could do) */
short GlobalMCNF::memory_store_cost(int c)
{
	int ret;
	/* assume its the cost of a load */
	if (alloc_classes->getType(c) != AllocationClassInfo::Memory)
		abort();
	ret = alloc_classes->getStoreCost(c);

	return ret;
}

/* Cost to load from memory class class.
 For now assume it's a constant (umm.. don't know what else we could do) */
short GlobalMCNF::memory_load_cost(int c)
{
	int ret;
	ret = alloc_classes->getMoveCost(c, alloc_classes->getRegisterClasses()[0]);

	return ret;
}

/* Cost to insert something after the control flow
 at the end of a basic block */
int GlobalMCNF::after_control_flow_cost(mcnf_block *b, int movecost)
{
	int out = b->succs.size();
	//don't currently support edge insertions.. should only
	//be a problem if jmps read vars, and we handle that specially
	return -1;

	if (!b->edge_insertable())
		return -1;
	if (out <= 1)
		return movecost; /* assume no problems with control flow */
	else
		return out * (movecost + 2); /* a basic block for each edge */
}

int GlobalMCNF::class_bandwidth(RAATInfo *ii, int c)
{
	if (alloc_classes->getType(c) == AllocationClassInfo::Memory)
		return ii->total_memory_bandwidth();
	else
		return -1;
}

/* Return an amount to multiply every cost in this block by
 (for speed, bias by loop level).
 */
int GlobalMCNF::cost_block_multiplier(mcnf_block *block)
{
	if (WhatMetric == Speed)
	{
		int mult = max_block_cost < 10000 ? max_block_cost : 10000;
		int cost = mult * (double) block->b->getFrequency()
				/ (double) max_block_cost;
		//normalize to 1000 to avoid astronomical costs
		if (cost <= 0)
			cost = 1;
		return cost;
	}
	else
	{
		return 1; //just size for now
	}
}

/* Allocates a node.  At some point we should use a freelist
 so always call through here */
mcnf_node* GlobalMCNF::mcnf_node_allocate(mcnf_block *block)
{
	mcnf_node *node = new mcnf_node();
	node->block_num = block->index;
	node->capacity = -1; // uncapacitated
	node->node_order = -1; //uninitialized
	num_nodes++;
	node->vecpos = block->allnodes.size();
	block->allnodes.push_back(node);

	return node;
}

/* Allocate and initialize a node containing a single commodity (sink or def) */
mcnf_node* GlobalMCNF::create_single_commodity_node(mcnf_block *block,
		mcnf_node_type type, commod_type q, int position, short subindex,
		short aclass)
{
	mcnf_node *node;
	std::vector<mcnf_node*> *vec;

	if (type == nSink)
		vec = &block->sinks;
	else if (type == nDef)
		vec = &block->defs;
	else
		abort();

	node = mcnf_node_allocate(block);
	node->type = type;
	node->position = position;
	node->subindex = subindex;
	node->allocClass = aclass;

	/* add single commodity */
	assert(node->commodities.size() == 0);
	node->commodities.push_back(q);
	node->cindex.add(q, 0);

	vec->push_back(node);

	return node;

}

/* Initializes the passed node (which must arleady be allocated and zeroed)
 to represent a specific class in an insn/xbar/entry/exit
 */
void GlobalMCNF::mcnf_initialize_node_for_class(mcnf_node *node, int c, std::vector<char>& available_regs)
{
	node->allocClass = c;

	/* If a memory class is capacitated the capacity is
	 paid for by the antis in the side network.
	 The capacity of the memory class is overload to
	 be the index of the corresponding side network.
	 */
	if (alloc_classes->getType(c) == AllocationClassInfo::Memory)
	{
		node->capacity = -1;
	}
	else if (node->type == nEntry || node->type == nExit)
	{
		node->capacity = -1;
	}
	else
	{
		int cap = alloc_classes->getCapacity(c);
		if(alloc_classes->getType(c) == AllocationClassInfo::Register)
		{
			assert(available_regs[c] <= cap);
			node->capacity = available_regs[c];
		}
		else
			node->capacity = cap;
	}
}

/* Add quantity q to node.

 This updates the commodities array, the cindex, num_commodites
 but does not expand any other arrays.

 If quantity is already present, does not modify commodities.
 */
void GlobalMCNF::mcnf_node_add_quantity(mcnf_node *node, commod_type q)
{
	if (node->cindex.get(q) < 0)
	{
		index_type index = node->commodities.size();
		node->commodities.push_back(q);
		node->cindex.add(q, index);

		/* insn nodes have individual costs for every commodity */
		if (node->type == nInsn)
		{
			node->nonanti_costs.push_back(0);
		}
	}
}

/* Allocate an edge */
mcnf_edge* GlobalMCNF::mcnf_edge_allocate(mcnf_block *block)
{
	mcnf_edge *edge = new mcnf_edge();
	block->alledges.push_back(edge);
	num_edges++;
	return edge;
}

/* If an edge already exists between src and dst, return it, otherwise create it */
mcnf_edge* GlobalMCNF::mcnf_edge_find_or_allocate(mcnf_block *block,
		mcnf_node* src, mcnf_node* dst)
{
	for (unsigned s = 0, ns = src->succs.size(); s < ns; s++)
	{
		if (src->succs[s]->dst == dst)
			return src->succs[s];
	}
	//create
	mcnf_edge *ret = mcnf_edge_allocate(block);
	mcnf_connect_nodes(src, dst, ret);
	return ret;
}
/* Adds a commodity to an edge */
void GlobalMCNF::mcnf_edge_add_commodity(mcnf_edge *edge, commod_type q)
{
	edge->commodities.push_back(q);
	edge->commod_map.set(q);
}

/* Initialize the commodities of edge to be the intersection
 of the commodities of src and dest */
void GlobalMCNF::mcnf_initialize_edge_node_commodities(mcnf_node *src,
		mcnf_node *dest, mcnf_edge *edge)
{
	int i;
	int nums = src->commodities.size();
	int numd = dest->commodities.size();
	int crossmem = 0;

	/* any edges that cross memory classes should not include memory vars */
	if (src->allocClass != dest->allocClass && src->allocClass >= 0
			&& dest->allocClass >= 0 && alloc_classes->getType(src->allocClass)
			== AllocationClassInfo::Memory && alloc_classes->getType(
			dest->allocClass) == AllocationClassInfo::Memory)
		crossmem = 1;

	if (nums < numd)
	{
		//first count, then size the vector and fill
		unsigned cnt = 0;
		for (i = 0; i < nums; i++)
		{
			int commod = src->commodities[i];

			if (crossmem && commod < 0)
				continue;

			if (dest->cindex.get(commod) >= 0)
			{
				cnt++;
			}
		}
		edge->commodities.resize(cnt);
		edge->commod_map.reserve(cnt);
		cnt = 0;
		for (i = 0; i < nums; i++)
		{
			int commod = src->commodities[i];

			if (crossmem && commod < 0)
				continue;

			if (dest->cindex.get(commod) >= 0)
			{
				edge->commodities[cnt++] = commod;
				edge->commod_map.set(commod);
			}
		}
		assert(cnt == edge->commodities.size());
	}
	else //go from dst
	{
		//first count, then size the vector and fill
		unsigned cnt = 0;
		for (i = 0; i < numd; i++)
		{
			int commod = dest->commodities[i];

			if (crossmem && commod < 0)
				continue;

			if (src->cindex.get(commod) >= 0)
			{
				cnt++;
			}
		}
		edge->commodities.resize(cnt);
		edge->commod_map.reserve(cnt);
		cnt = 0;
		for (i = 0; i < numd; i++)
		{
			int commod = dest->commodities[i];

			if (crossmem && commod < 0)
				continue;

			if (src->cindex.get(commod) >= 0)
			{
				edge->commodities[cnt++] = commod;
				edge->commod_map.set(commod);
			}
		}
		assert(cnt == edge->commodities.size());
	}
}

/* Connect src and dest using edge.
 Fills in edge's src/dest fields and
 the succ/pred of the nodes
 */
void GlobalMCNF::mcnf_connect_nodes(mcnf_node *src, mcnf_node *dest,
		mcnf_edge *edge)
{
	edge->src = src;
	edge->dst = dest;

	if (src == NULL)
		abort();
	if (dest == NULL)
		abort();

	/* add this edge as a successor to src */
	src->succs.push_back(edge);

	/* add this edge as a predecessor to dest */
	dest->preds.push_back(edge);
}

/* Creates a row of nodes of type.  The nodes are added
 to the current block and the row is output in the nodes array.
 */
void GlobalMCNF::create_nodes_from_bitmap(mcnf_block *block,
		mcnf_node_type type, int position, short subindex,
		std::vector<char>& available_regs,
		hash_map<unsigned, BitVector>& usable_classes, const BitVector& live,
		mcnf_node **nodes)
{
	int c;
	std::vector<mcnf_node*>* vec = NULL;

	if (type == nEntry)
		vec = &block->entries;
	else if (type == nExit)
		vec = &block->exits;
	else if (type == nXBar || type == nInsn)
		vec = &block->body;
	else
		abort();

	memset(nodes, 0, sizeof(mcnf_node*) * num_alloc_classes);
	if (type == nEntry || type == nExit)
	{
		//always have at least one entry/exit node so it's easier
		//to tell where the block starts/stops
		c = 0;
		if (nodes[c] == NULL)
		{
			nodes[c] = mcnf_node_allocate(block);
			nodes[c]->type = type;
			nodes[c]->position = position;
			nodes[c]->subindex = subindex;
			nodes[c]->allocClass = c;
			mcnf_initialize_node_for_class(nodes[c], c, available_regs);
			vec->push_back(nodes[c]);

			/* assert not anti class */
			assert (alloc_classes->getType(c) != AllocationClassInfo::AntiOnly);
		}
	}

	/* For every live variable */
	for (int r = live.find_first(); r >= 0; r = live.find_next(r))
	{
		/* For every class this variable might reside in,
		 and which has available registers... */
		int q = reg_quantity.get(r);

		if (q > 0)
		{
			const BitVector& usable = usable_classes[q];
			for (c = 0; c < num_alloc_classes; c++)
			{
				if (((usable.test(c)) && available_regs[c])
						|| alloc_classes->getType(c)
								== AllocationClassInfo::Memory
						|| alloc_classes->getType(c)
								== AllocationClassInfo::AntiOnly)
				{
					/* create the node if it doesn't exist yet */
					if (nodes[c] == NULL)
					{
						nodes[c] = mcnf_node_allocate(block);
						nodes[c]->type = type;
						nodes[c]->position = position;
						nodes[c]->subindex = subindex;
						nodes[c]->allocClass = c;
						mcnf_initialize_node_for_class(nodes[c], c, available_regs);
						vec->push_back(nodes[c]);

						/* set up mirror node, assumes anti's always come after memory */
						if (alloc_classes->getType(c)
								== AllocationClassInfo::AntiOnly)
						{
							int antic = alloc_classes->getAntiClass(c);
							nodes[c]->mirror_node = nodes[antic];
							nodes[antic]->mirror_node = nodes[c];
						}
					}

					/* add this commodity to it */
					if ((usable.test(c) && available_regs[c])
							|| alloc_classes->getType(c)
									== AllocationClassInfo::Memory)
					{
						if (alloc_classes->getType(c)
								== AllocationClassInfo::AntiOnly)
							abort();
						mcnf_node_add_quantity(nodes[c], q);
					}

					/* handle memory nodes which also have anti-quantities */
					if (alloc_classes->getType(c)
							== AllocationClassInfo::Memory
							|| alloc_classes->getType(c)
									== AllocationClassInfo::AntiOnly)
						mcnf_node_add_quantity(nodes[c], -q);

				}

			}
		}
	}
}

/* Creates a set of defs  (including antis) for
 quantity q. It initializes these nodes, but does not
 hook them up.

 The caller is expected to hook up the defs.
 */
void GlobalMCNF::create_def(mcnf_block *block, commod_type q)
{
	/* Create a def
	 for every memory class */
	mcnf_node *def;

	def = create_single_commodity_node(block, nDef, q, q,
			quantities[q].defs.size(), -1);
	quantities[q].defs.push_back(def);

	const std::vector<int>& memory_classes = alloc_classes->getMemoryClasses();
	for (int i = 0, n = memory_classes.size(); i < n; i++)
	{
		int c = memory_classes[i];
		int index = quantities[q].antidefs[i].size();
		def = create_single_commodity_node(block, nDef, -q, -q, index, c);
		quantities[q].antidefs[i].push_back(def);
	}
}

/* Creates a set of sinks  (including antis) for
 quantity q. It initializes these nodes, but does not
 hook them up.

 The caller is expected to hook up the sinks.
 */
void GlobalMCNF::create_sink(mcnf_block *block, commod_type q)
{
	/* Create a def, a sink, and a def/sink pair
	 for every memory class */
	mcnf_node *sink;

	sink = create_single_commodity_node(block, nSink, q, q,
			quantities[q].sinks.size(), -1);
	quantities[q].sinks.push_back(sink);

	const std::vector<int>& memory_classes = alloc_classes->getMemoryClasses();
	for (int i = 0, n = memory_classes.size(); i < n; i++)
	{
		int c = memory_classes[i];
		int index = quantities[q].antisinks[i].size();
		sink = create_single_commodity_node(block, nSink, -q, -q, index, c);
		quantities[q].antisinks[i].push_back(sink);
	}
}

/* Connects an already created def into a node vector.
 Only connects to classes specified by classes.
 Always connects antivars to memory
 const_equiv and mem_equiv are set to equivalent classes
 or -1 if not applicable
 insn is the defining insn
 */
void GlobalMCNF::connect_def(commod_type q, const BitVector& classes,
		mcnf_block *block, mcnf_node **nodes, int class_equiv)
{
	int c;
	mcnf_node *def = quantities[q].defs.back();

	for (c = 0; c < num_alloc_classes; c++)
	{
		if (classes.test(c) && nodes[c] != NULL)
		{
			mcnf_edge *edge = mcnf_edge_allocate(block);
			mcnf_initialize_edge_node_commodities(def, nodes[c], edge);
			mcnf_connect_nodes(def, nodes[c], edge);

			/* if we can define into a constant, the cost of
			 such an edge is negative the cost of the constant
			 load */
			if (alloc_classes->getType(c) == AllocationClassInfo::Constant
					&& class_equiv == c)
			{
				edge->cost = -const_load_cost(c, classes);
			}

			/* if we're defining into memory, the cost is minus the cost of eviction
			 If there's an equivalence we subtract the cost of
			 the load and the eviction.
			 */
			if (alloc_classes->getType(c) == AllocationClassInfo::Memory)
			{
				if (class_equiv == c)
				{
					edge->cost = -memory_load_cost(c) - memory_store_cost(c);
				}
				else
				{
					edge->cost = -memory_store_cost(c);
				}
			}

			edge->cost *= cost_block_multiplier(block);
			if (nodes[c]->type == nExit || nodes[c]->type == nEntry)
				edge->cost = 0;
		}

	}

	/* connect anti variables */
	const std::vector<int>& memory_classes = alloc_classes->getMemoryClasses();
	for (int i = 0, n = memory_classes.size(); i < n; i++)
	{
		mcnf_edge *edge;
		mcnf_node *antidef;
		int c = memory_classes[i];
		int ac = alloc_classes->getAntiClass(c);

		antidef = quantities[q].antidefs[i].back();

		edge = mcnf_edge_allocate(block);
		mcnf_initialize_edge_node_commodities(antidef, nodes[c], edge);
		mcnf_connect_nodes(antidef, nodes[c], edge);

		/* edge to antionly network (cost of store) */
		if (ac < 0 || ac >= num_alloc_classes)
			abort();

		edge = mcnf_edge_allocate(block);
		mcnf_initialize_edge_node_commodities(antidef, nodes[ac], edge);
		mcnf_connect_nodes(antidef, nodes[ac], edge);
		edge->cost = memory_store_cost(c) * cost_block_multiplier(block);

		if (nodes[c]->type == nExit || nodes[c]->type == nEntry)
			edge->cost = 0;
	}
}

/* Connects an already created sink from a node vector.
 Only connects to classes specified by classes.
 */
void GlobalMCNF::connect_sink(commod_type q, const BitVector& classes,
		mcnf_block *block, mcnf_node **nodes)
{
	int c;
	mcnf_node *sink = quantities[q].sinks.back();

	for (c = 0; c < num_alloc_classes; c++)
	{
		if (classes.test(c) && nodes[c] != NULL)
		{
			mcnf_edge *edge = mcnf_edge_allocate(block);

			mcnf_initialize_edge_node_commodities(nodes[c], sink, edge);
			mcnf_connect_nodes(nodes[c], sink, edge);
		}

	}

	/* connect anti variables */
	const std::vector<int>& memory_classes = alloc_classes->getMemoryClasses();
	for (int i = 0, n = memory_classes.size(); i < n; i++)
	{
		mcnf_edge *edge;
		mcnf_node *antisink;
		int c = memory_classes[i];
		int ac = alloc_classes->getAntiClass(c);

		antisink = quantities[q].antisinks[i].back();

		edge = mcnf_edge_allocate(block);
		mcnf_initialize_edge_node_commodities(nodes[c], antisink, edge);
		mcnf_connect_nodes(nodes[c], antisink, edge);

		/* edge to antionly network (cost of store) */
		if (ac < 0 || ac >= num_alloc_classes)
			abort();

		edge = mcnf_edge_allocate(block);
		mcnf_initialize_edge_node_commodities(nodes[ac], antisink, edge);
		mcnf_connect_nodes(nodes[ac], antisink, edge);
	}
}

/* Create entry nodes and put them in nodes.
 In addition to creating this row, hooks up a def to all
 live in variables, just to make shortest paths easier */
void GlobalMCNF::create_entry_nodes(mcnf_block *block,
		std::vector<char>& available_regs, std::set<unsigned>& unavailable_regs,
		hash_map<unsigned, BitVector>& usable_classes, const BitVector& live,
		mcnf_node **nodes)
{
	/* it's important that nodes are created in topological order,
	 so first we create the live in block defs */
	for (int var = live.find_first(); var >= 0; var = live.find_next(var))
	{
		int q = reg_quantity.get(var);
		create_def(block, q);
	}

	//some registers can never be live out of a block (x86 fp stack registers,
	//because the stackifier isn't very robust)
	std::vector<char> adjusted_available_regs = available_regs;
	std::set<unsigned> adjusted_unavailable_regs = unavailable_regs;
	for (unsigned c = 0, nc = num_alloc_classes; c < nc; c++)
	{
		if (!alloc_classes->usableAcrossBlocks(c))
		{
			adjusted_available_regs[c] = 0;
			const std::vector<unsigned>& regs = alloc_classes->getRegisters(c);
			for(unsigned i = 0, n = regs.size(); i < n; i++)
			{
				adjusted_unavailable_regs.erase(regs[i]);
			}
		}
	}

	/* then the entry nodes */
	create_nodes_from_bitmap(block, nEntry, -2, 0, adjusted_available_regs,
			 usable_classes, live, nodes);
	block->setHardRegUnavailability(-2, 0, adjusted_unavailable_regs, false);

	/* and finally connect them */
	for (int var = live.find_first(); var >= 0; var = live.find_next(var))
	{
		int q = reg_quantity.get(var);
		connect_def(q, usable_classes[q], block, nodes, -1);
	}
}
struct blocked_quantity_info
{
	int q;
	int allocClass;
	bool oldsetting;
	blocked_quantity_info() :
		q(-1), allocClass(0), oldsetting(false)
	{
	}
	blocked_quantity_info(int quant, int ac, bool old) :
		q(quant), allocClass(ac), oldsetting(old)
	{
	}
};

/* Explore the CFG looking for related blocks with live-in hard regs
 * (exception landing pads and weird blocks from tail merging)
 * and append to blocks.  */
void GlobalMCNF::get_exception_landingpads(mcnf_block *block, entry_exit type,
		std::vector<mcnf_block*>& blocks, hash_set<int>& visited)
{
	int bindex = block->index;
	if (type == Entry)
		bindex = -bindex;
	if (visited.count(bindex) > 0)
		return;
	visited.insert(bindex);

	if (type == Exit)
	{
		//just look at all the successors
		for (int s = 0, ns = block->succs.size(); s < ns; s++)
		{
			mcnf_block *succ = block->succs[s];
			get_exception_landingpads(succ, Entry, blocks, visited);
		}
	}
	else
	{
		//check for livein regs
		if (block->hard_liveins.size() > 0)
			blocks.push_back(block);
		//look at predecessors
		for (int p = 0, np = block->preds.size(); p < np; p++)
		{
			mcnf_block *pred = block->preds[p];
			get_exception_landingpads(pred, Exit, blocks, visited);
		}
	}
}

/* Create exit nodes and put them in nodes.
 In addition to creating this row, hooks up a sink to all
 live out variables, just to make shortest paths easier */
void GlobalMCNF::create_exit_nodes(mcnf_block *block,
		std::vector<char>& available_regs, std::set<unsigned>& unavailable_regs,
		hash_map<unsigned, BitVector>& usable_classes, const BitVector& live,
		mcnf_node **nodes, int position)
{
	hash_set<int> quantity_seen;

	std::vector<blocked_quantity_info> blocked_quantities;

	//if any successor block has livein hard regs, then we can't allocate
	//vars that are livein to the successor to those hard regs; this
	//happens at the landing pad blocks for exceptions
	std::vector<mcnf_block*> landingpads;
	hash_set<int> visited;
	get_exception_landingpads(block, Exit, landingpads, visited);
	for (int s = 0, ns = landingpads.size(); s < ns; s++)
	{
		mcnf_block *succ = landingpads[s];
		if (succ->hard_liveins.size() > 0)
		{
			for (unsigned i = 0, n = succ->hard_liveins.size(); i < n; i++)
			{
				unsigned r = succ->hard_liveins[i];
				int allocClass = alloc_classes->getRegisterIndex(r);
				assert(allocClass != -1);
				//everything livein to succ can't use allocClass at the end of block
				for (int r = 0, nr = succ->liveins_vec.size(); r < nr; r++)
				{
					int q = reg_quantity.get(succ->liveins_vec[r]);
					blocked_quantities.push_back(blocked_quantity_info(q,
							allocClass, usable_classes[q].test(allocClass)));
					usable_classes[q].reset(allocClass);
				}
			}
		}
	}

	//some registers can never be live out of a block (x86 fp stack registers,
	//because the stackifier isn't very robust)
	std::vector<char> adjusted_available_regs = available_regs;
	std::set<unsigned> adjusted_unavailable_regs = unavailable_regs;
	for (unsigned c = 0, nc = num_alloc_classes; c < nc; c++)
	{
		if (!alloc_classes->usableAcrossBlocks(c))
		{
			adjusted_available_regs[c] = 0;
			const std::vector<unsigned>& regs = alloc_classes->getRegisters(c);
			for (unsigned i = 0, n = regs.size(); i < n; i++)
			{
				adjusted_unavailable_regs.erase(regs[i]);
			}
		}
	}

	create_nodes_from_bitmap(block, nExit, position, 0, adjusted_available_regs,
			 usable_classes, live, nodes);
	block->setHardRegUnavailability(position, 0, adjusted_unavailable_regs, false);

	for (int var = live.find_first(); var >= 0; var = live.find_next(var))
	{
		int q = reg_quantity.get(var);
		if (q > 0)
		{
			if (quantity_seen.count(q) > 0)
				abort();
			quantity_seen.insert(q);
			create_sink(block, q);
			connect_sink(q, usable_classes[q], block, nodes);
		}
	}

	//undo usuable class settings
	//go backwards to we don't have to worry about overwriting a setting
	//with the new value (if here are duplicates in the vector)
	for (int i = blocked_quantities.size() - 1; i >= 0; i--)
	{
		blocked_quantity_info& bq = blocked_quantities[i];
		usable_classes[bq.q][bq.allocClass] = bq.oldsetting;
	}
}

/*
 Connect two NUM_ALLOC_CLASS arrays of nodes using
 simple pass through edges.  The interesection of the
 corresponding nodes commodities are allowed on the edge.
 */
void GlobalMCNF::link_straight(mcnf_block *block, mcnf_node **first,
		mcnf_node **second)
{
	for (int c = 0; c < num_alloc_classes; c++)
	{
		if (first[c] != NULL && second[c] != NULL
				&& first[c]->commodities.size() > 0
				&& second[c]->commodities.size() > 0)
		{
			mcnf_edge *edge = mcnf_edge_allocate(block);
			mcnf_initialize_edge_node_commodities(first[c], second[c], edge);
			mcnf_connect_nodes(first[c], second[c], edge);
		}
	}
}

/* just allocate all necessary arrays for dual prices and gradients */
static void setup_node_dual_prices(mcnf_node *node)
{
	int n = node->commodities.size();
	node->dual_prices = new double[n];
	memset(node->dual_prices, 0, sizeof(double) * n);
	node->dual_gradients = new double[n];
	memset(node->dual_gradients, 0, sizeof(double) * n);
}


/* Create an xbar

 First setup the top and bottom.
 Everything that is livein is allowed into the top; everything
 that is liveout is allowed out of the bottom.  Variables that
 are both are allowed through the middle.

 If a variable dies in the xbar, connect it to a sink.
 If a variable generated in the xbar (?), create a def and connect
 it to the top.

 */
void GlobalMCNF::create_xbar(mcnf_block *block, RAATInfo *ii, RAATInfo *nextii,
		int position, std::vector<char>& available_regs_top,
		std::vector<char>& available_regs_bottom,
		std::set<unsigned>& unavailable_regs_top,
		std::set<unsigned>& unavailable_regs_bottom,
		hash_map<unsigned, BitVector>& usable_classes, BitVector& livein,
		BitVector& liveout, mcnf_node **top, mcnf_node **bottom, int issimple)
{
	int i, n;
	BitVector inq(quantities.size()), outq(quantities.size());
	std::vector<commod_type> in, out;
	bool is_control_flow = false;
	bool next_is_control_flow = false;
	const std::vector<int>& memory_classes = alloc_classes->getMemoryClasses();

	/* disallow movement after control flow insns */
	if (ii != NULL && ii->isControlFlow())
		is_control_flow = true;
	if (nextii != NULL && nextii->isControlFlow())
		next_is_control_flow = true;
	if (ii != NULL && !is_control_flow && nextii == NULL)
		next_is_control_flow = true; //fall through block
	if(next_is_control_flow && !is_control_flow)
		block->lastXBarPos = position;
	/* figure out which quantities are just livein, just liveout, and both */
	for (int var = livein.find_first(); var >= 0; var = livein.find_next(var))
	{
		int q = reg_quantity.get(var);
		if (q > 0)
			inq.set(q);
	}

	for (int var = liveout.find_first(); var >= 0; var = liveout.find_next(var))
	{
		int q = reg_quantity.get(var);
		if (q > 0)
			outq.set(q);
	}

	/* convert in,out,both into list of quantities based bitmap setting */
	BitVector bothq = inq & outq;
	inq ^= bothq; //just live in
	outq ^= bothq; //just live out
	for (int var = inq.find_first(); var >= 0; var = inq.find_next(var))
	{
		in.push_back(var);
	}
	for (int var = outq.find_first(); var >= 0; var = outq.find_next(var))
	{
		out.push_back(var);
	}

	/* in is now a list of all in-only quantities,
	 out a list of all out-only quantities,
	 and both a list of all in and out quantities */

	/* Create and link the innards of the crossbar along with
	 the top and bottom */
	mcnf_node **bothn = (mcnf_node**) alloca(num_alloc_classes
			*sizeof(mcnf_node*));
	BitVector both = livein & liveout; //reg indexed

	create_nodes_from_bitmap(block, nXBar, position, XBAR_TOP_LEVEL,
			available_regs_top, usable_classes, livein, top);
	block->setHardRegUnavailability(position, XBAR_TOP_LEVEL, unavailable_regs_top, true);
	/* middle level */
	create_nodes_from_bitmap(block, nXBar, position, XBAR_TOP_LEVEL + 1,
			available_regs_bottom, usable_classes, both, bothn);
	block->setHardRegUnavailability(position, XBAR_TOP_LEVEL+1, unavailable_regs_bottom, true);

	/* first just allow moves to the right */
	for (i = 0; i < num_alloc_classes; i++)
	{
		for (int j = i; j < num_alloc_classes; j++)
		{
			int cost = alloc_classes->getMoveCost(i, j); /* neg cost means no edge */
			if (is_control_flow && i != j && cost >= 0)
				cost = after_control_flow_cost(block, cost);
			cost = cost * cost_block_multiplier(block);

			if (cost >= 0 && top[i] != NULL && bothn[j] != NULL
					&& top[i]->commodities.size() > 0
					&& bothn[j]->commodities.size() > 0)
			{
				/* network simplification: only allow stores after
				 defining or using instructions or at the end of the block
				 allow two crossbars at end so moves can be placed before
				 jumps..
				 */
				if (alloc_classes->getType(j) == AllocationClassInfo::Memory
						&& alloc_classes->getType(i)
								== AllocationClassInfo::Register && position
						< (int) block->RAATs.size() - 1 && ii != NULL && nextii
						!= NULL && !next_is_control_flow)
				{
					unsigned k;
					mcnf_edge *edge = NULL;

					for (k = 0; k < ii->getNumOps(); k++)
					{
						const RAATInfo::OperandInfo& opinfo =
								ii->getOpFromIndex(k);
						if ((opinfo.isDef || opinfo.isUse) && liveout.test(
								opinfo.regNum))
						{
							if (edge == NULL)
								edge = mcnf_edge_allocate(block);
							mcnf_edge_add_commodity(edge, reg_quantity.get(
									opinfo.regNum));
						}
					}

					if (edge != NULL)
					{
						mcnf_connect_nodes(top[i], bothn[j], edge);
						edge->cost = cost;
					}
				}
				else if(MoveModel == NoMoves && i != j && (alloc_classes->getType(i)
						== AllocationClassInfo::Register
						|| alloc_classes->getType(j)
								== AllocationClassInfo::Register))
				{
					//nada
				}
				else if (issimple && i != j && (alloc_classes->getType(i)
						== AllocationClassInfo::Register
						|| alloc_classes->getType(j)
								== AllocationClassInfo::Register) && ii != NULL
						&& !next_is_control_flow)
					; /* no edge needed */
				else if (MoveModel == LimitMoves && i != j && (alloc_classes->getType(i)
						== AllocationClassInfo::Register
						|| alloc_classes->getType(j)
								== AllocationClassInfo::Register) && ii != NULL
						&& !next_is_control_flow)
				{/*
				 //only allow vars from ii or nextii to participate in the crossbar
				 unsigned k;
				 mcnf_edge *edge= NULL;

				 //ii
				 for (k = 0; k < ii->getNumOps(); k++)
				 {
				 const RAATInfo::OperandInfo& opinfo =
				 ii->getOpFromIndex(k);
				 if ((opinfo.isDef || opinfo.isUse)
				 && liveout.test(opinfo.regNum))
				 {
				 if (edge == NULL)
				 edge = mcnf_edge_allocate(block);
				 mcnf_edge_add_commodity(edge,
				 reg_quantity.get(opinfo.regNum));
				 }
				 }
				 //nextii
				 for (k = 0; k < nextii->getNumOps(); k++)
				 {
				 const RAATInfo::OperandInfo& opinfo =
				 nextii->getOpFromIndex(k);
				 if (opinfo.isUse
				 && liveout.test(opinfo.regNum))
				 {
				 if (edge == NULL)
				 edge = mcnf_edge_allocate(block);
				 int q = reg_quantity.get(opinfo.regNum);
				 if(!edge->commodity_valid(q))
				 {
				 mcnf_edge_add_commodity(edge, q);
				 }
				 }
				 }
				 if (edge != NULL)
				 {
				 mcnf_connect_nodes(top[i], bothn[j], edge);
				 edge->cost = cost;
				 }
				 */
				}
				else
				{
					mcnf_edge *edge = mcnf_edge_allocate(block);
					mcnf_initialize_edge_node_commodities(top[i], bothn[j],
							edge);
					mcnf_connect_nodes(top[i], bothn[j], edge);
					edge->cost = cost;
				}
			}
		}
	}

	/* let antivars flow back into the memory network even in the middle */
	for (i = 0, n = memory_classes.size(); i < n; i++)
	{
		int m = memory_classes[i];
		int a = alloc_classes->getAntiClass(m);
		int cost = alloc_classes->getMoveCost(a, m);
		if (is_control_flow && cost >= 0)
			cost = after_control_flow_cost(block, cost);
		cost = cost * cost_block_multiplier(block);

		if (top[a] && bothn[m] && cost >= 0)
		{
			mcnf_edge *edge = mcnf_edge_allocate(block);
			mcnf_initialize_edge_node_commodities(top[a], bothn[m], edge);
			mcnf_connect_nodes(top[a], bothn[m], edge);

			edge->cost = cost;
		}
	}

	create_nodes_from_bitmap(block, nXBar, position, XBAR_BOTTOM_LEVEL,
			available_regs_bottom, usable_classes, liveout, bottom);
	block->setHardRegUnavailability(position, XBAR_BOTTOM_LEVEL, unavailable_regs_bottom, true);

	/* now moves to the left */
	for (i = 0; i < num_alloc_classes; i++)
	{
		for (int j = 0; j <= i; j++)
		{
			int cost = alloc_classes->getMoveCost(i, j); /* neg cost means no edge */
			if (is_control_flow && i != j && cost >= 0)
				cost = after_control_flow_cost(block, cost);

			cost = cost * cost_block_multiplier(block);

			if (cost >= 0 && bothn[i] != NULL && bottom[j] != NULL
					&& bothn[i]->commodities.size() > 0
					&& bottom[j]->commodities.size() > 0)
			{
				if (alloc_classes->getType(i) == AllocationClassInfo::Memory
						&& alloc_classes->getType(j)
								== AllocationClassInfo::Register && position
						< (int) block->RAATs.size() - 1 && ii != NULL && nextii
						!= NULL && !next_is_control_flow)
				{
					unsigned k;
					mcnf_edge *edge = NULL;

					for (k = 0; k < nextii->getNumOps(); k++)
					{
						const RAATInfo::OperandInfo& opinfo =
								nextii->getOpFromIndex(k);

						if (opinfo.isUse && bottom[j]->cindex.get(
								reg_quantity.get(opinfo.regNum)) >= 0)
						{
							if (edge == NULL)
								edge = mcnf_edge_allocate(block);
							mcnf_edge_add_commodity(edge, reg_quantity.get(
									opinfo.regNum));
						}
					}

					if (edge != NULL)
					{
						mcnf_connect_nodes(bothn[i], bottom[j], edge);
						edge->cost = cost;
					}
				}
				else if (issimple && i != j && (alloc_classes->getType(i)
						== AllocationClassInfo::Register
						|| alloc_classes->getType(j)
								== AllocationClassInfo::Register) && ii != NULL
						&& !next_is_control_flow)
					; /* no edge needed */
				else if (MoveModel == LimitMoves && i != j && (alloc_classes->getType(i)
						== AllocationClassInfo::Register
						|| alloc_classes->getType(j)
								== AllocationClassInfo::Register) && ii != NULL
						&& !next_is_control_flow)
				{
					/*
					 //only allow vars from ii or nextii to participate in the crossbar
					 unsigned k;
					 mcnf_edge *edge= NULL;

					 //ii
					 for (k = 0; k < ii->getNumOps(); k++)
					 {
					 const RAATInfo::OperandInfo& opinfo =
					 ii->getOpFromIndex(k);
					 if ((opinfo.isDef || opinfo.isUse)
					 && liveout.test(opinfo.regNum))
					 {
					 if (edge == NULL)
					 edge = mcnf_edge_allocate(block);
					 mcnf_edge_add_commodity(edge,
					 reg_quantity.get(opinfo.regNum));
					 }
					 }
					 //nextii
					 for (k = 0; k < nextii->getNumOps(); k++)
					 {
					 const RAATInfo::OperandInfo& opinfo =
					 nextii->getOpFromIndex(k);
					 if (opinfo.isUse
					 && liveout.test(opinfo.regNum))
					 {
					 if (edge == NULL)
					 edge = mcnf_edge_allocate(block);
					 int q = reg_quantity.get(opinfo.regNum);
					 if(!edge->commodity_valid(q))
					 {
					 mcnf_edge_add_commodity(edge, q);
					 }
					 }
					 }
					 if (edge != NULL)
					 {
					 mcnf_connect_nodes(bothn[i], bottom[j], edge);
					 edge->cost = cost;
					 }
					 */
				}
				else
				{
					mcnf_edge *edge = mcnf_edge_allocate(block);
					mcnf_initialize_edge_node_commodities(bothn[i], bottom[j],
							edge);
					mcnf_connect_nodes(bothn[i], bottom[j], edge);
					edge->cost = cost;
				}
			}
		}
	}

	/* let antivars flow back out the memory network even in the middle */
	for (i = 0, n = memory_classes.size(); i < n; i++)
	{
		int m = memory_classes[i];
		int a = alloc_classes->getAntiClass(m);
		int cost = alloc_classes->getMoveCost(m, a);
		if (is_control_flow && cost >= 0)
			cost = after_control_flow_cost(block, cost);
		cost = cost * cost_block_multiplier(block);

		if (bothn[m] && bottom[a] && cost >= 0)
		{
			mcnf_edge *edge = mcnf_edge_allocate(block);
			mcnf_initialize_edge_node_commodities(bothn[m], bottom[a], edge);
			mcnf_connect_nodes(bothn[m], bottom[a], edge);

			edge->cost = cost;
		}
	}

	/* memory nodes need dual constraints and nonanti costs */
	for (i = 0, n = memory_classes.size(); i < n; i++)
	{
		int c = memory_classes[i];

		if (bothn[c] != NULL && bothn[c]->commodities.size() > 0)
		{
			setup_node_dual_prices(bothn[c]);
		}
	}

	/* handle livein onlies
	 Put an edge to the corresponding sink from
	 every compatible class in the top row.
	 */
	for (i = 0, n = in.size(); i < n; i++)
	{
		int q = in[i];
		const BitVector& usable = usable_classes[q];

		create_sink(block, q);
		connect_sink(q, usable, block, top);
	}

	/* handle liveout onlies
	 Must create def/sink/antidef/antisinks.  Hook the defs up
	 to the bottom row
	 */
	for (i = 0, n = out.size(); i < n; i++)
	{
		int q = out[i];
		const BitVector& usable = usable_classes[q];

		create_def(block, q);
		connect_def(q, usable, block, bottom, -1);
	}
}

#define NODE_BLOCKED_IN 1
#define NODE_BLOCKED_OUT 2

/* Updates the various state effected by ii
 This is split into before and after.  The insn is actually created inbetween.

 Updates the available regs based on what this insn does.
 Going in, they refer to the previous crossbar, out to the next.

 Updates the usable_classes (if some quantity can now use/
 no longer use a certain class).  Assumes these changes
 only happen at defs.

 Updates the bitmaps.  Remove unextended dead vars from bottom,
 add defs.  This is the new top.  Remove extended dead vars
 and get the new bottom.

 Update reg_quantity.
 */
void GlobalMCNF::update_insn_info_before(RAATInfo *ii,
		std::vector<char>& available_regs_top,
		std::vector<char>& available_regs_bottom,
		std::set<unsigned>& unavailable_regs_top,
		std::set<unsigned>& unavailable_regs_bottom,
		hash_map<unsigned, BitVector>& usable_classes, BitVector& top,
		BitVector& bottom)
{
	int i, nq, nr, nhr, c;
	std::vector<RAATInfo::QuantityInfo>& info = ii->getQuantityInfo();

	/* temporarily make usable any const equiv regs
	 so we have the correct usables for just this insn */
	for (i = 0, nr = ii->getNumOps(); i < nr; i++)
	{
		const RAATInfo::OperandInfo& opinfo = ii->getOpFromIndex(i);

		if (opinfo.isDef && opinfo.equivalentClass >= 0
				&& alloc_classes->getType(opinfo.equivalentClass)
						== AllocationClassInfo::Constant)
		{
			int q = ii->getOpQuantity(i);
			usable_classes[q].set(opinfo.equivalentClass);
		}
	}

	/* update reg_quantity of each def */
	/* temporarily make usable any const equiv regs
	 so we have the correct usables for just this insn */
	for (i = 0, nr = ii->getNumOps(); i < nr; i++)
	{
		const RAATInfo::OperandInfo& opinfo = ii->getOpFromIndex(i);
		if (opinfo.isDef)
			reg_quantity.put(opinfo.regNum, ii->getOpQuantity(i));
	}

	/* initialize the number of registers free at this insn,
	 need to do this even for insns which have no quantities
	 use the incoming available regs, later handle the outgoing
	 */
	for (c = 0; c < num_alloc_classes; c++)
	{
		if (alloc_classes->getType(c) == AllocationClassInfo::Memory
				&& ii->total_memory_bandwidth() >= 0)
		{
			ii->getClassFreeness(c).free_in = ii->total_memory_bandwidth();
		}
		else if (alloc_classes->getType(c) != AllocationClassInfo::Register
				|| available_regs_bottom[c])
		{
			ii->getClassFreeness(c).free_in = alloc_classes->getCapacity(c);
		}
	}

	/* update the available regs, just update the next top
	 (next bottom will be handled after) */
	available_regs_top = available_regs_bottom;
	unavailable_regs_top = unavailable_regs_bottom;

	for (i = 0, nhr = ii->getNumHardOps(); i < nhr; i++)
	{
		const RAATInfo::OperandInfo& opinfo = ii->getHardOpFromIndex(i);
		unsigned r = TM->getRegisterInfo()->getCanonicalReg(opinfo.regNum);

		if (opinfo.isLastUse && !opinfo.isExtend)
		{
			int c = alloc_classes->getRegisterIndex(r);
			available_regs_top[c]++;
			unavailable_regs_top.erase(r);
			assert(available_regs_top[c] <= alloc_classes->getCapacity(c));
		}
		if ((opinfo.isDef && !opinfo.isUse) || opinfo.isClobber)
		{
			int c = alloc_classes->getRegisterIndex(r);
			available_regs_top[c]--;
			unavailable_regs_top.insert(r);
			assert(available_regs_top[c] >= 0);
			if (!ii->hasSavedReg()) //ignore reg saving insns since we understand them
				clobberedClasses.insert(c);
		}
	}

	for (i = 0, nq = info.size(); i < nq; i++)
	{
		int j, nreg;
		bool livein = false, liveout = false;
		int q = info[i].q;
		BitVector usable = info[i].classes & usable_classes[q];
		if (!usable.any())
			abort();
		/* compute livein/liveout for each quantity*/

		/* sometimes an insn isn't marked as explicitly using
		 a variable (ie, setting a subword) but it is live in anyway
		 and so should be treated as a redefinition */
		nreg = quantities[q].regs.size();
		for (j = 0; j < nreg; j++)
		{
			if (bottom.test(quantities[q].regs[j]))
				livein = true;
		}

		liveout = !info[i].lastuse || info[i].extend;
		/* initialize insn_state info */
		info[i].livein = livein;
		info[i].liveout = liveout;
		info[i].classes = usable;
	}
}

void GlobalMCNF::update_insn_info_after(RAATInfo *ii,
		std::vector<char>& available_regs_top,
		std::vector<char>& available_regs_bottom,
		std::set<unsigned>& unavailable_regs_top,
		std::set<unsigned>& unavailable_regs_bottom,
		hash_map<unsigned, BitVector>& usable_classes, BitVector& top,
		BitVector& bottom)
{
	int i, nr, nhr, c;

	/* update the available regs at the bottom of the next xbar
	 basically just restore clobbered registers */
	available_regs_bottom = available_regs_top;
	unavailable_regs_bottom = unavailable_regs_top;

	for (i = 0, nhr = ii->getNumHardOps(); i < nhr; i++)
	{
		const RAATInfo::OperandInfo& opinfo = ii->getHardOpFromIndex(i);
		int r = TM->getRegisterInfo()->getCanonicalReg(opinfo.regNum);

		if ((!opinfo.isDef && opinfo.isClobber) || /* clober, result not used */
		(opinfo.isLastUse && opinfo.isExtend)) /* result could be used but wasn't */
		{
			int c = alloc_classes->getRegisterIndex(r);
			available_regs_bottom[c]++;
			unavailable_regs_bottom.erase(r);
			assert(available_regs_bottom[c] <= alloc_classes->getCapacity(c));
		}
	}

	/* outgoing freely available regs */
	for (c = 0; c < num_alloc_classes; c++)
	{
		if (alloc_classes->getType(c) == AllocationClassInfo::Memory
				&& ii->total_memory_bandwidth() >= 0)
		{
			ii->getClassFreeness(c).free_out = ii->total_memory_bandwidth();
		}
		else if (alloc_classes->getType(c) != AllocationClassInfo::Register
				|| available_regs_top[c])
		{
			ii->getClassFreeness(c).free_out = alloc_classes->getCapacity(c);
		}
	}

	/* update the bitmaps
	 top is the liveness immediately out of the insn
	 */
	top = bottom;

	/* first kill last uses */
	for (i = 0, nr = ii->getNumOps(); i < nr; i++)
	{
		const RAATInfo::OperandInfo& opinfo = ii->getOpFromIndex(i);

		if (opinfo.isLastUse && !opinfo.isExtend)
			top.reset(opinfo.regNum);
	}
	/* gen defs */
	for (i = 0, nr = ii->getNumOps(); i < nr; i++)
	{
		const RAATInfo::OperandInfo& opinfo = ii->getOpFromIndex(i);

		if (opinfo.isDef)
			top.set(opinfo.regNum);
	}

	bottom = top;
	/* now kill extended last uses */
	for (i = 0, nr = ii->getNumOps(); i < nr; i++)
	{
		const RAATInfo::OperandInfo& opinfo = ii->getOpFromIndex(i);

		if (opinfo.isExtend)
			bottom.reset(opinfo.regNum);
	}

	/* update the usable classes */
	/* make unusable usable classes where the reg is dying */
	for (i = 0, nr = ii->getNumOps(); i < nr; i++)
	{
		const RAATInfo::OperandInfo& opinfo = ii->getOpFromIndex(i);

		if (opinfo.isLastUse)
		{
			int j, n;
			int q = ii->getOpQuantity(i);

			const std::vector<int>& constant_classes =
					alloc_classes->getConstantClasses();
			for (j = 0, n = constant_classes.size(); j < n; j++)
			{
				if (usable_classes[q].test(constant_classes[j]))
					usable_classes[q].reset(constant_classes[j]);
			}
		}
	}
	/* then make usable any const equiv regs */
	for (i = 0, nr = ii->getNumOps(); i < nr; i++)
	{
		const RAATInfo::OperandInfo& opinfo = ii->getOpFromIndex(i);

		if (opinfo.isDef && opinfo.equivalentClass >= 0
				&& alloc_classes->getType(opinfo.equivalentClass)
						== AllocationClassInfo::Constant)
		{
			int q = ii->getOpQuantity(i);
			usable_classes[q].set(opinfo.equivalentClass);
		}
	}
}

/* Create an insn

 Fills out nodes with a row of properly initialized nodes
 (which are added to block).

 Updates the available regs based on what this insn does.
 Going in, they refer to the previous crossbar, out to the next.

 Updates the usable_classes (if some quantity can now use/
 no longer use a certain class).  Assumes these changes
 only happen at defs.

 Updates the bitmaps.  Remove unextended dead vars from bottom,
 add defs.  This is the new top.  Remove extended dead vars
 and get the new bottom.

 Update reg_quantity.

 Creates and links in any necessary defs or sinks.
 */
void GlobalMCNF::create_insn(mcnf_block *block, int insn_number, RAATInfo *ii,
		std::vector<char>& available_regs_top,
		std::vector<char>& available_regs_bottom,
		std::set<unsigned>& unavailable_regs_top,
		std::set<unsigned>& unavailable_regs_bottom,
		hash_map<unsigned, BitVector>& usable_classes, BitVector& top,
		BitVector& bottom, mcnf_node **nodes)
{
	int i, nq, c;
	std::vector<char> single_quantity_node(num_alloc_classes, 0); /* true if a quantity (or two!) must use this node */
	std::vector<RAATInfo::QuantityInfo>& info = ii->getQuantityInfo();

	memset(nodes, 0, sizeof(mcnf_node*) * num_alloc_classes);

	update_insn_info_before(ii, available_regs_top, available_regs_bottom,
			unavailable_regs_top, unavailable_regs_bottom,
			usable_classes, top, bottom);

	for (i = 0, nq = info.size(); i < nq; i++)
	{
		const BitVector& usable = info[i].classes;
		if (!usable.any())
			abort();
		/* for any quantity that has to use a specific node,
		 mark that node as special */
		if (usable.count() == 1)
		{
			c = usable.find_first();
			if (c >= num_alloc_classes)
				abort();

			/* must be exclusively used (into and out of insn) by quantity */
			if (info[i].livein)
				single_quantity_node[c] |= NODE_BLOCKED_IN;
			if (info[i].liveout)
				single_quantity_node[c] |= NODE_BLOCKED_OUT;
		}
	}

	/* for each quantity */
	for (i = 0, nq = info.size(); i < nq; i++)
	{
		int livein, liveout;
		int q = info[i].q;
		BitVector usable = info[i].classes & usable_classes[q];

		livein = info[i].livein;
		liveout = info[i].liveout;

		/* add the quantity to the appropriate node
		 (alloc if node doesn't exist) */

		for (c = 0; c < num_alloc_classes; c++)
		{
			if ((usable.test(c) && (available_regs_top[c]
					|| available_regs_bottom[c])) || (alloc_classes->getType(c)
					== AllocationClassInfo::Memory || alloc_classes->getType(c)
					== AllocationClassInfo::AntiOnly))
			{
				if (nodes[c] == NULL)
				{
					nodes[c] = mcnf_node_allocate(block);
					nodes[c]->type = nInsn;
					nodes[c]->position = insn_number;
					nodes[c]->subindex = 0;
					nodes[c]->allocClass = c;
					mcnf_initialize_node_for_class(nodes[c], c, available_regs_bottom);
					block->body.push_back(nodes[c]);
					nodes[c]->capacity = class_bandwidth(ii, c);
				}

				if (info[i].classes.test(c))
				{
					if (single_quantity_node[c]
							&& alloc_classes->getCapacity(c) == 1)
					{
						/* block all other classes, they can't possibly
						 be allocated here */
						if (usable.count() == 1 && usable.find_first() == c)
							mcnf_node_add_quantity(nodes[c], q);
						else if (!(single_quantity_node[c] & NODE_BLOCKED_IN)
								&& !liveout)
							mcnf_node_add_quantity(nodes[c], q);
						else if (!(single_quantity_node[c] & NODE_BLOCKED_OUT)
								&& !livein)
							mcnf_node_add_quantity(nodes[c], q);
					}
					else
						mcnf_node_add_quantity(nodes[c], q);
				}

				if (alloc_classes->getType(c) == AllocationClassInfo::Memory)
				{
					mcnf_node_add_quantity(nodes[c], -q);
				}

				if (alloc_classes->getType(c) == AllocationClassInfo::AntiOnly
						&& alloc_classes->getCapacity(c) >= 0)
				{
					int antic = alloc_classes->getAntiClass(c);
					nodes[c]->mirror_node = nodes[antic];
					nodes[antic]->mirror_node = nodes[c];
				}

				/* if the var is redefined, do not let it propagate
				 through the antionly network */
				if (!(info[i].def && info[i].use) && alloc_classes->getType(c)
						== AllocationClassInfo::AntiOnly)
					mcnf_node_add_quantity(nodes[c], -q);

				/* now that quantity has been added to node, register any
				 quantity specific costs for this insn */
				if (info[i].equivalentClass == c)
					; /* no cost on this node */
				else
				{
					int qpos = nodes[c]->cindex.get(q);
					if (qpos >= 0)
					{
						int cost = 0;
						for (unsigned r = 0; r < info[i].opindex.size(); r++)
						{
							cost += ii->getClassCost(info[i].opindex[r], c);
						}
						nodes[c]->nonanti_costs[qpos] = cost_block_multiplier(
								block) * cost;
						if (alloc_classes->getType(c)
								== AllocationClassInfo::Memory && info[i].use
								&& info[i].def && (info[i].classes.test(c)))
						{
							/* if a quantity is being redefined in memory,
							 we need to subtract the cost of a store that the
							 anti variable is being forced to pay because of
							 the redefinition (can't just leave the anti alone
							 because of the case where the var is in a register
							 but the anti has been evicted previously)
							 */
							nodes[c]->nonanti_costs[qpos]
									-= cost_block_multiplier(block)
											* memory_store_cost(c);
						}
					}
				}

			}
		}
	}

	/* create defs and sinks as necessary and hook them up */
	for (i = 0, nq = info.size(); i < nq; i++)
	{
		int livein = 0, liveout = 0;
		int q = info[i].q;
		BitVector usable = info[i].classes;

		livein = info[i].livein;
		liveout = info[i].liveout;

		if (info[i].def && !livein) /* create def */
		{
			create_def(block, q);
			connect_def(q, usable, block, nodes, info[i].equivalentClass);
		}
		if (!liveout)
		{
			create_sink(block, q);
			connect_sink(q, usable, block, nodes);
		}
	}

	/* initialize tied constraint info, always do this
	 * so we can adjust predicted solution values  */
	for (i = 0; i < (int) ii->getNumTiedConstraints(); i++)
	{
		const RAATInfo::TiedConstraint& TC = ii->getTiedConstraint(i);
		int srcQ = reg_quantity.get(TC.getInReg());
		int dstQ = reg_quantity.get(TC.getOutReg());

		//I currently can't handle a quantity that is tied at both ends
		if (TiedConstraints)
		{
			bool skip = false;
			for (int j = block->TC_RAATs.size() - 1; j >= 0; j--)
			{
				TC_RAATInfo& TCR = block->TC_RAATs[j];
				if (TCR.dstq == srcQ) //screwed
				{
					std::cerr << "SKIPPING TIED CONSTRAINT\n";
					skip = true;
					break;
				}
				else if (TCR.srcq == srcQ)
					break;
			}
			if (skip)
				continue;
		}
		block->TC_RAATs.push_back(TC_RAATInfo(srcQ, dstQ, TC.getCost()));
		for (c = 0; c < num_alloc_classes; c++)
		{
			if (nodes[c] != NULL && TC.goodAllocClass(c))
			{
				block->TC_RAATs.back().inodes.push_back(nodes[c]);
				block->TCRAATMap[nodes[c]->position] = block->TC_RAATs.size()
						- 1;
			}
		}
	}

	update_insn_info_after(ii, available_regs_top, available_regs_bottom,
			unavailable_regs_top, unavailable_regs_bottom, usable_classes, top, bottom);

}

/* Link an already created insn and xbars */
void GlobalMCNF::link_insn(mcnf_block *block, RAATInfo *ii, mcnf_node **before,
		mcnf_node **insn, mcnf_node **after)
{
	int i, nq, c;
	std::vector<RAATInfo::QuantityInfo>& info = ii->getQuantityInfo();
	hash_set<int> accessed;
	nq = info.size();

	/* mark every quantity that is accessed in the insn
	 (these don't get passed through)
	 */
	for (i = 0; i < nq; i++)
	{
		accessed.insert(info[i].q);
	}

	/* get all the use/defs through */
	link_straight(block, before, insn);
	link_straight(block, insn, after);

	for (i = 0; i < nq; i++)
	{
		if (info[i].def)
		{
			/* shunt all antivars through the insn
			 if a var is redefined, push the anti through a memory node
			 with an eviction edge to get back into the antionly  network */
			int q = info[i].q;

			const std::vector<int>& memory_classes =
					alloc_classes->getMemoryClasses();
			for (int m = 0, n = memory_classes.size(); m < n; m++)
			{
				int c = memory_classes[m];
				int ac = alloc_classes->getAntiClass(c);

				if (before[c] != NULL && after[c] != NULL
						&& before[c]->commodities.size() > 0
						&& after[c]->commodities.size() > 0)
				{
					mcnf_edge *edge;

					if (before[ac] == NULL || after[ac] == NULL)
						abort();

					/* if the quantity is live before the insn, this
					 is a redefinition */
					if (before[c]->cindex.get(q) >= 0)
					{
						int cost = alloc_classes->getMoveCost(ac, c);
						/* edge cost negative if this edge shouldn't exist */
						if (cost < 0)
							abort();

						cost = cost * cost_block_multiplier(block);
						edge = mcnf_edge_allocate(block);
						mcnf_edge_add_commodity(edge, -q);
						edge->cost = cost;
						mcnf_connect_nodes(before[ac], insn[c], edge);

						/* ..and an eviction edge back */
						edge = mcnf_edge_allocate(block);
						mcnf_edge_add_commodity(edge, -q);
						edge->cost = memory_store_cost(c)
								* cost_block_multiplier(block);
						mcnf_connect_nodes(insn[c], after[ac], edge);
					}
				}
			}
		}
		else if (info[i].use && !info[i].extend && !info[i].def && info[i].q
				> 0)
		{
			/* if a quantity is used in an instruction, but is
			 * not defined, then it is okay to overwrite the
			 * register it is in with the defining value of the
			 * instruction as long as the value is clean in memory
			 * so put in an edge to a memory node from each register */
			int q = info[i].q;

			const std::vector<int>& memory_classes =
					alloc_classes->getMemoryClasses();
			for (int c = 0; c < num_alloc_classes; c++)
			{
				if (alloc_classes->getType(c) == AllocationClassInfo::Register)
				{
					for (unsigned m = 0; m < memory_classes.size(); m++)
					{
						int mc = memory_classes[m];
						if (insn[c] != NULL && after[mc] != NULL)
						{
							mcnf_node *src = insn[c];
							mcnf_node *dst = after[mc];
							int sindex = src->cindex.get(q);
							int dindex = dst->cindex.get(q);
							if (sindex >= 0 && dindex >= 0)
							{ //create edge
								mcnf_edge *edge = mcnf_edge_find_or_allocate(
										block, src, dst);
								mcnf_edge_add_commodity(edge, q);
							}
						}
					}
				}
			}
		}
	}

	/* connect vars that weren't accessed to next xbar
	 skip classes that are clobbered */
	for (c = 0; c < num_alloc_classes; c++)
	{
		if (before[c] != NULL && after[c] != NULL
				&& before[c]->commodities.size() > 0
				&& after[c]->commodities.size() > 0)
		{
			mcnf_edge *edge = mcnf_edge_allocate(block);
			int n = before[c]->commodities.size();

			for (i = 0; i < n; i++)
			{
				int q = before[c]->commodities[i];
				if (accessed.count(abs(q)) == 0)
				{
					mcnf_edge_add_commodity(edge, q);
				}
			}
			mcnf_connect_nodes(before[c], after[c], edge);
		}
	}

}

/*
 * Create vector of livein vars
 */
void GlobalMCNF::setup_livesets()
{
	//setup liveins vec
	unsigned numb = blocks.size();
	for (unsigned i = 0; i < numb; i++)
	{
		mcnf_block *block = blocks[i];
		const BitVector& livein = LV->getLiveIn(block->b);
		for (int r = livein.find_first(); r != -1; r = livein.find_next(r))
		{
			if (MRegisterInfo::isVirtualRegister(r))
			{
				block->liveins_vec.push_back(r);
			}
		}

		std::vector<unsigned> tlive;
		LV->setTrulyLiveIn(block->b, tlive);
		hash_set<unsigned> seen;
		for (unsigned j = 0, nlive = tlive.size(); j < nlive; j++)
		{
			unsigned r = TM->getRegisterInfo()->getCanonicalReg(tlive[j]);
			if (alloc_classes->getRegisterIndex(r) >= 0)
			{
				if (seen.count(r))
					continue;
				seen.insert(r);
				block->hard_liveins.push_back(r);
			}
		}
	}
}

bool GlobalMCNF::hasConflict(unsigned r1, unsigned r2)
{
	assert(conflicts.count(r1)> 0);
	assert(conflicts.count(r2)> 0);
	return conflicts[r1].test(r2) || conflicts[r2].test(r1);
}

/*
 * Setup the  conflict map of what variable conflict with others
 */
void GlobalMCNF::setup_conflicts()
{
	const TargetInstrInfo *TII = TM->getInstrInfo();
	int nvars = MF->getSSARegMap()->getLastVirtReg() + 1;
	//initialize conflict sets
	for (unsigned c = 0, nc = num_alloc_classes; c < nc; c++)
	{
		unsigned r = alloc_classes->getRegister(c);
		if (r > 0)
			conflicts[r].resize(nvars);
	}
	for (unsigned r = MRegisterInfo::FirstVirtualRegister, n =
			MF->getSSARegMap()->getLastVirtReg(); r <= n; r++)
	{
		conflicts[r].resize(nvars);
	}

	maxLive = 0;
	maxLiveAcrossCalls = 0;
	maxIntLive = 0;
	maxFPLive = 0;

	unsigned numb = blocks.size();
	for (unsigned i = 0; i < numb; i++)
	{
		mcnf_block *block = blocks[i];
		MachineBasicBlock *MBB = block->b;
		MachineBasicBlock::iterator I = MBB->end();
		BitVector live = LV->getLiveOut(MBB);

		//initialize counts
		int liveCnt = 0;
		int intLiveCnt = 0;
		int fpLiveCnt = 0;
		for (int r = live.find_first(); r >= 0; r = live.find_next(r))
		{
			if (MRegisterInfo::isVirtualRegister(r))
			{
				const TargetRegisterClass *RC =
						MF->getSSARegMap()->getRegClass(r);
				if (MVT::isInteger(*RC->vt_begin()))
				{
					intLiveCnt++;
				}
				else if (MVT::isFloatingPoint(*RC->vt_begin()))
				{
					fpLiveCnt++;
				}
				liveCnt++;
			}
		}

		while (I != MBB->begin())
		{
			I--;
			MachineInstr *MI = I;

			if (TII->isCall(MI->getOpcode()))
			{
				if (liveCnt > (int) maxLiveAcrossCalls)
					maxLiveAcrossCalls = liveCnt;
			}

			for (unsigned op = 0, n = MI->getNumOperands(); op < n; ++op)
			{
				//register operand?
				MachineOperand &MO = MI->getOperand(op);
				unsigned reg;
				if (MO.isReg() && (reg = MO.getReg()))
				{
					if (MO.isDef())
					{
						//remove from live
						live.reset(reg);
						//conflicts with everything alive
						if (MRegisterInfo::isVirtualRegister(reg))
						{
							const TargetRegisterClass *RC =
									MF->getSSARegMap()->getRegClass(reg);
							if (!MO.isDead())
							{
								if (MVT::isInteger(*RC->vt_begin()))
								{
									intLiveCnt--;
								}
								else if (MVT::isFloatingPoint(*RC->vt_begin()))
								{
									fpLiveCnt--;
								}
								liveCnt--;
							}
							conflicts[reg] |= live;
						}
						else
						{
							int c = alloc_classes->getRegisterIndex(reg);
							if (c >= 0)
							{
								unsigned r = alloc_classes->getRegister(c);
								conflicts[r] |= live;
							}
						}
					}
					if (MO.isUse())
					{
						if (!live.test(reg)
								&& MRegisterInfo::isVirtualRegister(reg))
						{
							const TargetRegisterClass *RC =
									MF->getSSARegMap()->getRegClass(reg);
							if (MVT::isInteger(*RC->vt_begin()))
							{
								intLiveCnt++;
							}
							else if (MVT::isFloatingPoint(*RC->vt_begin()))
							{
								fpLiveCnt++;
							}
							liveCnt++;
						}

						//is live
						live.set(reg);
					}
				}
			}
			assert(liveCnt >= 0 && intLiveCnt >= 0 && fpLiveCnt >= 0);
			if (liveCnt > (int) maxLive)
				maxLive = liveCnt;
			if (intLiveCnt > (int) maxIntLive)
				maxIntLive = intLiveCnt;
			if (fpLiveCnt > (int) maxFPLive)
				maxFPLive = fpLiveCnt;
		}
	}
}

/* Set any register preferences globally in the quantity info.  The idea
 * is the preference that's baked into the model might take a while to
 * propagate so heuristical start with the preference.
 * */
void GlobalMCNF::setup_reg_prefs()
{
	typedef hash_map<int, int> iimap;
	iimap prefs;
	//iterate over all the tied constraints
	for (unsigned bnum = 0, n = blocks.size(); bnum < n; bnum++)
	{
		mcnf_block *block = blocks[bnum];
		for (unsigned r = 0, nr = block->RAATs.size(); r < nr; r++)
		{
			RAATInfo *ii = &block->RAATs[r];
			for (unsigned op = 0, n = ii->getNumOps(); op < n; op++)
			{
				int pref = ii->getClassPref(op);
				if (pref >= 0)
				{
					int q = ii->getOpQuantity(op);
					if (prefs.count(q) && prefs[q] != pref)
					{
						//already exists, no pref
						prefs[q] = -1;
					}
					else
					{
						prefs[q] = pref;
					}
				}
			}
		}
	}

	for (iimap::iterator itr = prefs.begin(); itr != prefs.end(); itr++)
	{
		int q = itr->first;
		int pref = itr->second;
		if (pref >= 0)
		{
			quantities[q].prefClass = pref;
		}
	}
}

/* Setup the quantity mappings.  This is where we (partially) deal with tied
 * constraints.  If it is safe to do so, a tied instruction is assigned the
 * same quantity.
 */
void GlobalMCNF::setup_quantities()
{
	hash_map<unsigned, int> indexMap; //index of vars equiv class
	std::vector<std::set<unsigned> > equivClasses;
	unsigned numb = blocks.size();
	for (unsigned b = 0; b < numb; b++)
	{
		mcnf_block *block = blocks[b];
		if (DebugPRA > 3)
			std::cerr << "Merging block " << block->index << "\n";
		// loop over each instruction
		for (unsigned r = 0, nr = block->RAATs.size(); r < nr; r++)
		{
			RAATInfo *ii = &block->RAATs[r];
			//first create a class for every operand
			for (unsigned op = 0, n = ii->getNumOps(); op < n; op++)
			{
				const RAATInfo::OperandInfo& opinfo = ii->getOpFromIndex(op);
				unsigned var = opinfo.regNum;
				if (indexMap.count(var) == 0)
				{
					indexMap[var] = equivClasses.size();
					equivClasses.push_back(std::set<unsigned>());
					equivClasses.back().insert(var);
				}
			}

			//if possible, merge equivclasses of tied vars
			std::vector<int> removedTCs;
			for (unsigned i = 0, n = ii->getNumTiedConstraints(); i < n; i++)
			{
				const RAATInfo::TiedConstraint c = ii->getTiedConstraint(i);
				unsigned src = c.getInReg();
				unsigned dst = c.getOutReg();

				if(DebugPRA > 5)
				{
					std::cerr << "Considering merging ";
					ii->begin()->dump();
					std::cerr << "\n";
				}
				//can implement this constraint as a merge/coalesce of quantities
				if (!c.isMergeable())
					continue;
				if(DisableMerging)
					continue;

				assert(indexMap.count(src)> 0 && indexMap.count(dst)> 0);
				//possible candidate for coalescing

				int sindex = indexMap[src];
				int dindex = indexMap[dst];

				if (sindex == dindex) //already coalesced

				{
					if (DebugPRA > 3)
						std::cerr << "Already MERGED " << src << " " << dst
								<< "\n";
					coalesceableMoves.insert(ii->begin());
					removedTCs.push_back(i);
				}
				else //can only merge if no conflicts between equivclasses

				{
					bool canMerge = true;
					for (std::set<unsigned>::iterator sitr =
							equivClasses[sindex].begin(), sitrE =
							equivClasses[sindex].end(); sitr != sitrE; sitr++)
					{
						for (std::set<unsigned>::iterator ditr =
								equivClasses[dindex].begin(), ditrE =
								equivClasses[dindex].end(); ditr != ditrE; ditr++)
						{
							if (hasConflict(*sitr, *ditr))
							{
								if (DebugPRA > 3)
									std::cerr << "CONFLICT " << *sitr << " "
											<< *ditr << "\n";
								canMerge = false;
								break;
							}
						}
						if (!canMerge)
							break;
					}

					if (canMerge)
					{
						if (DebugPRA > 3)
							std::cerr << "MERGING " << src << " " << dst
									<< "\n";
						coalesceableMoves.insert(ii->begin());
						removedTCs.push_back(i);
						//combine equivclasses into src
						equivClasses[sindex].insert(
								equivClasses[dindex].begin(),
								equivClasses[dindex].end());
						for (std::set<unsigned>::iterator ditr =
								equivClasses[dindex].begin(), ditrE =
								equivClasses[dindex].end(); ditr != ditrE; ditr++)
						{
							indexMap[*ditr] = sindex; //point to merged class
						}
						equivClasses[dindex].clear();
					}
					else if (DebugPRA > 3)
						std::cerr << "CANT MERGE " << src << " " << dst << "\n";
				}
			}

			//remove the tied constraints that we converted to merged quantities
			//do this in reverse order so indexes remain valid
			for (int i = removedTCs.size() - 1; i >= 0; i--)
			{
				ii->removeTiedConstraint(i);
			}
		}
	}
	//now setup quantity mappings
	if (DebugPRA > 1)
		std::cerr << "Quantity Mappings:\n";
	assert(quantities.size() == 1);
	for (unsigned i = 0, n = equivClasses.size(); i < n; i++)
	{
		if (equivClasses[i].size() > 0)
		{
			int q = quantities.size();
			quantities.push_back(quantity(
					alloc_classes->getMemoryClasses().size()));
			if (DebugPRA > 1)
				std::cerr << q << ":";
			for (std::set<unsigned>::iterator itr = equivClasses[i].begin(),
					itrE = equivClasses[i].end(); itr != itrE; itr++)
			{
				int var = *itr;
				reg_quantity.put(var, q);
				quantities.back().regs.push_back(var);
				if (LI->hasInterval(var))
				{
					quantities.back().weight += LI->getInterval(var).weight;
				}
				if (DebugPRA > 1)
					std::cerr << " " << var;
			}
			if (DebugPRA > 1)
				std::cerr << "  weight " << quantities[q].weight << "\n";
		}
	}

	//now update RAAT info
	for (unsigned b = 0; b < numb; b++)
	{
		mcnf_block *block = blocks[b];
		// loop over each instruction
		for (unsigned r = 0, nr = block->RAATs.size(); r < nr; r++)
		{
			RAATInfo *ii = &block->RAATs[r];
			//first create a class for every operand
			for (unsigned op = 0, n = ii->getNumOps(); op < n; op++)
			{
				const RAATInfo::OperandInfo& opinfo = ii->getOpFromIndex(op);
				unsigned var = opinfo.regNum;
				ii->setOpQuantity(op, reg_quantity.get(var));
			}
		}
	}
}

// basic cfg initialization, setup blocks and control flow
void GlobalMCNF::mcnf_cfg_init()
{
	int num_blocks = MF->size();

	int total_insns = 0;

	num_edges = 0;
	num_nodes = 0;
	num_RAATs = 0;

	hash_map<int, int> block_index;
	blocks.resize(num_blocks);
	block_order.resize(num_blocks, 0);
	block_order_index.resize(num_blocks, 0);

	int bnum = 0;
	for (MachineFunction::iterator MBB = MF->begin(), MBBe = MF->end(); MBB
			!= MBBe; ++MBB, ++bnum)
	{
		int index = MBB->getNumber();
		mcnf_block* block = new mcnf_block(this);
		blocks[bnum] = block;

		//initialize block
		block->index = bnum;
		block->b = MBB;

		block_index[index] = bnum;

		total_insns += MBB->size();

	}

	/* now compute control flow information */
	for (int i = 0, n = blocks.size(); i < n; i++)
	{
		mcnf_block* block = blocks[i];
		MachineBasicBlock *b = block->b;

		for (MachineBasicBlock::const_succ_iterator SI = b->succ_begin(), E =
				b->succ_end(); SI != E; ++SI)
		{
			MachineBasicBlock *SuccMBB = *SI;
			mcnf_block *succ = blocks[block_index[SuccMBB->getNumber()]];
			block->succs.push_back(succ);
			succ->preds.push_back(block);
		}
	}

	/* set blocks where the entry and exit are mutually constrained */
	for (int i = 0, n = blocks.size(); i < n; i++)
	{
		mcnf_block* block = blocks[i];
		block->compute_entry_exit_constraints();
	}

	/* setup block order arrays */
	dfo_blockorder();
	for (int i = 0, n = blocks.size(); i < n; i++)
	{
		block_order_index[block_order[i]] = i;
	}
}

/* Populate the mcnf_block with the RAATs from the instruction stream
 * in the MachineBasicBlock.  Right now we do a conversion from insns to
 * RAATs here, but eventually the insn selector will generate RAATs and
 * this will just be a copy.
 *
 * TODO: merge mergeable RAATs into one
 */
void GlobalMCNF::generateRAATs()
{
	unsigned numb = blocks.size();
	for (unsigned b = 0; b < numb; b++)
	{
		mcnf_block *block = blocks[b];
		MachineBasicBlock *MBB = block->b;
		for (MachineBasicBlock::iterator I = MBB->begin(), E = MBB->end(); I
				!= E; I++)
		{
			block->RAATs.push_back(RAATInfo(I));
		}
		num_RAATs += block->RAATs.size();
	}
}

/* Return true if we do not need to model ii.  For example, a coalesced move
 * or an insn with no operands */
bool GlobalMCNF::can_ignore_insn(RAATInfo *ii)
{
	//simple test, isn't actually checking pressure; if an instruction
	//has been coalesced, can ignore one of the crossbars
	if (ii->getNumOps() == 2 && ii->getNumHardOps() == 0
			&& ii->getOpQuantity(0) == ii->getOpQuantity(1))
	{
		//this is not true if the insn is imposing class restrictions, ie
		//MOV16to16_ and MOV32to32_
		//being unnecessarily conservative, but if we don't include potentially
		// constraining, non-move insns tranformations get put in the wrong place;
		//can fix this possibly be renumbering xbars?
		const TargetInstrInfo *TII = TM->getInstrInfo();
		unsigned src, dst;
		if (TII->isMoveInstr(*ii->begin(), src, dst))
		{
			num_ignoreable++;
			return true;
		}
	}
	//no operands and not a control flow insn (which affect allocation)
	if (ii->getNumHardOps() == 0 && ii->getNumOps() == 0
			&& !ii->isControlFlow())
	{
		num_ignoreable++;
		return true;
	}

	return false;
}

static int count_available_regs(std::vector<char>& available_regs)
{
	int ret = 0;
	for (unsigned i = 0, n = available_regs.size(); i < n; i++)
	{
		if (alloc_classes->getType(i) == AllocationClassInfo::Register
				&& available_regs[i])
			ret += available_regs[i];
	}
	return ret;
}

/* Return true if a full crossbar is not needed.
 */
bool GlobalMCNF::no_pressure_between_insns(RAATInfo *ii, RAATInfo *nextii,
		BitVector& live, std::vector<char>& available_regs, hash_map<unsigned,
				BitVector>& usable_classes)
{
	int ilive, nlive, avail;
	/* count number of variables made live (presumably) by
	 these insns, conservatively ignore last uses since order matters */
	ilive = ii->getNumDefs();
	nlive = live.count();
	avail = count_available_regs(available_regs);

	if (avail > (nlive + ilive))
	{ //potentially no pressure, basically do a quick allocation to
		//make sure the right kinds of registers are available
		/* For each of nextii's regs */
		int ndefs = nextii->getNumDefs();
		for (unsigned i = 0, n = nextii->getNumOps(); i < n; i++)
		{
			int c;
			std::vector<char> regavail(num_alloc_classes);
			const RAATInfo::OperandInfo& op = nextii->getOpFromIndex(i);
			/* compute which registers are available for this var and how many */
			int numavail = 0;

			for (c = 0; c < num_alloc_classes; c++)
			{
				regavail[c] = 0;
				if (alloc_classes->getType(c) == AllocationClassInfo::Register
						&& op.allowableClasses.test(c))
				{
					regavail[c] = available_regs[c];
					numavail += available_regs[c];
				}
			}

			/* now for every reg that is live and could possible occupy a
			 register this var could use, reduce the number available
			 */
			for (int r = live.find_first(); r >= 0; r = live.find_next(r))
			{
				int q = reg_quantity.get(r);
				if (q > 0)
				{
					if ((usable_classes[q] & op.allowableClasses).any())
						numavail--;
				}
			}
			/* for every def in ii that isn't also a use (so is presumably
			 alive into nextii and wasn't in the live set) decrement avail
			 if conflicting */
			for (unsigned d = 0, n2 = ii->getNumOps(); d < n2; d++)
			{
				const RAATInfo::OperandInfo& op2 = ii->getOpFromIndex(d);
				if (op2.isDef && !op2.isUse && (op2.allowableClasses
						& op.allowableClasses).any())
					numavail--;
			}

			/* if nothing is available, no good */
			if (numavail < 0)
				return false;
			/* defs need something to be available */
			if (op.isDef && !op.isUse && numavail < ndefs)
				return false;
		}
		return true;
	}

	return false;
}

/* Assign a topological ordering to the nodes of a block
 Relies on nondef and sink nodes being allocated in order.
 */
void GlobalMCNF::mcnf_block_order_nodes(mcnf_block *block)
{
	int n, num;
	int currnum = 0;
	num = block->defs.size();

	/* all the defs go first */
	for (n = 0; n < num; n++)
	{
		mcnf_node *node = block->defs[n];
		node->node_order = currnum++;
	}

	/* then entry nodes */
	num = block->entries.size();
	for (n = 0; n < num; n++)
	{
		mcnf_node *node = block->entries[n];
		assert(node != NULL);
		node->node_order = currnum++;
	}

	/* xbar and insn nodes */
	num = block->body.size();
	for (n = 0; n < num; n++)
	{
		mcnf_node *node = block->body[n];
		assert(node != NULL);
		node->node_order = currnum++;
	}

	/* exit nodes */
	num = block->exits.size();
	for (n = 0; n < num; n++)
	{
		mcnf_node *node = block->exits[n];
		assert(node != NULL);
		node->node_order = currnum++;
	}

	/* finally sinks */
	num = block->sinks.size();
	for (n = 0; n < num; n++)
	{
		mcnf_node *node = block->sinks[n];
		assert(node != NULL);
		node->node_order = currnum++;
	}

}
/* Compute class occupancy (which hard registers are available)

 Any hard registers that are alloc classes that are live at the
 start of the block get set to zero.
 */
void GlobalMCNF::initialize_block_available_regs(mcnf_block *block,
		std::vector<char>& available_regs, std::set<unsigned>& unavailable_regs)
{
	available_regs.clear();
	available_regs.resize(num_alloc_classes, 1);
	unavailable_regs.clear();

	for (int i = 0; i < num_alloc_classes; i++)
	{
		available_regs[i] = alloc_classes->getCapacity(i);
	}
	for (unsigned i = 0, n = block->hard_liveins.size(); i < n; i++)
	{
		unsigned r = TM->getRegisterInfo()->getCanonicalReg(block->hard_liveins[i]);
		int c = alloc_classes->getRegisterIndex(r);
		if (c >= 0)
		{
			unavailable_regs.insert(r);
			available_regs[c]--;
			assert(available_regs[c] >= 0);
		}
	}
}

/* Make sure any constant classes are properly set at block entry.

 Any pseudos in the livein set which are reg_equiv to a const
 have the appropriate constant classes added to their usable set.


 */
void GlobalMCNF::update_block_usable_classes(mcnf_block *block, hash_map<
		unsigned, BitVector>& usable_classes)
{
	//todo: handle constants
}

/* Compute possible allocation classes for every variable.
 If the corresponding bit is set, the quantity may be
 allocated to that class somewhere (if not everywhere).

 Usability is determined by taking the union of the acceptable
 classes of each variable making up a quantity.  Constant classes
 are always considered unacceptable by default and nonconstant noreg
 classes are considered acceptable.
 */
void GlobalMCNF::initialize_usable_classes(
		hash_map<unsigned, BitVector>& usable_classes)
{
	usable_classes.clear();

	//for each quantity
	for (unsigned q = 1, nq = quantities.size(); q < nq; q++)
	{
		BitVector& val = usable_classes[q];
		val.resize(num_alloc_classes, false);
		//consider each var in the quantity
		for (unsigned i = 0, last = quantities[q].regs.size(); i < last; i++)
		{
			unsigned var = quantities[q].regs[i];

			//for each allocation class
			for (unsigned c = 0, nac = num_alloc_classes; c < nac; c++)
			{
				if (alloc_classes->getType(c) == AllocationClassInfo::Memory)
				{
					val.set(c);
				}
				else if (alloc_classes->getType(c)
						== AllocationClassInfo::Register)
				{
					//can var be allocated to this allocation class?
					const TargetRegisterClass* RC =
							MF->getSSARegMap()->getRegClass(var);

					if (alloc_classes->isCompatible(*MF, RC, c))
					{
						val.set(c);
					}
				}
			}
		}
	}
}

// initialize next use information
void GlobalMCNF::initialize_quantity_next_use(mcnf_block *block)
{
	//scan backwards propagating info
	int nq = quantities.size();
	int ni = block->RAATs.size();
	block->quantity_next_use.resize(ni + 1);
	block->quantity_next_use[ni].resize(nq, ni);
	for (int i = ni - 1; i >= 0; i--)
	{
		block->quantity_next_use[i] = block->quantity_next_use[i + 1];

		RAATInfo *ii = &block->RAATs[i];
		for (int op = 0, n = ii->getNumOps(); op < n; op++)
		{
			block->quantity_next_use[i][ii->getOpQuantity(op)] = i;
		}
	}

}

/* Link the entry and exit nodes of blocks.
 This also allocates the boundary_prices in the entry/exit nodes.
 */
void GlobalMCNF::mcnf_link_blocks(void)
{
	int b, nb;

	nb = blocks.size();
	/* link each block's exit nodes to each of
	 its successors entry nodes of the same class */
	for (b = 0; b < nb; b++)
	{
		int s, ns;
		mcnf_block *block = blocks[b];
		int n = block->exits.size();
		ns = block->succs.size();

		for (s = 0; s < ns; s++)
		{
			int i;
			int j;
			mcnf_block *succ = block->succs[s];
			int nentries = succ->entries.size();

			for (i = 0; i < n; i++)
			{
				mcnf_node *a, *b = NULL;
				mcnf_edge *edge;
				a = block->exits[i];

				for (j = 0; j < nentries; j++)
				{
					b = succ->entries[j];

					if (a->allocClass == b->allocClass)
						break;
				}

				if (j == nentries) /* couldn't find an equivalent node */
					continue;

				edge = mcnf_edge_allocate(block);
				mcnf_initialize_edge_node_commodities(a, b, edge);
				mcnf_connect_nodes(a, b, edge);

				// allow antivars to freely switch from antinodes to the
				//memory class
				if (CrossBlockAntis && alloc_classes->getType(a->allocClass)
						== AllocationClassInfo::AntiOnly && b->mirror_node
						!= NULL)
				{
					mcnf_edge *edge = mcnf_edge_allocate(block);
					mcnf_initialize_edge_node_commodities(a, b->mirror_node,
							edge);
					mcnf_connect_nodes(a, b->mirror_node, edge);
				}
			}
		}
	}

	/* allocate boundary_prices and gradients*/
	for (b = 0; b < nb; b++)
	{
		int i;
		mcnf_block *block = blocks[b];
		int n = block->exits.size();

		/* exits */
		for (i = 0; i < n; i++)
		{
			int nc;
			mcnf_node *node = block->exits[i];
			nc = node->commodities.size();

			if (node->succs.size() > 0)
			{
				int ns = node->succs.size();
				node->boundary_prices = new double*[ns];
				node->boundary_gradients = new double*[ns];
				for (int s = 0; s < ns; s++)
				{
					node->boundary_prices[s] = new double[nc];
					node->boundary_gradients[s] = new double[nc];
					memset(node->boundary_prices[s], 0, nc * sizeof(double));
					memset(node->boundary_gradients[s], 0, nc * sizeof(double));
				}
			}

			node->used_classes = new usable_info[nc];
			memset(node->used_classes, 0, sizeof(usable_info) * nc);
		}

		/* entries */
		n = block->entries.size();
		for (i = 0; i < n; i++)
		{
			int nc;
			mcnf_node *node = block->entries[i];
			nc = node->commodities.size();

			if (node->preds.size() > 0)
			{
				int np = node->preds.size();
				node->boundary_prices = new double*[np];
				node->boundary_gradients = new double*[np];
				for (int p = 0; p < np; p++)
				{
					node->boundary_prices[p] = new double[nc];
					node->boundary_gradients[p] = new double[nc];
					memset(node->boundary_prices[p], 0, nc * sizeof(double));
					memset(node->boundary_gradients[p], 0, nc * sizeof(double));
				}
			}

			node->used_classes = new usable_info[nc];
			memset(node->used_classes, 0, sizeof(usable_info) * nc);
		}

	}
}

//output a bit mask of the life variables
void GlobalMCNF::dump_live_quantities_map(const BitVector& live)
{
	for (unsigned i = 1, n = quantities.size(); i < n; i++)
	{
		bool on = false;
		for (unsigned r = 0, nr = quantities[i].regs.size(); r < nr; r++)
		{
			unsigned reg = quantities[i].regs[r];
			if (live.test(reg))
				on = true;
		}
		std::cerr << "\t" << on;
	}
	std::cerr << "\n";
}

//if profile info is valid, find the largest block cost so we can normalize
void GlobalMCNF::compute_max_block_cost()
{

	unsigned max = 0;
	for (unsigned bnum = 0, n = blocks.size(); bnum < n; bnum++)
	{
		mcnf_block *block;
		block = blocks[bnum];
		MachineBasicBlock *MBB = block->b;
		unsigned cnt = MBB->getFrequency();
		if (cnt > max)
			max = cnt;

		if (DebugPRA > 1)
		{
			std::cerr << "Block " << bnum << " cost " << cnt << "\n";
		}
	}
	max_block_cost = max;

}

/* Construct a global MCNF representation of the register allocation
 * problem for function mf. */
void GlobalMCNF::build(MachineFunction *mf, DataFlowLiveVariables *lv,
		LiveIntervals *li, LoopInfo *loops)
{
	clear();
	MF = mf;
	TM = &MF->getTarget();
	LV = lv;
	LI = li;
	Loops = loops;
	num_alloc_classes = alloc_classes->numClasses();

	std::vector<char> available_regs_top(num_alloc_classes, 0);
	std::vector<char> available_regs_bottom(num_alloc_classes, 0);
	std::vector<char> available_regs_saved(num_alloc_classes, 0);
	std::set<unsigned> unavailable_regs_top;
	std::set<unsigned> unavailable_regs_bottom;

	/* vectors representing the slice of the MCNF we are building,
	 implemented as pointers to avoid copying as we shift the window */
	mcnf_node **current_top;
	mcnf_node **current_bottom;
	mcnf_node **previous_bottom;
	mcnf_node **current_insn;
	mcnf_node **swap_tmp;
	current_top = (mcnf_node**) alloca(num_alloc_classes*sizeof(mcnf_node*));
	current_bottom = (mcnf_node**) alloca(num_alloc_classes*sizeof(mcnf_node*));
	previous_bottom = (mcnf_node**) alloca(num_alloc_classes*sizeof(mcnf_node*));
	current_insn = (mcnf_node**) alloca(num_alloc_classes*sizeof(mcnf_node*));

	BitVector livein_top, liveout_bottom;
	BitVector hard_regsc(MF->getSSARegMap()->getLastVirtReg() + 1);
	hash_map<unsigned, BitVector> usable_classes; /* for each quantity, permitted classes */

	for (unsigned r = 0; r < MRegisterInfo::FirstVirtualRegister; r++)
	{
		hard_regsc.set(r);
	}
	hard_regsc.flip(); //complement for easy intersection

	mcnf_cfg_init();
	setup_livesets();
	setup_conflicts();
	compute_max_block_cost();

	if (DebugPRA > 1)
	{
		for (unsigned bnum = 0, n = blocks.size(); bnum < n; bnum++)
		{
			mcnf_block *block;
			block = blocks[bnum];
			std::cerr << "Block " << bnum << " normalized cost "
					<< cost_block_multiplier(block) << "\n";
		}
	}
	/* generate RAATInfos for each block */
	generateRAATs();
	if (CPLEXCoalesce)
		cplex_output_coalesce_model(*this);
	//create the quantity maps
	setup_quantities();
	//compute dominators
	setup_def_dominators();
	//hard reg prefs - doesn't actually help. oh well.
	//setup_reg_prefs();

	//initilize usable classes based on var types
	initialize_usable_classes(usable_classes);

	/* Build MCNF for each block */
	for (unsigned bnum = 0, n = blocks.size(); bnum < n; bnum++)
	{
		mcnf_block *block;
		block = blocks[bnum];
		MachineBasicBlock *MBB = block->b;

		livein_top = LV->getLiveIn(MBB) & hard_regsc;
		liveout_bottom = livein_top;

		if (DebugPRA > 1)
		{
			std::cerr << "Block " << block->index << " LIVEIN: ";
			livein_top.print(std::cerr);
			std::cerr << "\n";
		}
		/* Compute class occupancy (which hard registers are available) */
		initialize_block_available_regs(block, available_regs_top, unavailable_regs_top);
		available_regs_bottom = available_regs_top;
		unavailable_regs_bottom = unavailable_regs_top;

		/* Make sure any fickle classes are properly set at block entry */
		update_block_usable_classes(block, usable_classes);

		/* Create entry nodes.  For each class, if it
		 is conceivable that a livein quantity could be
		 allocated to that class, create an entry node for
		 that class (if one doesn't exist already) and
		 allow that commodity in that node. */
		create_entry_nodes(block, available_regs_bottom, unavailable_regs_bottom,
				usable_classes,
				livein_top, previous_bottom);

		/* Create the initial xbar. */
		create_xbar(block, NULL, NULL, -1, available_regs_top,
				available_regs_bottom, unavailable_regs_top,
				unavailable_regs_bottom, usable_classes, livein_top,
				liveout_bottom, current_top, current_bottom, 0);

		/* Link xbar with entry nodes */
		link_straight(block, previous_bottom, current_top);

		/* set current xbar to be the previous xbar */
		/* swap pointers (avoid memcpy) */
		swap_tmp = previous_bottom;
		previous_bottom = current_bottom;
		current_bottom = swap_tmp;

		/* process each instruction */
		unsigned i, n, nexti;
		for (i = 0, n = block->RAATs.size(); i < n; i = nexti)
		{
			RAATInfo *ii = &block->RAATs[i];
			RAATInfo *nextii = NULL;
			nexti = i + 1;
			//skip uninteresting instructions
			while (nexti < n && can_ignore_insn(&block->RAATs[nexti]))
				nexti++;

			if (nexti < n)
				nextii = &block->RAATs[nexti];

			if (can_ignore_insn(ii)) //ignore an insn even if its first
			{
				//if we're skipping instructions, in particular moves, we
				//need to keep the live sets uptodate
				for (unsigned i2 = i; i2 < nexti; i2++)
				{
					RAATInfo *skippedii = &block->RAATs[i2];
					update_insn_info_before(skippedii, available_regs_top,
							available_regs_bottom, unavailable_regs_top,
							unavailable_regs_bottom, usable_classes, livein_top,
							liveout_bottom);
					update_insn_info_after(skippedii, available_regs_top,
							available_regs_bottom, unavailable_regs_top,
							unavailable_regs_bottom, usable_classes, livein_top,
							liveout_bottom);
				}
				continue;
			}

			if (DebugPRA > 1)
			{
				if (DebugPRA > 3)
				{
					std::cerr << "CREATE; live size " << liveout_bottom.count();
					std::cerr << "; avail size " << count_available_regs(
							available_regs_top) << ", "
							<< count_available_regs(available_regs_bottom);
					std::cerr << " unavail: ";
					copy(unavailable_regs_top.begin(), unavailable_regs_top.end(), std::ostream_iterator<int>(std::cerr," "));
					std::cerr << "\n";
					ii->dump();
				}
				if (DebugPathsPRA)
				{
					std::cerr << "LIVESET " << i << "b";
					dump_live_quantities_map(liveout_bottom);
				}
			}
			bool supports_simple_xbar = false;
			if (nextii != NULL)
				supports_simple_xbar = no_pressure_between_insns(ii, nextii,
						liveout_bottom, available_regs_bottom, usable_classes);

			/* Create insn.  This will setup defs and sinks as necessary.
			 Also 	modifies BOTH livein_insn and liveout_insn.  Upon exit,
			 the liveness bitmaps refer the the top and bottom of the
			 next xbar as do the available_regs.
			 Updates the available hard registers and usable_classes
			 */
			create_insn(block, i, ii, available_regs_top,
					available_regs_bottom, unavailable_regs_top,
					unavailable_regs_bottom, usable_classes, livein_top,
					liveout_bottom, current_insn);

			/* Create following xbar. Fills in current_top/bottom */
			create_xbar(block, ii, nextii, i, available_regs_top,
					available_regs_bottom, unavailable_regs_top,
					unavailable_regs_bottom, usable_classes, livein_top,
					liveout_bottom, current_top, current_bottom,
					supports_simple_xbar);

			if (DebugPathsPRA)
			{
				std::cerr << "LIVESET " << i << "a";
				dump_live_quantities_map(livein_top);
			}

			/* Link insn with previous and this xbar.
			 This will handle any side constraints (eviction edge)
			 but not evictions to sinks (that's handled in insn creation)
			 unless it's a variable that dies in the xbar (unused vars).
			 */
			link_insn(block, ii, previous_bottom, current_insn, current_top);

			//if we're skipping instructions, in particular moves, we
			//need to keep the live sets uptodate
			for (unsigned i2 = i + 1; i2 < nexti; i2++)
			{
				RAATInfo *skippedii = &block->RAATs[i2];
				update_insn_info_before(skippedii, available_regs_top,
						available_regs_bottom, unavailable_regs_top,
						unavailable_regs_bottom, usable_classes, livein_top,
						liveout_bottom);
				update_insn_info_after(skippedii, available_regs_top,
						available_regs_bottom, unavailable_regs_top,
						unavailable_regs_bottom, usable_classes, livein_top,
						liveout_bottom);
			}
			/* set current xbar to be the previous xbar */
			/* swap pointers (avoid memcpy) */
			swap_tmp = previous_bottom;
			previous_bottom = current_bottom;
			current_bottom = swap_tmp;
		} /*for each insn*/

		/* Create exit nodes. */
		create_exit_nodes(block, available_regs_bottom, unavailable_regs_bottom,
				usable_classes, liveout_bottom, current_top, n);

		/* Link exit nodes to previous xbar */
		link_straight(block, previous_bottom, current_top);

		/* assign topo ordering */
		mcnf_block_order_nodes(block);

		/* initialize quantity next use information */
		initialize_quantity_next_use(block);

	} /*blocks*/

	/* Now link together block MCNF's */
	mcnf_link_blocks();

	if (DebugPRA > 1)
		dump_info();

}

/* reserve memory and initialize shortest path info for the solver */
class nodeAllocAndInitializeShortest
{
	//these are bump pointers
	char *binfo_pool;
	char *sinfo_pool[MAX_K];
	char *fsinfo_pool;
	//the end of the pools
	char *binfo_end;
	char *sinfo_end[MAX_K];
	char *fsinfo_end;

public:
	nodeAllocAndInitializeShortest(unsigned num, char *bpool, char **spool,
			char *fspool) :
		binfo_pool(bpool), fsinfo_pool(fspool)
	{
		binfo_end = binfo_pool + num * sizeof(balanced_info);
		fsinfo_end = fsinfo_pool + num * sizeof(feasible_solution_info);
		for (unsigned i = 0; i < MAX_K; i++)
		{
			sinfo_pool[i] = spool[i];
			sinfo_end[i] = spool[i] + num * sizeof(shortest_info);
		}
	}

	void operator()(mcnf_node *node)
	{
		int nc = node->commodities.size();
		node->binfo = (balanced_info*) binfo_pool;
		binfo_pool += nc * sizeof(balanced_info);
		assert(binfo_pool <= binfo_end);

		for (int i = 0; i < MAX_K; i++)
		{
			if (node->sinfo[i] == NULL)
			{
				node->sinfo[i] = (shortest_info*) sinfo_pool[i];
				sinfo_pool[i] += nc * sizeof(shortest_info);
				assert(sinfo_pool[i] <= sinfo_end[i]);
			}
		}

		node->fsinfo = (feasible_solution_info*) fsinfo_pool;
		fsinfo_pool += nc * sizeof(feasible_solution_info);
		assert(fsinfo_pool <= fsinfo_end);
		node->feas_marked.assign(nc, false);
	}
};

class tallyCommodities
{
	unsigned total;

public:
	tallyCommodities() :
		total(0)
	{
	}

	void operator()(mcnf_node *node)
	{
		total += node->commodities.size();
	}

	unsigned getTotal()
	{
		return total;
	}
};

/* zero out all price data in a node */
static void clear_node_prices(mcnf_node *node)
{
	node->price = 0;
	if (node->dual_prices)
	{
		int nc = node->commodities.size();
		memset(node->dual_prices, 0, sizeof(double) * nc);
	}

	if (node->boundary_prices)
	{
		int n = 0;
		int nc = node->commodities.size();
		if (node->type == nExit)
		{
			n = node->succs.size();
		}
		else if (node->type == nEntry)
		{
			n = node->preds.size();
		}
		else
			abort();

		for (int i = 0; i < n; i++)
		{
			memset(node->boundary_prices[i], 0, sizeof(double) * nc);
		}
	}
}

//clear price data from nodes
void GlobalMCNF::clear_prices()
{
	traverse_nodes(clear_node_prices);
}

void GlobalMCNF::clear_solver_data()
{
	free(binfo_pool);
	binfo_pool = NULL;
	free(fsinfo_pool);
	fsinfo_pool = NULL;
	for (unsigned i = 0; i < MAX_K; i++)
	{
		free(sinfo_pool[i]);
		sinfo_pool[i] = NULL;
	}
}

//initialize all data used by the solver
void GlobalMCNF::initialize_solver_data()
{
	tallyCommodities tally;
	traverse_nodes(tally);
	unsigned num = tally.getTotal();

	if (DebugPRA)
	{
		std::string fname =
				MF->getFunction()->getParent()->getModuleIdentifier() + "_"
						+ MF->getFunction()->getName();
		std::cerr << fname << " TotalCommod = " << num << "\n";
	}
	//allocate pools used for solver data
	assert(binfo_pool == NULL);
	binfo_pool = (char*) malloc(sizeof(balanced_info) * num);
	for (unsigned i = 0; i < MAX_K; i++)
	{
		assert(sinfo_pool[i] == NULL);
		sinfo_pool[i] = (char*) malloc(sizeof(shortest_info) * num);
	}
	assert(fsinfo_pool == NULL);
	fsinfo_pool = (char*) malloc(sizeof(feasible_solution_info) * num);

	//assign memory in each node
	nodeAllocAndInitializeShortest initer(num, binfo_pool, sinfo_pool,
			fsinfo_pool);
	traverse_nodes(initer);

	for (int b = 0, nb = blocks.size(); b < nb; b++)
	{
		mcnf_block *block = blocks[b];
		block->quantity_prices.clear();
	}
}

/* Do the necessary initialization for the path algorithms on all nodes
 clears the visited bit
 sets the balanced info to zero */
void GlobalMCNF::initialize_path_info()
{
	int b, nb;
	nb = blocks.size();

	for (b = 0; b < nb; b++)
	{
		mcnf_block *block = blocks[b];
		block->initialize_path_info();
	}
	for (unsigned q = 0, nq = quantities.size(); q < nq; q++)
	{
		quantities[q].priced_cost = 0.0;
		quantities[q].unpriced_cost = 0;
		quantities[q].anti_priced_cost = 0.0;
		quantities[q].anti_unpriced_cost = 0;
		quantities[q].lastAlloced = -1;
		quantities[q].spilled = false;
	}
}

double GlobalMCNF::shortest_balanced_paths()
{
	int b, nb;
	double ret = 0.0;
	nb = blocks.size();

	initialize_path_info();
	/* compute the shortest paths for each block */
	for (b = 0; b < nb; b++)
	{
		mcnf_block *block = blocks[b];

		/* forward analysis, finds shortest paths
		 and count the number of paths */
		ret += block->shortest_paths(mcnf_edge_price, mcnf_edge_cost, true);
		if (TiedConstraints)
		{
			double TC_adjust = block->shortest_tied_constraint_adjust();
			ret += TC_adjust;

			if (DebugPRA > 4)
			{
				std::cerr << "Unconstrainted TC_adjust " << TC_adjust << "\n";
			}
		}
		/* backwards, propagates back the total paths
		 info so we get a ratio at each node that participates
		 in a shortest path */
		block->set_balanced_flows();
	}

	if (DebugPathsPRA)
	{
		double unpriced_cost = 0.0;
		double prices_only = 0.0;
		fprintf(stderr, "Unfeasible price/costs:\n");
		for (unsigned c = 1; c < quantities.size(); c++)
		{
			fprintf(stderr, "\t%d:  %f(%d)\n", c, quantities[c].priced_cost,
					quantities[c].unpriced_cost);
			fprintf(stderr, "\t-%d:  %f(%d)\n", c,
					quantities[c].anti_priced_cost,
					quantities[c].anti_unpriced_cost);

			unpriced_cost += quantities[c].unpriced_cost
					+ quantities[c].anti_unpriced_cost;
			prices_only += quantities[c].priced_cost
					- quantities[c].unpriced_cost;
			prices_only += quantities[c].anti_priced_cost
					- quantities[c].anti_unpriced_cost;
		}

		fprintf(stderr, "UNPRICED COST %f\n", unpriced_cost);
		fprintf(stderr, "PRICES ONLY %f\n", prices_only);
	}

	return ret;
}

//initializes just info used by feasible solution finder, leaves shortest alone
void initialize_node_for_feasible_solution_finder(mcnf_node *node)
{
	int nc = node->commodities.size();

	node->visited = VISITED_CLEAR;
	node->flowcnt = 0;
	if (node->used_classes != NULL)
	{
		memset(node->used_classes, 0, nc * sizeof(usable_info));
	}
	memset(node->fsinfo, 0, nc * sizeof(feasible_solution_info));
	node->feas_marked.assign(nc, false);
}

void sanity_check_node(mcnf_node *node)
{
	if (node->capacity > 0)
	{
		assert(node->flowcnt <= node->capacity && "Overflowed node");
	}

	if (node->dual_prices != NULL)
	{
		//make sure an anti and nonanti variable aren't allocated to the same node
		for (unsigned i = 0, nq = node->commodities.size(); i < nq; i++)
		{
			if (node->feas_marked[i])
			{
				int q = node->commodities[i];
				int index = node->cindex.get(-q);
				if (index >= 0)
				{
					assert(!node->feas_marked[index]
							&& "anti/nonanti overlap");
				}
			}
		}
	}
}

static void clear_flowcnt(mcnf_node *node)
{
	node->flowcnt = 0;
}

// make sure that the current feasible solution in the graph is valid
//returns the cost of the solution
int GlobalMCNF::sanity_check_feasible_solution()
{
	for (unsigned b = 0, nb = blocks.size(); b < nb; b++)
	{
		mcnf_block *block = blocks[b];
		block->check_feasible_links();
	}

	traverse_nodes(clear_flowcnt);
	/* compute cost of allocation */
	tally_path_cost tally;
	set_flow_count flower;
	for (unsigned q = 1, nq = quantities.size(); q < nq; q++)
	{
		trace_all_feasible_paths(q, tally);
		trace_all_feasible_paths(q, flower);
		trace_all_feasible_paths(-q, tally);
	}

	traverse_nodes(sanity_check_node);
	return tally.getCost();
}

//to arbitrarily break ties
static bool could_care_less(mcnf_edge *newedge, mcnf_edge *oldedge, int q)
{
	return false;
}

/* initialize for shortest backwars
 */
static void clear_backwards(mcnf_node *node)
{
	int nc = node->commodities.size();
	node->visited = VISITED_CLEAR;
	memset(node->sinfo[1], 0, sizeof(shortest_info) * nc);

	for (int i = 0; i < nc; i++)
	{
		node->sinfo[1][i].value = HUGE_VAL;
	}
}

class tied_constraint_price
{
	hash_map<mcnf_node *, std::pair<int,int> > &badnodes;
public:
	tied_constraint_price(hash_map<mcnf_node *, std::pair<int,int> > &baddies) :
		badnodes(baddies)
	{

	}

	double operator()(mcnf_edge *edge, int q)
	{
		if (badnodes.count(edge->dst) && (badnodes[edge->dst].first == q
				|| badnodes[edge->dst].second == q))
		{
			return UNREACHABLE;
		}
		return mcnf_edge_price(edge, q);
	}
};

/* return amount unconstrained shortest path problem should be adjusted
 * to account for tied constraints; at each tie look at the best untied
 * cost and compare to the best tied cost
 * We assume shortest forward paths are computed, but not backwards (which we will selective compute)
 * Adjusts the shortest paths to be those paths chosen by the tied constraint
 * so that the computed balanced paths will reflect these constraints (scary).
 *  */
double mcnf_block::shortest_tied_constraint_adjust()
{
	double ret = 0;
	hash_map<mcnf_node *, std::pair<int,int> > badnodes;

	if (TC_RAATs.size() > 0)
	{
		//need to compute shortest backwards paths
		traverse_nodes(clear_backwards);
		shortest_paths_backwards(mcnf_edge_price, mcnf_edge_cost,
				could_care_less);
	}
	for (unsigned t = 0, nt = TC_RAATs.size(); t < nt; t++)
	{
		TC_RAATInfo& TC = TC_RAATs[t];
		int s = TC.srcq;
		int d = TC.dstq;

		//compute the unconstrained minimum cost of the shortest paths
		//of s and d
		double mins = HUGE_VAL;
		double mind = HUGE_VAL;
		double minc = HUGE_VAL;
		hash_set<mcnf_node*> should_be_shortest;

		for (unsigned i = 0, n = TC.inodes.size(); i < n; i++)
		{
			mcnf_node *node = TC.inodes[i];
			int sindex = node->cindex.get(s);
			int dindex = node->cindex.get(d);
			if (sindex >= 0)
			{
				double val = node->sinfo[0][sindex].value
						+ node->sinfo[1][sindex].value;
				if (val < mins)
					mins = val;
			}
			if (dindex >= 0)
			{
				double val = node->sinfo[0][dindex].value
						+ node->sinfo[1][dindex].value;
				if (val < mind)
					mind = val;
			}
			if (sindex >= 0 && dindex >= 0)
			{
				double val = node->sinfo[0][sindex].value
						+ node->sinfo[1][sindex].value
						+ node->sinfo[0][dindex].value
						+ node->sinfo[1][dindex].value + TC.cost;
				if (val < minc)
				{
					should_be_shortest.clear();
					minc = val;
					should_be_shortest.insert(node);
				}
				else if (val == minc)
				{
					should_be_shortest.insert(node);
				}
			}
		}
		double diff = mins + mind - minc;
		if (diff > 0)
		{
			//combining is better
			ret -= diff; //this must improvement

			//record the nodes s and d should not go through
			for (unsigned i = 0, n = TC.inodes.size(); i < n; i++)
			{
				mcnf_node *node = TC.inodes[i];
				if (!should_be_shortest.count(node))
				{
					badnodes[node] = std::pair<int,int>(s, d);
				}
			}
		}
	}

	//now recomptue forward paths forcing paths to go through the good nodes
	//found above (there's probably a much more efficient way to do this..)
	initialize_path_info();
	tied_constraint_price tcp(badnodes);
	shortest_paths(tcp, mcnf_edge_cost, true);

	return ret;
}

/* look for cases where the feasible solution is fulfilling
 * tied constraints and add in the costs (benefits) */
int GlobalMCNF::feas_sol_tied_constraint_adjust()
{
	int ret = 0;
	for (unsigned b = 0, nb = blocks.size(); b < nb; b++)
	{
		mcnf_block *block = blocks[b];
		for (unsigned t = 0, nt = block->TC_RAATs.size(); t < nt; t++)
		{
			TC_RAATInfo& TC = block->TC_RAATs[t];
			int s = TC.srcq;
			int d = TC.dstq;
			for (unsigned i = 0, n = TC.inodes.size(); i < n; i++)
			{
				mcnf_node *node = TC.inodes[i];
				int sindex = node->cindex.get(s);
				int dindex = node->cindex.get(d);
				if (sindex >= 0 && dindex >= 0)
				{
					if (node->feas_marked[sindex] && node->feas_marked[dindex])
					{
						ret += TC.cost;
						break;
					}
				}
			}
		}
	}
	return ret;
}

class cost_and_flow: public std::binary_function<mcnf_edge *, int, void>
{
	int cost;
public:
	cost_and_flow() :
		cost(0)
	{
	}
	bool operator()(mcnf_edge* edge, int q)
	{
		if (q > 0) //ignore antis
			edge->dst->flowcnt++;
		cost += edge->getCost(q);
		return true;
	}
	int getCost()
	{
		return cost;
	}
};

class priceOfConstraintsComputer
{
	double total;
public:
	priceOfConstraintsComputer() :
		total(0)
	{
	}

	void operator()(mcnf_node *node)
	{
		double price = 0;
		if (node->capacity > 0)
		{
			price += node->capacity * node->price;
		}
		else
			assert(node->price == 0);
		if (node->dual_prices != NULL)
		{
			int i;
			int nc = node->commodities.size();
			for (i = 0; i < nc; i++)
			{
				assert(node->dual_prices[i] >= 0);
				if (node->dual_prices[i] > 0)
					assert(node->commodities[i]> 0);
				price += node->dual_prices[i];
			}
		}
		total += price;
		//std::cerr << "PoC " << node->getName() << " " << price << "\n";
	}

	double getTotal()
	{
		return total;
	}
};

/* return the sum of the product of each bundle constraint with
 the corresponding price */
double GlobalMCNF::price_of_constraints()
{
	priceOfConstraintsComputer pricer;
	traverse_nodes(pricer);
	return pricer.getTotal();
}

//set savedRegHardRegs
void GlobalMCNF::setSavedReg(unsigned r, unsigned hardreg)
{
	int q = reg_quantity.get(r);
	if (q > 0) //save/restore go away if function doesnt' return for example
		savedRegHardRegs[q] = hardreg;
}

/* copy prices from one commodity indexed array to another between two nodes */
static void copyNodePrices(mcnf_node *snode, mcnf_node *dnode, double *src,
		double *dst)
{
	for (unsigned i = 0, n = snode->commodities.size(); i < n; i++)
	{
		int index = dnode->cindex.get(snode->commodities[i]);
		if (index >= 0)
		{
			dst[index] = src[i];
		}
	}
}

//add edges to blocked edges to constrain the current full model by the solution in simple
void GlobalMCNF::constrainModel(GlobalMCNF& simple,
		const AllocationClassInfo *simple_alloc_classes)
{
	assert(blocks.size() == simple.blocks.size());
	for (int b = 0, bn = blocks.size(); b < bn; b++)
	{
		mcnf_block *block = blocks[b];
		mcnf_block *sblock = simple.blocks[b];

		assert(sblock->allnodes.size() <= block->allnodes.size());
		unsigned curr = 0;
		mcnf_node *curr_node = block->allnodes[curr];
		curr++;

		assert(sblock->defs.size() == block->defs.size());
		//constrain defs into memory
		for(int i = 0, n = sblock->defs.size(); i < n; i++)
		{
			mcnf_node *def = block->defs[i];
			mcnf_node *sdef = sblock->defs[i];
			assert(def->position == sdef->position); //q
			int q = def->commodities[0];
			mcnf_node *snext = sdef->fsinfo[0].next_edge->dst;
			bool nextisMem = simple_alloc_classes->getType(snext->allocClass) == AllocationClassInfo::Memory;
			for(int s = 0, ns = def->succs.size(); s < ns; s++)
			{
				mcnf_edge *nedge = def->succs[s];
				if(nextisMem != (alloc_classes->getType(nedge->dst->allocClass) == AllocationClassInfo::Memory))
				{
					blockedEdges[nedge].push_back(q);
				}
			}
		}

		//deal with leveled nodes
		for (int i = 0, n = sblock->allnodes.size(); i < n; i++)
		{
			mcnf_node *node = sblock->allnodes[i];
			if (!node->has_class())
				continue;

			//sync levels
			while (curr < block->allnodes.size() && (!curr_node->has_class()
					|| curr_node->position < node->position || curr_node->type
					!= node->type))
			{
				curr_node = block->allnodes[curr];
				curr++;
			}


			assert(curr_node->position == node->position && curr_node->type == node->type && curr_node->subindex == node->subindex);
			while (curr_node->position == node->position && curr_node->type
					== node->type && curr_node->subindex == node->subindex
					&& alloc_classes->isSubClassOf(curr_node->allocClass,
							simple_alloc_classes, node->allocClass))
			{
				//for every quantity in node
				for(int c = 0, nc = node->commodities.size(); c < nc; c++)
				{
					int q = node->commodities[c];
					int index = node->cindex.get(q);
					int cindex = curr_node->cindex.get(q);
					mcnf_edge *nexte =  node->fsinfo[index].next_edge;
					if(nexte == NULL)
						continue;
					mcnf_node *next = nexte->dst;
					if(cindex >= 0 && next->has_class())
					{
						if(q > 0) //force loads and stores
						{
							if(next->allocClass != node->allocClass &&
									(simple_alloc_classes->getType(next->allocClass) == AllocationClassInfo::Memory ||
											simple_alloc_classes->getType(node->allocClass) == AllocationClassInfo::Memory))
							{
								bool blockNonMemDst = simple_alloc_classes->getType(next->allocClass) == AllocationClassInfo::Memory;
								for(int s = 0, ns = curr_node->succs.size(); s < ns; s++)
								{
									mcnf_edge *nedge = curr_node->succs[s];
									bool nextisMem = alloc_classes->getType(nedge->dst->allocClass) == AllocationClassInfo::Memory;
									if(nedge->commodity_valid(q) && nextisMem == !blockNonMemDst)
									{
										blockedEdges[nedge].push_back(q);
									}
								}
							}
						}
						else //antivars, force stoers
						{
							if(next->allocClass != node->allocClass &&
									simple_alloc_classes->getType(next->allocClass) == AllocationClassInfo::AntiOnly )
							{
								for(int s = 0, ns = curr_node->succs.size(); s < ns; s++)
								{
									mcnf_edge *nedge = curr_node->succs[s];
									if(nedge->commodity_valid(q) &&
											alloc_classes->getType(nedge->dst->allocClass) != AllocationClassInfo::AntiOnly)
									{
										blockedEdges[nedge].push_back(q);
									}
								}
							}
						}
					}
				}


				curr_node = block->allnodes[curr];
				curr++;
				if (curr >= block->allnodes.size())
					break;
			}
		}
	}

}

/* copy the prices from a model built using simple_alloc_classes into this model
 * Assume alloc_classes is set appropriately for this model
 */
void GlobalMCNF::copyPrices(GlobalMCNF& simple,
		const AllocationClassInfo *simple_alloc_classes)
{
	assert(blocks.size() == simple.blocks.size());
	for (int b = 0, bn = blocks.size(); b < bn; b++)
	{
		mcnf_block *block = blocks[b];
		mcnf_block *sblock = simple.blocks[b];

		assert(sblock->allnodes.size() <= block->allnodes.size());
		unsigned curr = 0;
		mcnf_node *curr_node = block->allnodes[curr];
		curr++;

		for (int i = 0, n = sblock->allnodes.size(); i < n; i++)
		{
			mcnf_node *node = sblock->allnodes[i];
			if (!node->has_class())
				continue;

			//sync levels
			while (curr < block->allnodes.size() && (!curr_node->has_class()
					|| curr_node->position < node->position || curr_node->type
					!= node->type))
			{
				curr_node = block->allnodes[curr];
				curr++;
			}

			assert(curr_node->position == node->position && curr_node->type == node->type && curr_node->subindex == node->subindex);
			while (curr_node->position == node->position && curr_node->type
					== node->type && curr_node->subindex == node->subindex
					&& alloc_classes->isSubClassOf(curr_node->allocClass,
							simple_alloc_classes, node->allocClass))
			{
				curr_node->price = node->price;
				curr_node->gradient = node->gradient;

				if (curr_node->dual_prices)
				{
					assert(node->dual_prices);
					copyNodePrices(node, curr_node, node->dual_prices,
							curr_node->dual_prices);
					copyNodePrices(node, curr_node, node->dual_gradients,
							curr_node->dual_gradients);
				}
				if (curr_node->boundary_prices)
				{
					assert(node->boundary_prices);

					if (node->type == nEntry)
					{
						unsigned p;
						unsigned np = node->preds.size();
						unsigned cnp = curr_node->preds.size();

						if (cnp == np)
						{
							for (p = 0; p < np; p++)
							{
								copyNodePrices(node, curr_node, node->boundary_prices[p], curr_node->boundary_prices[p]);
								copyNodePrices(node, curr_node, node->boundary_gradients[p], curr_node->boundary_gradients[p]);
							}
						}
						else //edges to correspond exactly in full model (not live in some regs?)
						{
							for (p = 0; p < np; p++)
							{
								mcnf_node *src = node->preds[p]->src;
								if (src->type != nDef)
								{
									int sblock = src->block_num;
									//find dblock in curr_node.. this could be more efficient
									for (unsigned cp = 0; cp < cnp; cp++)
									{
										mcnf_node *csrc =
												curr_node->preds[cp]->src;
										if (csrc->type != nDef
												&& csrc->block_num == sblock)
										{
											copyNodePrices(node, curr_node, node->boundary_prices[p], curr_node->boundary_prices[cp]);
											copyNodePrices(node, curr_node, node->boundary_gradients[p], curr_node->boundary_gradients[cp]);
											break;
										}
									}
								}
							}
						}
					}
					else if (node->type == nExit)
					{
						unsigned s;
						unsigned ns = node->succs.size();
						unsigned cns = curr_node->succs.size();
						if (cns == ns)
						{
							for (s = 0; s < ns; s++)
							{
								copyNodePrices(node, curr_node, node->boundary_prices[s], curr_node->boundary_prices[s]);
								copyNodePrices(node, curr_node, node->boundary_gradients[s], curr_node->boundary_gradients[s]);
							}
						}
						else //edges to correspond exactly in full model (not live in some regs?)
						{
							for (s = 0; s < ns; s++)
							{
								mcnf_node *dst = node->succs[s]->dst;
								if (dst->type != nSink)
								{
									int dblock = dst->block_num;
									//find dblock in curr_node.. this could be more efficient
									for (unsigned cs = 0; cs < cns; cs++)
									{
										mcnf_node *cdst =
												curr_node->succs[cs]->dst;
										if (cdst->type != nSink
												&& cdst->block_num == dblock)
										{
											copyNodePrices(node, curr_node, node->boundary_prices[s], curr_node->boundary_prices[cs]);
											copyNodePrices(node, curr_node, node->boundary_gradients[s], curr_node->boundary_gradients[cs]);
											break;
										}
									}
								}
							}
						}
					}
				}
				curr_node = block->allnodes[curr];
				curr++;
				if (curr >= block->allnodes.size())
					break;
			}
		}
	}
}

int GlobalMCNF::getNumTiedconstraints()
{
	int ret = 0;
	for(int b = 0, nb = blocks.size(); b < nb; b++)
	{
		mcnf_block *block = blocks[b];
		ret += block->TC_RAATs.size();
	}
	return ret;
}
}
