/*
 *  A heuristic solver which computes a valid solution to the global MCNF
 *  problem using a simultaneous allocator on traces.
 */
#include "HeuristicSolver.h"
#include <iterator>
#include <iostream>

namespace llvm
{
cl::opt<bool> TraceUpdateMin("trace-update-min", cl::Hidden,
		cl::desc("Minimal amount of trace updating"));
cl::opt<bool> TraceUpdateAll("trace-update-all", cl::Hidden,
		cl::desc("Fully update all traces"));
cl::opt<bool> CrossBlockEvictions("crossblock-evict", cl::init(true), cl::Hidden,
		cl::desc("Allow evictions to span block boundaries within a trace"));
cl::opt<bool> PRAExperiment("pra-ex", cl::init(false), cl::Hidden,
		cl::desc("Experimental flag for PRA"));
cl::opt<bool> CrossBlockAntis("crossblock-antis", cl::init(true), cl::Hidden,
		cl::desc("Allow antivariables to switch allocation classes cross blocks"));
cl::opt<bool> TiedConstraints("tied-constraints", cl::init(false), cl::Hidden,
		cl::desc("Support weighted tied constraints"));

std::ostream &operator<<(std::ostream &OS, const AssignRegLiveInterval &interval)
{
	OS << "[" << interval.q << ", " << interval.startpos << "-" << interval.endpos << "]";
	return OS;
}

/* Keeps track of what quantities are mapped to a single class.  Necessary
 * if a class has more than one register. */
class ClassAlloc
{
	UnReservedVector<int> quantities; //same size as class
	unsigned pos; //first empty position
public:
	ClassAlloc(): pos(0) {}

	void setSize(int size)
	{
		if(size >= 0)
			quantities.resize(size, -1);
	}

	void clear()
	{
		pos = 0;
	}
	//add quantities to class
	void add(int q)
	{
		if(quantities.size() == 0) //uncapacitated
			return;

		if(pos >= quantities.size())
			abort();
		quantities[pos++] = q;
	}

	//remove quantity from class
	void remove(int q)
	{
		int i;
		int n = quantities.size();
		if(n == 0) //uncapacitated
			return;
		for(i = 0; i < (int)pos; i++)
		{
			if(quantities[i] == q)
			{
				quantities[i] = -1;
				break;
			}
		}
		if(i == (int)pos)
			return; //not present
		for( ; i < (int)pos-1; i++)
		{
			quantities[i] = quantities[i+1]; //shift down
		}
		assert(pos > 0);
		pos--;
	}

	//true if q in this class
	bool contains(int q)
	{
		for(unsigned i = 0, n = pos; i < n; i++)
		{
			if(quantities[i] == q)
				return true;
		}
		return false;
	}

	//number of quantities allocated in this class
	int size()
	{
		return pos;
	}

	//not the same as !isFull
	bool isEmpty()
	{
		return pos == 0;
	}

	typedef UnReservedVector<int>::const_iterator iterator;
	iterator begin() const
	{
		return quantities.begin();
	}
	iterator end() const
	{
		return quantities.begin()+pos;
	}

	void print(std::ostream &out) const
	{
		if(pos > 0)
		{
			out << quantities[0];
			for(unsigned i = 1; i < pos; i++)
			{
				out << " " << quantities[i];
			}
		}
	}
};

inline std::ostream& operator<<(std::ostream &OS, const ClassAlloc &CA) {
	CA.print(OS);
	return OS;
}

/* A map from quantity to allocation class.  Returns -1 if quantity unmapped. */
class QuantityAlloc
{
	hash_map<int, int> map;

public:
	QuantityAlloc()
	{
	}

	int get(int q)
	{
		if (map.count(q) > 0)
			return map[q];
		return -1;
	}
	void put(int q, int c)
	{
		if(DebugPRA > 9)
		{
			std::cerr << q << " is now in " << c << "\n";
		}
		map[q] = c;
	}
	void erase(int q)
	{
		map.erase(q);
	}
	void clear()
	{
		map.clear();
	}

	typedef hash_map<int,int>::const_iterator iterator;
	iterator begin() const
	{
		return map.begin();
	}
	iterator end() const
	{
		return map.end();
	}
};

//used to accumulate what classes aquantity is allocated to
//ignore callee save register quantities in their own class
class accum_allocClasses : public std::binary_function<mcnf_edge *, int, bool>
{
	GlobalMCNF &mcnf_cfg;
	hash_set<int> allocClasses;
public:
	accum_allocClasses(GlobalMCNF &cfg) : mcnf_cfg(cfg)
	{

	}

	bool operator()(mcnf_edge *edge, int q)
	{
		mcnf_node *node = edge->dst;
		if(node->has_class())
		{
			if(mcnf_cfg.savedRegHardRegs.count(q) &&
					alloc_classes->getRegisterIndex(mcnf_cfg.savedRegHardRegs[q]) == node->allocClass)
				; //skip these
			else
				allocClasses.insert(node->allocClass);
		}
		return true;
	}

	const hash_set<int>& getClasses() const
	{
		return allocClasses;
	}
};

int print_path(mcnf_edge* edge, int q)
{
	std::cerr << edge->dst->getName() << "(" << edge->getPrice(q)
	 << ";" << edge->getCost(q) << ")"
			<< "<-";
	if(edge->src->type == nDef)
		std::cerr << "\t";
	return 1;
}

class betterTiedEdge
{
	GlobalMCNF& mcnf_cfg;

public:
	betterTiedEdge(GlobalMCNF& cfg) :
		mcnf_cfg(cfg)
	{

	}

	bool operator()(mcnf_edge *newedge, mcnf_edge *oldedge, int q);
};

/* Return the price of an edge useful for trace allocation.
 In addition to the normal price, obey all boundary constraints
 and, for exits, add in the price from any successor
 entries. */
class trace_price
{
	HeuristicSolver *solver;
public:
	trace_price(HeuristicSolver *s) :
		solver(s)
	{

	}

	double operator()(mcnf_edge *edge, int q)
	{
		mcnf_node *src = edge->src;
		mcnf_node *dst = edge->dst;
		double ret;
		BlockTraces& traces = solver->traces;
		/* if quantity has already been allocated at boundary,
		 obey that allocation */

		if (!mcnf_boundary_node_valid(src, q))
		{
			return UNREACHABLE;
		}

		if (!mcnf_boundary_node_valid(dst, q))
		{
			return UNREACHABLE;
		}

		ret = edge->getPrice(q);
		if (dst->type == nExit)
		{
			int s, ns;
			/* current position in trace, find the the closest successor
			 with q live-in in the trace and add in computed cost */
			int currpos = traces.block_trace_index(dst->block_num);
			int nextb = solver->mcnf_cfg->blocks.size();
			double nextblockcost = 0;

			ns = dst->succs.size();
			for (s = 0; s < ns; s++)
			{
				mcnf_node *succ = dst->succs[s]->dst;
				if (succ->type == nEntry)
				{
					int index = succ->cindex.get(q);
					if (index >= 0)
					{
						int strace = traces.blocks_trace(succ->block_num);
						if (strace == solver->current_trace)
						{
							int tindex =
									traces.block_trace_index(succ->block_num);
							if (tindex > currpos && tindex < nextb)
							{
								nextb = tindex;
								if (succ->sinfo[1][index].prev_edge) /* has computed value */
								{
									nextblockcost = succ->sinfo[1][index].value;
									if (!succ->sinfo[0][index].prev_edge)
										abort();
									/* include price on entry node */
									nextblockcost
											+= succ->sinfo[0][index].prev_edge->getPrice(q);
								}
							}
						}
						else if (traces.block_trace_index(succ->block_num) == 0
								&& traces.trace_ancestor(traces.blocks_trace(succ->block_num))
										== dst->block_num)
						{
							/* include price of trace connected to this exit */
							if (succ->sinfo[1][index].prev_edge)
							{
								ret += succ->sinfo[1][index].value;
								if (!succ->sinfo[0][index].prev_edge)
									abort();
								/* include price on entry node */
								ret
										+= succ->sinfo[0][index].prev_edge->getPrice(q);
							}
						}
					}
				}
			}
			ret += nextblockcost;
		}

		return ret;
	}
};

/* Return the cost of the edge.  Look across block boundaries
 to accumulate the cost of the whole trace. */
class trace_cost
{
	HeuristicSolver *solver;
public:
	trace_cost(HeuristicSolver *s) :
		solver(s)
	{

	}

	int operator()(mcnf_edge *edge, int q)
	{
		int ret = edge->getCost(q);
		mcnf_node *dst = edge->dst;
		BlockTraces& traces = solver->traces;

		if (dst->type == nExit)
		{
			int s, ns;
			/* current position in trace, find the the closest successor
			 with q live-in in the trace and add in computed cost */
			int currpos = traces.block_trace_index(dst->block_num);
			int nextb = solver->mcnf_cfg->blocks.size();
			int nextblockcost = 0;

			ns = dst->succs.size();
			for (s = 0; s < ns; s++)
			{
				mcnf_node *succ = dst->succs[s]->dst;
				if (succ->type == nEntry)
				{
					int index = succ->cindex.get(q);
					if (index >= 0)
					{
						int strace = traces.blocks_trace(succ->block_num);
						if (strace == solver->current_trace)
						{
							int tindex =
									traces.block_trace_index(succ->block_num);
							if (tindex> currpos && tindex < nextb)
							{
								nextb = tindex;
								if (succ->sinfo[1][index].prev_edge)
								{
									nextblockcost = succ->sinfo[1][index].cost;
									if (!succ->sinfo[0][index].prev_edge)
										abort();
									/* include price on entry node */
									nextblockcost
											+= succ->sinfo[0][index].prev_edge->getCost(q);
								}
							}
						}
						else if (traces.block_trace_index(succ->block_num) == 0
								&& traces.trace_ancestor(traces.blocks_trace(succ->block_num))
										== dst->block_num)
						{
							/* include price of trace connected to this exit */
							if (succ->sinfo[1][index].prev_edge)
							{
								ret += succ->sinfo[1][index].cost;
								if (!succ->sinfo[0][index].prev_edge)
									abort();
								/* include price on entry node */
								ret
										+= succ->sinfo[0][index].prev_edge->getCost(q);
							}
						}
					}
				}
			}
			ret += nextblockcost;
		}

		return ret;
	}
};

/* Returns true if newedge is better for q than oldedge if
 all else is equal
 This is for the backwards paths computations (the dst changes between new and old edges
 */
bool betterTiedEdge::operator()(mcnf_edge *newedge, mcnf_edge *oldedge, int q)
{
	mcnf_node *newdst = newedge->dst;

	if (newdst->type == nExit || newdst->type == nEntry)
	{
		if (q < 0)
		{
			return alloc_classes->getType(newdst->allocClass)
			== AllocationClassInfo::Memory;
		}
		else
		{
			/* keep vars that span calls out of temps */
			unsigned dstr = alloc_classes->getRegister(newdst->allocClass);
			if(dstr> 0)
			{
				for(unsigned i = 0, n = mcnf_cfg.quantities[q].regs.size(); i < n; i++)
				{
					unsigned r = mcnf_cfg.quantities[q].regs[i];
					if(mcnf_cfg.hasConflict(r, dstr))
						return false;
				}
				return true;
			}
		}
	}
	else if (newdst->has_class() && newedge->src->has_class())
	{
		/* prefer to keep things in the same class */
		if (newdst->allocClass == newedge->src->allocClass)
			return true;
	}

	/* prefer the shortest unpriced path */
	// this has almost no impact unless there are perfect prices..
	mcnf_node *olddst = oldedge->dst;
	int nindex = newdst->cindex.get(q);
	int oindex = olddst->cindex.get(q);
	int ncost = newdst->sinfo[1][nindex].cost + newedge->getCost(q);
	int ocost = olddst->sinfo[1][oindex].cost + oldedge->getCost(q);

	if(ncost < ocost)
	{
		if(DebugPRA> 11)
		{
			std::cerr << "betterTied " << q << " " << "old " << oldedge->getName(q) <<
			" new " << newedge->getName(q) << " " << ncost << " o" << ocost << "\n";
		}
		return true;
	}

	return false;
}

/* Return true if q1 shoudl be propagated before q2 */
bool liveq_compare::operator()(int q1, int q2)
{
	int class1, class2;
	mcnf_node *node1, *node2, *dst1, *dst2;
	mcnf_edge *edge1, *edge2;
	int index1, index2;

	if(q1 == q2)
		return false;

	/* first, make sure we allocate antivars last */
	if(q1 < 0)
	{
		if(q2 < 0)
		return q1 < q2 ? false : true;
		else
		return false;
	}
	else if(q2 < 0)
	{
		return true;
	}

	if(q1 == 0 || q2 == 0)
		abort();
	class1 = current_quantity_alloc.get(q1);
	class2 = current_quantity_alloc.get(q2);
	if(class1 < 0 || class2 < 0)
		abort();
	node1 = current_prev_nodes[class1];
	node2 = current_prev_nodes[class2];

	/* sort by the intended next class */

	index1 = node1->cindex.get(q1);
	index2 = node2->cindex.get(q2);

	if(index1 < 0 || index2 < 0)
	abort();

	edge1 = solver->next_edge_for_allocation(node1, q1);
	edge2 = solver->next_edge_for_allocation(node2, q2);

	dst1 = edge1->dst;
	dst2 = edge2->dst;

	//update class variables to refer to next class
	class1 = dst1->allocClass;
	class2 = dst2->allocClass;

	if(class1 < class2)
		return true;
	else if(class1> class2)
		return false;
	else if(alloc_classes->getCapacity(class1) < 0)
		return false;

	/* if there is an overlap, allocate the more evictable
	 quantity first */

	//anything with zero weight goes first
	if(solver->mcnf_cfg->quantities[q1].weight == 0)
	{
		if(solver->mcnf_cfg->quantities[q2].weight > 0)
			return true;
	}
	else if(solver->mcnf_cfg->quantities[q2].weight == 0)
	{
		return false;
	}

#if 0
	//doesn't appear to work properly (the insn case is missed...)
	int level = node1->position;
	if(level < 0)
		level = 0;
	mcnf_block *block = solver->mcnf_cfg->blocks[node1->block_num];
	if (q1 > 0 && q2 > 0 && level >= 0 && level < (int)block->RAATs.size())
	{
		int next1 = block->quantity_next_use[level][q1];
		int next2 = block->quantity_next_use[level][q2];

		/* put most evictable things first, so whatever is farthest away
		 * is less */
		if (next1> next2)
		return true;
		if (next1 < next2)
		return false;
	}
#endif

	mcnf_node *inode1 = NULL, *inode2 = NULL;
	/* allocate variables that go on to insns after xbar insns;
	 we want to allocate evictable things _first_ */
	index1 = dst1->cindex.get(q1);
	index2 = dst2->cindex.get(q2);

	if(dst1->sinfo[1][index1].prev_edge != NULL)
		inode1 = dst1->sinfo[1][index1].prev_edge->dst;
	if(dst2->sinfo[1][index2].prev_edge != NULL)
		inode2 = dst2->sinfo[1][index2].prev_edge->dst;

	if(inode1 == NULL && inode2 == NULL)
	{
		//quantity with lower weight should be more evictable
		return solver->mcnf_cfg->quantities[q1].weight <
		solver->mcnf_cfg->quantities[q2].weight;
	}
	if(inode1 == NULL)
		return true;
	if(inode2 == NULL)
		return false;
	if(inode1->type == inode2->type)
	{
		if(current_quantity_alloc.get(q1) < current_quantity_alloc.get(q2))
			return true;
		else if(current_quantity_alloc.get(q1)> current_quantity_alloc.get(q2))
			return false;
		else
		//quantity with lower weight should be more evictable
			return solver->mcnf_cfg->quantities[q1].weight <
			solver->mcnf_cfg->quantities[q2].weight;
	}

	if(inode1->type == nXBar)
		return true;
	if(inode2->type == nXBar)
		return false;
	//quantity with lower weight should be more evictable
	return solver->mcnf_cfg->quantities[q1].weight <
	solver->mcnf_cfg->quantities[q2].weight;
}

/* Returns the first quantity (if any) that insn insn_number in block defines
 and can be defined into class
 -1 if none */
static int get_insn_class_define(mcnf_block *block, int insn_number, int allocClass)
{
	unsigned i, numq;
	RAATInfo *ii = &block->RAATs[insn_number];

	std::vector<RAATInfo::QuantityInfo>& qinfo = ii->getQuantityInfo();
	numq = qinfo.size();
	for(i = 0; i < numq; i++)
	{
		if(qinfo[i].classes.test(allocClass) && qinfo[i].def)
		return qinfo[i].q;
	}
	return -1;
}

/* initialize the dualprices of the MCNF,
 these represent (roughly) the cost of a
 store (anti-variable evicition) which is complicated
 because only dirty values need to be stored

 Assumes other prices are already initialized.
 */
void HeuristicSolver::initialize_dual_prices(void)
{
	int b, bn;
	bn = mcnf_cfg->blocks.size();
	for(b = 0; b < bn; b++)
	{
		int n;
		mcnf_block *block = mcnf_cfg->blocks[b];
		int nn = block->allnodes.size();

		for(n = 0; n < nn; n++)
		{
			mcnf_node *node = block->allnodes[n];

			/* initialize prices of an xbar node after a defining insn to
			 reflect cost of antivar for purposes of getting a better
			 initial solution.
			 */
			if(node->type == nXBar && node->subindex == 1 && node->dual_prices != NULL
					&& node->position >= 0 && node->position < (int)block->RAATs.size())
			{
				int q = get_insn_class_define(block, node->position, node->allocClass);
				if(q> 0 && mcnf_cfg->quantities[q].weight> 0) //bias towards spilling if cheap

				{
					int index = node->cindex.get(q);
					if(index >= 0)
						node->dual_prices[index] = mcnf_cfg->memory_store_cost(node->allocClass);
				}
			}
		}
	}
}
/* like initialize node_for_fsf_all, but don't touch boundary conditions
 and leave forward info alone; we just want to recompute backwards
 no need to reinitialize fsinfo either
 */
void reinitialize_node_for_fsf_trace(mcnf_node *node)
{
	int nc = node->commodities.size();
	node->visited = VISITED_CLEAR;
	memset(node->sinfo[1], 0, sizeof(shortest_info)*nc);

	for(int i = 0; i < nc; i++)
	{
		node->sinfo[1][i].value = HUGE_VAL;
	}
}

static void initialize_node_for_fsf_all(mcnf_node *node)
{
	int nc = node->commodities.size();

	node->visited = VISITED_CLEAR;
	memset(node->sinfo[0], 0, sizeof(shortest_info)*nc);
	memset(node->sinfo[1], 0, sizeof(shortest_info)*nc);

	for(int i = 0; i < nc; i++)
	{
		node->sinfo[0][i].value = HUGE_VAL;
		node->sinfo[1][i].value = HUGE_VAL;
	}
	node->flowcnt = 0;

	if (node->used_classes != NULL)
	{
		memset(node->used_classes, 0, sizeof(usable_info)*nc);
	}

	memset(node->fsinfo, 0, sizeof(feasible_solution_info)*nc);
	node->feas_marked.assign(nc, false);
}

int HeuristicSolver::construct_feasible_solution(bool initial)
{
	unsigned i, numb, numt;

	initial_heuristic = initial;
	/* allocate the traces in order */
	mcnf_cfg->traverse_nodes(initialize_node_for_fsf_all);
	alloced_blocks.clear();

	/* compute all forwards paths (this is used by eviction, probalby not necessary */
	numb = mcnf_cfg->blocks.size();
	for (i = 0; i < numb; i++)
	{
		mcnf_block *block = mcnf_cfg->blocks[i];
		block->shortest_paths(mcnf_edge_price, mcnf_edge_cost);
		block->quantities_used_in_class.clear();
	}

	/* compute backwards paths through the trace tree */
	numt = traces.num_traces();

	if(TraceUpdateAll)
	{
		for(int j = numt-1; j >= 0; j--)
		{
			update_trace_paths(j, &fixed_blocks);
		}
	}

	for (i = 0; i < numt; i++)
	{
		if(DebugPRA > 3)
		std::cerr << "Allocating trace " << i << "\n";
		allocate_trace(i);
	}

	//improve allocation of register saving variables
	//first see what classes non-reg saving vars use
	accum_allocClasses acc(*mcnf_cfg);
	for(unsigned q = 1, nq = mcnf_cfg->quantities.size(); q < nq; q++)
	{
		mcnf_cfg->trace_all_feasible_paths(q, acc);
	}

	for(hash_map<int, int>::iterator itr = mcnf_cfg->savedRegHardRegs.begin(),
			eitr = mcnf_cfg->savedRegHardRegs.end(); itr != eitr; itr++)
	{
		int q = itr->first;
		int r = itr->second;
		int c = alloc_classes->getRegisterIndex(r);

		//if c was used, make sure q is globally evicted
		if(acc.getClasses().count(c) || mcnf_cfg->clobberedClasses.count(c))
		{
			globally_evict_to_class(-q, alloc_classes->getAntiClass(alloc_classes->getMemoryClasses()[0]));
			globally_evict_to_class(q, alloc_classes->getMemoryClasses()[0]);
		}
		else
		{
			//otherwise, it should be possible to keep it in c
			globally_evict_to_class(q, c);
		}
	}

	/* compute cost of allocation */
	int total_cost = 0;
	double total_price = 0;
	int TC_adjust = 0;

	if(TiedConstraints)
	{
		TC_adjust = mcnf_cfg->feas_sol_tied_constraint_adjust();
	}
	/* compute cost of allocation */
	if(DebugPRA > 1)
	{
		std::cerr << "Feasible Totals:\n";
	}
	for(unsigned q = 1, nq = mcnf_cfg->quantities.size(); q < nq; q++)
	{
		tally_path_cost tally, atally;

		mcnf_cfg->trace_all_feasible_paths(q, tally);
		mcnf_cfg->trace_all_feasible_paths(-q, atally);
		total_cost += tally.getCost();
		total_cost += atally.getCost();

		if(DebugPRA > 1)
		{

			tally_path_price ptally;
			tally_path_price aptally;
			mcnf_cfg->trace_all_feasible_paths(q, ptally);
			mcnf_cfg->trace_all_feasible_paths(-q, aptally);
			total_price += ptally.getPrice() + aptally.getPrice();

			if(DebugPathsPRA)
			{
				std::cerr << q << ": " << tally.getCost() << " a" << atally.getCost() << "  p" << ptally.getPrice() << " ap" << aptally.getPrice() << "\n";
				std::cerr << "\t";
				mcnf_cfg->trace_all_feasible_paths(q, print_path);
				std::cerr << "\n\t";
				mcnf_cfg->trace_all_feasible_paths(-q, print_path);
				std::cerr << "\n";
			}
		}
	}
	if(DebugPRA)
	{
		if(DebugPRA > 1) std::cerr << "TC Adjust: " << TC_adjust << "\n";
		std::cerr << "Total Cost: " << total_cost << "\n";
		std::cerr << "Total Price: " << total_price << "\n";
	}
	return total_cost+TC_adjust;
}

/* Find the quantities that are defined into this level.
 Fill out defs (which is allocated accordingly) with the actual defs
 TODO: make this much more efficient
 */
void HeuristicSolver::get_level_defs(mcnf_node **level, mcnf_node **ilevel,
		std::vector<mcnf_node*>& defs)
{
	int i;
	int which;
	int numClasses = alloc_classes->numClasses();
	defs.clear();

	for (which = 0; which < 2; which++)
	{
		/* for each node, look at all predecessors, looking for an unmarked def */
		for (i = 0; i < numClasses; i++)
		{
			mcnf_node *node;
			if (which)
			node = ilevel[i];
			else
			node = level[i];

			if (node != NULL)
			{
				int p, nump;
				nump = node->preds.size();
				for (p = 0; p < nump; p++)
				{
					mcnf_node *prev = node->preds[p]->src;
					if (prev->type == nDef && !prev->feas_marked[0])
					{
						prev->feas_marked[0] = 1;
						defs.push_back(prev);
					}
				}
			}
		}
	}
}

/* figure out the allocation for all live variables transitioning
 from prev_nodes to next_nodes, updating current alloc info.
 This handles shifts in allocation.
 */
void HeuristicSolver::propagate_allocations(mcnf_node **prev_nodes,
		mcnf_node **insn_nodes, mcnf_node **next_nodes,
		QuantityAlloc& quantity_alloc, std::vector<ClassAlloc>& class_alloc,
		hash_set<int>& live)
{
	int q;
	std::vector<int> liveqs;
	liveqs.reserve(live.size());
	liveqs.insert(liveqs.end(),live.begin(), live.end());
	std::vector<int> problems;
	problems.reserve(live.size());

	if(DebugPRA > 9)
	{
		std::cerr << "Staring propagating with live: ";
		copy(liveqs.begin(), liveqs.end(), std::ostream_iterator<int>(std::cerr," "));
		std::cerr << "\n";
	}

	/* sort liveqs for good allocation */
	liveq_compare comp(this, prev_nodes, quantity_alloc);
	std::sort(liveqs.begin(), liveqs.end(), comp);

	/* clear out class allocs */
	for(unsigned i = 0, n = class_alloc.size(); i < n; i++)
	{
		class_alloc[i].clear();
	}

	/* take the next step in the shortest path, skip over problematic
	 allocations for now */
	for (int i = 0, numlive = liveqs.size(); i < numlive; i++)
	{
		int index, allocClass;
		mcnf_node *node, *dst;
		mcnf_edge *next, *iedge = NULL;
		q = liveqs[i];
		allocClass = quantity_alloc.get(q);
		if (allocClass < 0)
			abort();
		node = prev_nodes[allocClass];
		if (node == NULL)
			abort();
		index = node->cindex.get(q);
		if (index < 0)
			abort();
		if (!node->feas_marked[index])
			abort();
		if(DebugPRA > 9)
		{
			std::cerr << "Propagating " << q << " currently in class " << allocClass <<
			" at node " << node->getName();
			if(node->primal_solution != NULL)
			{
				std::cerr << "  primal " << node->primal_solution[index];
			}
			std::cerr << "\n";
		}
		next = next_edge_for_allocation(node, q);
		dst = next->dst;

		if (dst->type == nInsn)
		{
			/* allocate through insns, keep track of the insn edge */
			iedge = next;
			if (q> 0 && dst->capacity> 0 && dst->flowcnt >= dst->capacity)
			{
				abort(); /* this is suppose to be avoided */
			}
			if (dst != insn_nodes[dst->allocClass])
			abort();
			index = dst->cindex.get(q);
			next = dst->sinfo[1][index].prev_edge;
			dst = next->dst;

			/* antivars don't necessarily pass through insns
			 (defining insns force them through memory)
			 since non antis can't be evicted from memory (usually),
			 just pick the mem network eviction edge of the pass through
			 isn't valid */
			if (q < 0 && node_not_valid_for_allocation(dst, q, 0, class_alloc))
			{
				int s;
				mcnf_node *inode = iedge->dst;
				int ns = inode->succs.size();
				for (s = 0; s < ns; s++)
				{
					mcnf_edge *edge = inode->succs[s];
					if (edge->commodity_valid(q))
					{
						if (!node_not_valid_for_allocation(edge->dst, q, 0, class_alloc))
						{
							next = edge;
							dst = next->dst;
							break;
						}
					}
				}

			}
		}

		if (dst->type == nSink)
		{
			if (iedge != NULL)
				iedge->mark_feasible_edge(q);
			next->mark_feasible_edge(q);
			live.erase(q);
			quantity_alloc.erase(q);
			continue;
		}

		if (dst->type == nXBar && dst != next_nodes[dst->allocClass])
			abort();

		if (node_not_valid_for_allocation(dst, q, 0, class_alloc))
		{
			/* punt on allocating this for now, allocate
			 all the nonproblematic ones first */
			problems.push_back(q);
		}
		else
		{
			/* accept this allocation */
			if (iedge != NULL)
				iedge->mark_feasible_edge(q);
			next->mark_feasible_edge(q);
			assign_quantity_to_class(q, dst, class_alloc, quantity_alloc);
		}
	}

	int numClasses = alloc_classes->numClasses();
	std::vector<double> pricesto(numClasses);
	std::vector<mcnf_edge *> edges(numClasses, (mcnf_edge*)NULL);

	/* handle problematic allocations
	 for each problem case, select the best class for it, allowing evictions
	 and put it there */
	for (int i = 0, numprobs = problems.size(); i < numprobs; i++)
	{
		int c, s, ns, allocClass, index;
		mcnf_node *node, *next;
		mcnf_edge *nexte;
		mcnf_edge *eedge = NULL;
		int ec = -1;
		int eq = 0;

		q = problems[i];
		node = prev_nodes[quantity_alloc.get(q)];
		index = node->cindex.get(q);
		nexte = next_edge_for_allocation(node, q);
		next = nexte->dst;

		if(DebugPRA > 3)
		{
			std::cerr << "Problem progating " << q << "\n";
		}
		if (q < 0)
		{
			/* antivar, the var is now in the way, have to evict */
			/* find best eviction for antivar */
			const std::vector<int>& memory_classes =
			alloc_classes->getMemoryClasses();
			for (int i = 0, n = memory_classes.size(); i < n; i++)
			{
				int allocClass = memory_classes[i];
				int anticlass = alloc_classes->getAntiClass(allocClass);

				if (!evict_antivar(q, next_nodes[anticlass],
								node->sinfo[1][index].prev_edge))
					abort();
				else
				{
					quantity_alloc.put(q,anticlass);
					break;
				}
			}
		}
		else
		{
			int antievict = 0;
			/* nonantivar, figure out what to evict to fix it */
			for (c = 0; c < numClasses; c++)
			{
				pricesto[c] = HUGE_VAL;
				edges[c] = NULL;
			}

			/* consider all valid successors of the current node as
			 possible allocation points */
			ns = node->succs.size();
			for (s = 0; s < ns; s++)
			{
				mcnf_edge *edge = node->succs[s];
				mcnf_node *dst = edge->dst;

				double price = mcnf_edge_price_fixed_boundary_exits(edge, q);
				int dindex = dst->cindex.get(q);
				if (dindex < 0 || !edge->commodity_valid(q) ||
						node_blocked_by_insn(dst, q))
					continue;

				pricesto[dst->allocClass] = node->fsinfo[index].priced_cost	+ price;
				edges[dst->allocClass] = edge;
			}

			if (class_alloc[quantity_alloc.get(q)].contains(q))
				class_alloc[quantity_alloc.get(q)].remove(q); /* don't try to evict ourselves */
			allocClass = select_alloc_class(q, insn_nodes, next_nodes,
					class_alloc, &eedge, &eq, &ec, &antievict, pricesto);
			if (eedge != NULL) /* have to evict something */
			{
				if (class_alloc[allocClass].isEmpty())
					abort();
				evict_quantity(eq, eedge, ec, next_nodes,
						quantity_alloc, class_alloc);
			}

			if (edges[allocClass] == NULL)
				abort();
			assign_quantity_to_class(q, next_nodes[allocClass], class_alloc, quantity_alloc);
			edges[allocClass]->mark_feasible_edge(q);
			if(edges[allocClass]->dst->type == nInsn)
			{
				//mark through insn
				mcnf_node *dst = edges[allocClass]->dst;
				int index = dst->cindex.get(q);
				mcnf_edge *next = dst->sinfo[1][index].prev_edge;
				next->mark_feasible_edge(q);
			}

			if (antievict)
			{
				if (q < 0)
					abort();
				if (antievict < 0 || next_nodes[antievict] == NULL)
					abort();
				if(DebugPRA > 9)
				{
					std::cerr << "Antievict problem " << q << " to class " << antievict << "\n";
				}
				if (!evict_antivar(-q, next_nodes[antievict], edges[allocClass]) )
					abort();
				quantity_alloc.put(-q, antievict);
			}
		}
	}

	sanity_check_allocs(quantity_alloc, class_alloc, next_nodes);
}

/* eedge is an eviction edge of an anti var into the anti network,
 * this function will evict q, changing entry/exit allocation as needed,
 * in the block to the anti-network.  It simultaneously unmarks the
 * previous feasible path and only stops when a sink is reached or
 * the eviction path intersects the current feasible path
 *
 * This only works if the block is already allocated and has a feasible path.
 * Also, we assume if the var is pushed back into memory it is safe to stop.
 */
void HeuristicSolver::change_anti_alloc(mcnf_edge *eedge, int q)
{
	assert(q < 0);
	mcnf_edge *oldedge, *newedge;
	newedge = eedge;
	//old edge is the current feasible path
	int index = eedge->src->cindex.get(q);
	assert(index >= 0);
	oldedge = eedge->src->fsinfo[index].next_edge;
	mcnf_block *block = mcnf_cfg->blocks[eedge->src->block_num];
	assert(oldedge != NULL); //need feasible path

	while(oldedge != NULL && newedge != NULL)
	{
		assert(newedge->commodity_valid(q));
		if(newedge->dst->type == nEntry)
		{
			//change entry allocation
			unset_entry_exit_condition(block, q, Entry, false);
			visited_blocks.clear();
			set_entry_exit_condition(block, q, newedge->dst->allocClass, Entry);
		}
		else if(newedge->dst->type == nExit)
		{
			//change exit allocation
			unset_entry_exit_condition(block, q, Exit, false);
			visited_blocks.clear();
			set_entry_exit_condition(block, q, newedge->dst->allocClass, Exit);
		}

		//unmark oldedge, but first save next edge (not actually necessary..)
		mcnf_edge *next_oldedge = NULL;
		int oindex = oldedge->dst->cindex.get(q);
		next_oldedge = oldedge->dst->fsinfo[oindex].next_edge;
		oldedge->unmark_feasible_edge(q);

		//mark new edge
		newedge->mark_feasible_edge(q);

		if(newedge->dst == oldedge->dst) //paths are converged
		{
			//I don't think we can automatically assume that just because
			//the paths are converged that the previous path will stay in
			//the anti network, however, if the new path has been forced
			//out of the antinetwork we should stop
			if(newedge->dst->has_class() &&
					alloc_classes->getType(newedge->dst->allocClass) != AllocationClassInfo::AntiOnly)
			{
				break;
			}
		}

		//update edges
		oldedge = next_oldedge;

		if(newedge->dst->type == nSink)
		{
			//at the end, old path must be ending as well
			newedge = NULL;
			assert(oldedge == NULL);
		}
		else
		{
			int nindex = newedge->dst->cindex.get(q);
			//the shortest path isn't always through the anti network because
			//of exit constraints, but check it first
			mcnf_edge *next = newedge->dst->sinfo[1][nindex].prev_edge;
			if(next == NULL || (next->dst->has_class() &&
							alloc_classes->getType(next->dst->allocClass) != AllocationClassInfo::AntiOnly))
			{
				//look for an edge to the anti network
				for(int s = 0, ns = newedge->dst->succs.size(); s < ns; s++)
				{
					mcnf_edge *succ = newedge->dst->succs[s];
					if(succ->dst->type == nSink && succ->dst->position == q)
					{
						next = succ;
						break;
					}
					else if(succ->src->type != nExit &&
							alloc_classes->getType(succ->dst->allocClass) == AllocationClassInfo::AntiOnly &&
							succ->commodity_valid(q))
					{
						next = succ;
						break;
					}
				}
			}

			newedge = next;
		}
	}
}

/* return the antionly node in vec that contains q */
static mcnf_node* find_anti_node(const std::vector<mcnf_node*>& vec, int q)
{
	mcnf_node *ret = NULL;
	for(int i = vec.size()-1; i >= 0; i--)
	{
		ret = vec[i];
		if(ret->cindex.get(q) >= 0)
			break;
		ret = NULL;
	}
	return ret;
}

/* find the first edge that will evict q from memory into the antionly node;
 * search is necessary since the last xbar may be constrained in the prsence
 * of control flow (hmmm) */
/*
static mcnf_edge* find_first_anti_evict_edge(mcnf_node *exit, int q)
{
	mcnf_node *xbar = exit;
	mcnf_edge *eedge = NULL;

	while(eedge == NULL)
	{
		assert(xbar->preds.size() == 1); //more than one edge, should be an eviction edge
		xbar = xbar->preds[0]->src;
		//look for an eviction edge
		for(int p = 0, np = xbar->preds.size(); p < np; p++)
		{
			eedge = xbar->preds[p];
			int sindex = eedge->src->cindex.get(q);
			if(sindex >= 0 && eedge->src->feas_marked[sindex])
				break;
			if(eedge->src->type == nDef)
				break;
			eedge = NULL;
		}
	}
	return eedge;
}
*/

/* Put q in memory everywhere.  Assume it is legal to do so. */
void HeuristicSolver::globally_evict_to_class(int q, int allocClass)
{
	if(DebugPRA >4)
	{
		std::cerr << "Globally evicting " << q << " to " << allocClass << "\n";
	}
	//change entry/exit allocations
	visited_blocks.clear();
	for(int b = 0, nb = mcnf_cfg->blocks.size(); b < nb; b++)
	{
		mcnf_block *block = mcnf_cfg->blocks[b];
		unset_entry_exit_condition(block, q, Exit, false);
		unset_entry_exit_condition(block, q, Entry, false);
		set_entry_exit_condition(block, q, allocClass, Exit,false);
		set_entry_exit_condition(block, q, allocClass, Entry,false);
	}

	//change any existing feasible paths
	std::vector<mcnf_node*> defs;
	if(q > 0)
	{
		defs = mcnf_cfg->quantities[q].defs;
	}
	else //anti
	{
		for(int i = 0, n = mcnf_cfg->quantities[-q].antidefs.size(); i < n; i++)
		{
			defs.insert(defs.end(), mcnf_cfg->quantities[-q].antidefs[i].begin(), mcnf_cfg->quantities[-q].antidefs[i].end());
		}
	}
	for(int d = 0, nd = defs.size(); d < nd; d++)
	{
		mcnf_node *def = defs[d];

		if(def->feas_marked[0])
		{
			mcnf_edge *edge = def->fsinfo[0].next_edge;

			while(edge != NULL)
			{
				if(edge->dst->type == nSink)
					break;
				while(edge != NULL &&
						edge->dst->allocClass == allocClass) //no need to evict
				{
					mcnf_node *node = edge->dst;
					int index = node->cindex.get(q);
					edge = node->fsinfo[index].next_edge;
				}

				if(edge == NULL || edge->dst->type == nSink)
					break;
				if(edge->dst->allocClass != allocClass)
				{
					mcnf_node *stop = NULL;
					mcnf_edge *oldedge = edge;
					//need to change feasible path
					//first unmark current path
					while(edge != NULL)
					{
						mcnf_node *node = edge->dst;
						int index = node->cindex.get(q);
						mcnf_edge *next = node->fsinfo[index].next_edge;
						edge->unmark_feasible_edge(q);
						edge = next;
						if(node->allocClass == allocClass && node->block_num == current_block_num)
						{
							stop = node;
							break;
						}
					}

					//now find and follow new path
					edge = oldedge;
					for(int s = 0, ns = edge->src->succs.size(); s < ns; s++)
					{
						if(edge->src->succs[s]->dst->allocClass == allocClass)
						{
							edge = edge->src->succs[s];
							break;
						}
					}
					assert(edge != NULL);

					//follow edge to a sink
					while(edge != NULL)
					{
						mcnf_node *node = edge->dst;
						int index = node->cindex.get(q);
						assert(index >= 0);
						if(node->has_class()) assert(node->allocClass == allocClass);

						if(node == stop)
						{
							edge->mark_feasible_edge(q);
							break;
						}
						mcnf_edge *next = node->sinfo[1][index].prev_edge;
						if(next != NULL && next->dst->allocClass != allocClass)
							next = NULL;
						if(next == NULL && node->type != nSink)
						{
							//entry/exit constraints blocking the path, find a good next edge
							for(int s = 0, ns = node->succs.size(); s < ns; s++)
							{
								mcnf_edge *succ = node->succs[s];
								if(succ->dst->block_num != def->block_num)
								continue;
								if(succ->src->type == nExit && succ->dst->type == nEntry)
								continue;
								if(succ->dst->allocClass == allocClass && succ->dst->has_class()
										&& succ->commodity_valid(q))
								{
									next = succ;
									break;
								}
								else if(succ->dst->type == nSink && succ->dst->position == q &&
										succ->commodity_valid(q))
								{
									next = succ;
								}
							}
							assert(next != NULL);

						}
						edge->mark_feasible_edge(q);
						edge = next;
					}
				}
			}
		}
	}

	if(fixed_blocks.count(current_block_num))
	{
		//we changed the exit allocation of the current block, recompte
		//backwards shortest paths
		trace_cost tc(this);
		trace_price tp(this);
		betterTiedEdge tiebreaker(*mcnf_cfg);
		mcnf_block *curr = mcnf_cfg->blocks[current_block_num];
		curr->traverse_nodes(reinitialize_node_for_fsf_trace);
		curr->shortest_paths_backwards(tp, tc, tiebreaker);
	}
}

/* Evict q coming out of block, but only if the eviction can be done with
 * cost equal or less than cost.  Evicts q from this
 * block, then forward evicts q on all descendants of block that are dominated.
 * If curr_ends_in_mem is true, then even if the curren block starts with an anti
 * allocation, we still have to change the block internal allocations
 */
void HeuristicSolver::global_anti_evict(int q, mcnf_block *block, int cost, bool curr_ends_in_mem)
{
	//find the exit antinode
	mcnf_node *exit = find_anti_node(block->exits, q);
	assert(exit);
	int index = exit->cindex.get(q);

	if(DebugPRA> 3)
	{
		std::cerr << "globalevict " << q << " from " << block->index << "\n";
	}
	visited_blocks.clear();
	if(alloced_blocks.count(block->index)> 0)
	{
		if(!exit->feas_marked[index])
		{ //needs to be evicted
			assert(exit->preds.size() == 1);
			//don't span traces since we could be in some weird place
			mcnf_edge *eedge = cheapest_eviction_edge(exit, q, NULL, NULL, true);
				//find_first_anti_evict_edge(exit, q);
			//check the cost of evicting at the end of this block
			if(eedge == NULL) return;
			if(eedge->getCost(q)> cost)
			return; //give up

			//evict q from this block
			change_anti_alloc(eedge, q);
		}
	}
	else //not allocated already (strange, but possible) just change exit allocation
	{
		unset_entry_exit_condition(block, q, Exit, false);
		visited_blocks.clear();
		//go recursively.. will only touch unallocated blocks
		//don't change the exit of the current block since we won't recompute
		visited_blocks.insert(current_block_num);
		set_entry_exit_condition(block, q, exit->allocClass, Exit);
	}
	//traverse all successors and forward evict
	hash_set<mcnf_block*> processed;
	std::queue<mcnf_block*> worklist;
	processed.insert(block);
	for(int s = 0, ns = block->succs.size(); s < ns; s++)
	{
		worklist.push(block->succs[s]);
	}

	while(!worklist.empty())
	{
		mcnf_block *curr = worklist.front();
		worklist.pop();

		if(processed.count(curr)> 0)
			continue;
		processed.insert(curr);

		//is curr dominated by block?
		if(curr->def_dominator_in.count(-q)> 0 &&
				curr->def_dominator_in[-q] == block)
		{
			//if so, see if we need to change the anti allocation
			if(alloced_blocks.count(curr->index) || curr->index == current_block_num)
			{
				mcnf_node *entry = find_anti_node(curr->entries, q);
				int index = entry->cindex.get(q);
				if(!entry->feas_marked[index] || (curr_ends_in_mem && curr->index == current_block_num))
				{
					//have a feasible path
					if(DebugPRA > 3)
					{
						std::cerr << "GE changing block " << curr->index << " at entry\n";
					}
					//find the def edge
					mcnf_edge *defedge = NULL;
					for(int p = 0, np = entry->preds.size(); p < np; p++)
					{
						mcnf_node *pred = entry->preds[p]->src;
						if(pred->type == nDef && pred->position == q)
						{
							defedge = entry->preds[p];
							break;
						}
					}
					assert(defedge);
					change_anti_alloc(defedge, q);
				}
				//check exit
				if(curr->def_dominator_out.count(-q)> 0 &&
					curr->def_dominator_out[-q] == block)
				{ //the exit allocation should stay in anti
					mcnf_node *exit = find_anti_node(curr->exits, q);
					if(exit != NULL)
					{
						int xindex = exit->cindex.get(q);
						if(!exit->feas_marked[xindex])
						{
							if(DebugPRA > 3)
							{
								std::cerr << "GE changing block " << curr->index << " at exit\n";
							}

							if(curr->index == current_block_num)
							{
								//exit shouldn't be allocated, just reset
								unset_entry_exit_condition(curr, q, Exit, false);
								visited_blocks.clear();
								set_entry_exit_condition(curr, q, exit->allocClass, Exit);
							}
							else
							{
								//need to change allocation
								change_anti_alloc(entry->fsinfo[index].next_edge, q);
							}
						}
					}
				}
			}
			else //unallocated, just change entry exit as necessary
			{
				unset_entry_exit_condition(curr, q, Entry, false);
				visited_blocks.clear();
				//don't change the exit of the current block since we won't recompute
				visited_blocks.insert(current_block_num);
				set_entry_exit_condition(curr, q, exit->allocClass, Entry);
				//if exit is still dominated, change that too
				if(curr->def_dominator_out.count(-q)> 0 &&
						curr->def_dominator_out[-q] == block)
				{
					unset_entry_exit_condition(curr, q, Exit, false);
					visited_blocks.clear();
					set_entry_exit_condition(curr, q, exit->allocClass, Exit);
				}
			}

			//add successors, the check above will take care of bad blocks
			//add successors even of unalloced blocks so storeness is fully
			//propagated
			for(int s = 0, ns = curr->succs.size(); s < ns; s++)
			{
				worklist.push(curr->succs[s]);
			}

			//if the output of this block is still dominated by the same block
			//as the input, then q should be in the anti-network at block exit
			if(curr->def_dominator_out.count(-q)> 0 &&
					curr->def_dominator_out[-q] == block &&
					alloced_blocks.count(curr->index))
			{
				mcnf_node *exit = find_anti_node(curr->exits, q);
				if(exit != NULL)
				{
					int xindex = exit->cindex.get(q);
					assert(exit->feas_marked[xindex]);
				}
			}
		}
	}

	if(fixed_blocks.count(current_block_num))
	{
		//we changed the exit allocation of the current block, recompte
		//backwards shortest paths
		trace_cost tc(this);
		trace_price tp(this);
		betterTiedEdge tiebreaker(*mcnf_cfg);
		mcnf_block *curr = mcnf_cfg->blocks[current_block_num];
		curr->traverse_nodes(reinitialize_node_for_fsf_trace);
		curr->shortest_paths_backwards(tp, tc, tiebreaker);
	}
}

/* Evict antivar q where necessary up to eedge starting at node.
 * Node might not be an antinode, for the rare case where a quantity's
 * new eviction path only temporarily goes into memory.
 * The non-antiquantity is being evicted along eedge.
 * Return true if we are successful
 */
bool HeuristicSolver::evict_antivar(int q, mcnf_node *node, mcnf_edge *eedge)
{
	evict_anti_memoize.clear();
	evict_anti_terminate = NULL;

	int index = node->cindex.get(q);

	if(DebugPRA > 9)
	{
		std::cerr << "Evicting anti " << q << " to node " << node->getName() << "\n";
	}
	//don't allow evictions into a place we can't get out of
	if(node->sinfo[1][index].prev_edge == NULL)
		return false;

	if(!evict_antivar_helper(q, node, eedge->src))
		return false;

	if(DebugPRA > 9)
	{
		std::cerr << "Antivar eviction for " << q << " stopped at " << evict_anti_terminate->getName() << "\n";
	}
	//we need cross block anti edges to extend antivar eviction
	//beyond a single trace (otherwise the crossblock constraints get too hard)
	if(CrossBlockAntis)
	{
		assert(traces.block_trace_index(evict_anti_terminate->block_num) <=
				traces.block_trace_index(eedge->src->block_num));
		//consider where the anti eviction terminated, it was
		//at a block boundary, perform a "global" dominator based eviction
		if(evict_anti_terminate->type == nXBar &&
				evict_anti_terminate->position < 0)
		{
			mcnf_block *block = mcnf_cfg->blocks[evict_anti_terminate->block_num];
			//now make sure we're actually evicting and get the cost
			int index = evict_anti_terminate->cindex.get(q);
			assert(index >= 0);
			mcnf_edge *eedge = evict_anti_terminate->fsinfo[index].next_edge;

			if(eedge == NULL)
			{
				abort();
			}
			int cost = eedge->getCost(q);
			//if this block is dominated by a definer of q, globally evict
			if(block->def_dominator_in.count(-q) > 0 &&
					block->def_dominator_in[-q] != block)
			{
				mcnf_block* dom = block->def_dominator_in[-q];
				global_anti_evict(q, dom, cost,
						alloc_classes->getType(node->allocClass) != AllocationClassInfo::AntiOnly);
			}
		}

	}
	return true;
}


/* Evict antivar q where necessary up to the location specifed by stop.
 Return false if not successful.
 Memoize false results to avoid retracing paths.
 Does a DFS (possibly crossblock) to find the best cost valid eviction path.
 Follows the shortest path, which should result in antivar staying in anti
 network as long as possible.
 Stores the terminating node in evict_anti_terminate
 */
bool HeuristicSolver::evict_antivar_helper(int q, mcnf_node *node, mcnf_node *stop)
{
	mcnf_node *prev;
	mcnf_edge *edge;
	int pindex;
	int index = node->cindex.get(q);
	int vindex = node->cindex.get(-q);

	if(evict_anti_memoize.count(node))
		return false;

	if (index < 0)
		abort();
	if (vindex >= 0 && node->feas_marked[vindex] && node->type != nInsn)
	{
		/* the normal is present in this node, cannot be evicted in this node */
		evict_anti_memoize.insert(node);
		evict_anti_terminate = NULL;
		return false;
	}

	if (node->type == nDef)
	{
		/* a base case */
		evict_anti_terminate = node;
		return true;
	} //don't keep looking past the original eviction site

	else if (stop->type != nDef && (node->type == nXBar || node->type == nInsn) &&
			traces.block_trace_index(node->block_num) <=
			traces.block_trace_index(stop->block_num) &&
			 ((node->position < stop->position) || (node->position
							== stop->position && node->subindex
							<= stop->subindex)))
	{
		assert(traces.blocks_trace(node->block_num) == traces.blocks_trace(stop->block_num));
		if (node->feas_marked[index])
		{
			evict_anti_terminate = node;
			return true;
		}
	}

	edge = node->sinfo[0][index].prev_edge;

	if (edge->src->type == nDef && edge->dst->type == nEntry &&
			CrossBlockEvictions)
	{
		/* crossing block boundaries, assume DFS and go to
		 the previous allocated block exit*/
		int p, np;
		mcnf_node *entry = edge->dst;
		mcnf_node *exit = NULL;
		np = entry->preds.size();
		for (p = 0; p < np; p++)
		{
			mcnf_node *pred = entry->preds[p]->src;
			if (pred->type == nExit && pred->allocClass == entry->allocClass && alloced_blocks.count(pred->block_num))
			{
				exit = pred;
				if (traces.blocks_trace(pred->block_num) != current_trace || !safe_to_change_exit_alloc(pred, entry, q))
				{
					evict_anti_memoize.insert(node);
					evict_anti_terminate = NULL;
					return false;
				}
				break;
			}
		}

		if (exit == NULL)
		{
			evict_anti_memoize.insert(node);
			evict_anti_terminate = NULL;
			return false;
		}

		if (!evict_antivar_helper(q, exit, stop))
		{
			evict_anti_memoize.insert(node);
			evict_anti_terminate = NULL;
			return false;
		}
		/* assume recursive call will take care of fixing up pred block,
		 keep edge the def edge, clear the block constraints */
		visited_blocks.clear();
		unset_entry_exit_condition(mcnf_cfg->blocks[exit->block_num], q, Exit);
		visited_blocks.clear();
		set_entry_exit_condition(mcnf_cfg->blocks[exit->block_num], q,
				exit->allocClass, Exit);
		if(q < 0 && CrossBlockAntis)
		{
			//cross block antis means we have to try both directions since
			//some constraints are one way
			visited_blocks.clear();
			set_entry_exit_condition(mcnf_cfg->blocks[entry->block_num], q,
				entry->allocClass, Entry);
		}
		/* mark edge to sink */
		exit_edge_to_sink(exit, q)->mark_feasible_edge(q);

	}
	else if (!evict_antivar_helper(q, edge->src, stop)) /* uh oh, shortest path is no good, look for another one */
	{
		int p, np;
		np = node->preds.size();
		for (p = 0; p < np; p++)
		{
			mcnf_edge *e = node->preds[p];
			if (e != edge && e->commodity_valid(q))
			{
				edge = e;
				break;
			}
		}

		if (p == np)
		{
			evict_anti_memoize.insert(node);
			evict_anti_terminate = NULL;
			return false;
		}

		if (!evict_antivar_helper(q, edge->src, stop))
		{
			evict_anti_memoize.insert(node);
			evict_anti_terminate = NULL;
			return false;
		}
	}
	else if (edge->src->type == nDef && edge->dst->type == nEntry)
	{
		evict_anti_memoize.insert(node);
		evict_anti_terminate = NULL;
		return false; /* not allowed to change boundary without CrossBlockEvictions */
	}
	/* this is a good edge; if previous path
	 went somewhere else, unmark it and mark this edge */
	prev = edge->src;
	pindex = prev->cindex.get(q);
	if (!prev->feas_marked[pindex])
	abort();
	if (prev->fsinfo[pindex].next_edge == NULL)
	{
		edge->mark_feasible_edge(q);
	}
	else if (prev->fsinfo[pindex].next_edge->dst != node)
	{
		mcnf_edge *next = prev->fsinfo[pindex].next_edge;
		while (next != NULL)
		{
			int nindex = next->dst->cindex.get(q);
			mcnf_edge *e = next->dst->fsinfo[nindex].next_edge;
			next->unmark_feasible_edge(q);
			next = e;
		}

		edge->mark_feasible_edge(q);
	}

	return true;
}

/* Return the edge from node (an exit) to q's sink, NULL if n/a */
mcnf_edge* HeuristicSolver::exit_edge_to_sink(mcnf_node *node, int q)
{
	int s, ns;

	ns = node->succs.size();
	for (s = 0; s < ns; s++)
	{
		mcnf_node *sink = node->succs[s]->dst;
		if (sink->type == nSink && sink->position == q)
		{
			return node->succs[s];
		}
	}
	return NULL;
}

/* Recursively mark the eviction path for q upto eedge from node.
 As we come back, mark the feasible edges. Return true if
 we need to do an antieviction.
 reg is the register we're currently allocated to (only necessary for mulireg classes)
 */
int HeuristicSolver::mark_new_eviction_path(int q, mcnf_edge *eedge,
		mcnf_node *node)
{
	int antievict = 0;
	mcnf_edge *next;
	int index = node->cindex.get(q);
	if (index < 0)
	abort();

	/* see if it is necessary to evict antivar */
	if (q> 0 && node->mirror_node != NULL)
	{
		int aindex = node->cindex.get(-q);
		if (aindex < 0)
		abort();
		if (node->feas_marked[aindex]) /* have to evict antivar */
		{
			antievict = 1;
		}
	}

	if (node == eedge->dst)
	{
		/* found the eviction point */
		eedge->mark_feasible_edge(q);
		return antievict;
	}
	else
	{
		/* otherwise trace back up the shortest path */
		next = node->sinfo[0][index].prev_edge;
		if (!can_evict_to_node(next->src, q))
		{
			int p;
			double mineprice = HUGE_VAL;
			int which = -1;
			int np = node->preds.size();
			for (p = 0; p < np; p++)
			{
				mcnf_node *pred = node->preds[p]->src;
				if (can_evict_to_node(pred, q))
				{
					double ep = mcnf_edge_price(node->preds[p], q);
					if (ep < mineprice)
					{
						mineprice = ep;
						which = p;
					}
				}
			}

			if (which >= 0)
			{
				next = node->preds[which];
				node = next->src;
			}
			else
			abort();
		}

		if (next == NULL)
			abort();
		if (next->src == NULL)
			abort();

		if (next->src->type == nDef && next->dst->type == nEntry)
		{
			/* assume a DFS global eviction strategy, look for the
			 already alloced previous node */
			int p, np, index;
			mcnf_node *entry = next->dst;
			mcnf_node *exit = NULL;
			np = entry->preds.size();
			for (p = 0; p < np; p++)
			{
				mcnf_edge *pedge = entry->preds[p];
				mcnf_node *pred = pedge->src;
				if (pred->type == nExit
						&& alloced_blocks.count(pred->block_num))
				{
					exit = pred;
					break;
				}
			}
			visited_blocks.clear();

			set_entry_exit_condition(mcnf_cfg->blocks[entry->block_num], q, entry->allocClass, Entry);

			antievict |= mark_new_eviction_path(q, eedge, exit);
			/* mark edge to sink */
			exit_edge_to_sink(exit, q)->mark_feasible_edge(q);

			/* mark edge from def in next block */
			index = entry->cindex.get(q);
			entry->sinfo[0][index].prev_edge->mark_feasible_edge(q);
		}
		else
		{
			antievict |= mark_new_eviction_path(q, eedge, next->src);
			next->mark_feasible_edge(q);
		}
	}

	return antievict;
}

/* Return true if q can be globally evicted (pushed to node's class every).
 * For example, caller save register vars. */
bool HeuristicSolver::quantity_is_globally_evictable(int q)
{
	if(mcnf_cfg->savedRegHardRegs.count(q) > 0)
		return true;

	return false;
}

/* Evict quantity q from its current path using eviction edge eedge.
 Erase the old feasible path and create a new one ending
 at level next_nodes. Updates quantity and class alloc.

 If necessary, will also evict an antiquantity.
 */
void HeuristicSolver::evict_quantity(int q, mcnf_edge *eedge, int ec,
		mcnf_node **next_nodes, QuantityAlloc& quantity_alloc,
		std::vector<ClassAlloc>& class_alloc)
{
	int antievict = 0;
	mcnf_node *node = eedge->src;

	if(DebugPRA > 9)
	{
		std::cerr << "Evicting quantity " << q << " to " << eedge->getName(q) << "\n";
	}
	quantity_alloc.erase(q);
	/* follow the feasible path, unmarking as we go */
	while (node != NULL)
	{
		mcnf_edge *next;
		int index = node->cindex.get(q);

		if (index < 0)
		abort();
		next = node->fsinfo[index].next_edge;

		if (next != NULL)
		{
			if (node->type == nExit)
			{
				int ns, s, index;
				if (node->block_num == current_block_num)
				abort();

				/* clear edge to sink */
				if (next->dst->type != nSink)
				abort();
				next->unmark_feasible_edge(q);
				visited_blocks.clear();
				unset_entry_exit_condition(mcnf_cfg->blocks[node->block_num], q,
						Exit);

				/*  assume we are crossing a block and
				 look for a valid successor */
				next = NULL;
				ns = node->succs.size();
				for (s = 0; s < ns; s++)
				{
					mcnf_edge *e = node->succs[s];
					if (e->dst->type == nEntry)
					{
						if (alloced_blocks.count(e->dst->block_num)>0)
						{
							//we can safely ignore blocks with no live in vars
							//(usually exception handling landing pads)
							if(!e->commodity_valid(q)) //q must live into next block
								continue;
							if (next != NULL)
								abort();
							next = e;
						}
						else if (e->dst->block_num == current_block_num)
						{
							next = e;
						}
					}
				}
				if (next == NULL)
					abort();
				node = next->dst;

				if (node->type != nEntry)
					abort();
				/* need to unmark edge from def to entry */
				index = node->cindex.get(q);
				if (node->fsinfo[index].prev_edge != NULL) /* null when at entry node level */
				node->fsinfo[index].prev_edge->unmark_feasible_edge(q);
			}
			else
			{
				node = next->dst;
				next->unmark_feasible_edge(q);
			}
		}
		else
		{
			if (next_nodes[node->allocClass] != node)
				abort();

			class_alloc[node->allocClass].remove(q);
			node = NULL;
		}
	}

	/* now mark the new path.  Trace back up from
	 the resulting eviction class, following the shortest
	 path but looking for edges equal to eedge */
	node = next_nodes[ec];

	antievict = mark_new_eviction_path(q, eedge, node);

	quantity_alloc.put(q,ec);
	if (node->capacity> 0)
	{
		if (class_alloc[ec].size() == node->capacity)
			abort();
		if (quantity_alloc.get(q) >= 0)
			class_alloc[quantity_alloc.get(q)].remove(q);
		quantity_alloc.put(q, ec);
		class_alloc[ec].add(q);
	}
	else
		class_alloc[ec].remove(q);

	if (antievict)
	{
		int i, n;
		if(DebugPRA > 9)
		{
			std::cerr << "Performing antieviction for " << -q << "\n";
		}
		/* find best eviction for antivar */
		const std::vector<int>& memory_classes =
		alloc_classes->getMemoryClasses();
		for (i = 0, n = memory_classes.size(); i < n; i++)
		{
			int allocClass = memory_classes[i];
			int anticlass = alloc_classes->getAntiClass(allocClass);
			mcnf_node *mem = next_nodes[allocClass];
			mcnf_node *anti = next_nodes[anticlass];
			if (!evict_antivar(-q, mem, eedge))
			{
				if(DebugPRA > 9)
				{
					std::cerr << "Evicting antivar " << -q << " to anti network\n";
				}
				if (!evict_antivar(-q, anti, eedge))
					abort();
			}
			//check where antivar actually got evicted to.. with crossblock antis,
			//we may have said to keep it in memory and it ends up in anti
			int mindex = mem->cindex.get(-q);
			int aindex = anti->cindex.get(-q);
			assert(mindex >= 0 && aindex >= 0);
			if(mem->feas_marked[mindex])
				quantity_alloc.put(-q, allocClass);
			else if(anti->feas_marked[aindex])
				quantity_alloc.put(-q, anticlass);
			else
				abort();

		}
	}

	if(quantity_is_globally_evictable(q) &&
			alloc_classes->getType(node->allocClass) == AllocationClassInfo::Memory)
	{
		//antivar might be stuck in mem somewhere else, so do global eviction
		//this might be redundant, but better safe than sorry
		mcnf_block *block = mcnf_cfg->blocks[node->block_num];
		if(block->def_dominator_in.count(q)> 0 &&
				block->def_dominator_in[q] != block)
		{
			mcnf_block* dom = block->def_dominator_in[q];
			global_anti_evict(-q, dom, INT_MAX, false);
		}

		globally_evict_to_class(q, node->allocClass);
	}

	if (node == NULL)
		abort();
}

/* return 0 if, recursively, node isn't connected to an already alloced
 unvisited block or the current block in a weird way*/
bool HeuristicSolver::safe_to_change_helper(mcnf_node *node, int q)
{
	int index;
	int bindex = node->block_num;
	if(node->type == nEntry)
		bindex = -bindex;
	if (visited_blocks.count(bindex)> 0)
		return true;

	visited_blocks.insert(bindex);

	if(node->block_num == current_block_num)
		return false;
	if (alloced_blocks.count(node->block_num))
		return false;

	index = node->cindex.get(q);
	if (index >= 0 && node->feas_marked[index])
		return false;

	if (node->type == nExit)
	{
		int s, ns;
		ns = node->succs.size();
		for (s = 0; s < ns; s++)
		{
			mcnf_node *succ = node->succs[s]->dst;
			if (succ->type == nEntry && succ->allocClass == node->allocClass
					&& !safe_to_change_helper(succ, q))
				return false;
		}
	}
	else if (node->type == nEntry)
	{
		int p, np;
		np = node->preds.size();
		for (p = 0; p < np; p++)
		{
			mcnf_node *pred = node->preds[p]->src;
			if (pred->type == nExit && pred->allocClass == node->allocClass
					&& !safe_to_change_helper(pred, q))
				return false;
		}
	}
	return true;
}

/* Returns true if the only block connected to exit that is allocated
 is the block corresponding to entry.  Also be aware that we shouldn't
 change the exit allocation of the block currently being processed since
 we won't recompute it. */
bool HeuristicSolver::safe_to_change_exit_alloc(mcnf_node *exit,
		mcnf_node *entry, int q)
{
	int s, ns, p, np;

	visited_blocks.clear();

	/* set bits for exit and entry because we handle them specially */
	visited_blocks.insert(exit->block_num);
	visited_blocks.insert(-entry->block_num);

	ns = exit->succs.size();
	for (s = 0; s < ns; s++)
	{
		if(!exit->succs[s]->commodity_valid(q))
			continue; //ignore if not live acrossboundary
		mcnf_node *succ = exit->succs[s]->dst;
		if (succ != entry && succ->type == nEntry && succ->allocClass == exit->allocClass)
		{
			if (succ->block_num == exit->block_num)
			return false; /* self loop */
			if (alloced_blocks.count(succ->block_num)> 0 || succ->block_num == current_block_num)
				return false;
			if (traces.num_traces()> 0)
			{
				/* can't change a boundary that connects to the
				 current trace */
				/* why not?
				int tnum = traces.blocks_trace(succ->block_num);
				if (tnum == current_trace)
					return false;
					*/
			}

			if (safe_to_change_helper(succ, q) == 0)
				return false;
		}
	}

	np = entry->preds.size();
	for (p = 0; p < np; p++)
	{
		mcnf_node *pred = entry->preds[p]->src;
		if (pred != exit && pred->type == nExit && pred->allocClass == entry->allocClass)
		{
			if (pred->block_num == entry->block_num)
				return false; /* self loop, too hard */
			if (alloced_blocks.count(pred->block_num)> 0 || pred->block_num == current_block_num)
				return false;
			if (traces.num_traces()> 0)
			{
				/* can't change a boundary that connects to the
				 current trace */
				/* why not?
				int tnum = traces.blocks_trace(pred->block_num);
				if (tnum == current_trace)
				return false;
				*/
			}
			if (safe_to_change_helper(pred, q) == 0)
			return false;
		}
	}

	return true;
}

/* Return true if node can be along q's eviction path */
bool HeuristicSolver::can_evict_to_node(mcnf_node *node, int q)
{
	int nindex = node->cindex.get(q);
	/* stop if this is no longer a valid path */
	if (nindex < 0)
		return false;
	if (node->capacity> 0 && node->flowcnt >= node->capacity)
		return false;
	if (!finite(node->sinfo[0][nindex].value))
		return false;
	if (!finite(node->sinfo[1][nindex].value))
		return false;
	if (node->feas_marked[nindex])
		return false;

	return true;
}

/*
 Find the cheapest edge that moves q from its current allocation to
 a shortest path to evictto.  Return NULL if no such edge. cost is cost of
 the eviction
 */
mcnf_edge* HeuristicSolver::cheapest_eviction_edge(mcnf_node *evictto, int q,
		double *cost, int *unpriced_cost, bool block_only = false)
{
	mcnf_edge *edge = NULL;
	double minval = HUGE_VAL;
	double price;
	int min_up_cost = 0;
	int up_cost, antiup_cost;
	mcnf_node *node = evictto;
	int index = evictto->cindex.get(q);
	mcnf_node *antinode = NULL;
	double antiprice = 0.0;
	double origantiprice = 0.0;

	if (index < 0)
		return NULL;

	price = evictto->sinfo[1][index].value;
	if (!finite(price))
		return NULL;
	up_cost = evictto->sinfo[1][index].cost;
	antiup_cost = 0;

	if(DebugPRA > 10)
	{
		std::cerr << "Finding cheapest edge for " << q << " from " << evictto->getName();
		std::cerr << " price " << price << " up_cost " << up_cost << "\n";
	}

	while (node != NULL)
	{
		int p, np;
		mcnf_edge *next = NULL;
		mcnf_edge *aedge = NULL;

		int nindex = node->cindex.get(q);
		/* stop if this is no longer a valid path */
		if (!can_evict_to_node(node, q))
			break;

		/* if necessary, simultaneously evict antivar */
		if (q> 0 && node->mirror_node != NULL)
		{
			int antiq = -q;
			int aindex = node->cindex.get(antiq);
			if (aindex < 0)
				abort();
			if (node->feas_marked[aindex]) /* have to evict antivar */
			{
				if (antinode == NULL)
				{
					int a2index;
					antinode = node->mirror_node;
					origantiprice = node->sinfo[0][aindex].value
					+ node->sinfo[1][aindex].value;
					a2index = antinode->cindex.get(antiq);
					if (a2index < 0)
						abort();
					antiprice = antinode->sinfo[1][a2index].value;
					antiup_cost = antinode->sinfo[1][a2index].cost;

					if(DebugPRA > 10) std::cerr << "  antiprice " << antiprice << " antiup_cost " << antiup_cost << " origantiprice " << origantiprice <<"\n";
				}
				else if (node == antinode)
					abort(); /* if this happens need to be more sophisticated in tracing antieviction path */
			}
		}

		/* look at all predecessors of node that are q-reachable */
		np = node->preds.size();
		for (p = 0; p < np; p++)
		{
			mcnf_node *pred = node->preds[p]->src;
			int pindex = pred->cindex.get(q);

			if (pred->block_num != node->block_num || (pred->type == nExit
							&& node->type == nEntry))
				continue;
			if (node->type == nEntry)
				continue; /* not legal to change entry alloc from a def */

			/* if the predecessor is on the current feasible path for q,
			 this edge is a potential eviction edge */
			if (pindex >= 0 && pred->feas_marked[pindex] && node->preds[p]->commodity_valid(q))
			{
				/* compute the cost of this eviction path */
				int antieup_cost = 0;
				double eprice = mcnf_edge_price(node->preds[p], q);
				eprice += pred->fsinfo[pindex].priced_cost;
				eprice += price;

				if (antinode != NULL)
				{
					int ap, anp;
					/* add in cost of anti eviction */
					int aindex = antinode->cindex.get(-q);
					double antievictprice = antiprice;

					/* look for an eviction edge for the antivar */
					anp = antinode->preds.size();
					for (ap = 0; ap < anp; ap++)
					{
						mcnf_node *apred = antinode->preds[ap]->src;
						int apindex = apred->cindex.get(-q);

						if (apindex >= 0 && apred->feas_marked[apindex])
						{
							/* an actual eviction edge */
							if(DebugPRA > 10)
							{
								std::cerr << "  antievictprice " << antievictprice << "\n";
							}
							antievictprice += mcnf_edge_price(
									antinode->preds[ap], -q);
							if(DebugPRA > 10)
							{
								std::cerr << "  antievictprice " << antievictprice << "\n";
								std::cerr << " fspc " <<  apred->fsinfo[apindex].priced_cost;
								std::cerr << " apsinfo " << apred->sinfo[0][apindex].value;
								std::cerr << " apsinfo[1] " << apred->sinfo[1][apindex].value;
								std::cerr << " antinode0 " << antinode->sinfo[0][aindex].value;
								std::cerr << " antinode1 " << antinode->sinfo[1][aindex].value;
							}
							antievictprice += apred->fsinfo[apindex].priced_cost;
							antieup_cost = antinode->preds[ap]->getCost(-q);
							break;
						}
					}

					if (ap == anp) /* this probably shouldn't happen.. */
						antievictprice += antinode->sinfo[0][aindex].value;

					antievictprice -= origantiprice;

					if(DebugPRA > 10)
					{
						std::cerr << "  antievictprice " << antievictprice << "\n";
					}
					eprice += antievictprice;
				}

				if(DebugPRA > 10)
				{
					std::cerr << "  eprice to " << pred->getName() << " " << eprice << "\n";
				}
				int total_up_cost = up_cost + node->preds[p]->getCost(q) + antiup_cost + antieup_cost;
				if (eprice < minval || (eprice == minval && total_up_cost <= min_up_cost) ) //do earliest cheapest eviction
				{
					minval = eprice;
					edge = node->preds[p];
					min_up_cost = total_up_cost;

					if (unpriced_cost)
					{
						*unpriced_cost = total_up_cost;
					}

					if(DebugPRA > 10 && unpriced_cost)
					{
						std::cerr << "  minval " << minval << " unpriced_cost " << *unpriced_cost << "\n";
					}
					if (pred->fsinfo[pindex].prev_edge == NULL && pred->type
							!= nDef)
					abort();
				}
			}
		}

		next = node->sinfo[0][nindex].prev_edge;

		if (next == NULL)
		{
			node = NULL;
		}
		else
		{
			/* can't evict along shortest path, look for another path
			 if there's no eviction edge currently found */
			if (!can_evict_to_node(next->src, q) && edge == NULL)
			{
				double mineprice = HUGE_VAL;
				int which = -1;
				np = node->preds.size();
				for (p = 0; p < np; p++)
				{
					mcnf_node *pred = node->preds[p]->src;
					if (pred->type != nExit && can_evict_to_node(pred, q))
					{
						double ep = mcnf_edge_price(node->preds[p], q);
						if (ep < mineprice)
						{
							mineprice = ep;
							which = p;
						}
					}
				}

				if (which >= 0)
				{
					next = node->preds[which];
					node = next->src;
					price += mineprice;
					up_cost += next->getCost(q);
				}
				else
					node = NULL;
			}
			else
			{
				price += mcnf_edge_price(next, q);
				up_cost += next->getCost(q);
				node = next->src;
			}
		}

		if (node != NULL && antinode != NULL)
		{
			/* simultaneously keep track of the antivar eviction */
			int aindex = antinode->cindex.get(-q);
			aedge = antinode->sinfo[0][aindex].prev_edge;
			if (aedge->src == node && node->type != nInsn)
			{
				/* look for an edge to an anti class */
				int p, np;
				np = antinode->preds.size();
				for (p = 0; p < np; p++)
				{
					mcnf_edge *e = antinode->preds[p];
					if (e->src != node && e->commodity_valid(-q))
					{
						aedge = e;
					}
				}
			}

			antiprice += mcnf_edge_price(aedge, -q);
			antinode = aedge->src;
			antiup_cost += aedge->getCost(-q);
		}

		if (node != NULL && node->type == nDef && next->dst->type == nEntry
				&& CrossBlockEvictions && !block_only)
		{
			/* DFS only non-local eviction
			 only continue searching for an eviction path if node's block is
			 the only allocated (if it's allocated) successor of pred
			 */
			int p, np;
			mcnf_node *entry = next->dst;
			mcnf_node *exit = NULL;

			np = entry->preds.size();
			for (p = 0; p < np; p++)
			{
				mcnf_node *pred = entry->preds[p]->src;
				/* node has to be the only successor of pred that is alloced */
				if (pred->type == nExit
						&& pred->allocClass == entry->allocClass
						&& alloced_blocks.count(pred->block_num)> 0
						&& traces.blocks_trace(pred->block_num) == current_trace)
				{
					if (exit != NULL)
					abort();
					exit = pred;
					if (!safe_to_change_exit_alloc(pred, entry, q))
					{
						if(DebugPRA > 5)
						{
							std::cerr << "Not safe to change alloc from block " << pred->block_num << " to block " << entry->block_num << "\n";
						}
						exit = NULL;
						break;
					}
				}
			}
			node = exit;

			if (antinode != NULL && node != NULL)
			{
				antinode = aedge->dst;

				if (antinode->type != nEntry)
				abort();
				np = antinode->preds.size();
				for (p = 0; p < np; p++)
				{
					mcnf_node *pred = antinode->preds[p]->src;
					if (pred->block_num == node->block_num)
					{
						antinode = pred;
						break;
					}
				}

				if (p == np)
				abort();
			}
		}
	}

	if (edge != NULL && edge->dst->type == nEntry)
	abort();

	if (cost)
	*cost = minval;
	return edge;
}

/* Return the total unpriced cost of the current feasible path from
 node (spanning blocks if necessary) */
int HeuristicSolver::feasible_forward_path_unpriced_cost(int q, mcnf_node *node)
{
	int ret = 0;

	while (node != NULL)
	{
		int index = node->cindex.get(q);
		mcnf_edge *edge = node->fsinfo[index].next_edge;

		if (edge == NULL)
		break;
		ret += edge->getCost(q);

		if (edge->src->type == nExit)
		{
			/* cross to next allocated block, assuming DFS */
			int s, ns;
			ns = edge->src->succs.size();
			int curtrace = traces.block_trace_index(node->block_num);
			for (s = 0; s < ns; s++)
			{
				mcnf_node *succ = edge->src->succs[s]->dst;
				if (succ->type == nEntry &&
						curtrace+1 == traces.block_trace_index(succ->block_num))
				{
					int sindex = succ->cindex.get(q);
					if (sindex >= 0 &&
							 succ->feas_marked[sindex])
					{
						/* get price from def to entry */
						ret += succ->fsinfo[sindex].prev_edge->getCost(q);
						node = succ;
						break;
					}
				}
			}
			if (s == ns)
			abort();
		}
		else
		{
			node = edge->dst;
		}
	}

	/* if we're in the middle of allocation, approximate the
	 remaining cost */
	if (node != NULL)
	{
		int index = node->cindex.get(q);
		ret += node->sinfo[1][index].cost;
	}
	return ret;
}

/* at the end of the block, if q does not have an exit allocation constraint,
 * but the number of quantities using the class is >= the capacity, can't
 * allocate to this node without evicting
 */
bool HeuristicSolver::node_blocked_by_exit_allocations(mcnf_node *node, int q,  std::vector<ClassAlloc>& class_alloc)
{
	int c = node->allocClass;
	int cap = alloc_classes->getCapacity(c);
	if(cap <= 1)
		return false;
	if(node->type != nExit)
		return false;
	int index = node->cindex.get(q);
	if(index < 0)
		return false;
	if(node->used_classes[index].set)
		return false;
	mcnf_block *block = mcnf_cfg->blocks[node->block_num];
	if(block->quantities_used_in_class[c].count(q))
		return false;

	//if a quantity is allocated to this class, but isn't in quantities_used,
	//it just hasn't been set
	ClassAlloc& ca = class_alloc[c];
	int unset = 0;
	for(ClassAlloc::iterator itr = ca.begin(); itr != ca.end(); itr++)
	{
		int cq = *itr;
		if(block->quantities_used_in_class[c].count(cq) == 0 && cq != q)
			unset++;
	}

	int tot = block->quantities_used_in_class[c].size() + unset;
	if(tot < cap) //less than, since q is accounted for in class_alloc
		return false;
	return true;
}
/* Return true if we can't allocate q to this node because
 of an upcoming capacitated insn. */
bool HeuristicSolver::node_blocked_by_insn(mcnf_node *node, int q)
{
	/* handle special case where next node is a constrained insn node */
	if (node->mirror_node != NULL) /* only mem is constrained */
	{
		int index = node->cindex.get(q);
		if (index >= 0 && finite(node->sinfo[1][index].value))
		{
			mcnf_node *next = node->sinfo[1][index].prev_edge->dst;
			if (next->type == nInsn && next->capacity> 0)
			{
				if (q> 0)
				{
					int i, nc;
					int cnt = 0;
					nc = next->commodities.size();
					for (i = 0; i < nc; i++)
					{
						int c = next->commodities[i];
						if (c> 0 && c != q)
						{
							int cindex = node->cindex.get(c);
							if (cindex >= 0 && node->feas_marked[cindex])
							cnt++;
						}
					}
					if (cnt >= next->capacity)
					return true;
				}
			}
		}
	}
	return false;
}



/* Return true if, given the current allocation, q can't be
 allocated to node */
bool HeuristicSolver::node_not_valid_for_allocation(mcnf_node *node, int q,
		int ignoreantis, std::vector<ClassAlloc>& class_alloc)
{
	if (q> 0 && node->capacity> 0 && node->flowcnt >= node->capacity)
		return true;
	if (q> 0 && node->capacity >= 1)
		return false;

	/* antivars aren't part of flowcnt, have to check them specifically;
	 mem/antis don't conflict at insns */
	if (!ignoreantis && node->mirror_node != NULL && node->type != nInsn)
	{
		int antiq = -q;
		int index = node->cindex.get(antiq);
		if (index >= 0 && node->feas_marked[index])
		return true;
	}

	/* handle special case where next node is a constrained insn node */
	if (node_blocked_by_insn(node, q))
		return true;

	if(node_blocked_by_exit_allocations(node, q, class_alloc))
	{
		return true;
	}

	return false;
}

/* Return the cost of evicting q from its current allocation.
 Only consider evictions that don't require additional changes
 to existing allocations.
 Also returns the eviction edge and the class q ends up in at this level
 and the unpriced additional cost of performing this eviction. */
double HeuristicSolver::eviction_cost(int q, int allocClass,
		mcnf_node **next_nodes, mcnf_edge** eedge, int *ec, int *unpriced_cost, std::vector<ClassAlloc>& class_alloc)
{
	int c;
	int up_cost;
	double minval= UNREACHABLE;
	mcnf_edge *best_edge = NULL;
	mcnf_node *orig = next_nodes[allocClass];
	int index = orig->cindex.get(q);
	int best_class = -1;
	int numClasses = alloc_classes->numClasses();
	double base = orig->fsinfo[index].priced_cost + orig->sinfo[1][index].value;

	for (c = 0; c < numClasses; c++)
	{
		mcnf_node *node = next_nodes[c];
		if (c != allocClass && node != NULL)
		{
			if (!node_not_valid_for_allocation(node, q, 1, class_alloc))
			{
				/* we might be able to evict .. */
				double price;
				mcnf_edge *edge = cheapest_eviction_edge(node, q, &price,	&up_cost);

				if (edge != NULL && price < minval)
				{
					minval = price;
					best_edge = edge;
					best_class = c;
				}
			}

		}
	}

	if (eedge)
	*eedge = best_edge;
	if (ec)
	*ec = best_class;

	if (best_edge && unpriced_cost)
	{
		int bcost = feasible_forward_path_unpriced_cost(q, best_edge->src);
		up_cost -= bcost;
		*unpriced_cost = up_cost;
	}

	return minval - base;
}

/* if the classes, oldc and newc, are considered equivalent,
 return true if newc would be a better class to evict from
 (if according to some heuristic)
 */
bool HeuristicSolver::break_eviction_tie(int q, mcnf_node * oldn, mcnf_node* newn,
		std::vector<double>& evictcosts, std::vector<int>& up_costs,
		std::vector<ClassAlloc>& class_alloc)
{
	int var1, var2;
	int next1, next2;
	int level;
	mcnf_block *block;
	int oldc = oldn->allocClass;
	int newc = newn->allocClass;

	if(DebugPRA > 15)
	{
		std::cerr << "considering tie between " << oldc << " and new " << newc << " allocing quantity " << q << "\n";
	}

	if(oldn->primal_solution != NULL && newn->primal_solution != NULL)
	{
		//don't let flow go into a node that is not in the primal solution
		int oindex = oldn->cindex.get(q);
		int nindex = newn->cindex.get(q);
		double oldval = oldn->primal_solution[oindex];
		double newval = newn->primal_solution[nindex];

		if(oldval != newval)
		{
			if(oldval == 0)
				return true;
			if(newval == 0)
				return false;
			//if(newval > oldval)
			//	return true;
		}
	}

	/* prefer putting antivars in the mem network since
	 they are readily evicted to the anti network */
	if (alloc_classes->getType(newc) == AllocationClassInfo::AntiOnly)
		return false;
	else if (alloc_classes->getType(oldc) == AllocationClassInfo::AntiOnly)
		return true;

	if(DebugPRA > 15) std::cerr << "  1\n";

	//if the prices are the same, go with the lower cost
	// this only makes sense if backwards cost is the lowest cost..
	int oindex = oldn->cindex.get(q);
	int nindex = newn->cindex.get(q);
	double ocost = oldn->sinfo[1][oindex].cost+evictcosts[oldc];
	double ncost = newn->sinfo[1][nindex].cost+evictcosts[newc];

	if(!initial_heuristic)
	{
		//respect our initial prices (which are masking deceptive negative costs)
		ocost += up_costs[oldc];
		ncost += up_costs[newc];
	}

	if(DebugPRA > 15) std::cerr << "  ocost " << ocost << "  ncost " << ncost << "\n";

	if(ocost > ncost)
		return true;
	else if(ncost > ocost)
		return false;

	if((class_alloc[oldc].isEmpty() || class_alloc[oldc].size() < oldn->capacity) &&
			(class_alloc[newc].isEmpty() || class_alloc[newc].size() < newn->capacity))
	{
		//both classes unoccupied
		if(alloc_classes->getType(oldc) == AllocationClassInfo::Memory)
			return true;
		else if(alloc_classes->getType(newc) == AllocationClassInfo::Memory)
			return false;
		else if(alloc_classes->getType(newc) == AllocationClassInfo::Register &&
				alloc_classes->getType(oldc) == AllocationClassInfo::Register)
		{
			//do we have a prefered reg?
			unsigned oldr = alloc_classes->getRegister(oldc);
			unsigned newr = alloc_classes->getRegister(newc);
			unsigned newconf = 0;
			unsigned oldconf = 0;
			for(unsigned i = 0, nr = mcnf_cfg->quantities[q].regs.size(); i < nr; i++)
			{
				if(mcnf_cfg->hasConflict(newr, mcnf_cfg->quantities[q].regs[i]))
					newconf++;
				if(mcnf_cfg->hasConflict(oldr, mcnf_cfg->quantities[q].regs[i]))
					oldconf++;
			}
			if(newconf < oldconf)
				return true;

			//check preference register (from hard reg moves)
			int pref = mcnf_cfg->quantities[q].prefClass;
			if(pref >= 0)
			{
				if(pref == newc)
					return true; //new better
				else if(pref == oldc)
					return false; //old better
			}
			int last = mcnf_cfg->quantities[q].lastAlloced;
			if(last >= 0)
			{
				//don't change allocations for no reason
				if(last == newc)
					return true; //new better
				else if(last == oldc)
					return false; //old better
			}
		}
	}
	if(DebugPRA > 15) std::cerr << "  2\n";

	if (class_alloc[oldc].isEmpty() || class_alloc[oldc].size() < oldn->capacity)
		return false;
	if (class_alloc[newc].isEmpty() || class_alloc[newc].size() < newn->capacity)
		return true;

	//TODO: improve this for non unit classes
	var1 = *class_alloc[oldc].begin();
	var2 = *class_alloc[newc].begin();

	if(DebugPRA > 15) std::cerr << "  3\n";

	//if both choice necessitate an eviction, but exactly one of the
	//quantities to be evicted has been spilled before, choose that
	//one to minimize the number of spilled vars (and spill slots)
	if(var1 > 0 && var2 > 0 &&
			alloc_classes->getType(newc) == AllocationClassInfo::Register &&
							alloc_classes->getType(oldc) == AllocationClassInfo::Register &&
							mcnf_cfg->quantities[var1].spilled != mcnf_cfg->quantities[var2].spilled)
	{
		if(mcnf_cfg->quantities[var1].spilled)
			return false;
		else if(mcnf_cfg->quantities[var2].spilled)
			return true;
	}

	if(initial_heuristic) //try to emulate llvm linear scan
	{
		if (var1 <= 0 || var2 <= 0)
		return false;
		//use llvm spill weight
		float weight1 = mcnf_cfg->quantities[var1].weight;
		float weight2 = mcnf_cfg->quantities[var2].weight;

		if(weight1> weight2)
			return true;
		else if(weight1 < weight2)
			return false;
	}

	if(DebugPRA > 15) std::cerr << "  4\n";

	if (fabs(evictcosts[oldc] - evictcosts[newc]) > THRESHOLD)
	{
		return evictcosts[newc] < evictcosts[oldc];
	}
	if(DebugPRA > 15) std::cerr << "  5\n";
	if (up_costs[oldc] != up_costs[newc])
	{
		return up_costs[newc] < up_costs[oldc];
	}
	if(DebugPRA > 15) std::cerr << "  6\n";


	if (var1 <= 0 || var2 <= 0)
	return false;

	block = mcnf_cfg->blocks[oldn->block_num];
	level = oldn->position;

	//use llvm spill weight
	float weight1 = mcnf_cfg->quantities[var1].weight;
	float weight2 = mcnf_cfg->quantities[var2].weight;

	if(weight1> weight2)
	return true;
	else if(weight1 < weight2)
	return false;
	if(DebugPRA > 15) std::cerr << "  7\n";

	if (level >= 0 && level < (int)block->RAATs.size())
	{
		next1 = block->quantity_next_use[level][var1];
		next2 = block->quantity_next_use[level][var2];

		/* traditional linear scan heuristic, evict quantity that
		 is used next latest */
		if (next1> next2)
		return false;
		if (next1 < next2)
		return true;
	}

	//additional heuristics?
	return false;
}

/* return the number of quantities with allocated feasible paths through node
 * only consider non-antis*/
int HeuristicSolver::feasible_node_flow(mcnf_node *node)
{
	int i;
	int nc = node->commodities.size();
	int ret = 0;

	for (i = 0; i < nc; i++)
	{
		if (node->commodities[i]> 0 && node->feas_marked[i])
		ret++;
	}
	return ret;
}

/* make sure quantity_alloc and class_alloc agree */
void HeuristicSolver::sanity_check_allocs(QuantityAlloc& quantity_alloc,
		std::vector<ClassAlloc>& class_alloc, mcnf_node **nodes)
{
	for (QuantityAlloc::iterator itr = quantity_alloc.begin(); itr
			!= quantity_alloc.end(); itr++)
	{
		int q = itr->first;
		int allocClass = itr->second;
		if (allocClass >= 0)
		{
			mcnf_node *node = nodes[allocClass];
			int index = node->cindex.get(q);
			if (!node->feas_marked[index])
				abort();
			if (alloc_classes->getCapacity(allocClass) == 1)
			{
				if (!class_alloc[allocClass].contains(q))
					abort();
			}
			if (quantity_alloc.get(q) == quantity_alloc.get(-q))
				abort();
		}
	}
}

/* Given that q is currently allocated to node, return the edge that
 should be followed to get q's next allocation.  This usually
 just the shortest path.
 */
mcnf_edge* HeuristicSolver::next_edge_for_allocation(mcnf_node *node, int q)
{
	mcnf_edge *next = NULL;
	int index = node->cindex.get(q);
	if (index < 0)
	abort();
	if (!node->feas_marked[index])
	abort();

	next = node->sinfo[1][index].prev_edge;

	if(node->primal_solution != NULL && q > 0)
	{
		mcnf_node *dst = next->dst;
		int dindex = dst->cindex.get(q);
		double price = node->sinfo[1][index].value;
		if(dst->primal_solution[dindex] < THRESHOLD)
		{
			//look for an alternative edge that does lead to a feasible solution
			for(int s = 0, ns = node->succs.size(); s < ns; s++)
			{
				mcnf_edge *edge = node->succs[s];
				if(edge->commodity_valid(q))
				{
					mcnf_node *succ = edge->dst;
					int sindex = succ->cindex.get(q);
					double eprice = edge->getPrice(q);

					if(succ->primal_solution[sindex] > 0 &&
							succ->sinfo[1][sindex].value + eprice <= price+THRESHOLD)
					{
						if(DebugPRA > 10)
						{
							std::cerr << "Changing next edge for allocation from " <<
							next->getName(q) << " to " << edge->getName(q) << "\n";
						}
						next = edge;
						break;
					}
				}
			}
		}
	}

	return next;
}

/* find the best place to allocate q, lookign at the price of the path
 for q and the cost to evict something
 If eviction is necessary, sets eedge to the eviction edge and eq and ec
 to the eviction quantity and class.
 pricesto contains the price to each node from q's feasible path
 this may be null
 If it is necessary to evict an antivar, evictanti is set to the appropriate anti eviction class.
 prices to is the price to next_nodes
 */
int HeuristicSolver::select_alloc_class(int q, mcnf_node **insn_nodes,
		mcnf_node **next_nodes, std::vector<ClassAlloc>& class_alloc,
		mcnf_edge **eedge, int *eq, int *ec, int *evictanti, std::vector<double>& pricesto)
{
	mcnf_node *node;
	int c, index;
	int pick = -1;
	double minval= UNREACHABLE;
	int numClasses = alloc_classes->numClasses();
	std::vector<double> prices(numClasses, 0.0);
	std::vector<int> costs(numClasses, 0);
	std::vector<double> evictcosts(numClasses, 0.0);
	std::vector<int> antievict(numClasses, 0);
	std::vector<mcnf_edge*> edges(numClasses, (mcnf_edge*)NULL);
	std::vector<int> eclasses(numClasses, 0); /* where things get evicted to */
	std::vector<int> eqs(numClasses, 0); /* what gets evicted */
	std::vector<int> up_costs(numClasses, 0); /* unpriced costs of evictions */

	/* compute the cost of each possible allocation decision
	 This is the cost for this variable alone, plus an
	 eviction cost if a capacitated node already has something
	 allocated to it.  The costs may be infinite for unreachable nodes */
	for (c = 0; c < numClasses; c++)
	{
		node = next_nodes[c];
		if (node == NULL)
		prices[c] = UNREACHABLE;
		else
		{
			double price;
			index = node->cindex.get(q);
			if (index < 0)
			{
				prices[c] = UNREACHABLE;
				continue;
			}

			costs[c] = node->sinfo[1][index].cost;
			if (pricesto.size() == 0)
			{
				if (node->feas_marked[index])
				price = node->fsinfo[index].priced_cost
				+ node->sinfo[1][index].value;
				else
				price = node->sinfo[0][index].value
				+ node->sinfo[1][index].value;
			}
			else
			{
				price = pricesto[c] + node->sinfo[1][index].value;
			}

			if (node->type == nEntry)
			{
				if (!mcnf_boundary_node_valid(node, q))
					price = UNREACHABLE;
				else if (alloc_classes->getCapacity(c)> 0 && class_alloc[c].size() == node->capacity
						&& !class_alloc[c].contains(q))
					price = UNREACHABLE; /* uninitialized vars.. ? */
			}

			/* avoid allocating into a capacitated insn, capcity
			 only applies to nonantis */
			if (insn_nodes[c] != NULL && insn_nodes[c]->capacity> 0)
			{
				int iindex = insn_nodes[c]->cindex.get(q);
				if (iindex >= 0 && q> 0)
				{
					int flow = feasible_node_flow(insn_nodes[c]);
					if (flow >= insn_nodes[c]->capacity)
					price = UNREACHABLE;
				}
			}

			if (finite(price))
			{
				/* this class is capacitated and occupied */
				edges[c] = NULL;
				if (node->capacity> 0 && class_alloc[c].size() == node->capacity && !class_alloc[c].contains(q))
				{
					//need to evict something in order to allocate
					evictcosts[c] = HUGE_VAL;

					//figure out what we have to evict in order to allocate q
					for(ClassAlloc::iterator itr = class_alloc[c].begin(), eitr = class_alloc[c].end();
						itr != eitr; itr++)
					{
						mcnf_edge *minedge;
						int mineclass, minupcost;
						double cost = eviction_cost(*itr, c, next_nodes, &minedge, &mineclass, &minupcost, class_alloc);
						if(cost < evictcosts[c])
						{
							evictcosts[c] = cost;
							eqs[c] = *itr;
							edges[c] = minedge;
							eclasses[c] = mineclass;
							up_costs[c] = minupcost;
							if(eqs[c] == q)
							abort();
						}
					}
				}
				else if(node_blocked_by_exit_allocations(node, q, class_alloc))
				{
					//need to evict something in order to allocate
					evictcosts[c] = HUGE_VAL;

					//figure out what we have to evict in order to allocate q
					for(ClassAlloc::iterator itr = class_alloc[c].begin(), eitr = class_alloc[c].end();
							itr != eitr; itr++)
					{
						mcnf_edge *minedge;
						int mineclass, minupcost;
						double cost = eviction_cost(*itr, c, next_nodes, &minedge, &mineclass, &minupcost, class_alloc);
						if(cost < evictcosts[c])
						{
							evictcosts[c] = cost;
							eqs[c] = *itr;
							edges[c] = minedge;
							eclasses[c] = mineclass;
							up_costs[c] = minupcost;
							if(eqs[c] == q)
								abort();
						}
					}
				}
				else if (node->mirror_node != NULL && q> 0) /* antivar eviction */
				{
					int aindex = node->cindex.get(-q);
					if (aindex < 0)
						;
					else if (node->feas_marked[aindex]) /* have to evict */
					{
						mcnf_edge *aedge = NULL;
						int aclass;
						evictcosts[c] = eviction_cost(-q, c, next_nodes,
								&aedge, &aclass, NULL, class_alloc);
						if (aedge != NULL)
							antievict[c] = aclass;
					}
					else if(!initial_heuristic)
					{
						/* ignore case of unallocated anti at first
						 *  it turns out this
						 * just makes things worse, especially with respect to
						 * caller save registers */
						mcnf_node *mirror = node->mirror_node;
						int mindex = mirror->cindex.get(-q);
						assert(mindex >= 0);
						double anti_node_price = node->sinfo[0][aindex].value + node->sinfo[1][aindex].value;
						double anti_mirror_price = mirror->sinfo[0][mindex].value + mirror->sinfo[1][mindex].value;
						//if we allocate to this node, we force the anti to the mirror node, which may result
						//in extra cost
						double extra = anti_mirror_price - anti_node_price;
						price += extra;
					}
				}

				//if (evictcosts[c] == 0 && edges[c] != NULL)
				//	evictcosts[c] = THRESHOLD; /* avoid needless evictions */
				if(initial_heuristic && evictcosts[c]> 0 && finite(evictcosts[c]))
					evictcosts[c] = 3.0; // exclusively use llvm heuristic to choose evictions
				if (evictcosts[c]> 0) /* ignore negative evictcosts except when breaking ties */
					price += evictcosts[c];
			}

			prices[c] = price;
		}
	}

	/* find best (lowest price) path */
	for (c = 0; c < numClasses; c++)
	{
		if (prices[c] < minval-THRESHOLD/2.0)
		{
			if(DebugPRA > 9)
			{
				std::cerr << "Choosing class " << c << " over " << pick << " with price difference " << minval-prices[c] << "\n";
			}
			pick = c;
			minval = prices[c];
		}
		/* break ties with some heuristic */
		else if (pick >= 0 && fabs(prices[c] - minval) < THRESHOLD && break_eviction_tie(q,
						next_nodes[pick], next_nodes[c], evictcosts, up_costs,
						class_alloc))
		{
			if(DebugPRA > 9)
			{
				std::cerr << "Breaking tie by prefering " << c << " over " << pick << "\n";
			}
			pick = c;
			minval = prices[c];
		}

	}

	if (pick < 0)
	{
		int i;
		std::cerr << "FAILED to select alloc class for " << q << " in block " << current_block_num << "\n";
		std::cerr << "pricesto:";
		for( i = 0; i < numClasses; i++)
			std::cerr << " " << pricesto[i];
		std::cerr << "\nprices:";
		for( i = 0; i < numClasses; i++)
			std::cerr << " " << prices[i];
		std::cerr << "\nclass_allocs:";
		for(i=0; i < numClasses; i++)
			std::cerr << " " <<i <<"| " << class_alloc[i];
		std::cerr << "\n";
		mcnf_block *block = mcnf_cfg->blocks[next_nodes[0]->block_num];
		for(i=0; i < numClasses; i++)
		{
			std::cerr << i <<"| ";
			copy(block->quantities_used_in_class[i].begin(), block->quantities_used_in_class[i].end(), std::ostream_iterator<int>(std::cerr," "));
			std::cerr << "\t";
		}
		std::cerr << "\n";
		abort();
	}
	if (eedge != NULL)
		*eedge = edges[pick];
	if (ec != NULL)
		*ec = eclasses[pick];
	if (eq != NULL)
		*eq = eqs[pick];
	if (evictanti != NULL)
		*evictanti = antievict[pick];

	if(eqs[pick] == q)
		abort();
	if(DebugPRA > 8)
	{
		std::cerr << "Picking class " << pick << " for " << q;
		if(edges[pick])
		{
			std::cerr << " evicting " << eqs[pick] << " with cost " << evictcosts[pick] << "\n";
		}
		if(next_nodes[pick]->primal_solution)
		{
			std::cerr << " with primal " <<next_nodes[pick]->primal_solution[next_nodes[pick]->cindex.get(q)];
		}
		std::cerr << "\nPrices:";
		for(int i = 0; i < numClasses; i++)
		{
			std::cerr << " " << i << ":" << prices[i];
		}
		std::cerr << "\n";
		std::cerr << "ecosts =";
		for(int i = 0; i < numClasses; i++)
		{
			if(edges[i])
			{
				std::cerr << " [" << i << "," << class_alloc[i] << ":" << evictcosts[i];
				std::cerr << "," << up_costs[i];
				std::cerr << "(" << mcnf_cfg->quantities[*class_alloc[i].begin()].weight << ")]";
				if(next_nodes[i]->primal_solution)
				{
					std::cerr << " prim" << next_nodes[i]->primal_solution[next_nodes[i]->cindex.get(q)];
				}
			}
		}
		std::cerr << "\n";
	}

	if(q > 0)
	{
		if(alloc_classes->getType(pick) == AllocationClassInfo::Register)
			mcnf_cfg->quantities[q].lastAlloced = pick; //keep track of our last register pick
		else if(alloc_classes->getType(pick) == AllocationClassInfo::Memory)
			mcnf_cfg->quantities[q].spilled = true;
	}
	return pick;
}

/* Set the pricesto array to be the price from def to each class;
 if defining into an insn, include cost through to xbar */
void HeuristicSolver::fill_def_pricesto(mcnf_node *def,
		std::vector<double>& pricesto, std::vector<mcnf_edge*>& edges,
		int mirroralloc)
{
	int i, s, ns;
	int q = def->position;
	int numClasses = alloc_classes->numClasses();

	for (i = 0; i < numClasses; i++)
	{
		pricesto[i] = UNREACHABLE;
	}
	edges.assign(numClasses, (mcnf_edge*)NULL);

	ns = def->succs.size();
	for (s = 0; s < ns; s++)
	{
		mcnf_node *succ = def->succs[s]->dst;
		if (succ->type == nEntry && succ->allocClass == mirroralloc)
			pricesto[succ->allocClass] = UNREACHABLE; /* for uninitialized antivars */
		else
		{
			pricesto[succ->allocClass] = mcnf_edge_price(def->succs[s], q);
			edges[succ->allocClass] = def->succs[s];

			if (succ->type == nInsn) /* price through to xbar */
			{
				mcnf_edge *edge;
				int index = succ->cindex.get(q);
				if (index < 0)
					pricesto[succ->allocClass] = UNREACHABLE;
				else
				{
					edge = succ->sinfo[1][index].prev_edge;
					if (edge == NULL)
						pricesto[succ->allocClass] = UNREACHABLE;
					else
						pricesto[succ->allocClass] += mcnf_edge_price(edge, q);
				}

				if(TiedConstraints)
				{
					//check for tied constraint and modify cost accordingly
					mcnf_block *block = mcnf_cfg->blocks[def->block_num];
					int level = succ->position;
					if(block->TCRAATMap.count(level))
					{
						TC_RAATInfo& TC = block->TC_RAATs[block->TCRAATMap[level]];
						if(TC.dstq == q)
						{
							//check feas marked since quantity info has already been cleared of last uses
							int sindex = succ->cindex.get(TC.srcq);
							if(sindex >= 0 && succ->feas_marked[sindex])
							{
								pricesto[succ->allocClass] += TC.cost;
							}
						}
					}
				}
			}
		}
	}


}

/* allocate q (possibly evicting other quantities and updates arrays)
 assume that all other live variables are allocated in next_nodes
 q is being defined
 */
void HeuristicSolver::allocate_quantity_at_level(mcnf_node* def,
		mcnf_node **insn_nodes, mcnf_node **next_nodes,
		QuantityAlloc& quantity_alloc, std::vector<ClassAlloc>& class_alloc)
{
	mcnf_node *node;
	mcnf_edge *eedge;
	int ec, eq;
	int index, pick;
	int antievict;
	int numClasses = alloc_classes->numClasses();
	std::vector<double> pricesto(numClasses, 0.0);
	std::vector<mcnf_edge*> defedges(numClasses, (mcnf_edge*)NULL);
	mcnf_edge *defedge;
	int q = def->position;

	if (def->type != nDef)
	abort();

	/* figure out the cost of defining into each class using edges*/
	fill_def_pricesto(def, pricesto, defedges, quantity_alloc.get(-q));

	pick = select_alloc_class(q, insn_nodes, next_nodes, class_alloc, &eedge, &eq,
			&ec, &antievict, pricesto);
	/* perform necessary evictions */
	if (eedge != NULL)
	{
		/* have to evict a quantity */
		if (eedge == NULL)
			abort();
		evict_quantity(eq, eedge, ec, next_nodes, quantity_alloc, class_alloc);
	}

	defedge = defedges[pick];
	if (defedge == NULL)
	abort();

	if (antievict)
	{
		if (antievict < 0 || next_nodes[antievict] == NULL)
			abort();
		if (!evict_antivar(-q, next_nodes[antievict], defedge))
			abort();
		quantity_alloc.put(-q, antievict);
	}

	/* allocate q to pick */
	defedge->mark_feasible_edge(q);

	node = defedge->dst;
	if (node->type == nInsn) /* mark through to xbar */
	{
		mcnf_edge *edge;
		index = node->cindex.get(q);
		if (index < 0)
			abort();
		edge = node->sinfo[1][index].prev_edge;

		if (edge->dst != next_nodes[pick])
			abort();
		edge->mark_feasible_edge(q);
	}

	if(insn_nodes[pick])
		assign_quantity_to_class(q, insn_nodes[pick], class_alloc, quantity_alloc); //liveout only values are offset from the insn
	else
		assign_quantity_to_class(q, next_nodes[pick], class_alloc, quantity_alloc);
}

/*
 compute backwards paths for all marked blocks (abs value of) or
 all blocks if marked is NULL in the trace currtrace
 */
void HeuristicSolver::update_trace_paths(int currtrace, hash_set<int> *marked)
{
	int b, numb;
	const std::vector<mcnf_block*>& trace = traces.get_trace(currtrace);
	numb = trace.size();
	bool recompute = false;
	current_trace = currtrace;

	for (b = numb-1; b >= 0; b--)
	{
		mcnf_block *block = trace[b];
		if (alloced_blocks.count(block->index))
			break;
		if (marked == NULL)
		{
			trace_cost tc(this);
			trace_price tp(this);
			betterTiedEdge tiebreaker(*mcnf_cfg);
			block->traverse_nodes(reinitialize_node_for_fsf_trace);
			block->shortest_paths_backwards(tp, tc, tiebreaker);
			if(DebugPRA > 3)
			{
				std::cerr << "Updating (all) trace paths for block " << block->index << "\n";
			}
		}
		else if (marked->count(block->index) || marked->count(-block->index) || recompute)
		{
			//TODO: if just fixing entry this can be a lot more efficient
			recompute = true;
			trace_cost tc(this);
			trace_price tp(this);
			betterTiedEdge tiebreaker(*mcnf_cfg);

			if(DebugPRA > 3)
			{
				std::cerr << "Updating trace paths for block " << block->index << "\n";
			}
			block->traverse_nodes(reinitialize_node_for_fsf_trace);
			block->shortest_paths_backwards(tp, tc, tiebreaker);

			if (!TraceUpdateAll)
				recompute = false;
		}
	}
}

/* Return true if the feasible paths in block don't match up
 at the boundary */
bool HeuristicSolver::paths_fail_entry_exit_loop_constraints(mcnf_block *block)
{
	int i, len;
	std::vector<mcnf_node*>& entries = block->entries;
	std::vector<mcnf_node*>& exits = block->exits;

	len = entries.size();
	if (len != (int)exits.size())
	abort();

	for (i = 0; i < len; i++)
	{
		int j, nc;
		mcnf_node *entry = entries[i];
		mcnf_node *exit = exits[i];

		if (entry->allocClass != exit->allocClass)
		abort();

		/* for every commodity in entry, check exit's allocation
		 (only care about vars live in both places) */
		nc = entry->commodities.size();
		for (j = 0; j < nc; j++)
		{
			if (entry->feas_marked[j])
			{
				int q = entry->commodities[j];
				int index = exit->cindex.get(q);
				if (index >= 0 && !exit->feas_marked[index])
				return true;
			}
		}
	}

	return false;
}

/* Unmark entry/exit conditions for a single quantity.
 Assumes visited_blocks is cleared on first call
 */
void HeuristicSolver::unset_entry_exit_condition(mcnf_block *block, int q,
		enum entry_exit type, bool recurse /* = true */, int allocClass /* = -1 */)
{
	int i;
	int len;
	mcnf_node *classnode = NULL;
	int bindex = block->index;
	bool qalive = false;

	if (type == Entry)
		bindex = -bindex;

	/* distinguish between entry and exit of block by negating
	 index, presumably never set entry conditions of block 0 */
	if (recurse && visited_blocks.count(bindex))
		return;

	if(recurse)
		visited_blocks.insert(bindex);

	if(DebugPRA > 3)
	{
		std::cerr << "Unsetting Entry/Exit for " << q  << " at " << bindex << "\n";
	}

	std::vector<mcnf_node*> *vec = &block->entries;
	if (type == Exit)
		vec = &block->exits;

	len = vec->size();
	for (i = 0; i < len; i++)
	{
		mcnf_node *node = vec->at(i);

		int index = node->cindex.get(q);
		if (index >= 0)
		{
			qalive = true;
			if(q > 0)
			{
				if(allocClass < 0)
					allocClass = node->used_classes[index].allocClass;
				assert(allocClass == node->used_classes[index].allocClass);
			}

			if (node->used_classes[index].set)
			{
				node->used_classes[index].set = 0;
				node->used_classes[index].allocClass = -1;

				if (q > 0 && node->allocClass == allocClass
						&& alloc_classes->getCapacity(allocClass)> 0)
				{
					for(unsigned c = 0, nc = node->commodities.size(); c < nc; c++)
					{
						if(DebugPRA > 10)
						{
							std::cerr << "Unsetting unusable " << node->commodities[c] << " in block " << bindex << "\n";
						}
						if(node->commodities[c] != q)
						{
							node->used_classes[c].unusable = false;
						}
					}
				}
			}
			else
				return;
		}
		if (q > 0 && node->allocClass == allocClass
				&& alloc_classes->getCapacity(allocClass)> 0)
		{
			classnode = node;
			if( alloc_classes->getCapacity(allocClass)>1 && type == Exit)
			{
				block->quantities_used_in_class[allocClass].erase(q);
			}
		}
	}

	if(!recurse)
		return;

	if(!qalive)
	{
		if(q > 0 && allocClass >= 0 && alloc_classes->getCapacity(allocClass)>1)
		{
			hash_set<int> seen;
			unset_quantities_used_in_class(block, allocClass, type, seen, q);
		}

		if(classnode != NULL)
		{
			//unset unusability of current live vars
			std::vector<int> live(classnode->commodities.begin(), classnode->commodities.end());
			hash_set<int> marked(visited_blocks);
			marked.erase(bindex);
			set_entry_exit_usability(classnode, marked, live, false);
		}
	}
	/* now look at all preds/succ */
	else if (type == Entry)
	{
		len = block->preds.size();
		for (i = 0; i < len; i++)
		{
			mcnf_block *pred = block->preds[i];
			unset_entry_exit_condition(pred, q, Exit, recurse, allocClass);
		}
	}
	else
	{
		len = block->succs.size();
		for (i = 0; i < len; i++)
		{
			mcnf_block *succ = block->succs[i];
			unset_entry_exit_condition(succ, q, Entry, recurse, allocClass);
		}
	}
}

//recursivly explore connected exits and very conservatively block
//and that are "full" (blocked anywhere)
//this does not get undone
//TODO: be more accurate
void HeuristicSolver::set_multireg_usability(mcnf_node *node, hash_set<int>& marked)
{
	int bindex = node->block_num;
	if(node->type == nEntry)
		bindex = -bindex;
	else if(node->type != nExit)
		abort();

	if(marked.count(bindex) > 0)
		return;
	marked.insert(bindex);
	bool changed = false;
	if(node->type == nExit)
	{ //be very conservative with multireg classes.. block everything for now
		mcnf_block * block = mcnf_cfg->blocks[node->block_num];
		if(block->quantities_used_in_class[node->allocClass].size() > 1)
		{
			int cap = alloc_classes->getCapacity(node->allocClass);
			if((int)block->quantities_used_in_class[node->allocClass].size() >= cap)
			{
				for(int c = 0, n = node->commodities.size(); c < n; c++)
				{
					int q = node->commodities[c];
					if(block->quantities_used_in_class[node->allocClass].count(q) == 0)
					{
						if(!node->used_classes[c].unusable)
						{

							if(DebugPRA > 10)
							{
								std::cerr << "Setting multiunusable  " << q << " for class " << node->allocClass << " at " << bindex << "\n";
							}
							node->used_classes[c].unusable = true;
							changed = true;
						}
					}
				}
			}
		}
	}

	//now explore succs/preds if we've changed a fix
	if(changed)
	{
		fixed_blocks.insert(bindex); //may have to recalculate
	}
	//even if nothing changed, continue exploring since this block may
	//have been previously seen and fixed with a different live set
	if (node->type ==nEntry)
	{
		int len = node->preds.size();
		for (int i = 0; i < len; i++)
		{
			mcnf_node *pred = node->preds[i]->src;
			if(pred->type == nExit)
			{
				set_multireg_usability(pred, marked);
			}
		}
	}
	else
	{
		int len = node->succs.size();
		for (int i = 0; i < len; i++)
		{
			mcnf_node *succ = node->succs[i]->dst;
			if(succ->type == nEntry)
			{
				set_multireg_usability(succ, marked);
			}
		}
	}
}

/*
 * Recursively explore all the nodes connected to node and mark them
 * as unusable for those quantities in live.
 */
void HeuristicSolver::set_entry_exit_usability(mcnf_node *node,
		hash_set<int>& marked, const std::vector<int>& Live, bool unusable)
{
	int bindex = node->block_num;
	if(node->type == nEntry)
		bindex = -bindex;
	else if(node->type != nExit)
		abort();

	if(marked.count(bindex) > 0)
		return;
	marked.insert(bindex);

	if(DebugPRA > 3)
	{
		std::cerr << "Setting unusable  " << unusable << " for class " << node->allocClass << " at " << bindex << "\t";
	}

	std::vector<int> live; //kind of pass by value, but just copy once
	live.reserve(Live.size());

	//mark all the live commodities' usability
	bool changed = false;
	for(unsigned c = 0, nc = Live.size(); c < nc; c++)
	{
		int q = Live[c];
		int cindex = node->cindex.get(q);
		if(cindex >= 0)
		{
			if(DebugPRA > 3)
			{
				std::cerr << q << " ";
			}
			//mark as unusable
			if(node->used_classes[cindex].unusable != unusable)
			{
				node->used_classes[cindex].unusable = unusable;
				changed = true;
			}

			//add to live
			live.push_back(q);
		}
	}

	if(DebugPRA > 3)
	{
		std::cerr << "\n";
	}
	if(live.size() == 0)
		return;

	//now explore succs/preds if we've changed a fix
	if(changed)
	{
		fixed_blocks.insert(bindex); //may have to recalculate
	}
	//even if nothing changed, continue exploring since this block may
	//have been previously seen and fixed with a different live set
	if (node->type ==nEntry)
	{
		int len = node->preds.size();
		for (int i = 0; i < len; i++)
		{
			mcnf_node *pred = node->preds[i]->src;
			if(pred->type == nExit)
			{
				set_entry_exit_usability(pred, marked, live, unusable);
			}
		}
	}
	else
	{
		int len = node->succs.size();
		for (int i = 0; i < len; i++)
		{
			mcnf_node *succ = node->succs[i]->dst;
			if(succ->type == nEntry)
			{
				set_entry_exit_usability(succ, marked, live, unusable);
			}
		}
	}
}

void HeuristicSolver::set_quantities_used_in_class(mcnf_block *block, int allocClass,
		entry_exit type, hash_set<int>& marked, int q)
{
	int bindex = block->index;
	if(type == Entry)
		bindex = -bindex;

	if(marked.count(bindex))
		return;
	marked.insert(bindex);

	if(DebugPRA > 5)
	{
		std::cerr << "Setting quantities used in class " << allocClass << " for " << q << " at " << bindex << "\n";
	}

	if(type == Entry)
	{
		//just explore predeccessors
		int len = block->preds.size();
		for (int i = 0; i < len; i++)
		{
			mcnf_block *pred = block->preds[i];
			set_quantities_used_in_class(pred, allocClass, Exit, marked, q);
		}
	}
	else //exit
	{
		block->quantities_used_in_class[allocClass].insert(q);
		for(int i = 0, len = block->succs.size(); i < len; i++)
		{
			mcnf_block *succ = block->succs[i];
			set_quantities_used_in_class(succ, allocClass, Entry, marked, q);
		}

	}
}
void HeuristicSolver::unset_quantities_used_in_class(mcnf_block *block, int allocClass,
		entry_exit type, hash_set<int>& marked, int q)
{
	int bindex = block->index;
	if(type == Entry)
		bindex = -bindex;

	if(marked.count(bindex))
		return;
	marked.insert(bindex);

	if(DebugPRA > 5)
	{
		std::cerr << "unsetting quantities used in class " << allocClass << " for " << q << " at " << bindex << "\n";
	}
	if(type == Entry)
	{
		//just explore predeccessors
		int len = block->preds.size();
		for (int i = 0; i < len; i++)
		{
			mcnf_block *pred = block->preds[i];
			unset_quantities_used_in_class(pred, allocClass, Exit, marked, q);
		}
	}
	else //exit
	{
		block->quantities_used_in_class[allocClass].erase(q);

		int cap = alloc_classes->getCapacity(allocClass);
		if((int)block->quantities_used_in_class[allocClass].size() >= cap)
		{
			//make sure is now marked unusable (previously was allowed access to class)
			std::vector<mcnf_node*>& exits = block->exits;
			int len = exits.size();
			//for each exit/entry node mark what class q is alloced to
			for (int i = 0; i < len; i++)
			{
				mcnf_node *node = exits[i];
				if(node->allocClass == allocClass)
				{
					int index = node->cindex.get(q);
					if(index >= 0)
					{
						node->used_classes[index].unusable = true;
					}
					break;
				}
			}
		}

		for(int i = 0, len = block->succs.size(); i < len; i++)
		{
			mcnf_block *succ = block->succs[i];
			unset_quantities_used_in_class(succ, allocClass, Entry, marked, q);
		}

	}
}
/* Mark unusables in the entry nodes of block so that quantity q
 must be alloced to class.  Recursively decend on successors and
 predecessors.  visited_blocks must be initialized properly

 Along all edges where q is alive, set q and mark all quantities unusable.
 The blocks get marked as visited and stay that way.

 If q isn't alive along an edge, propagate all the vars that are alive and
 mark them as unusable if the class is a capacitated class.

 When we make the transition from q being alive to being dead we don't permanently
 mark blocks as visited since different paths to the same block might propagate
 different sets of live variables.
 */
void HeuristicSolver::set_entry_exit_condition(mcnf_block *block, int q,
		int allocClass, entry_exit type, bool recurse /* = true */, bool prevFull /* = false */ )
{
	int i;
	int len;
	mcnf_node *classnode = NULL;
	int bindex = block->index;
	if (type == Entry)
		bindex = -bindex;

	/* distinguish between entry and exit of block by negating
	 index, presumably never set entry conditions of block 0 */
	if (recurse && visited_blocks.count(bindex)> 0)
		return;

	if(recurse)
		visited_blocks.insert(bindex);

	std::vector<mcnf_node*> *vec = &block->entries;
	if (type == Exit)
		vec = &block->exits;

	bool qlive = false;
	bool isFull = false;
	len = vec->size();
	//for each exit/entry node mark what class q is alloced to
	for (i = 0; i < len; i++)
	{
		mcnf_node *node = vec->at(i);

		int index = node->cindex.get(q);
		if (index >= 0)
		{
			qlive = true;

			if (q> 0)
			{
				if (node->used_classes[index].set
						&& node->used_classes[index].allocClass != allocClass)
					abort();
				else if (node->used_classes[index].set)
					return;

				node->used_classes[index].set = 1;
				node->used_classes[index].allocClass = allocClass;

				if(DebugPRA > 5)
				{
					std::cerr << "Setting Entry/Exit for " << q  << " with class " << allocClass << " at " << bindex << "\n";
				}
			}
			else /* yuck, only set relevant classes for antivar */
			{
				int indexa = node->cindex.get(-q);
				if (node->allocClass == allocClass || node->allocClass
						== alloc_classes->getAntiClass(allocClass))
				{
					if (node->used_classes[index].set)
						return;

					if (indexa >= 0 && node->used_classes[indexa].set
							&& node->used_classes[indexa].allocClass
							== allocClass)
						abort();
					node->used_classes[index].set = 1;
					node->used_classes[index].allocClass = allocClass;
					if(DebugPRA > 5)
					{
						std::cerr << "Setting Entry/Exit for " << q  << " with class " << allocClass << " at " << bindex << "\n";
					}
				}
			}
		}

		if (node->allocClass == allocClass
				&& alloc_classes->getCapacity(allocClass) > 0)
		{
			assert(classnode == NULL);
			classnode = node;
			if(alloc_classes->getCapacity(allocClass) == 1)
			{
				isFull = true;
				for(unsigned c = 0, nc = node->commodities.size(); c < nc; c++)
				{
					if(node->commodities[c] != q)
					{
						node->used_classes[c].unusable = true;
					}
				}
			}
			else //multireg class
			{
				//count number of quantities using this node
				int cnt = 0;
				for(unsigned c = 0, nc = node->commodities.size(); c < nc; c++)
				{
					if(node->used_classes[c].set && node->used_classes[c].allocClass == allocClass)
					{
						cnt++;
					}
				}

				if(type == Exit)
					block->quantities_used_in_class[allocClass].insert(q);
				int cap = alloc_classes->getCapacity(allocClass);

				//even more conservative
				if((int)block->quantities_used_in_class[allocClass].size() >= cap)
				{
					for(unsigned c = 0, nc = node->commodities.size(); c < nc; c++)
					{
						if(block->quantities_used_in_class[allocClass].count(node->commodities[c]) == 0)
						{
							node->used_classes[c].unusable = true;
						}
					}
				}
				if(cnt == cap) //full
				{
					isFull = true;
					if(DebugPRA > 5)
					{
						std::cerr << "multireg node full at " << node->getName() << "\n";
					}
					for(unsigned c = 0, nc = node->commodities.size(); c < nc; c++)
					{
						if(!node->used_classes[c].set)
						{
							if(DebugPRA > 9) std::cerr << "setting " << node->commodities[c] << " as unusable in full node " << node->getName() << "\n";
							node->used_classes[c].unusable = true;
						}
					}
				}
			}
		}
	}

	//insert into fixed even if not recursing - may still need to update path info
	fixed_blocks.insert(bindex);
	if(!recurse)
		return;
	if(DebugPRA > 3)
	{
		std::cerr << "Fixing block " << bindex << "\n";
	}
	/* now look at all preds/succ */
	if(!qlive)
	{
		if(alloc_classes->getCapacity(allocClass) > 1)
		{
			hash_set<int> seen;
			set_quantities_used_in_class(block, allocClass, type, seen, q);
			if(classnode != NULL)
			{
				seen.clear();
				set_multireg_usability(classnode, seen);
			}
		}
		if(classnode != NULL && (isFull || prevFull))
		{
			//need to propagate unusability of this class for current live vars
			std::vector<int> live(classnode->commodities.begin(), classnode->commodities.end());
			hash_set<int> marked(visited_blocks);
			marked.erase(bindex);
			set_entry_exit_usability(classnode, marked, live, true);
		}
	}
	else if (type == Entry)
	{
		//don't fully constraint antis, but punt on self loops since they would require
		//recomputation after choosing the entry
		if(q < 0 && alloc_classes->getType(allocClass) == AllocationClassInfo::Memory)
			;
		else
		{
			len = block->preds.size();
			for (i = 0; i < len; i++)
			{
				mcnf_block *pred = block->preds[i];
				set_entry_exit_condition(pred, q, allocClass, Exit, recurse, isFull);
			}
		}
	}
	else
	{
		if(q < 0 && alloc_classes->getType(allocClass) == AllocationClassInfo::AntiOnly)
			;
		else
		{
			len = block->succs.size();
			for (i = 0; i < len; i++)
			{
				mcnf_block *succ = block->succs[i];
				set_entry_exit_condition(succ, q, allocClass, Entry, recurse, isFull);
			}
		}
	}
}

// just setup the entry constraints, useful for updating self loops
void HeuristicSolver::set_block_entry_boundaries(mcnf_block *block)
{
	int i, len;
	std::vector<mcnf_node*>& vecentry = block->entries;

	len = vecentry.size();

	for (i = 0; i < len; i++)
	{
		int j, nc;
		mcnf_node *node = vecentry[i];
		nc = node->commodities.size();
		for (j = 0; j < nc; j++)
		{
			if (node->feas_marked[j])
			{
				visited_blocks.clear();
				int q = node->commodities[j];
				set_entry_exit_condition(block, q, node->allocClass, Entry);
			}
		}
	}
}
/* Set up the entry/exit constraints for all livein/liveout
 variables in block.  Also links any liveout vars to their
 sinks.
 */
void HeuristicSolver::set_block_boundaries_and_sinks(mcnf_block *block)
{
	int i, len;

	set_block_entry_boundaries(block);
	std::vector<mcnf_node*>& vecexit = block->exits;
	len = vecexit.size();

	for (i = 0; i < len; i++)
	{
		int j, nc;
		mcnf_node *node = vecexit[i];
		nc = node->commodities.size();
		for (j = 0; j < nc; j++)
		{
			if (node->feas_marked[j])
			{
				mcnf_edge *sinkedge;
				int q = node->commodities[j];
				visited_blocks.clear();
				set_entry_exit_condition(block, q, node->allocClass, Exit);

				sinkedge = node->sinfo[1][j].prev_edge;

				if (sinkedge == NULL || sinkedge->dst->type != nSink)
					abort();
				sinkedge->mark_feasible_edge(q);
			}
		}
	}
}


/* Allocate an entire trace.
 Having already computed the shortest backwards paths, use this to
 simultaneously allocate with (intra-block only?) evictions.
 When a block boundary is fixed, fix all connected blocks
 and if any additional trace blocks are fixed, recompute
 shortest backwards paths and continue.
 */
void HeuristicSolver::allocate_trace(int whicht)
{
	int b, numb;
	std::vector<mcnf_node *> defs;
	const std::vector<mcnf_block*>& trace = traces.get_trace(whicht);
	numb = trace.size();

	current_trace = whicht;
	visited_blocks.clear();

	if(!TraceUpdateMin)
	{
		/* always recompute entire trace */
		update_trace_paths(whicht, NULL);
	}

	QuantityAlloc quantity_alloc; /* what class a quantity is currently alloced to */
	hash_set<int> live;

	/* now allocate going forwards */
	for (b = 0; b < numb; b++)
	{
		mcnf_block *block = trace[b];
		int i, n;
		int d, ndefs;
		int levelpos = 0;
		int numClasses = alloc_classes->numClasses();
		std::vector<ClassAlloc> class_alloc(numClasses); /* what quantity is in a given class */
		for(i = 0; i < numClasses; i++)
		{
			class_alloc[i].setSize(alloc_classes->getCapacity(i));
		}

		mcnf_node **prev_nodes = (mcnf_node**)alloca(numClasses
				*sizeof(mcnf_node*));
		mcnf_node **next_nodes = (mcnf_node**)alloca(numClasses
				*sizeof(mcnf_node*));
		mcnf_node **insn_nodes = (mcnf_node**)alloca(numClasses
				*sizeof(mcnf_node*));

		memset(prev_nodes, 0, sizeof(mcnf_node*)*numClasses);
		memset(next_nodes, 0, sizeof(mcnf_node*)*numClasses);
		memset(insn_nodes, 0, sizeof(mcnf_node*)*numClasses);

		//clear here, evictions may fix additional, less related, blocks
		fixed_blocks.clear();

		if(TraceUpdateMin)
		{
			//just update this block once
			fixed_blocks.insert(block->index);
			update_trace_paths(whicht, &fixed_blocks);
		}
		current_block_num = block->index;

		if(DebugPRA > 3)
			std::cerr << "Allocating block " << current_block_num << " at trace index " << b << "\n";
		/* TODO: pull this out, lighter-weight entry setup */
		defs.clear();
		live.clear();
		quantity_alloc.clear();

		/* scan through the block */
		/* start with entry nodes, intialize livein vars */
		n = block->entries.size();
		for (i = 0; i < n; i++)
		{
			mcnf_node *node = block->entries[i];
			if (prev_nodes[node->allocClass] != NULL)
			abort();
			prev_nodes[node->allocClass] = node;
		}

		/* setup all live in vars
		 */
		get_level_defs(prev_nodes, insn_nodes, defs);

		for (d = 0, ndefs = defs.size(); d < ndefs; d++)
		{
			int q = defs[d]->position;
			/* allocate q */
			allocate_quantity_at_level(defs[d], insn_nodes, prev_nodes,
					quantity_alloc, class_alloc);
			live.insert(q); /* unset in propagate_allocations */
		}

		if(block->is_self_loop)
		{
			//after setting entry, may have to set exit (then recalculate)
			set_block_entry_boundaries(block);
			trace_cost tc(this);
			trace_price tp(this);
			betterTiedEdge tiebreaker(*mcnf_cfg);
			block->traverse_nodes(reinitialize_node_for_fsf_trace);
			block->shortest_paths_backwards(tp, tc, tiebreaker);
		}

		while (!block->is_exit_level(prev_nodes) && levelpos >= 0)
		{
			/* assume prev_nodes is correct, compute next and insn nodes */
			levelpos = block->get_level(levelpos, next_nodes, insn_nodes);

			/* figure out the allocation for all live variables transitioning
			 from prev_nodes to next_nodes, updating current alloc info.
			 This handles shifts in allocation */
			propagate_allocations(prev_nodes, insn_nodes, next_nodes,
					quantity_alloc, class_alloc, live);

			/* find all the quantities that are defined going into next (not live in prev)
			 this routine might put the defs in a prefered allocation order
			 */
			get_level_defs(next_nodes, insn_nodes, defs);

			for (d = 0, ndefs = defs.size(); d < ndefs; d++)
			{
				int q = defs[d]->position;
				/* allocate q (possibly evicting other quantities and updates arrays)*/
				allocate_quantity_at_level(defs[d], insn_nodes, next_nodes,
						quantity_alloc, class_alloc);
				live.insert(q); /* unset in propagate_allocations */
			}

			/* shift nodes */
			memcpy(prev_nodes, next_nodes, sizeof(mcnf_node*) *numClasses);
		}

		/* fix connected boundary constraints as well as connecting sinks */
		set_block_boundaries_and_sinks(block);
		alloced_blocks.insert(block->index);

		if(TraceUpdateMin)
			continue; //only update a block at a time
		if(TraceUpdateAll)
		{
			for(int j = traces.num_traces() -1; j > whicht; j--)
			{
				update_trace_paths(j, &fixed_blocks);
			}
		}
		update_trace_paths(whicht, &fixed_blocks);
	}
}


/* Update maps so q is allocated to c.
 * May also choose a register if necessary for a multireg class.
 */
void HeuristicSolver::assign_quantity_to_class(int q, mcnf_node *node, std::vector<ClassAlloc>& class_alloc, QuantityAlloc& quantity_alloc, int offset /* = 0 */)
{
	if(DebugPRA > 9)
		std::cerr << "assignQ " << q << " " << node->getName() << "\n";
	int c = node->allocClass;
	if (quantity_alloc.get(q) >= 0)
		class_alloc[quantity_alloc.get(q)].remove(q);
	quantity_alloc.put(q, c);
	class_alloc[c].add(q);
}

//decompose the solution in block into live intervals for each quantity and put them in intervals
void HeuristicSolver::findLiveIntervals(mcnf_block *block, std::vector<AssignRegLiveInterval>& intervals)
{
	//go through quantity by quantity
	//for each def
	for(int d = 0, nd = block->defs.size(); d < nd; d++)
	{
		mcnf_node *def = block->defs[d];
		int q = def->commodities[0];
		if(q < 0)
			continue; //skip antis
		//follow the feasible path, recording a live interval everytime class alloc changes
		mcnf_node *node = def->fsinfo[0].next_edge->dst;
		mcnf_node *start = node;
		bool atStart = true;

		while(node != NULL)
		{
			int index = node->cindex.get(q);
			mcnf_node *next = node->fsinfo[index].next_edge->dst;

			if(next->type == nSink) //end of interval
			{
				intervals.push_back(AssignRegLiveInterval(q, start, node, atStart, true));
				break;
			}
			else if(next->allocClass != node->allocClass) //also end of interval
			{
				intervals.push_back(AssignRegLiveInterval(q, start, node, atStart, false));
				atStart = false;
				start = next; //begining of next interval
			}

			//keep going
			node = next;
		}
	}
}

//look through all the intervals conflict with interval and are live
//until the end of the block, if there is an outreg associated with that interval
//mark that register as conflicting
static void getForwardConflictingRegsForInterval(mcnf_block *block, const AssignRegLiveInterval &interval, std::vector<AssignRegLiveInterval>& intervals, hash_set<unsigned>& forwardConflicts)
{
	int end = interval.endpos;
	int start = interval.startpos;

	//assume intervals is sorted, starting from back look for conflicting intervals
	for(int i = intervals.size()-1; i >= 0; i--)
	{
		const AssignRegLiveInterval& conf = intervals[i];
		if(conf.startpos <= start)
			break; //too early
		if(conf.startpos > end)
			continue; //doesn't conflict with this interval
		if(conf.isExit) //goes to end of block, potentially conflicting
		{
			unsigned outreg = block->getRegisterForQuantityOut(conf.q);
			if(outreg != 0)
				forwardConflicts.insert(outreg);
		}
	}
}

/* return the memory node at pos
 * start looking at node
 */
static mcnf_node* getMemNodeAtLevelPos(mcnf_block *block, mcnf_node *node, int pos)
{
	int nm = block->allnodes.size();
	int subindex = getIntervalLevelSubindex(pos);
	//make sure we match subindex so we're at the right place
	for(int i = node->vecpos; i < nm; i++)
	{
		mcnf_node *n = block->allnodes[i];
		if(n->has_class() && alloc_classes->getType(n->allocClass) == AllocationClassInfo::Memory)
		{
			if(getNodeIntervalLevel(n) >= pos && n->subindex == subindex)
				return n;
		}
	}
	return NULL;
}

//generate a spill within interval that moves q from srcr at srcPos to dstr at dstPos
//dstpos is where teh register is loaded into dstr
//storePos is where the store to memory happens, transPos is where the quantity goes out
//of register (can be different)
//returns the cost of the spill code
int HeuristicSolver::generateSpill(const AssignRegLiveInterval& interval, int storePos, int transPos,
		unsigned srcr, int dstPos, unsigned dstr, std::vector<transformation>&  shuffleCode)
{
	int q = interval.q;
	int cost = 0;
	//first find and validate teh node at srcPos
	mcnf_block *block = mcnf_cfg->blocks[interval.start->block_num];
	mcnf_node *node = getMemNodeAtLevelPos(block, interval.start, storePos);
	assert(node != NULL);
	int srcc = alloc_classes->getRegisterIndex(srcr);

	if(transPos < storePos)
		abort();
	//now insert a store if necessary
	int aindex = node->cindex.get(-q);
	if(aindex < 0)
		abort(); //anti not live?
	//if anti is allocated to mem and not anti class, have to do a store
	if(node->feas_marked[aindex])
	{
		int pos = getIntervalLevelPosition(storePos);
		int sub = getIntervalLevelSubindex(storePos);
		int storeCost = alloc_classes->getMoveCost(node->allocClass, node->mirror_node->allocClass);
		shuffleCode.push_back(transformation(Store, pos, sub, storeCost, -1, 0,
												alloc_classes->getAntiClass(node->allocClass), 0, -q));
		cost += storeCost;
	}
	int pos = getIntervalLevelPosition(transPos);
	int sub = getIntervalLevelSubindex(transPos);
	//in any case, q no longer in reg after srcPos
	shuffleCode.push_back(transformation(Move, pos, sub, 0, srcc, srcr, node->allocClass, 0, q));
	block->setQuantityRegAtPos(q, 0, transPos+1);

	//now add a load at dstPos; we assume there is nothing inbetween transPos and dstPos
	//that would invalidate this (like a use in an insn)
	 pos = getIntervalLevelPosition(dstPos-1);
	 int loadCost = alloc_classes->getMoveCost(node->allocClass, srcc);
	 cost += loadCost;
	 if(!block->regAvailableForQuantityAtInterval(dstr, q, dstPos, dstPos, false))
		 abort();

	 //create the load
	 shuffleCode.push_back(transformation(Move, pos, 1, loadCost, node->allocClass, 0, srcc, dstr, q));
	 block->setQuantityRegAtPos(q, dstr, dstPos);

	 if(DebugPRA > 5)
	 {
		 std::cerr << "Generating spill " << storePos << ", " << transPos << ", " << dstPos << " from " << srcr << " to " << dstr << " with cost " << cost << "\n";
	 }
	 return cost;
}

static unsigned findBestAvailableRegAtPos(mcnf_block *block, int q, int c,
		bool outConstraint, bool isExit, int start, int end, int lastXBarPos, int& regEndPos)
{
	unsigned reg = 0;
	regEndPos = 0;
	//find the reg for q
	//first, if there's an out constraint see if we can now meet it
	unsigned outReg = 0;
	if(outConstraint)
	{
		unsigned outReg = block->getRegisterForQuantityOut(q);
		if(block->regAvailableForQuantityAtInterval(outReg, q, start, end, true))
		{
			reg = outReg;
			regEndPos = end;
		}
	}

	if(reg == 0)
	{
		const std::vector<unsigned>& regs = alloc_classes->getRegisters(c);
		for(int r = 0, nr = regs.size(); r < nr; r++)
		{
			unsigned tmpr = regs[r];
			int endPos = block->lastAvailablePosForQuantityInReg(q, tmpr, start, end, isExit);
			if(endPos> regEndPos)
			{
				reg = tmpr;
				regEndPos = endPos;
			}
		}

		if(regEndPos == end && outConstraint)
		{
			assert(outReg != reg);
			//we found a reg that is good all the way to the end,
			//but doesn't match the output constraint, so we still
			//have to breakup the interval
			regEndPos = lastXBarPos; //will move intout outreg next iteration
		}
		//if regEndPos is at an invalid spot, like in the middle of an insn,
		//backup if a move is needed
		if(regEndPos != end)
		{
			int pos = getIntervalLevelPosition(regEndPos+1);
			int sub = getIntervalLevelSubindex(regEndPos+1);
			if(getIntervalLevelType(regEndPos+1) == nInsn || sub == 0) //move spans from insn to xbar

			{
				//regEndPos+1 will be the dst of the next move
				pos--;
				regEndPos = compute_interval_level(pos, 1, true);
			}
		}
	}
	return reg;
}


//create a move within interval at position start between lastReg and reg
//return size of shuffle code created
int HeuristicSolver::generateMove(const AssignRegLiveInterval& interval, int start, unsigned lastReg, unsigned reg, std::vector<transformation>& shuffleCode)
{
	int cost = 0;
	//generate shuffle code for the move
	unsigned srcr = lastReg;
	unsigned dstr = reg;
	if(dstr == 0 || srcr == 0) //birth or death is free
		return 0;
	if(srcr == dstr)
		return 0;

	mcnf_block *block = mcnf_cfg->blocks[interval.start->block_num];
	int c = interval.start->allocClass;
	int pos = getIntervalLevelPosition(start);
	int sub = getIntervalLevelSubindex(start);
	int q = interval.q;
	if(getIntervalLevelType(start) == nInsn)
		abort(); //not sure how to handle this
	if(sub == 0) //move spans from insn to xbar, not sure how to handle it
		abort();
	//position is position of source of move
	sub--;
	if(sub > 0) sub = 1; //stupid gap

	//if the dst register is free in src position, can unconditionally
	//perform this move early (won't overwriter others)
	if(block->getQuantityForRegAtPos(dstr, start-1) == 0)
	{
		int movecost = alloc_classes->getSelfMoveCost(c);
		shuffleCode.push_back(transformation(Move, pos, sub, -1, movecost, c, srcr, c, dstr, q));
		cost += movecost;
	}
	//if the src register is free in the next position, can unconditionally
	//perform this move  last (nothing is going to overwrite srcr)
	else if(block->getQuantityForRegAtPos(srcr, start) == 0)
	{
		int movecost = alloc_classes->getSelfMoveCost(c);
		shuffleCode.push_back(transformation(Move, pos, sub, 1, movecost, c, srcr, c, dstr, q));
		cost += movecost;
	}
	else //problem move
	{
		int dq = block->getQuantityForRegAtPos(dstr, start-1);
		//if the quantity currently in the dst register is going to memory (or is unallocated),
		//we can savely do an early move since stores always happen first
		unsigned dqr = block->getRegisterForQuantityAtPos(dq, start);
		if(dqr == 0)
		{
			int movecost = alloc_classes->getSelfMoveCost(c);
			shuffleCode.push_back(transformation(Move, pos, sub, -1, movecost, c, srcr, c, dstr, q));
			cost += movecost;
		}
		else
		{
			//just spill this quantity (store and load to get it into the right register)
			//TODO: fancier move ordering
			cost += generateSpill(interval, start-1, start-1, srcr, start, dstr, shuffleCode);
		}
	}
	return cost;
}

//fix problem interval, return size of anything added shuffleCode
//updates intervals with any split intervals
int HeuristicSolver::handleProblemInterval(mcnf_block *block, const AssignRegLiveInterval& interval, OrderedIntervals& intervals, std::vector<transformation>& shuffleCode)
{
	int cost = 0; //cost of shuffle code
	int lastXBarPos = compute_interval_level(block->lastXBarPos, 1, true);

	//a problem interval cannot be allocated to a single register for
	//its entire length, so break it down to multiple registers in the same
	//class joined by a move
	int q = interval.q;
	int c = interval.start->allocClass;
	int start = interval.startpos;
	int end = interval.endpos;
	bool inConstraint = interval.isEntry;
	bool isExit = interval.isExit;
	bool outConstraint = interval.isExit && block->getRegisterForQuantityOut(q) != 0;
	unsigned lastReg = interval.lastReg;

	if(DebugPRA> 5)
	{
		std::cerr << "in block " << block->index << " fixing ProbInterval " << interval << "\n";
	}
	if(inConstraint)
	{
		//q has to be in the input reg
		lastReg = block->getRegisterForQuantityIn(q);
		assert(lastReg != 0);
		//extend past just the first bit
		block->setQuantityRegAtPos(q, lastReg, interval.startpos+offset_to_xbar_middle);
		start = block->lastAvailablePosForQuantityInReg(q, lastReg, start, end, isExit);
		//make sure start is a valid spot
		if(start < end)
		{
			//is the new starting position acceptable?
			int pos = getIntervalLevelPosition(start+1);
			int sub = getIntervalLevelSubindex(start+1);
			if(getIntervalLevelType(start+1) == nInsn || sub == 0) //move spans from insn to xbar

			{
				//start will be the dst of the next move
				pos--;
				start = compute_interval_level(pos, 1, true);
			}
		}
		else if(start >= end && outConstraint)
		{
			//we found a reg that is good all the way to the end,
			//but doesn't match the output constraint, so we still
			//have to breakup the interval
			start = lastXBarPos; //will move intout outreg next iteration
		}

		start++; //first new spot
		block->setQuantityRegAtPos(q, 0, start); //remainder of the interval
		assert(start < end);
		intervals.push(AssignRegLiveInterval(interval, start, lastReg));
	}
	else //not input constrained
	{
		unsigned reg = 0;
		int regEndPos = 0;

		reg = findBestAvailableRegAtPos(block, q, c, outConstraint, isExit, start, end, lastXBarPos, regEndPos);

		//reg and regEndPos should now be set
		//if reg isn't set, we probably backedup regEndPos out of an invalid
		//position and so the current start has no available reg (but it would be
		//invalid to switch inside of an insn)
		if(regEndPos < start)
		{
			//spill q to memory
			//find the very end of this interval (presumably a move invalid location)
			int endPos = block->lastAvailablePosForQuantityInReg(q, lastReg, start, end, isExit);
			if(endPos <= start)
				abort(); //problem isn't early endpos..
			if(endPos < 0)
				abort();
			int dstPos = endPos+1;

			//now find a free reg at dstPos
			unsigned newReg = 0;

			while(newReg == 0 && dstPos < end)
			{
				newReg = findBestAvailableRegAtPos(block, q, c, outConstraint, isExit, dstPos, end, lastXBarPos, regEndPos);
				dstPos++;
			}
			if(dstPos >= end || newReg == 0)
			abort();
			//spill, storing early and transitioning late
			cost += generateSpill(interval, start, endPos, lastReg, dstPos, newReg, shuffleCode);
			//q is now loaded into newReg at dstPos
			lastReg = newReg;
			start = regEndPos+1;
		}
		else
		{
			if(regEndPos < start || reg == 0)
				abort();
			if(!block->regAvailableForQuantityAtInterval(reg, q, start, regEndPos, false))
				abort();
			//set quantity to reg and generate move
			block->setQuantityRegAtPos(q, reg, start);

			if(lastReg == reg)
				abort(); //no need for move
			cost += generateMove(interval, start, lastReg, reg, shuffleCode);
			start = regEndPos+1;
			lastReg = reg;
		}

		//start and lastReg are now new starting point and current reg (lastreg of next interval)
		block->setQuantityRegAtPos(q, 0, start); //remainder of the interval
		if(start < end)
		{
			intervals.push(AssignRegLiveInterval(interval, start, lastReg));
		}
	}

	return cost;
}


//assign registers to all the live intervals
//if a register cannot be assigned (due to conflicts with hard regs or end conditions)
//then split the interval and color the remainder of the interval in order
//return cost of any shuffle code generated
int HeuristicSolver::colorLiveIntervals(mcnf_block *block, std::vector<AssignRegLiveInterval>& intervals, std::vector<transformation>& shuffleCode)
{
	int cost = 0;
	//allocate in sorted order
	OrderedIntervals ordered_intervals(intervals);

	while(!ordered_intervals.isEmpty())
	{
		const AssignRegLiveInterval interval = ordered_intervals.pop();
		int q = interval.q;
		int c = interval.start->allocClass;
		int spos = interval.startpos;
		int epos = interval.endpos;
		bool isExit = interval.isExit;

		if(DebugPRA > 5)
		{
			std::cerr << "in block " << block->index << " coloring Interval " << interval << " into class " << c << "\n";
		}
		//is this a register class?
		if(alloc_classes->getType(c) == AllocationClassInfo::Register)
		{
			if(interval.isEntry && interval.isExit &&
					 block->getRegisterForQuantityIn(q) != 0 &&
					 block->getRegisterForQuantityOut(q) != 0)
			{
				//interval spans whole block, make sure in matches out
				unsigned reg = block->getRegisterForQuantityIn(q);
				if(block->regAvailableForQuantityAtInterval(reg, q, spos, epos, isExit) &&
						reg == block->getRegisterForQuantityOut(q))
				{
					block->setQuantityRegAtPos(q, reg, spos);
				}
				else
				{
					//leave first bit of interval occupied
					block->setQuantityRegAtPos(q, 0, interval.startpos+offset_to_xbar_middle);
					cost += handleProblemInterval(block, interval, ordered_intervals, shuffleCode);
				}
			}
			else if(interval.isExit && block->getRegisterForQuantityOut(q) != 0)
			{
				//must match exit condition if one is set
				unsigned reg = block->getRegisterForQuantityOut(q);
				if(block->regAvailableForQuantityAtInterval(reg, q, spos, epos, isExit))
				{
					//just set to the proper outReg
					block->setQuantityRegAtPos(q, reg, spos);
					cost += generateMove(interval, spos, interval.lastReg, reg, shuffleCode);
				}
				else
				{
					//problematic, doesn't match end
					block->setQuantityRegAtPos(q, 0, spos);
					cost += handleProblemInterval(block, interval, ordered_intervals, shuffleCode);
				}
			}
			else if(interval.isEntry && block->getRegisterForQuantityIn(q) != 0)
			{
				//must match entry condition if one is set
				unsigned reg = block->getRegisterForQuantityIn(q);
				if(block->regAvailableForQuantityAtInterval(reg, q, spos, epos, isExit))
				{
					//just set to the proper reg
					block->setQuantityRegAtPos(q, reg, spos);
					block->setQuantityRegAtPos(q, 0, epos+1); //clear end
				}
				else
				{
					//problematic, can't allocate with correct start reg
					//leave first bit occupied
					block->setQuantityRegAtPos(q, 0, spos+offset_to_xbar_middle);
					cost += handleProblemInterval(block, interval, ordered_intervals, shuffleCode);
				}
			}
			else //pick an available register
			{
				const std::vector<unsigned>& regs = alloc_classes->getRegisters(c);
				hash_set<unsigned> forwardConflictingRegs;
				getForwardConflictingRegsForInterval(block, interval, intervals, forwardConflictingRegs);

				//look for an available reg
				int r, nr = regs.size();
				unsigned bestReg = 0;
				for(r = 0; r < nr; r++)
				{
					unsigned reg = regs[r];
					if(block->regAvailableForQuantityAtInterval(reg, q, spos, epos, isExit))
					{
						if(interval.isExit &&
								!block->regAvailableForQuantityAtOut(reg, q))
							continue; //can't use this reg
						//otherwise
						if(bestReg == 0)
						{ //pick first reg
							bestReg = reg;
						}
						else if(forwardConflictingRegs.count(bestReg) &&
								!forwardConflictingRegs.count(reg))
						{
							//discard previous pick if it will cause a problem
							//and this reg won't
							bestReg = reg;
						}

					}
				}
				if(bestReg != 0) //found one
				{
					block->setQuantityRegAtPos(q, bestReg, spos);
					//clear out interval end
					block->setQuantityRegAtPos(q, 0, epos+1);
					cost += generateMove(interval, spos, interval.lastReg, bestReg, shuffleCode);

				}
				else//could not find available reg
				{
					//make sure interval is clear
					block->setQuantityRegAtPos(q, 0, spos);
					cost += handleProblemInterval(block, interval, ordered_intervals, shuffleCode);
				}
			}
		}
		else //not register class
		{
			block->setQuantityReg(q, 0, interval.start);
		}
	}
	return cost;
}

//recursively explore entry/exits connected to block and mark reg as unusable
//for any vars in Live (which shrinking due to intersection)
//these are vars that woudl conflict with reg
void HeuristicSolver::setUnavailability(mcnf_node *node,
		hash_set<int>& marked, const std::vector<int>& Live, unsigned reg)
{
	mcnf_block *block = mcnf_cfg->blocks[node->block_num];
	int bindex = node->block_num;
	if(node->type == nEntry)
		bindex = -bindex;
	else if(node->type != nExit)
		abort();

	if(marked.count(bindex) > 0)
		return;
	marked.insert(bindex);

	//mark all the live commodities' availibility, but only on block exit
	if(node->type == nExit)
	{

		std::vector<int> live; //kind of pass by value, but just copy once
		live.reserve(Live.size());

		for(unsigned c = 0, nc = Live.size(); c < nc; c++)
		{
			int q = Live[c];
			int cindex = node->cindex.get(q);
			if(cindex >= 0)
			{
				//mark reg as unavailable for q
				block->setRegUnavailableForQuantityOut(q, reg);
				//add to live
				live.push_back(q);
			}
		}

		if(live.size() == 0)
			return;

		int len = node->succs.size();
		for (int i = 0; i < len; i++)
		{
			mcnf_node *succ = node->succs[i]->dst;
			if(succ->type == nEntry)
			{
				setUnavailability(succ, marked, live, reg);
			}
		}
	}
	else //entry, just recursively call
	{
		int len = node->preds.size();
		for (int i = 0; i < len; i++)
		{
			mcnf_node *pred = node->preds[i]->src;
			if(pred->type == nExit)
			{
				setUnavailability(pred, marked, Live, reg);
			}
		}
	}

}

//set in and out registers, as well as register unavailability
void HeuristicSolver::setEntryExitAllocations(mcnf_node *node, int q, unsigned reg)
{
	mcnf_block *block = mcnf_cfg->blocks[node->block_num];
	int bindex = block->index;
	if(node->type == nEntry)
		bindex = -bindex;

	if(visited_blocks.count(bindex))
		return;
	visited_blocks.insert(bindex);

	if(node->type == nExit)
	{
		//need to process this block?
		int index = node->cindex.get(q);
		if(index < 0) //not live
		{
			abort();
		}

		unsigned currout = block->getRegisterForQuantityOut(q);
		if(currout == reg)
			return;
		if(currout != 0)
			abort();

		//set this block's outreg
		block->setQuantityRegOut(q, reg);

		//now look at all the successors
		for(int s = 0, ns = node->succs.size(); s < ns; s++)
		{
			mcnf_node *succ = node->succs[s]->dst;
			if(succ->type == nEntry)
			{
				int sindex = succ->cindex.get(q);
				if(sindex < 0) //not live
				{
					//need to propagate unusability of this class for current live vars
					std::vector<int> live(succ->commodities.begin(), succ->commodities.end());
					hash_set<int> marked(visited_blocks);
					setUnavailability(succ, marked, live, reg);
				}
				else
				{
					setEntryExitAllocations(succ, q, reg);
				}
			}
		}

	}
	else //type == Entry
	{
		int index = node->cindex.get(q);
		if(index < 0) //not live
		{
			abort();
		}

		unsigned currin = block->getRegisterForQuantityIn(q);
		if(currin == reg)
			return;
		if(currin != 0)
			abort();

		//set block inreg
		block->setQuantityRegIn(q, reg);

		//now processes predecessors
		for(int p = 0, np = node->preds.size(); p < np; p++)
		{
			mcnf_node *pred = node->preds[p]->src;
			if(pred->type == nExit)
			{
				int pindex = pred->cindex.get(q);
				if(pindex < 0) //not live
				{
					//need to propagate unusability of this class for current live vars
					std::vector<int> live(pred->commodities.begin(), pred->commodities.end());
					hash_set<int> marked(visited_blocks);
					setUnavailability(pred, marked, live, reg);
				}
				else
				{
					setEntryExitAllocations(pred, q, reg);
				}
			}
		}

	}
}

//fix the register allocations of this block in all relevant block
//sets regOut
void HeuristicSolver::propagateExitAllocations(mcnf_block *block)
{
	if(DebugPRA > 5) std::cerr << "Propagating exit allocations for block " << block->index << "\n";
	//for everything that is allocated to a register class at block exit
	//recursive explore related blocks and set outreg
	for(int i = 0, n = block->exits.size(); i < n; i++)
	{
		mcnf_node *exit = block->exits[i];
		if(alloc_classes->getType(exit->allocClass) == AllocationClassInfo::Register)
		{
			for(int j = 0, nc = exit->commodities.size(); j < nc; j++)
			{
				if(exit->feas_marked[j])
				{
					int q = exit->commodities[j];
					//must have a register assigned
					unsigned reg = block->getRegisterForQuantity(q, exit);
					if(reg == 0)
						abort();
					if(block->getRegisterForQuantityOut(q) == 0) //not already set
					{
						//recursively set
						visited_blocks.clear();
						setEntryExitAllocations(exit, q, reg);
					}
				}
			}
		}
	}
}

/* Assign registers to classes and generate any necessary end of block shuffle code.
 * Return the overhead of the shuffle code (todo: also compute changes do to optimistic pricing?)
 */
int HeuristicSolver::assign_registers_to_classes(std::vector< std::vector<transformation> >& shufflecode)
{
	shufflecode.clear();
	shufflecode.resize(mcnf_cfg->blocks.size());
	if(!alloc_classes->hasMultiRegClasses())
		return 0; 	//do nothing

	int shuffleSize = 0;

	//clear reg mappings
	for(int b = 0, nb = mcnf_cfg->blocks.size(); b < nb; b++)
	{
		mcnf_block *block = mcnf_cfg->blocks[b];
		block->clearQuantityRegs();
	}

	//otherwise allocate each block, iterating over traces
	int numt = traces.num_traces();
	for(int t = 0; t < numt; t++)
	{
		const std::vector<mcnf_block*>& trace = traces.get_trace(t);
		int numb = trace.size();
		for(int b = 0; b < numb; b++)
		{
			mcnf_block *block = trace[b];
			int bindex = block->index;

			//first decompose solution into live intervals
			std::vector<AssignRegLiveInterval> intervals;
			findLiveIntervals(block, intervals);

			//than greedily color live intervals in sorted order
			//split problematic intervals and introduce shuffle and spill code as necessary
			shuffleSize += colorLiveIntervals(block, intervals, shufflecode[bindex]);

			propagateExitAllocations(block);
		}
	}
	return shuffleSize;
}


}
