#ifndef HEURISTICSOLVER_H_
#define HEURISTICSOLVER_H_


#include "llvm/PrincipledBackend/GlobalMCNF.h"
#include "llvm/PrincipledBackend/Solution.h"
#include <iostream>
#include <stdio.h>
#include "llvm/Function.h"
#include "llvm/Module.h"
#include "llvm/ADT/StringExtras.h"
#include "llvm/Analysis/LoopInfo.h"
#include "BlockTraces.h"
/*
 *  A heuristic solver which computes a valid solution to the global MCNF
 *  problem using a simultaneous allocator on traces.
 */
namespace llvm
{

class QuantityAlloc;
class ClassAlloc;


/* an interval in a block where q is in the same allocation class */
struct AssignRegLiveInterval
{
	int q;
	int startpos; //level num of start
	int endpos;// and end
	mcnf_node *start; //first node with alloc class
	mcnf_node *end; //last node of live interval (inclusive)
	unsigned lastReg; //for a split interval, previous allocation
	bool first; //true if this is the first interval for this quantity
	bool last; //true if this is the last interval for this quantity
	bool isEntry; //true if live into block
	bool isExit; //true if live out of block
	bool operator<(const AssignRegLiveInterval& rhs) const
	{
		return startpos < rhs.startpos;
	}

	bool operator==(const AssignRegLiveInterval& rhs) const
	{
		return startpos == rhs.startpos;
	}

	bool operator<=(const AssignRegLiveInterval& rhs) const
	{
		return startpos <= rhs.startpos;
	}

	AssignRegLiveInterval(int quant, mcnf_node *s, mcnf_node *e, bool isFirst, bool isLast):
		q(quant), start(s), end(e), lastReg(0), first(isFirst), last(isLast), isEntry(false), isExit(false)
	{
		assert(start->has_class() && end->has_class());
		startpos = getNodeIntervalLevel(start);
		if(start->type == nInsn && isFirst) //definition, just live out
			startpos++;
		endpos = getNodeIntervalLevel(end);
		if(start->type == nEntry)
			isEntry = true;
		if(end->type == nExit)
			isExit = true;
	}

	/* create a new split interval from an old interval (a subinterval with a laster starting point) */
	AssignRegLiveInterval(const AssignRegLiveInterval& oldi, int newstart, unsigned lastr)
	{
		*this = oldi;
		assert(newstart > startpos);
		startpos = newstart;
		lastReg = lastr;
		first = false;
		isEntry = false;
	}
};

std::ostream &operator<<(std::ostream &OS, const AssignRegLiveInterval &interval);


//keep track of intervals in order, allow split intervals to be added
class OrderedIntervals
{
	//keep original intervals in fixed sorted order
	std::vector<AssignRegLiveInterval>& orig;
	//separately maintain additional split intervals
	std::multiset<AssignRegLiveInterval> splits;

	unsigned opos;
public:
	OrderedIntervals(std::vector<AssignRegLiveInterval>& intervals): orig(intervals), opos(0)
	{
		std::sort(intervals.begin(), intervals.end());
	}

	bool isEmpty()
	{
		if(opos == orig.size() && splits.empty())
			return true;
		return false;
	}

	//return lowest starting interval
	//just do copy return for simplicity now
	AssignRegLiveInterval pop()
	{
		if(DebugPRA > 10)
		{
			std::cerr << "popping ordered with opos " << opos << " split size " << splits.size() << "\n";
		}
		if(opos < orig.size() && (splits.empty() || orig[opos] <= *(splits.begin())))
		{
			return orig[opos++];
		}
		else if(splits.size() > 0)
		{
			AssignRegLiveInterval ret = *splits.begin();
			splits.erase(splits.begin());
			return ret;
		}
		else
			abort();
	}

	/* add a new interval (a split orig interval) */
	void push(const AssignRegLiveInterval& split)
	{
		if(DebugPRA > 5)
		{
			std::cerr << "Adding split interval " << split << " with lastreg " << split.lastReg << "\n";
		}
		splits.insert(split);
	}
};


class HeuristicSolver
{
	GlobalMCNF *mcnf_cfg;
	LoopInfo *LI;
	BlockTraces traces;

	//a couple of "global" variables for anti eviction
	hash_set<mcnf_node*> evict_anti_memoize;
	mcnf_node *evict_anti_terminate;

	bool initial_heuristic; //true for first run only
	int current_trace;
	int current_block_num;
	hash_set<int> alloced_blocks; //set if allocated
	hash_set<int> visited_blocks; //by block index, entry and exit -/+
	hash_set<int> fixed_blocks;

	//allocate the ith trace
	void allocate_trace(int i);
	mcnf_edge *next_edge_for_allocation(mcnf_node *node, int q);
	void sanity_check_allocs(QuantityAlloc& quantity_alloc,
			std::vector<ClassAlloc>& class_alloc, mcnf_node **nodes);
	mcnf_edge* exit_edge_to_sink(mcnf_node *node, int q);
	int mark_new_eviction_path(int q, mcnf_edge *eedge, mcnf_node *node);
	bool safe_to_change_helper(mcnf_node *node, int q);
	bool safe_to_change_exit_alloc(mcnf_node *exit, mcnf_node *entry, int q);
	bool can_evict_to_node(mcnf_node *node, int q);
	bool node_blocked_by_insn(mcnf_node *node, int q);
	bool node_blocked_by_exit_allocations(mcnf_node *node, int q,  std::vector<ClassAlloc>& class_alloc);
	bool node_not_valid_for_allocation(mcnf_node *node, int q, int ignoreantis, std::vector<ClassAlloc>& class_alloc);
	int feasible_forward_path_unpriced_cost(int q, mcnf_node *node);
	mcnf_edge* cheapest_eviction_edge(mcnf_node *evictto, int q,
			double *cost, int *unpriced_cost, bool block_only);
	int feasible_node_flow(mcnf_node *node);
	bool break_eviction_tie(int q, mcnf_node * oldn, mcnf_node* newn,
			std::vector<double>& evictcosts,
			std::vector<int>& up_costs,
			std::vector<ClassAlloc>& class_alloc);
	double eviction_cost(int q, int allocClass,
			mcnf_node **next_nodes, mcnf_edge** eedge, int *ec,
			int *unpriced_cost, std::vector<ClassAlloc>& class_alloc);
	int select_alloc_class(int q, mcnf_node **insn_nodes, mcnf_node **next_nodes,
			std::vector<ClassAlloc>& class_alloc, mcnf_edge **eedge, int *eq, int *ec,
			int *evictanti, std::vector<double>& pricesto);
	void fill_def_pricesto(mcnf_node *def, std::vector<double>& pricesto,
			std::vector<mcnf_edge*>& edges, int mirroralloc);
	void change_anti_alloc(mcnf_edge *eedge, int q);
	bool evict_antivar(int q, mcnf_node *node, mcnf_edge *eedge);
	bool evict_antivar_helper(int q, mcnf_node *node, mcnf_node *stop);
	void global_anti_evict(int q, mcnf_block *block, int cost, bool curr_ends_in_mem);
	void evict_quantity(int q, mcnf_edge *eedge, int ec, mcnf_node **next_nodes,
			QuantityAlloc& quantity_alloc, std::vector<ClassAlloc>& class_alloc);
	void set_entry_exit_usability(mcnf_node *node,
			hash_set<int>& marked, const std::vector<int>& Live, bool unusable);
	void unset_entry_exit_condition(mcnf_block *block, int q, enum entry_exit type,
			bool recurse = true, int allocClass = -1);
	void set_entry_exit_condition(mcnf_block *block, int q, int allocClass,
			enum entry_exit type, bool recurse = true, bool prevFull = false);
	void set_quantities_used_in_class(mcnf_block *block, int allocClass, entry_exit type, hash_set<int>& marked, int q);
	void unset_quantities_used_in_class(mcnf_block *block, int allocClass,
			entry_exit type, hash_set<int>& marked, int q);
	void set_multireg_usability(mcnf_node *node, hash_set<int>& marked);
	void set_block_entry_boundaries(mcnf_block *block);
	void set_block_boundaries_and_sinks(mcnf_block *block);
	bool paths_fail_entry_exit_loop_constraints(mcnf_block *block);
	void get_level_defs(mcnf_node **level, mcnf_node **ilevel,
			std::vector<mcnf_node*>& def);
	bool update_block_entry_prices(mcnf_block *block);
	void update_trace_paths(int currtrace, hash_set<int> *marked);
	void allocate_quantity_at_level(mcnf_node* def,
			mcnf_node **insn_nodes, mcnf_node **next_nodes,
			QuantityAlloc&  quantity_alloc,
			std::vector<ClassAlloc>& class_alloc);
	void propagate_allocations(mcnf_node **prev_nodes, mcnf_node **insn_nodes,
		mcnf_node **next_nodes, QuantityAlloc& quantity_alloc,
		std::vector<ClassAlloc>& class_alloc, hash_set<int>& live);

	bool quantity_is_globally_evictable(int q);
	void globally_evict_to_class(int q, int allocClass);

	void assign_quantity_to_class(int q, mcnf_node *node, std::vector<ClassAlloc>& class_alloc, QuantityAlloc& quantity_alloc, int offset=0);

	void findLiveIntervals(mcnf_block *block, std::vector<AssignRegLiveInterval>& intervals);
	int colorLiveIntervals(mcnf_block *block, std::vector<AssignRegLiveInterval>& intervals, std::vector<transformation>& shuffleCode);
	int handleProblemInterval(mcnf_block *block, const AssignRegLiveInterval& interval, OrderedIntervals& intervals, std::vector<transformation>& shuffleCode);
	void setUnavailability(mcnf_node *node,
			hash_set<int>& marked, const std::vector<int>& Live, unsigned reg);
	void setEntryExitAllocations(mcnf_node *node, int q, unsigned reg);
	void propagateExitAllocations(mcnf_block *block);
	int generateSpill(const AssignRegLiveInterval& interval, int storePos, int transPos, unsigned srcr, int dstPos, unsigned dstr, std::vector<transformation>&  shuffleCode);
	int generateMove(const AssignRegLiveInterval& interval, int start, unsigned lastReg, unsigned reg, std::vector<transformation>& shuffleCode);
public:
	HeuristicSolver(GlobalMCNF *mcfg, LoopInfo *li) :
		mcnf_cfg(mcfg), LI(li), traces(mcfg, li)
	{
		if(DebugPRA)
			std::cerr << "NumTraces = " << traces.num_traces() << "\n";
	}

	void changeModel(GlobalMCNF *newcfg)
	{
		mcnf_cfg = newcfg;
		traces.changeModel(newcfg);
	}

	//simply heuristic initialization of prices
	void initialize_dual_prices();

	//find a feasible solution using cost and price info,
	//return unpriced cost
	int construct_feasible_solution(bool initial = false);
	//assign registers, only necessary for multireg classes
	int assign_registers_to_classes(std::vector< std::vector<transformation> >& shufflecode);
	friend class liveq_compare;
	friend class trace_cost;
	friend class trace_price;
};

// comparator object for quantities
class liveq_compare
{
	HeuristicSolver *solver;
	mcnf_node **current_prev_nodes;
	QuantityAlloc& current_quantity_alloc;

public:
	liveq_compare(HeuristicSolver *s, mcnf_node **pn, QuantityAlloc& qa):
		solver(s), current_prev_nodes(pn), current_quantity_alloc(qa)
		{

		}

	bool operator()(int lhs, int rhs);

};

}
#endif /*HEURISTICSOLVER_H_*/
