#include "LagrangianSolver.h"
#include "llvm/ADT/Statistic.h"
#include "llvm/ADT/STLExtras.h"
#include "llvm/Support/Debug.h"
#include "llvm/Support/Compiler.h"
#include "llvm/PrincipledBackend/ProgCommon.h"
#include "llvm/Support/CommandLine.h"
#include <sys/time.h>
#include <time.h>

namespace llvm
{

/* check the shortest path info in sinfo for a node
 * look at both predecessor and successors for every commodity */
void check_shortest(mcnf_node *node)
{
	for(unsigned c = 0, nc = node->commodities.size(); c < nc; c++)
	{
		int q = node->commodities[c];

		double min = UNREACHABLE;
		if(node->type == nDef)
			min = 0;
		//for each predecessor
		for(unsigned p = 0, np = node->preds.size(); p < np; p++)
		{
			mcnf_edge *pred = node->preds[p];
			if(pred->commodity_valid(q))
			{
				int pindex = pred->src->cindex.get(q);
				double dist = pred->src->sinfo[0][pindex].value + pred->getPrice(q);
				if(dist < min)
				{
					min = dist;
				}
			}
		}

		//check shortest distance
		if(node->sinfo[0][c].value != min)
		{
			std::cerr << "WRONG SHORTEST " << node->sinfo[0][c].value << "\t" << min << "\n";
			//abort();
		}
	}
}

// Return the largest cost block multiplier in the model
int LagrangianSolver::max_cost_block_multiplier()
{
	int max = 1;
	for(int b = 0, nb = mcnf_cfg->blocks.size(); b < nb; b++)
	{
		mcnf_block *block = mcnf_cfg->blocks[b];
		int mult = mcnf_cfg->cost_block_multiplier(block);
		if(mult > max)
			max = mult;
	}
	if(max > 1)
		max /= 10;
	if(DebugPRA > 8)
	{
		std::cerr << "max block mult " << max << "\n";
	}
	return max;
}

cl::opt<bool> flag_prog_ra_max_iter("prog-ra-max-iter", cl::init(false), cl::Hidden,
		cl::desc("Keep computing lagrangian bound until maximum iterations are reached"));

int LagrangianSolver::solve(int startIter, int endIter, bool pricesOnly /*=false*/)
{
	double value;
	int k;
	double best_lagrangian = -HUGE_VAL;
	double prev_lagrangian = -HUGE_VAL;
	int best_solution = INT_MAX;
	int add = 10;
	double mult = max_cost_block_multiplier();
    struct timeval start, end;
    std::vector< std::vector<transformation> > shufflecode;

	mcnf_step = 0.05;

	if(!pricesOnly)
	{
		best_solution = solver.construct_feasible_solution(true);
		best_solution += solver.assign_registers_to_classes(shufflecode);
		solution.record_solution(*mcnf_cfg, shufflecode);
	}

    if(DebugPRA) gettimeofday(&start,NULL);

	for (k = startIter; k <= endIter; k++)
	{
		int solution_value = 0;
		int changed = 0;

		current_iteration = k;

		/* find the value of all the shortest paths
		 ignore bundle and boundary constraints */
		value = mcnf_cfg->shortest_balanced_paths();
		//mcnf_cfg->traverse_nodes(check_shortest);

		if (DebugPRA)
			fprintf(stderr, "Shortest value = %f\n", value);

		/* subtract the bundleconstraints multiplied by the price */
		double p_of_c = mcnf_cfg->price_of_constraints();
		value -= p_of_c;
		if(DebugPRA)
			fprintf(stderr, "Price of constraints = %f\n", p_of_c);
		if (value > best_lagrangian)
			best_lagrangian = value;

		/* update step size */
		mcnf_step = mult/(double)(k+add);

		/*update prices */
		changed = update_prices(norm);

		if (!pricesOnly)
		{
			/* find a feasible solution, using prices as a guide */
			solution_value = solver.construct_feasible_solution();
			solution_value += solver.assign_registers_to_classes(shufflecode);
			//	mcnf_cfg->sanity_check_feasible_solution();

			if (solution_value < best_solution)
			{
				best_solution = solution_value;
				solution.record_solution(*mcnf_cfg, shufflecode);
			}
		}
		if (!finite(value))
			abort();

		prev_lagrangian = value;

		if (DebugPRA)
		{
			fprintf(stderr, "****ITERATION %d\tSTEP %.10f *******\n", k,
					mcnf_step);
			fprintf(stderr, "L(w) = %.20f\n", value);
			fprintf(stderr, "Lower bound = %.20f\n", best_lagrangian);
			fprintf(stderr, "Norm = %.20f\n", norm);
			if(!pricesOnly)
			{
				fprintf(stderr, "Feasible solution = %d\n", solution_value);
				fprintf(stderr, "Best solution = %d\n", best_solution);
			}
		}

		if (!pricesOnly && !flag_prog_ra_max_iter && best_lagrangian > (double)(best_solution - 1.0))
			break;

		if (!changed)
			break;
	}
    if(DebugPRA)
    {
        gettimeofday(&end,NULL);
        if(k < endIter)
        	k++;

        double time = (double)(end.tv_sec - start.tv_sec) + (end.tv_usec-start.tv_usec)/1000000.0;
        std::cerr << "TIME = " << time << " per iteration " << time/(k-startIter) << "\n";
    }
	return best_solution;
}


/* Update the prices in the mcnf_cfg.
 Assumes the old norm is in norm_ret and places the new norm into it.
 w_new = [w + step * over/underflow]*[adjusted theta]
 returns zero if nothing is changed
 */
bool LagrangianSolver::update_prices(double& norm_ret)
{
	int b, bn;
	double diff, price, oldprice, sum;
	int truncgrad = 1;
	double step = mcnf_step;
	double psi = sqrt(compute_grad_norm(truncgrad))/sqrt(norm_ret);
	double norm = 0.0;
	int changed = 0;

	if (psi < 0)
		psi = 0;

	/*   psi = .5; */

	if (DebugPRA)
	{
		fprintf(stderr, "psi = %f\n", psi);
	}

	bn = mcnf_cfg->blocks.size();

	for (b = 0; b < bn; b++)
	{
		int n;
		mcnf_block *block = mcnf_cfg->blocks[b];
		int nn = block->allnodes.size();
		for (n = 0; n < nn; n++)
		{
			mcnf_node *node = block->allnodes[n];

			if (node->capacity > 0)
			{
				sum = node->node_flow();

				diff = sum - node->capacity;

				/* weight diff by the previous gradient direction */
				diff += psi*node->gradient;

				if (truncgrad && diff < 0)
				{
					if (node->price == 0)
						diff = 0;
					else if (node->price < -diff*step)
						diff = -node->price/(step)-THRESHOLD;
				}

				norm += diff*diff;

				oldprice = node->price;
				price = node->price + diff*step;
				if (price < 0)
					price = 0;

				node->price = price;
				node->gradient = diff;

				if (fabs(price - oldprice) > THRESHOLD)
					changed = true;
			}

			/* handle dual constraints */
			if (node->dual_prices != NULL)
			{
				int i, nc;

				if (node->capacity >= 0)
					abort();

				nc = node->commodities.size();
				/* for each positive commodity */
				for (i = 0; i < nc; i++)
				{
					int c = node->commodities[i];
					if (c > 0 && node->dual_constraint_matters(c))
					{
						/* get the index for the corresponding anti commodity */
						int ai = node->cindex.get(-c);
						if (ai >= 0)
						{
							sum = node->binfo[i].ratio + node->binfo[ai].ratio;
							diff = sum - 1.0;

							if (truncgrad && diff < 0)
							{
								if (node->dual_prices[i] == 0)
									diff = 0;
								else if (node->dual_prices[i] < -diff*step)
									diff = -node->dual_prices[i]/(step)
											- THRESHOLD;
							}

							diff += psi*node->dual_gradients[i];

							norm += diff*diff;

							oldprice = node->dual_prices[i];
							price = oldprice + diff*step;

							if (price < 0)
								price = 0;

							node->dual_prices[i] = price;
							node->dual_gradients[i] = diff;

							if (fabs(price - oldprice) > THRESHOLD)
								changed = true;
						}
					}
				}
			}

			if (node->boundary_prices != NULL)
			{
				if (node->type == nEntry)
				{
					int i, nc;

					nc = node->commodities.size();
					/* for each commodity */
					for (i = 0; i < nc; i++)
					{
						int p, np;
						int c = node->commodities[i];
						/* look at all the predecessors and enforce
						 the equality constraint with the price */
						np = node->preds.size();
						for (p = 0; p < np; p++)
						{
							mcnf_node *pred = node->preds[p]->src;
							int pindex =
									pred->cindex.get(c);
							if (pindex >= 0 && pred->type == nExit && pred->allocClass == node->allocClass)
							{
								/* always put this node first, that way, if there's more
								 flow than the pred, the price will increase */
								oldprice = node->boundary_prices[p][i];

								diff = node->binfo[i].ratio
								- pred->binfo[pindex].ratio;

								diff += psi*node->boundary_gradients[p][i];

								price = oldprice + diff*step;
								//support crossblock antis
								if(CrossBlockAntis && c < 0)
								{
									//since these are <= constraints, truncate price
									//however, still the negative price on other node,
									//since the lagrangian price should multiply both
									//variables in the constraint (and the other has -1
									//as a coefficent)
									//exit nodes get positive prices on mem nodes,
									//entry nodes get positive prices on antis
									if(alloc_classes->getType(node->allocClass) == AllocationClassInfo::AntiOnly)
									{
										if(price < 0)
											price = 0;
									}
									else if(price > 0) //inverse, mem node
										price = 0;
								}
								norm += diff*diff;

								node->boundary_gradients[p][i] = diff;
								node->boundary_prices[p][i] = price;
								assert(finite(price));
								if (fabs(price - oldprice) > THRESHOLD)
									changed = true;
							}
						}
					}
				}
				else if (node->type == nExit)
				{
					int i, nc;

					nc = node->commodities.size();
					/* for each commodity */
					for (i = 0; i < nc; i++)
					{
						int s, ns;
						int c = node->commodities[i];
						/* look at all the successors and enforce
						 the equality constraint with the price */
						ns = node->succs.size();
						for (s = 0; s < ns; s++)
						{
							mcnf_node *succ = node->succs[s]->dst;
							int sindex =
									succ->cindex.get(c);
							if (sindex >= 0 && succ->type == nEntry && succ->allocClass == node->allocClass)
							{
								oldprice = node->boundary_prices[s][i];

								diff = node->binfo[i].ratio
								- succ->binfo[sindex].ratio;

								diff += psi*node->boundary_gradients[s][i];

								price = oldprice + diff*step;
								//support crossblock antis
								if(CrossBlockAntis && c < 0)
								{
									//since these are <= constraints, truncate price
									//however, still the negative price on other node,
									//since the lagrangian price should multiply both
									//variables in the constraint (and the other has -1
									//as a coefficent)
									//exit nodes get positive prices on mem nodes,
									//entry nodes get positive prices on antis
									if(alloc_classes->getType(node->allocClass) == AllocationClassInfo::Memory)
									{
										if(price < 0)
											price = 0;
									}
									else if(price > 0) //inverse, antinode
										price = 0;

								}
								norm += diff*diff;

								node->boundary_gradients[s][i] = diff;
								node->boundary_prices[s][i] = price;
								assert(finite(price));
								if (fabs(price - oldprice) > THRESHOLD)
									changed = true;
							}
						}
					}
				}
			}
		}
	}

	norm_ret = norm;

	return changed;

}

/* Compute the norm of the gradient using the current
 prices and flow information
 If truncgrad is true, truncate gradients that
 would put the price below zero
 */
double LagrangianSolver::compute_grad_norm(int truncgrad)
{
	int b, bn;
	double norm = 0;
	double step = mcnf_step;
	bn = mcnf_cfg->blocks.size();

	for (b = 0; b < bn; b++)
	{
		int n;
		mcnf_block *block = mcnf_cfg->blocks[b];
		int nn = block->allnodes.size();

		for (n = 0; n < nn; n++)
		{
			mcnf_node *node = block->allnodes[n];

			if (node->capacity > 0)
			{
				double diff;
				double sum = node->node_flow();

				diff = sum - node->capacity;

				if (truncgrad && diff < 0)
				{
					double price;
					price = node->price;
					if (price == 0)
						diff = 0;
					else if (price < -diff*step)
						diff = -price/step;
				}

				norm += diff*diff;
			}
			/* handle dual constraints */
			if (node->dual_prices != NULL)
			{
				int i, nc;

				if (node->capacity >= 0)
					abort();

				nc = node->commodities.size();
				/* for each positive commodity */
				for (i = 0; i < nc; i++)
				{
					int c = node->commodities[i];
					if (c > 0)
					{
						/* get the index for the corresponding anti commodity */
						int ai = node->cindex.get(-c);
						if (ai >= 0)
						{
							double sum = node->binfo[i].ratio
									+ node->binfo[ai].ratio;
							double diff = sum - 1.0;

							if (truncgrad && diff < 0)
							{
								double price;
								price = node->dual_prices[i];
								if (price == 0)
									diff = 0;
								else if (price < -diff*step)
									diff = -price/step;
							}

							norm += diff*diff;
						}
					}
				}

			}

			if (node->boundary_prices != NULL)
			{
				double diff;
				if (node->type == nEntry)
				{
					int i, nc;

					nc = node->commodities.size();
					/* for each commodity */
					for (i = 0; i < nc; i++)
					{
						int p, np;
						int c = node->commodities[i];
						/* look at all the predecessors and enforce
						 the equality constraint with the price */
						np = node->preds.size();
						for (p = 0; p < np; p++)
						{
							mcnf_node *pred = node->preds[p]->src;
							int pindex =
									pred->cindex.get(c);
							if (pindex >= 0 && pred->type == nExit)
							{
								/* always put this node first, that way, if there's more
								 flow than the pred, the price will increase */
								diff = node->binfo[i].ratio
										- pred->binfo[pindex].ratio;
								norm += diff*diff;

							}
						}
					}
				}
				else if (node->type == nExit)
				{
					int i, nc;

					nc = node->commodities.size();
					/* for each commodity */
					for (i = 0; i < nc; i++)
					{
						int s, ns;
						int c = node->commodities[i];
						/* look at all the successors and enforce
						 the equality constraint with the price */
						ns = node->succs.size();
						for (s = 0; s < ns; s++)
						{
							mcnf_node *succ = node->succs[s]->dst;
							int sindex =
									succ->cindex.get(c);
							if (sindex >= 0 && succ->type == nEntry)
							{
								diff = node->binfo[i].ratio
										- succ->binfo[sindex].ratio;
								norm += diff*diff;

							}
						}
					}
				}
			}

		}
	}

	if (norm == 0)
		norm = 1.0;
	return norm;
}

}
