/* ProgressiveAllocator.cpp
 * Defines a progressive register allocator pass.
 */
#include <ostream>
#include "llvm/Analysis/LoopInfo.h"
#include "llvm/CodeGen/LiveVariables.h"
#include "llvm/CodeGen/LiveIntervalAnalysis.h"
#include "llvm/CodeGen/SimpleRegisterCoalescing.h"
#include "llvm/Function.h"
#include "llvm/CodeGen/MachineFunctionPass.h"
#include "llvm/CodeGen/MachineInstr.h"
#include "llvm/CodeGen/MachineJumpTableInfo.h"
#include "llvm/CodeGen/Passes.h"
#include "llvm/CodeGen/RegAllocRegistry.h"
#include "llvm/CodeGen/RegisterCoalescer.h"
#include "llvm/CodeGen/SSARegMap.h"
#include "llvm/CodeGen/SimpleRegisterCoalescing.h"
#include "llvm/CodeGen/MachineFrameInfo.h"
#include "llvm/CodeGen/SimpleDCE.h"
#include "llvm/CodeGen/DataFlowLiveVariables.h"
#include "llvm/Target/MRegisterInfo.h"
#include "llvm/Target/TargetInstrInfo.h"
#include "llvm/Target/TargetMachine.h"
#include "llvm/ADT/EquivalenceClasses.h"
#include "llvm/ADT/Statistic.h"
#include "llvm/ADT/STLExtras.h"
#include "llvm/ADT/SmallSet.h"
#include "llvm/Support/Debug.h"
#include "llvm/Support/Compiler.h"
#include "llvm/Support/CommandLine.h"
#include "llvm/PrincipledBackend/GlobalMCNF.h"
#include "llvm/PrincipledBackend/Solution.h"
#include "llvm/PrincipledBackend/AllocClassMap.h"

#include "HeuristicSolver.h"
#include "LagrangianSolver.h"
#include "ColumnGenSolver.h"
#include <algorithm>
#include <set>
#include <queue>
#include <memory>
#include <cmath>

#ifdef FPU_CONTROL_AVAILABLE
# include <fpu_control.h>
#endif

using namespace llvm;

namespace llvm
{
enum SolverToUse
{
	Lagrangian, CPLEX, ColumnGen, CPLEX2Phase
};
static cl::opt<SolverToUse> WhatSolver(cl::Optional, cl::desc(
		"What solver to use for the progressive allocator."), cl::init(
		Lagrangian), cl::values(clEnumValN(Lagrangian, "lagrangian-solve", "Lagrangian solver (default)"),
		clEnumValN(CPLEX,  "cplex-solve",   "CPLEX solver"),clEnumValN (ColumnGen,"col-gen-solve", "Column generation (uses CPLEX)"),
		clEnumValN(CPLEX2Phase,  "cplex-2phase-solve",   "Use cplex to solve the spilling problem (simple model) then a constrained full model"),
		clEnumValEnd
		)
		);
cl::opt<bool> InitCPLEXSolution("init-cplex-sol", cl::Hidden,
		cl::desc("Provide cplex with the starting initial solution"));
cl::opt<bool> OutputCPLEX("output-cplex", cl::Hidden,
		cl::desc("Output global MCNF model to CPLEX"));
cl::opt<bool> UseSimpleModel("simple-model", cl::Hidden,
		cl::desc("Use a simplified model of the register allocation problem"));
cl::opt<bool> SimpleOnly("simple-model-only", cl::Hidden,
		cl::desc("Only use the simplified model of the register allocation problem"));


#ifndef NDEBUG
cl::opt<bool> OutputMST("output-mst", cl::Hidden,
		cl::desc("Output mst file of found solution"));
cl::opt<bool> BuildAndQuit("build-and-quit", cl::Hidden,
		 cl::desc("Build the model for the first function and then quit"));
cl::opt<bool> DumpCFG("dumpcfg-pra", cl::Hidden,
		 cl::desc("Write out the cfg prior to progressive allocation"));
#else
const bool OutputMST = 0;
const bool BuildAndQuit = 0;
const bool DumpCFG = 0;
#endif

cl::opt<int> flag_prog_ra_iter("prog-ra-iter", cl::init(-1), cl::Hidden,
		cl::desc("Number of iterations of lagrangian relaxation"));
cl::opt<int> flag_prog_ra_warmup_iter("prog-ra-warmup-iter", cl::init(-1), cl::Hidden,
		cl::desc("Number of iterations to spend computing just prices for lagrangian relaxation"));

cl::list<std::string >
InterestingFunctions("focus-on-functions", cl::desc("Focus on provided functions"),
          cl::Hidden, cl::value_desc("focus functions"), cl::CommaSeparated);

static bool isInterestingFunction(const std::string& str)
{
	if(InterestingFunctions.size() == 0)
		return true; //should not focus on any particular funtions
	for (unsigned i = 0; i != InterestingFunctions.size(); ++i)
	{
		const std::string& ifunc =  InterestingFunctions[i];
		if(str == ifunc)
			return true;
	}
	return false;
}
static RegisterRegAlloc progresiveRegAlloc("progressive",
		"  progressive register allocator", createProgressiveRegisterAllocator);
const AllocationClassInfo *alloc_classes;

class ProgressiveAllocator : public llvm::MachineFunctionPass
{
public:
	static char ID;
	ProgressiveAllocator() :
		MachineFunctionPass((intptr_t)&ID)
	{
	}

	virtual ~ProgressiveAllocator();

	virtual const char* getPassName() const
	{
		return "Progressive Register Allocator";
	}

	virtual void getAnalysisUsage(AnalysisUsage &AU) const
	{
		//some of these will no longer be necessary once we integrate
		//insn selection w/reg alloc
		AU.addRequired<SimpleRegisterCoalescing>();
		//register coalescing pulls in live intervals which requires these two
		AU.addPreservedID(PHIEliminationID);
		AU.addPreservedID(TwoAddressInstructionPassID);
		AU.addRequiredID(BranchFoldingID);
		AU.addRequired<DataFlowLiveVariables>();
		AU.addRequired<LoopInfo>();
		AU.addRequired<LiveIntervals>();
		MachineFunctionPass::getAnalysisUsage(AU);
	}

	/// runOnMachineFunction - register allocate the whole function
	bool runOnMachineFunction(MachineFunction&);

private:

	MachineFunction* MF;
	const TargetMachine* TM;
	const MRegisterInfo* RegInfo;
	const TargetInstrInfo *TII;
	DataFlowLiveVariables* LV;
	LoopInfo* LI;
	LiveIntervals *LInt;
	GlobalMCNF mcnf_cfg;
	GlobalMCNF simple_mcnf_cfg;
	Solution solution;
	int solution_cost;
	hash_map<unsigned, unsigned> savedRegs;

	void Solve();
	void ApplySolution();
	int FixCalleeSaves(GlobalMCNF *cfg);
	void SaveCalleeSaves();
	void FixJumpTables();
	int RemoveDumbMoves();
	void FixBlockFrequencies();
	int RemoveUnnecessarySaveRestore(GlobalMCNF *cfg);
private:

	// StackSlotForVirtReg - Maps SSA Regs => frame index on the stack where
	// these values are spilled
	std::map<unsigned, int> StackSlotForVirtReg;
	std::vector<int> StackSlotForQuantity; //indexed by global quantity
	int getStackSpaceFor(GlobalMCNF *cfg, unsigned VirtReg, const TargetRegisterClass *RC);

};

struct TallySolutionComponents
{
	GlobalMCNF *mcnf_cfg;
	//count up the contribution to the solution cost of moves, loads, stores, etc
	int loadCost;
	int storeCost;
	int moveCost;
	int memOpCost;
	int regPrefCost;
	int negMemOpCost; //probably load remat
	int miscCost;

	TallySolutionComponents(GlobalMCNF *cfg): mcnf_cfg(cfg), loadCost(0), storeCost(0), moveCost(0), memOpCost(0), regPrefCost(0), negMemOpCost(0), miscCost(0) {}

	void clear()
	{
		loadCost = storeCost = moveCost = memOpCost = regPrefCost = miscCost = 0;
	}
	void print(std::ostream& OS)
	{
		OS << "loadCost " << loadCost << "\n";
		OS << "storeCost " << storeCost << "\n";
		OS << "moveCost " << moveCost << "\n";
		OS << "memOpCost " << memOpCost << "\n";
		OS << "regPrefCost " << regPrefCost << "\n";
		OS << "negMemOpCost " << negMemOpCost << "\n";
		OS << "miscCost " << miscCost << "\n";
	}

	void dump() { print(std::cerr); }

	void operator()(mcnf_node *node)
	{
		//look at all the feasible solutions out of this node
		for(int c = 0, nc = node->commodities.size(); c < nc; c++)
		{
			int q = node->commodities[c];
			mcnf_edge *edge = node->fsinfo[c].next_edge;
			if(mcnf_cfg->savedRegHardRegs.count(q) || mcnf_cfg->savedRegHardRegs.count(-q))
				continue; //ignore any costs associated with saving callee save registers
			if(edge != NULL) //has feas solution through this node that costs something
			{
				mcnf_node *src = edge->src;
				mcnf_node *dst = edge->dst;
				int cost = edge->getCost(q);
				if(cost == 0)
					; //do nothing
				else if(src->has_class() && dst->has_class())
				{
					if(alloc_classes->getType(src->allocClass) == AllocationClassInfo::Memory &&
							alloc_classes->getType(dst->allocClass) == AllocationClassInfo::AntiOnly)
					{
						if(src->type == nXBar && dst->type == nXBar)
							storeCost += cost;
						else
							memOpCost += cost;
					}
					else if(alloc_classes->getType(src->allocClass) == AllocationClassInfo::Memory &&
							alloc_classes->getType(dst->allocClass) == AllocationClassInfo::Register)
					{
						loadCost += cost;
					}
					else if(alloc_classes->getType(src->allocClass) == AllocationClassInfo::Register &&
							alloc_classes->getType(dst->allocClass) == AllocationClassInfo::Register)
					{
						if(src->type == nXBar && dst->type == nXBar)
							moveCost+=cost;
						else if(src->type == nXBar && dst->type == nInsn)
							regPrefCost += cost;
						else
							miscCost += cost;
					}
					else if(alloc_classes->getType(src->allocClass) == AllocationClassInfo::Memory &&
							alloc_classes->getType(dst->allocClass) == AllocationClassInfo::Memory &&
							src->type == nXBar && dst->type == nInsn)
					{
						memOpCost += cost;
					}
					else
					{
						miscCost+=cost; //huh?
					}
				}
				else if(src->type == nDef)
				{
					if(alloc_classes->getType(dst->allocClass) == AllocationClassInfo::Memory ||
							alloc_classes->getType(dst->allocClass) == AllocationClassInfo::AntiOnly)
					{
						if(cost < 0)
							negMemOpCost += cost;
						else
							memOpCost += cost;
					}
					else if(alloc_classes->getType(dst->allocClass) == AllocationClassInfo::Register)
					{
						regPrefCost += cost;
					}
					else
						miscCost += cost;
				}
				else
					miscCost += cost;
			}
		}
	}
};

ProgressiveAllocator::~ProgressiveAllocator()
{
}

/* jump table jumps use a variable inside of branch at the end of
 * the block which means the exit allocation cannot be fixed.  To prevent
 * this from being a problem, make sure all jump target blocks have only
 * the jump block as a predecessor.
 * */
void ProgressiveAllocator::FixJumpTables()
{
	MachineJumpTableInfo *JTI = MF->getJumpTableInfo();
	const std::vector<MachineJumpTableEntry> &JTs = JTI->getJumpTables();
	if (!JTs.empty())
	{
		for (MachineFunction::iterator BB = MF->begin(), E = MF->end();
		   	BB != E; ++BB)
		{
			MachineBasicBlock *MBB = BB;
			for (MachineBasicBlock::iterator I = BB->begin(), E = BB->end();
				I != E; ++I)
			{
				for (unsigned op = 0, e = I->getNumOperands(); op != e; ++op)
				{
					MachineOperand &Op = I->getOperand(op);
					if (!Op.isJumpTableIndex()) continue;
					unsigned Idx = Op.getJumpTableIndex();
					std::vector<MachineBasicBlock*> NewTable;
					const MachineJumpTableEntry& JTE = JTs[Idx];
					hash_map<MachineBasicBlock*, MachineBasicBlock*> processed;
					//each block in the table must have only MBB as pred
					for (unsigned b = 0, nb = JTE.MBBs.size(); b < nb; b++)
					{
						MachineBasicBlock *targ = JTE.MBBs[b];
						//fallthrough cases will all point to the same targ, process
						//it but once
						if(processed.count(targ))
						{
							NewTable.push_back(processed[targ]);
						}
						else if(targ->pred_size() == 1)
						{
							NewTable.push_back(targ);
						}
						else
						{
							//have to create a new block in between
							//MBB and targ
							MachineBasicBlock *NewMBB = new MachineBasicBlock(targ->getBasicBlock());
							MF->getBasicBlockList().push_back(NewMBB);
							NewMBB->addSuccessor(targ);
							if(MBB->isSuccessor(targ))
								MBB->removeSuccessor(targ);
							MBB->addSuccessor(NewMBB);
							TII->InsertBranch(*NewMBB, targ, NULL, std::vector<MachineOperand>());
							NewTable.push_back(NewMBB);
							processed[targ] = NewMBB;
						}
					}
					unsigned NewIdx = JTI->getJumpTableIndex(NewTable);
					Op.setJumpTableIndex(NewIdx);
				}
			}
		}
	}
}

//attempt to assign frequencies to any blocks that have no information
//(were created after compute block freq)
void ProgressiveAllocator::FixBlockFrequencies()
{
	for (MachineFunction::iterator BB = MF->begin(), E = MF->end(); BB != E; ++BB)
	{
		MachineBasicBlock *B = BB;
		//compute sum of successors frequencies
		unsigned bfreq = B->getFrequency();
		unsigned sfreq = 0;
		int zerocnt = 0;
		for(MachineBasicBlock::succ_iterator S = B->succ_begin(), E = B->succ_end(); S != E; S++)
		{
			MachineBasicBlock *succ = *S;
			sfreq += succ->getFrequency();
			if(succ->getFrequency() == 0)
				zerocnt++;
		}

		//if sfreq is less than b's freq, something is wrong, give difference
		//to block's with zero cnt
		if(sfreq < bfreq && zerocnt > 0)
		{
			unsigned f = (bfreq-sfreq)/zerocnt;
			for(MachineBasicBlock::succ_iterator S = B->succ_begin(), E = B->succ_end(); S != E; S++)
			{
				MachineBasicBlock *succ = *S;
				if(succ->getFrequency() == 0)
				{
					succ->setFrequency(f);
					if(DebugPRA > 2)
					{
						std::cerr << "Fixing block " << succ->getNumber() << " frequency to " << f << "\n";
					}
				}
			}
		}
	}
}

//remove dead code returning how much removed
int ProgressiveAllocator::RemoveDumbMoves()
{
	int ret = 0;
	int sz = 0;
	LV->update_full();
	while((sz = SimpleDCE::removeDeadCode(*MF)) > 0)
	{
		ret += sz;
		LV->update_full();
	}
	return ret;
}

void ProgressiveAllocator::Solve()
{

    /* use double precision across the board, otherwise
       we have rounding issues */
#ifdef FPU_CONTROL_AVAILABLE
    fpu_control_t cw, oldcw;
    _FPU_GETCW(cw);
    _FPU_GETCW(oldcw);
    cw &= ~_FPU_EXTENDED;
    cw |= _FPU_DOUBLE;
    _FPU_SETCW(cw);
#endif
	solution_cost = 0;

	int maxiter = MAX_ITER;
	if(flag_prog_ra_iter >= 0)
		maxiter = flag_prog_ra_iter;
	if(flag_prog_ra_warmup_iter >= flag_prog_ra_iter)
		flag_prog_ra_warmup_iter = flag_prog_ra_iter-1;

	if(!isInterestingFunction(MF->getFunction()->getName()))
	{
		//compile very quickly
		maxiter = 0;
	}
	if(UseSimpleModel)
	{
		int middle = flag_prog_ra_warmup_iter;
		if(middle < 0) middle = maxiter;

		alloc_classes = TM->getSimpleAllocationClassInfo();

		simple_mcnf_cfg.initialize_solver_data();
		LagrangianSolver solver(&simple_mcnf_cfg, LI, solution);
		solution_cost = solver.solve(1, middle, !SimpleOnly);

		if(!SimpleOnly)
		{
			alloc_classes = TM->getAllocationClassInfo();
			mcnf_cfg.copyPrices(simple_mcnf_cfg, TM->getSimpleAllocationClassInfo());

			simple_mcnf_cfg.clear();
			mcnf_cfg.initialize_solver_data();

			solver.changeModel(&mcnf_cfg);
			solution_cost = solver.solve(middle+1, maxiter);
		}
	}
	else if(WhatSolver == CPLEX2Phase && isInterestingFunction(MF->getFunction()->getName()))
	{
		mcnf_cfg.initialize_solver_data();
		simple_mcnf_cfg.initialize_solver_data();
		//first solve simple model
		mcnf_cplex_solve(simple_mcnf_cfg, false);

		if(DebugPRA)
		{
			TallySolutionComponents tally(&mcnf_cfg);
			simple_mcnf_cfg.traverse_nodes(tally);
			tally.dump();
			tally.clear();
		}

		//constrain full model with results of simple model solution
		alloc_classes = TM->getAllocationClassInfo();
		mcnf_cfg.constrainModel(simple_mcnf_cfg, TM->getSimpleAllocationClassInfo());
		//now solve
		solution_cost = mcnf_cplex_solve(mcnf_cfg,ConstrainToSolution);
		solution.record_solution(mcnf_cfg);

		if(DebugPRA)
		{
			TallySolutionComponents tally(&mcnf_cfg);
			mcnf_cfg.traverse_nodes(tally);
			tally.dump();
			tally.clear();
		}

	}
	else if(WhatSolver == CPLEX && isInterestingFunction(MF->getFunction()->getName()))
	{
		mcnf_cfg.initialize_solver_data();
		if(InitCPLEXSolution)
		{
			//get an initial solution to speed thing along
			LagrangianSolver solver(&mcnf_cfg, LI, solution);
			int val = solver.solve(1, 0);
			if(DebugPRA > 3) std::cerr << "Solval " << val << "\n";
		}
		solution_cost = mcnf_cplex_solve(mcnf_cfg, InitCPLEXSolution);
		solution.record_solution(mcnf_cfg);

		if(DebugPRA)
		{
			TallySolutionComponents tally(&mcnf_cfg);
			mcnf_cfg.traverse_nodes(tally);
			tally.dump();
			tally.clear();
			std::cerr << "TCAdjust " << mcnf_cfg.feas_sol_tied_constraint_adjust() << "\n";
		}

	}
	else if(WhatSolver == ColumnGen)
	{
		TiedConstraints = false;
		ColumnGenSolver solver(mcnf_cfg, LI, solution);
		solution_cost = solver.solve();
	}
	else
	{
		mcnf_cfg.initialize_solver_data();
		LagrangianSolver solver(&mcnf_cfg, LI, solution);
		if(flag_prog_ra_warmup_iter > 0)
		{
			//do a prices only warmup period
			solver.solve(1, flag_prog_ra_warmup_iter, true);
			solution_cost = solver.solve(flag_prog_ra_warmup_iter+1, maxiter);
		}
		else
		{
			solution_cost = solver.solve(1, maxiter);
		}
	}
	if(OutputMST)
	{
		mcnf_cplex_output_mst(mcnf_cfg);
	}
	if(DebugPRA)
	{
		std::string fname = MF->getFunction()->getParent()->getModuleIdentifier() + "_" + MF->getFunction()->getName();
		std::cerr << fname << " total cost = " << solution_cost << "\n";
	}
#ifdef FPU_CONTROL_AVAILABLE
    _FPU_SETCW(oldcw);
#endif

}

void ProgressiveAllocator::ApplySolution()
{
	unsigned b, nb;
	int adjust = 0;
	int initialSize = 0;
	std::string fname = MF->getFunction()->getParent()->getModuleIdentifier() + "_" + MF->getFunction()->getName();

	if (DebugPRA > 1)
		solution.print_solution(*cerr.stream());

	GlobalMCNF *cfg = &mcnf_cfg;
	if(SimpleOnly)
		cfg = &simple_mcnf_cfg;

	if(DebugPRA > 0)
	{
		initialSize = MF->getSize();
		std::cerr << "BEFORERA " << fname <<  "\t" << initialSize << "\n";
		//add up cost of forcably coalesced moves before we remove them
		for(hash_set<MachineInstr*>::iterator itr = cfg->coalesceableMoves.begin(),
				eitr = cfg->coalesceableMoves.end(); itr != eitr; itr++)
		{
			MachineInstr *move = *itr;
			adjust += TII->instrSize(*move);
		}
	}
	nb = solution.block_sols.size();
	//go through and rename operands to their assigned phys registers
	for (b = 0; b < nb; b++)
	{
		block_solution& bsol = solution.block_sols[b];
		mcnf_block *block = bsol.block;

		if(DebugPRA > 1)
		{
			std::cerr << "Block " << block->index << "\n";
		}
		//indexed by LLVM register, records what class each register
		//is allocated to
		AllocClassMap alloc_class_map(block);
		bsol.setup_block_in(alloc_class_map);

		unsigned trans = 0;
		unsigned maxtrans = bsol.transforms.size();
		unsigned i, n;
		for (i = 0, n = block->RAATs.size(); i < n; i++)
		{
			RAATInfo& RI = block->RAATs[i];

			//apply all transformations before this instruction
			for (; trans < maxtrans && bsol.transforms[trans].position < (int)i; trans++)
			{
				bsol.transforms[trans].apply(RI, alloc_class_map);
			}
			RI.update_regs(alloc_class_map);

			if(DebugPRA > 1)
			{
				RI.dump();
			}
		}

		//apply all block end transformations
		for (; trans < maxtrans; trans++)
		{
			bsol.transforms[trans].applyEnd(block, alloc_class_map);
		}

	}

	//get rid of any stupid stores we may have inserted
	//(hopefully none); also remove self moves
	RemoveDumbMoves();
	adjust += RemoveUnnecessarySaveRestore(cfg);

	//only mark callee saves that actually need saving
	FixCalleeSaves(cfg);

	double memAccesses = 0;
	StackSlotForQuantity.resize(cfg->quantities.size(), UNALLOCED);
	//now generate stack references for unallocated operands
	for (b = 0; b < nb; b++)
	{
		mcnf_block *block = solution.block_sols[b].block;
		MachineBasicBlock *MBB = block->b;
		// loop over each instruction
		MachineBasicBlock::iterator nextMI;
		for (MachineBasicBlock::iterator MI = MBB->begin(); MI != MBB->end(); MI = nextMI )
		{
			unsigned srcReg, dstReg;
			nextMI = MI;
			nextMI++;
			MachineInstr *insn = MI;
			if (TII->isCoalesceable(*MI, srcReg, dstReg)
					&& MRegisterInfo::isVirtualRegister(srcReg)
					&& MRegisterInfo::isVirtualRegister(dstReg)
				&& cfg->coalesceableMoves.count(insn))
			{
				//this is a move between two virtual registers, presumably
				//of the same quantity, where both are in memory
				//so just assign the destination the same stack location and
				//omit the move

				//note that if the src and dst are of different sizes we could
				//have a problem, although I think maybe because of x86 being
				//little endian we can avoid it.. maybe..
				const TargetRegisterClass* RC = MF->getSSARegMap()->getRegClass(srcReg);
				int FrameIdx = getStackSpaceFor(cfg, srcReg, RC);

				//make sure the destination doesn't have an assigned index
				std::map<unsigned, int>::iterator I = StackSlotForVirtReg.lower_bound(dstReg);
				assert(I->first != dstReg || I->second == FrameIdx);
				//assign destReg to virtreg's stack slot
				StackSlotForVirtReg.insert(std::make_pair(dstReg, FrameIdx));
				//and update quantity
				int q = cfg->reg_quantity.get(dstReg);
				StackSlotForQuantity[q] = FrameIdx;
				assert(MF->getFrameInfo()->getObjectSize(FrameIdx) >= MF->getSSARegMap()->getRegClass(dstReg)->getSize());
				//and delete this instruction
				 MBB->erase(MI);
				continue;
			}

			//identify save/restore code for callee save registers we inserted
			//remove it and let the prolog/epilog generator do its thing instead
			if(TII->isMoveInstr(*MI, srcReg, dstReg) && MI->hasSavedReg())
			{
				assert(MI->getOperand(0).getReg() == dstReg);
				assert(MI->getOperand(1).getReg() == srcReg);
				//this is a hack for x86 sizing
				adjust += 1;
				MBB->erase(MI);
				continue;
			}

			// Loop over uses, move from memory into registers.
			for (int i = MI->getNumOperands() - 1; i >= 0; --i)
			{
				MachineOperand &op = MI->getOperand(i);
				if (op.isRegister() && op.getReg()
						&& MRegisterInfo::isVirtualRegister(op.getReg()))
				{
					unsigned virtualReg = (unsigned) op.getReg();
					int FrameIdx = 0;
					memAccesses += MBB->getFrequency();

					//if this is a def from a stack location, leave it in mem
					if (op.isDef() && TII->isLoadFromStackSlot(MI, FrameIdx)
							&& MF->getFrameInfo()->isFixedObjectIndex(FrameIdx))
					{
						// Find the location VirtReg would belong...
						std::map<unsigned, int>::iterator I =
								StackSlotForVirtReg.lower_bound(virtualReg);
						// Assign the slot...
						StackSlotForVirtReg.insert(I, std::make_pair(
								virtualReg, FrameIdx));
						// and update quantity
						int q = cfg->reg_quantity.get(virtualReg);
						StackSlotForQuantity[q] = FrameIdx;

						// and delete this instruction
						MBB->erase(MI);
						break;
					}
					else //access mem directly in insn
					{
						//this should be allocated to memory
						const TargetRegisterClass* RC = MF->getSSARegMap()->getRegClass(virtualReg);
						FrameIdx = getStackSpaceFor(cfg, virtualReg, RC);
						MachineInstr *ninsn = RegInfo->foldMemoryOperand(MI, i,
								FrameIdx);
						if(ninsn == NULL)
						{
							std::cerr << "Couldn't vold memory op!\n";
							MI->dump();
							std::cerr << "Block " << b << "\n";
							RegInfo->foldMemoryOperand(MI, i,
															FrameIdx);
							abort();
						}
						MBB->insert(MBB->erase(MI), ninsn); //replace insn
						MI = ninsn;
						i = MI->getNumOperands();
						continue;
					}
				}
			}
		}
	}
	RemoveDumbMoves();

	if(DebugPRA > 0)
	{
		if(DebugPRA > 1) MF->dump();
		int boilerplate = 4; //push ebp, mov esp,ebp, pop ebp
		std::cerr << "WeightedMemAccess " << fname << "\t" << memAccesses << "\n";
		std::cerr << "b,i,sc,adj:\t" << boilerplate << " " << initialSize << " " << solution_cost << " " << adjust<<"\n";
		std::cerr << "PREDICTED " << fname << "\t" << boilerplate+initialSize+solution_cost-adjust << "\n";
		std::cerr << "ACTUAL " << fname << "\t" << MF->getSize()+boilerplate << "\n";
	}
}

/// getStackSpaceFor - This allocates space for the specified virtual register
/// to be held on the stack.
int ProgressiveAllocator::getStackSpaceFor(GlobalMCNF *cfg, unsigned VirtReg, const TargetRegisterClass *RC) {
  // Find the location Reg would belong...
  std::map<unsigned, int>::iterator I =StackSlotForVirtReg.lower_bound(VirtReg);

  if (I != StackSlotForVirtReg.end() && I->first == VirtReg)
    return I->second;          // Already has space allocated?

  int FrameIdx = UNALLOCED;
  //check global quantity
  int q = cfg->reg_quantity.get(VirtReg);
  FrameIdx = StackSlotForQuantity[q];

  // Allocate a new stack object for this spill location...
  if(FrameIdx == UNALLOCED)
  {
	  FrameIdx = MF->getFrameInfo()->CreateStackObject(RC->getSize(),
                                                       RC->getAlignment());
	  StackSlotForQuantity[q] = FrameIdx;
  }
  // Assign the slot...
  StackSlotForVirtReg.insert(I, std::make_pair(VirtReg, FrameIdx));
  return FrameIdx;
}

//If the allocator was dumb and inserted explict save and restores
//of callee save registers, remove them now (assume prolog/epilog will clean up)
int ProgressiveAllocator::RemoveUnnecessarySaveRestore(GlobalMCNF *cfg)
{
	int cnt = 0;
	for (unsigned b = 0, nb = cfg->blocks.size(); b < nb; b++)
	{
		mcnf_block *block = cfg->blocks[b];
		MachineBasicBlock *MBB = block->b;
		// loop over each instruction
		MachineBasicBlock::iterator MI = MBB->begin();

		while(MI != MBB->end())
		{
			unsigned srcReg, dstReg;
			if(TII->isMoveInstr(*MI, srcReg, dstReg) && (savedRegs.count(srcReg) || savedRegs.count(dstReg)))
			{
				if(DebugPRA > 7)
				{
					std::cerr << "Erasing " << *MI;
				}
				MI = MBB->erase(MI);
				cnt++;
				continue;
			}
			MI++;
		}
	}
	return cnt;
}

// go through and mark the used physical registers,
//ignore the save/restore code; this will get deleted later
//and then the prolog/epilog code will put it back in (hopefully
//using more efficient insns)
// This is slightly broken if the regalloc does not use the original
// save restore code (shouldn't be a problem for code size).
// We perform this somewhat convoluted process so the initial heuristic
// can ignore caller save issues.
// returns the number of callee save registers which do not need to be saved
int ProgressiveAllocator::FixCalleeSaves(GlobalMCNF *cfg)
{
	const unsigned *regs = RegInfo->getCalleeSavedRegs(MF);
	for (; *regs; regs++)
	{
		unsigned reg = *regs;
		MF->setPhysRegUnused(reg);
	}

	for (unsigned b = 0, nb = cfg->blocks.size(); b < nb; b++)
	{
		mcnf_block *block = cfg->blocks[b];
		MachineBasicBlock *MBB = block->b;
		// loop over each instruction
		for (MachineBasicBlock::iterator MI = MBB->begin(); MI != MBB->end(); MI++ )
		{
			if(MI->hasSavedReg())
				continue;

			for (unsigned i = 0, e = MI->getNumOperands(); i != e; ++i)
			{
				//register operand?
				MachineOperand &MO = MI->getOperand(i);
				if (MO.isReg())
				{
					unsigned reg = MO.getReg();
					if(reg > 0 && MRegisterInfo::isPhysicalRegister(reg))
					{
						MF->setPhysRegUsed(reg);
					}
				}
			}
		}
	}

	int ret = 0;
	regs = RegInfo->getCalleeSavedRegs(MF);
	for (; *regs; regs++)
	{
		unsigned reg = *regs;
		if(!MF->isPhysRegUsed(reg))
			ret++;
	}
	return ret;
}
// insert moves at the start and end of the function that save
// all the callee save registers; if the registers are not needed,
// the virtuals will be assigned the same register and the moves will
// disappear
void ProgressiveAllocator::SaveCalleeSaves()
{
	std::vector<MachineBasicBlock::iterator> exitis;
	std::vector<MachineBasicBlock*> exitbs;
	//first find all the exiting blocks in the function
	for (MachineFunction::iterator FI = MF->begin(), E = MF->end(); FI != E; ++FI)
	{
		// look for return instruction
		if (!FI->empty() && TII->isReturn(FI->back().getOpcode()))
		{
			MachineBasicBlock *MBB = FI;
			MachineBasicBlock::iterator I = MBB->end();
			--I;

			// Skip over all terminator instructions, which are part of the return
			// sequence.
			MachineBasicBlock::iterator I2 = I;
			while (I2 != MBB->begin() && TII->isTerminatorInstr((--I2)->getOpcode()))
				I = I2;

			exitis.push_back(I);
			exitbs.push_back(MBB);
		}
	}

	const TargetRegisterClass * const *RCs =
			RegInfo->getCalleeSavedRegClasses(MF);
	const unsigned *regs = RegInfo->getCalleeSavedRegs(MF);
	for (; *regs; regs++, RCs++)
	{
		unsigned reg = *regs;
		assert(*RCs && "Empty reg class for callee save");

		if (alloc_classes->getRegisterIndex(reg) < 0) //not allocable
			continue;
		MachineBasicBlock *MBB = MF->begin();
		MachineBasicBlock::iterator I = MBB->begin();
		//store reg to a virtual at top of function
		unsigned virtReg = MF->getSSARegMap()->createVirtualRegister(*RCs);
		RegInfo->copyRegToReg(*MBB, I, virtReg, reg, *RCs);
		MachineInstr *copy = prior(I);
		copy->findRegisterDefOperand(virtReg)->setSavedReg();
		savedRegs[virtReg] = reg;

		if(DebugPRA > 10)
		{
			std::cerr << "SavedReg " << virtReg << " " << *copy;
		}
		//load back the value at each exit
		for (unsigned i = 0, n = exitis.size(); i < n; i++)
		{
			MBB = exitbs[i];
			I = exitis[i];
			RegInfo->copyRegToReg(*MBB, I, reg, virtReg, *RCs);
			copy = prior(I);
			copy->getOperand(copy->findRegisterUseOperandIdx(virtReg)).setSavedReg();
			if(DebugPRA > 10)
			{
				std::cerr << *copy;
			}
		}
	}

	//update live variable infor
	LV->update_full();
}

bool ProgressiveAllocator::runOnMachineFunction(MachineFunction &fn)
{
	MF = &fn;
	TM = &fn.getTarget();
	RegInfo = TM->getRegisterInfo();
	alloc_classes = TM->getAllocationClassInfo();
	LV = &getAnalysis<DataFlowLiveVariables>();
	TII = TM->getInstrInfo();
	LI = &getAnalysis<LoopInfo>();
	LInt = &getAnalysis<LiveIntervals>();

	if(SimpleOnly)
		UseSimpleModel = true;

	std::string fname = MF->getFunction()->getParent()->getModuleIdentifier() + "_" + MF->getFunction()->getName();

	if(DumpCFG)
		MF->viewCFGOnly();

	if(DebugPRA > 1)
		fn.dump();
	//fix jumptable control flow
	FixJumpTables();
	//fix up missing freq information
	FixBlockFrequencies();
	//insert pre-alloc prologue/epilogue code
	SaveCalleeSaves();

	RemoveDumbMoves();

	if(DebugPRA > 1)
		fn.dump();

	 if(WhatSolver == ColumnGen)
	 { //colgen doesn't yet support these
		 TiedConstraints = false;
	 }

	 if(UseSimpleModel || WhatSolver == CPLEX2Phase)
	 {
		 //create a simple model and solve it first
		 alloc_classes = TM->getSimpleAllocationClassInfo();
		 simple_mcnf_cfg.build(MF, LV, LInt, LI);

		 //let model know what regs are just saving callee-save registers
		 for(hash_map<unsigned,unsigned>::iterator itr = savedRegs.begin(), e = savedRegs.end(); itr != e; itr++)
		 {
			 simple_mcnf_cfg.setSavedReg(itr->first, itr->second);
		 }
		 if(DebugPRA)
		 {
			 std::cerr << fname << " SimpleNumNodes " << simple_mcnf_cfg.num_nodes << "\n";
			 std::cerr << fname << " SimpleNumEdges " << simple_mcnf_cfg.num_edges << "\n";
			 std::cerr << fname << " SimpleNumQuantities " << simple_mcnf_cfg.quantities.size() << "\n";
			 std::cerr << fname << " SimpleNumBlocks " << simple_mcnf_cfg.blocks.size() << "\n";
			 std::cerr << fname << " SimpleNumEmptyBlocks " << simple_mcnf_cfg.numEmptyBlocks() << "\n";
			 std::cerr << fname << " SimpleNumRAATs " << simple_mcnf_cfg.num_RAATs << "\n";
			 std::cerr << fname << " SimpleNumIgnoreable " << simple_mcnf_cfg.num_ignoreable << "\n";
		 }

		 if(BuildAndQuit)
			 exit(0);
		 if(OutputCPLEX)
		 {
			mcnf_cplex_output(simple_mcnf_cfg, "_simple");
		 }

		 alloc_classes = TM->getAllocationClassInfo();
	 }

	 if(!SimpleOnly)
	 {
		 //build the MCNF model from the RAATs
		 //for now we generate RAATs directly from LLVM insns, eventually
		 //RAATs will be generated directly by instruction selection
		 mcnf_cfg.build(MF, LV, LInt, LI);

		 //let model know what regs are just saving callee-save registers
		 for(hash_map<unsigned,unsigned>::iterator itr = savedRegs.begin(), e = savedRegs.end(); itr != e; itr++)
		 {
			 mcnf_cfg.setSavedReg(itr->first, itr->second);
		 }

		if(DebugPRA)
		{
			std::cerr << fname << " NumNodes " << mcnf_cfg.num_nodes << "\n";
			std::cerr << fname << " NumEdges " << mcnf_cfg.num_edges << "\n";
			std::cerr << fname << " NumBitSets " << MemoizedBitSet::numUnique() << "\n";
			std::cerr << fname << " NumQuantities " << mcnf_cfg.quantities.size() << "\n";
			std::cerr << fname << " NumBlocks " << mcnf_cfg.blocks.size() << "\n";
			std::cerr << fname << " NumEmptyBlocks " << mcnf_cfg.numEmptyBlocks() << "\n";
			std::cerr << fname << " NumRAATs " << mcnf_cfg.num_RAATs << "\n";
			std::cerr << fname << " NumIgnoreable " << mcnf_cfg.num_ignoreable << "\n";
			std::cerr << fname << " NumTied " << mcnf_cfg.getNumTiedconstraints() << "\n";
		}

		if(BuildAndQuit)
			exit(0);
		if(OutputCPLEX)
		{
			mcnf_cplex_output(mcnf_cfg, "");
		}
	 }

	//solve the MCNF problem
	Solve();

	//apply the allocation to the machine function
	ApplySolution();

	if(DebugPRA)
	{
		std::cerr << fname << " NumSpillSlots " << MF->getFrameInfo()->getObjectIndexEnd() << "\n";
	}
	//deallocate memory
	mcnf_cfg.clear();
	simple_mcnf_cfg.clear();
	StackSlotForVirtReg.clear();
	StackSlotForQuantity.clear();
	solution.clear();
	savedRegs.clear();
	if(DebugPRA > 1)
	{
		std::cerr << "After alloc:\n";
		fn.dump();
	}
	return true;
}

FunctionPass* createProgressiveRegisterAllocator()
{
	return new ProgressiveAllocator();
}

char ProgressiveAllocator::ID = 0;


} //llvm namespace
