#include "llvm/PrincipledBackend/RAATInfo.h"
#include "llvm/CodeGen/MachineFunction.h"
#include "llvm/CodeGen/MachineFrameInfo.h"
#include "llvm/CodeGen/SSARegMap.h"
#include "llvm/Target/TargetMachine.h"
#include "llvm/Target/TargetInstrInfo.h"
#include "llvm/ADT/StringExtras.h"
#include "llvm/Module.h"
#include <iostream>

/* Code that converst an LLVM machine instruction into a RAAT for use
 * by the progressive allocator.  Eventually we will replace the instruction
 * selector by something RAATier (or will we?).
 */

cl::opt<bool> DisableLoadRemat("disable-load-remat", cl::Hidden,
		cl::desc("Don't rematerialize loads from fixed stack objects"));


namespace llvm
{
// dump to stderr
void RAATInfo::dump()
{
	print(std::cerr);
}

//print RAAT out
void RAATInfo::print(std::ostream& out)
{
	unsigned i, n;
	out << *insn;

	for (i = 0, n = info.size(); i < n; i++)
	{
		out << "[" << itostr(info[i].q) << " (";
		for (unsigned j = 0, ops = info[i].opindex.size(); j < ops; j++)
		{
			if (j > 0)
				out << " ";
			out << itostr(info[i].opindex[j]);
		}
		out << ") ";

		if (info[i].def)
			out << "d";
		if (info[i].lastuse)
			out << "l";
		if (info[i].use)
			out << "u";
		if (info[i].extend)
			out << "e";
		out << "] ";
	}
	out << "\n";
}

// base cost of this rate independent of allocation, need to somehow make this
// flexible since it depends on the code quality metric used
int RAATInfo::computeBaseCost(MachineInstr *MI)
{
	MachineFunction *MF = MI->getParent()->getParent();
	const TargetInstrInfo *TII = MF->getTarget().getInstrInfo();
	return TII->instrSize(*MI);
}

//returns the indx of the first operand which uses reg, -1 if nonexistant
static int regOpIndex(MachineInstr *MI, unsigned reg)
{
	for (unsigned i = 0, e = MI->getNumOperands(); i != e; ++i)
	{


		const MachineOperand &MO = MI->getOperand(i);
		if (MO.isReg() && MO.getReg() == reg)
			return i;
	}
	return -1;
}

void RAATInfo::initializeFromInsn(MachineInstr *MI)
{
	MachineFunction *MF = MI->getParent()->getParent();
	const TargetInstrInfo *TII = MF->getTarget().getInstrInfo();

	insn = MI;
	baseCost = computeBaseCost(MI);
	//analyze each operand
	for (unsigned i = 0, e = MI->getNumOperands(); i != e; ++i)
	{
		//register operand?
		MachineOperand &MO = MI->getOperand(i);
		if (MO.isReg())
		{
			unsigned reg = MO.getReg();
			if (reg)
			{
				OperandInfo *OI;

				if (MRegisterInfo::isVirtualRegister(reg))
				{
					if (reg2index.count(reg) > 0) //already exists
					{
						OI = &ops[reg2index[reg]];
					}
					else
					{
						reg2index[reg] = ops.size();
						ops.push_back(OperandInfo());
						opClassCosts.resize(ops.size(), std::vector<short>(alloc_classes->numClasses(),0));
						OI = &ops.back();
					}
				}
				else /* hard reg */
				{
					reg = MF->getTarget().getRegisterInfo()->getCanonicalReg(reg);
					if (alloc_classes->getRegisterIndex(reg) < 0)
						continue; /* not a register used for allocation */
					if (reg2index.count(reg)> 0) /* already exists */
					{
						OI = &hardops[reg2index[reg]];
						if(MO.isImplicit() && !OI->isImplicit)
							continue; //ignore implicit info if there's explicit info .. of dubious goodness
					}
					else
					{
						reg2index[reg] = hardops.size();
						hardops.push_back(OperandInfo());
						OI = &hardops.back();
					}
				}

				int index = reg2index[reg];

				//if two views of a reg (ie ah/al) are accessed diferently, need to correctly merge
				if(!OI->isDef && MO.isDead())
				{
					OI->isLastUse = true;
					OI->isExtend = true;
				}
				else if(OI->isExtend && MO.isDef() && !MO.isDead())
				{
					OI->isLastUse = false;
					OI->isExtend = false;
				}
				OI->isUse |= MO.isUse();
				OI->isDef |= MO.isDef();
				OI->isImplicit &= MO.isImplicit();
				OI->isLastUse |= MO.isKill();
				OI->regNum = reg;
				OI->allowableClasses.resize(alloc_classes->numClasses(), false);

				if (MRegisterInfo::isVirtualRegister(reg))
				{
					/* //determine acceptable allocation classes */
					const TargetRegisterClass *RC = MF->getSSARegMap()->getRegClass(reg);
					for (unsigned c = 0, e = alloc_classes->numClasses(); c < e; c++)
					{
						if (alloc_classes->isCompatible(*MF, RC, c))
						{
							//can allocate to this class
							OI->allowableClasses.set(c);
						}
						else if (alloc_classes->getType(c)
								== AllocationClassInfo::Memory)
						{
							const MRegisterInfo *MRI = MF->getTarget().getRegisterInfo();
							MachineInstr *tmpI = MRI->foldMemoryOperand(MI, i,
									0);
							//check to see if this reg is used multiple time in insn
							if (tmpI)
							{
								int opIdx;
								while ((opIdx = regOpIndex(tmpI, reg)) >= 0)
								{
									MachineInstr *tmp2 = MRI->foldMemoryOperand(MI,
											opIdx, 0);
									if (tmp2)
									{
										delete tmpI;
										tmpI = tmp2;
									}
									else
									{
										//memory isn't an option
										delete tmpI;
										tmpI = NULL;
										break;
									}
								}
							}
							if (tmpI)
							{
								//can access directly in memory
								delete tmpI;
								OI->allowableClasses.set(c);
							}

							if (OI->isDef)
							{
								//check for a load from a fixed stack slot
								int FrameIdx = 0;
								if (!DisableLoadRemat && TII->isLoadFromStackSlot(MI, FrameIdx)
										&& MF->getFrameInfo()->isFixedObjectIndex(FrameIdx))
								{
									//if we leave the var in memory we can avoid this insn
									OI->allowableClasses.set(c);
									opClassCosts[index][c] = -alloc_classes->getInsnRemovalCost(MI, baseCost);
								}
							}
						}

						if(OI->allowableClasses.test(c) && opClassCosts[index][c] == 0)
							opClassCosts[index][c] = alloc_classes->getInsnCost(MI, i, c);
						//TODO: handle constants and rematerialization
					}
					assert(OI->allowableClasses.any()
							&& "need some allowable class");

				}
				else //hard reg
				{
					OI->isHard = true;
					OI->allowableClasses.set(alloc_classes->getRegisterIndex(reg));
				}

			}
		}
	}

	//check for tied constraint cases
	//conservatively only merge moves that don't change the register class
	unsigned src, dst;
	if (TII->isCoalesceable(*MI, src, dst))
	{
		if (ops.size() == 2)
		{
			assert(ops[0].isDef);
			if (ops[1].isLastUse)
			{
				BitVector ac = ops[0].allowableClasses&ops[1].allowableClasses;
				//if src and dst have the same size class they can share a stack slot
				const TargetRegisterClass *sRC = MF->getSSARegMap()->getRegClass(src);
				const TargetRegisterClass *dRC = MF->getSSARegMap()->getRegClass(dst);
				const std::vector<int>& memory_classes = alloc_classes->getMemoryClasses();
				for (int m = 0, nm = memory_classes.size(); m < nm; m++)
				{
					int c = memory_classes[m];
					if(sRC == dRC)
						ac.set(c);
					else
						ac.reset(c);
				}
				bool mergeable = sRC->getSize() == dRC->getSize();
				if(sRC->hasType(MVT::i32) || dRC->hasType(MVT::i32))
				{
					//target specific hack, can share stack slot if little endian
					if(MF->getFunction()->getParent()->getEndianness()  == Module::LittleEndian)
						mergeable = true;
				}
				tiedConstraints.push_back(TiedConstraint(src, dst,
						-alloc_classes->getInsnRemovalCost(MI, baseCost), ac, !mergeable));
			}
		}
		else if (ops.size() == 1 && hardops.size() >= 1)
		{
			//prefer with a negative cost the hard register class
			assert(!hardops[0].isImplicit);
			int c = alloc_classes->getRegisterIndex(hardops[0].regNum);
		//	if(alloc_classes->getCapacity(c) == 1) //only let cost contribute if class is 1 reg big
			if(!DisableCoalescing || TiedConstraints) //
				opClassCosts[0][c] = -alloc_classes->getInsnRemovalCost(MI, baseCost);
			classPref[0] = c;
		}
	}
}

void RAATInfo::initializeForAllocator()
{
	opQuantities = std::vector<int>(ops.size(), -1);
}

bool RAATInfo::isControlFlow()
{
	MachineFunction *MF = insn->getParent()->getParent();
	const TargetInstrInfo *TII = MF->getTarget().getInstrInfo();
	return TII->isBranch(insn->getOpcode());
}

/* Consolidate operand info into quantity info.  Assumes quantity
 * mapping is setup */
void RAATInfo::analyze_RAAT_quantities()
{
	int n = getNumOps();
	hash_map<int, int> seen; //indexed by quantity

	frees.resize(alloc_classes->numClasses());
	for(int i = 0; i < n; i++)
	{
		int q = getOpQuantity(i);
		const OperandInfo& opinfo = getOpFromIndex(i);
		if(seen.count(q)> 0) /* already seen this quantity */
		{
			int j = seen[q];
			const OperandInfo& jinfo = getOpFromIndex(j);

			if(info[j].q != q)
				abort();

			if(opinfo.isLastUse && !jinfo.isDef)
				info[j].lastuse = true;
			if(opinfo.isDef)
			{
				info[j].lastuse = false;
				/* can only be one equivalence */
				if(info[j].equivalentClass >= 0)
					abort();

				info[j].equivalentClass = opinfo.equivalentClass;
			}
			else
			{
				if(info[j].equivalentClass != opinfo.equivalentClass)
					info[j].equivalentClass = -2; /* no longer mem equiv */
			}

			info[j].use |= opinfo.isUse;
			info[j].def |= opinfo.isDef;
			info[j].extend |= opinfo.isExtend;
			info[j].classes &= opinfo.allowableClasses;
			info[j].opindex.push_back(i);
		}
		else
		{
			QuantityInfo qinfo;
			const OperandInfo& opinfo = getOpFromIndex(i);
			qinfo.q = getOpQuantity(i);
			qinfo.use = opinfo.isUse;
			qinfo.def = opinfo.isDef;
			qinfo.lastuse = opinfo.isLastUse;
			qinfo.extend = opinfo.isExtend;
			qinfo.classes = opinfo.allowableClasses;
			assert(qinfo.classes.any() && "need an allowable class");
			qinfo.equivalentClass = opinfo.equivalentClass;
			qinfo.opindex.push_back(i);
			seen[q] = info.size();
			info.push_back(qinfo);
		}
	}
}

//modify insn to use mappings present in map
void RAATInfo::update_regs(AllocClassMap& map)
{
	for (int i = insn->getNumOperands() - 1; i >= 0; --i)
	{
		MachineOperand &op = insn->getOperand(i);

		if (op.isRegister() && op.getReg() &&
				MRegisterInfo::isVirtualRegister(op.getReg()))
		{
			unsigned virtualReg = (unsigned) op.getReg();

			if(op.isDef())
			{
				int q = getOpQuantity(getOpIndex(virtualReg));
				map.set_quantity_reg(q, virtualReg);
			}
			unsigned physReg = map.getMappedReg(virtualReg);
			if(physReg != virtualReg)
			{
				insn->getOperand(i).setReg(physReg);
			}
		}
	}
}

} //namespace
