#include "llvm/PrincipledBackend/Solution.h"
#include "llvm/PrincipledBackend/GlobalMCNF.h"
#include "llvm/ADT/StringExtras.h"
#include <algorithm>

namespace llvm
{
/* Implementation of solution class which stores transformations
 * extracted from an MCNF solution */


bool transformation::operator<(const transformation& r) const
{
	return compare(r) < 0;
}

bool transformation::operator==(const transformation& r) const
{
	return compare(r) == 0;
}

/* sort transforms first by position, then by type (death, move, birth) */
int transformation::compare(const transformation& r) const
{
	const transformation *lhs = this;
	const transformation *rhs = &r;

	if(lhs->position < rhs->position)
	return -1;
	else if(lhs->position> rhs->position)
	return 1;

	if(lhs->level < rhs->level)
		return -1;
	if(lhs->level> rhs->level)
		return 1;

	if(lhs->type == rhs->type)
	{
		/*
		 do stores before everything else and loads after everything else
		 */
		bool lhsIsLoad = lhs->src_class >= 0 ? alloc_classes->getType(lhs->src_class) == AllocationClassInfo::Memory : false;
		bool rhsIsLoad = rhs->src_class >= 0 ? alloc_classes->getType(rhs->src_class) == AllocationClassInfo::Memory : false;
		bool lhsIsStore = lhs->dst_class >= 0 ? alloc_classes->getType(lhs->dst_class) == AllocationClassInfo::Memory : false;
		bool rhsIsStore = rhs->dst_class >= 0 ? alloc_classes->getType(rhs->dst_class) == AllocationClassInfo::Memory : false;

		if(lhsIsStore && !rhsIsStore)
			return -1;
		if(rhsIsStore && !lhsIsStore)
			return 1;
		if(lhsIsLoad && !rhsIsLoad)
			return 1;
		if(rhsIsLoad && !lhsIsLoad)
			return -1;

		if(lhs->order < rhs->order)
			return -1;
		else if(lhs->order > rhs->order)
			return 1;

		if(lhs->level> 0) /* go left to right */
		{
			if(lhs->src_class < rhs->src_class)
				return -1;
			if(lhs->src_class> rhs->src_class)
				return 1;
		}
		else if(lhs->level == 0) /* right to left */
		{
			if(lhs->src_class < rhs->src_class)
				return 1;
			if(lhs->src_class> rhs->src_class)
				return -1;
		}

		return 0;
	}

	if(lhs->q == -rhs->q)
	{
		/* if a value is defined by an instruction but is not live out,
		 we may still have a store (this is dubious; presumably from
		 extended the live range?).  Do that first.
		 */
		if(lhs->type == Store)
		return -1;
		if(rhs->type == Store)
		return 1;
	}

	if(lhs->type == Death)
	return -1;
	if(rhs->type == Death)
	return 1;
	if(lhs->type == Store)
	return -1;
	if(rhs->type == Store)
	return 1;
	if(lhs->type == Move)
	return -1;
	if(rhs->type == Move)
	return 1;

	abort();
	return 0;
}

void transformation::apply(MachineBasicBlock *MBB, MachineBasicBlock::iterator MI, AllocClassMap& map)
{
	MachineFunction *MF = MBB->getParent();
	const MRegisterInfo* RegInfo = MF->getTarget().getRegisterInfo();

	switch(type)
	{
		/* if a pseudo is born its mapping should be updated */
		case Birth:
		map.set_quantity_alloc_class(q, dst_class, dstr);
		break;
		case Death: /* remove mapping */
		map.unset_quantity_alloc_class(q);
		break;
		case Store:
		{
			/* perform an actual store, but don't change the mapping
			 * since the var may still be used in a register */
			unsigned r = map.getRegForQuantity(-q);
			unsigned src = map.getMappedReg(r);
			//r is the dst, we leave vars in memory unallocated

			if(src == r)
			; /* no insns for mem to mem movement */
			else
			{
				const TargetRegisterClass *RC = MF->getSSARegMap()->getRegClass(r);
				if(!RC->contains(src))
				{
					//trust our solution (cross fingers)
					RC = alloc_classes->getTargetClass(alloc_classes->getRegisterIndex(src));
				}
				RegInfo->copyRegToReg(*MBB, MI, r, src, RC);
				if(DebugPRA>1)
				{
					std::cerr << "Inserting Store: ";
					std::cerr << *prior(MI);
				}
			}

		}
		break;
		case Move:
		{
			/* need to change the mappings and insert a move */
			unsigned r = map.getRegForQuantity(q);
			unsigned src = map.getMappedReg(r);
			map.set_quantity_alloc_class(q, dst_class, dstr);
			unsigned dst = map.getMappedReg(r);

			if(alloc_classes->getType(src_class) == AllocationClassInfo::Register &&
					alloc_classes->getRegisterIndex(src) != src_class)
				abort();
			if(alloc_classes->getType(dst_class) == AllocationClassInfo::Register &&
					alloc_classes->getRegisterIndex(dst) != dst_class)
				abort();
			if(alloc_classes->getType(dst_class) != AllocationClassInfo::Register)
			{
				/* Store should take care of these */
			}
			else if(src == dst) //??
				abort();
			else
			{
				const TargetRegisterClass *TRC = MF->getSSARegMap()->getRegClass(r);

				assert(src != 0 && dst != 0);

				//handle moves between quantities of different register classes
				//where one is presumably a subset of the other; we don't want
				//to generate a movb esi,al so promote that to esi,eax or vice versa
				if(MRegisterInfo::isPhysicalRegister(src) && !TRC->contains(src))
				{
					//trust our solution (cross fingers)
					const TargetRegisterClass *NewTRC = alloc_classes->getTargetClass(src_class);
					assert(NewTRC->getSize() >= TRC->getSize()); //don't want to lose data
					TRC = NewTRC;
					if(!TRC->contains(dst)
							&& alloc_classes->getType(dst_class) == AllocationClassInfo::Register) //need up upsize
					{
						dst = dstr;
					}
					assert(dst != 0);
				}
				if(MRegisterInfo::isPhysicalRegister(dst) && !TRC->contains(dst))
				{
					//trust our solution (cross fingers)
					const TargetRegisterClass *NewTRC = alloc_classes->getTargetClass(dst_class);
					assert(NewTRC->getSize() >= TRC->getSize()); //don't want to lose data
					TRC = NewTRC;
					if(!TRC->contains(src)
							&& alloc_classes->getType(src_class) == AllocationClassInfo::Register) //need up upsize
					{
						src = srcr;
					}
					assert(src != 0);
				}
				RegInfo->copyRegToReg(*MBB, MI, dst, src, TRC);
				if(DebugPRA>1)
				{
					std::cerr << "Inserting Move: ";
					std::cerr << *prior(MI);
				}
			}

			break;
		}
		default:
		abort();

	}
}

//apply a transformation before the given RAAT
void transformation::apply(RAATInfo& RI, AllocClassMap& map)
{
	apply(RI.begin()->getParent(), RI.begin(), map);
}

//apply a transformation at the end of a block
void transformation::applyEnd(mcnf_block *block, AllocClassMap& map)
{
	if(type == Store || type == Move)
	{
		if(type == Move && alloc_classes->getType(dst_class) != AllocationClassInfo::Register)
			; //does not actually generate code
		else if(block->succs.size() > 1 && block->RAATs.size() > 0
				&& block->RAATs.back().isControlFlow())
			abort(); //requires edge insertion
	}
	//otherwise we rely on control flow instructions not using vars
	apply(block->b, block->b->getFirstTerminator(),map);
}

void transformation::print(std::ostream& out)
{
	out << itostr(q);
	out << ":";

    if(src_class >= 0)
    	out << alloc_classes->getName(src_class);

    if(type == Store)
    	out << "*";
    out << "->";

    if(dst_class >= 0)
    	out << alloc_classes->getName(dst_class);

    if(cost != 0)
    	out << "(" << itostr(cost) << ")";

    out << " (" << position << ", " << level << ", " << order << ")";
}

//record the solution present in the GMCNF model
//overwrites any existing solution, invalidates regmapping data
void Solution::record_solution(GlobalMCNF& mcnf_cfg,  std::vector< std::vector<transformation> >& shufflecode)
{
	unsigned b, nb;
	block_sols.resize(mcnf_cfg.blocks.size());
	for(b = 0, nb = mcnf_cfg.blocks.size(); b < nb; b++)
	{
		mcnf_block *block = mcnf_cfg.blocks[b];
		block_sols[b].block = block;
		block_sols[b].record_solution(shufflecode[block->index]);
	}
}

//record the solution present in the GMCNF model
//overwrites any existing solution, invalidates regmapping data
void Solution::record_solution(GlobalMCNF& mcnf_cfg)
{
	unsigned b, nb;
	block_sols.resize(mcnf_cfg.blocks.size());
	std::vector<transformation> dummy;
	for(b = 0, nb = mcnf_cfg.blocks.size(); b < nb; b++)
	{
		mcnf_block *block = mcnf_cfg.blocks[b];
		block_sols[b].block = block;
		block_sols[b].record_solution(dummy);
	}
}

//return the register for a quantity (for multireg classes)
unsigned block_solution::getRegisterForQuantity(int q, mcnf_node *node, int offset /*=0*/)
{
	if(q < 0)
		return 0;
	if(alloc_classes->getCapacity(node->allocClass) == 1)
	{
		return alloc_classes->getRegister(node->allocClass);
	}
	else if(quantityRegMap.count(q))
	{
		RegMap& map = quantityRegMap[q];
		int level = getNodeIntervalLevel(node)+offset;
		RegMap::iterator pos = map.lower_bound(level);
		if(pos->first != level) //overshot
			pos--;
		return pos->second;
	}
	else
	{
		return 0;
	}
}

unsigned block_solution::getRegisterForQuantityIn(int q, int c) //register into block
{
	if(q < 0)
		return 0;
	if(alloc_classes->getCapacity(c) == 1)
	{
		return alloc_classes->getRegister(c);
	}
	else if(quantityRegMap.count(q))
	{
		RegMap& map = quantityRegMap[q];
		assert(map.begin() != map.end());
		return map.begin()->second;
	}
	else
	{
		return 0;
	}
}


//record the solution in this block overwriting any existing solution
void block_solution::record_solution(std::vector<transformation>& shufflecode)
{
	transforms.clear();
	q_in_alloc.clear();
	q_out_alloc.clear();

	int i,n;
	int prevpos = -1;
	int currpos = -1;

	/* TODO: support multiple memory classes with the antivars */

	/* save regmaps, do this efficiently be swapping
	 */
	quantityRegMap.clear();
	quantityRegMap.swap(block->quantityRegMap);

	/* set the initial and final allocations using
	 the enter and exit nodes */
	n = block->entries.size();
	for(i = 0; i < n; i++)
	{
		int c, nc;
		mcnf_node *node = block->entries[i];

		/* for every quantity that might be allocated to this node */
		nc = node->commodities.size();
		for(c = 0; c < nc; c++)
		{
			if(node->feas_marked[c])
			{
				int q = node->commodities[c];

				if(q> 0)
				{
					assert(q_in_alloc.count(q) == 0 && "Quantitly allocated twice");
					q_in_alloc[q] = node->allocClass;
				}
			}
		}
	}

	n = block->exits.size();
	for(i = 0; i < n; i++)
	{
		int c, nc;
		mcnf_node *node = block->exits[i];

		/* for every quantity that might be allocated to this node */
		nc = node->commodities.size();
		for(c = 0; c < nc; c++)
		{
			if(node->feas_marked[c])
			{
				int q = node->commodities[c];

				if(q> 0)
				{
					assert(q_out_alloc.count(q) == 0 && "Quantitly allocated twice");
					q_out_alloc[q] = node->allocClass;
				}
			}
		}
	}

	/* go through the body and compute the list of
	 transformations */
	n = block->body.size();
	for(i = 0; i < n; i++)
	{
		int c, nc;
		mcnf_node *node = block->body[i];

		/* keep track of what the previous positions was
		 (merged insns make this not a constant offset */
		if(currpos != node->position)
		{
			prevpos = currpos;
			currpos = node->position;
		}

		nc = node->commodities.size();
		/* for every commodity that is allocated through this node */
		for(c = 0; c < nc; c++)
		{
			int q = node->commodities[c];
			if(node->feas_marked[c])
			{

				mcnf_edge *next = node->fsinfo[c].next_edge;
				mcnf_edge *prev = node->fsinfo[c].prev_edge;

				assert(next != NULL && prev != NULL && "broken solution");

				if(q> 0)
				{
					/* see if it is a death */
					if(next->dst->type == nSink)
					{
						transforms.push_back(transformation(Death,
										node->position, node->subindex, next->cost,
										node->allocClass, getRegisterForQuantity(q, node), -1, 0, q));
					}

					/* then if it is a birth */
					if(prev->src->type == nDef)
					{
						int cost = prev->cost;
						if(q> 0 && node->nonanti_costs.size()> 0)
						{
							int pos = node->cindex.get(q);
							cost += node->nonanti_costs[pos];
						}

						transforms.push_back(transformation(Birth, prevpos,
										node->subindex, cost, -1, 0, node->allocClass,
										getRegisterForQuantity(q, node, 1), q));
					}

					/* finally, look for a move */
					if((next->dst->type == nInsn || next->dst->type == nXBar)
							&& next->dst->allocClass != node->allocClass)
					{
						int cost = next->cost;
						if(q> 0 && next->dst->nonanti_costs.size()> 0)
						{
							int pos = next->dst->cindex.get(q);
							cost += next->dst->nonanti_costs[pos];
						}

						unsigned srcr = getRegisterForQuantity(q, node);
						unsigned dstr = getRegisterForQuantity(q, next->dst);
						transforms.push_back(transformation(Move, node->position,
										node->subindex, cost, node->allocClass,
										srcr, next->dst->allocClass, dstr, q));
					}
				}
				else /* q < 0, look for eviction/store */
				{
					if(prev->src->type == nDef &&
							alloc_classes->getType(node->allocClass)
							== AllocationClassInfo::AntiOnly )
					{
						/* starts out as anti, so must store resut of defining insn */
						transforms.push_back(transformation(Store,
										node->position, node->subindex, prev->cost, -1, 0,
										alloc_classes->getAntiClass(node->allocClass), 0, q));
					}
					else if((next->dst->type == nInsn || next->dst->type == nXBar)
							&& next->dst->allocClass != node->allocClass &&
							alloc_classes->getType(next->dst->allocClass) ==
							AllocationClassInfo::AntiOnly &&
							alloc_classes->getType(node->allocClass) ==
							AllocationClassInfo::Memory)
					{
						transforms.push_back(transformation(Store, node->position,
										node->subindex, next->cost, -1, 0, node->allocClass,
										getRegisterForQuantity(-q, node), q));
					}

				}
			}
		}
	}

	transforms.insert(transforms.end(), shufflecode.begin(), shufflecode.end());
	std::sort(transforms.begin(), transforms.end());
}

/* setup the alloc class map for the live in pseudos
 */
void block_solution::setup_block_in(AllocClassMap& map)
{
	for(unsigned i = 0, n = block->liveins_vec.size(); i < n; i++)
	{
		int var = block->liveins_vec[i];
		int q = block->cfg->reg_quantity.get(var);
		assert(q_in_alloc.count(q) > 0);
		int c = q_in_alloc[q];
		map.set_quantity_alloc_class(q, c, getRegisterForQuantityIn(q, c));
		map.set_quantity_reg(q, var);
	}
}

void block_solution::print_solution(std::ostream& out)
{
	int trans, maxtrans;
	int s, ns;
	out << "Block " << block->index << "\t" << block->b->getBasicBlock()->getName() << "\n";
	out << "predecessors:";
	ns = block->preds.size();
	for(s = 0; s < ns; s++)
	{
		mcnf_block *pred = block->preds[s];
		out << " " << pred->index;
	}
	out << "\n";

	out << "successors:";
	ns = block->succs.size();
	for(s = 0; s < ns; s++)
	{
		mcnf_block *succ = block->succs[s];
		out << " " << succ->index;
	}
	out << "\n";

	/* output starting state */

	for(hash_map<commod_type, index_type>::iterator itr = q_in_alloc.begin();
	itr != q_in_alloc.end(); itr++)
	{
		int q = itr->first;
		int allocClass = itr->second;

		block->print_quantity(out, q);
		out << ":" << alloc_classes->getName(allocClass) << " ";
	}

	out << "\n---\n";

	/* output insns with transformations */
	trans = 0;
	maxtrans = transforms.size();

	/* output start of block crossbar transformations */
	while(trans < maxtrans)
	{
		transformation *t = &transforms[trans];
		if(t->position == -1)
		{
			t->print(out);
			out << "\n";
			trans++;
		}
		else
		break;
	}

	for(int i = 0, n = block->RAATs.size(); i < n; i++)
	{
		/* output insn */
		block->RAATs[i].print(out);

		/* output relevant transformations */
		while(trans < maxtrans)
		{
			transformation *t = &transforms[trans];
			if(t->position == i)
			{
				t->print(out);
				out << "\n";
				trans++;
			}
			else if(t->position < i)
				abort();
			else
				break;
		}
	}

	/* output ending state */
	out << "\n---\n";
	for(hash_map<commod_type, index_type>::iterator itr = q_out_alloc.begin();
	itr != q_out_alloc.end(); itr++)
	{
		int q = itr->first;
		int allocClass = itr->second;

		block->print_quantity(out, q);
		out << ":" << alloc_classes->getName(allocClass) << " ";
	}
	out << "\n\n";
}


void Solution::print_solution(std::ostream& out)
{
	int b, numb;
	numb = block_sols.size();

	for (b = 0; b < numb; b++)
	{
		block_solution *bsol = &block_sols[b];
		bsol->print_solution(out);
	}
}

}
