#include "X86AllocationClassInfo.h"
#include "X86TargetMachine.h"

namespace llvm
{


void X86AllocationClassInfo::setupClasses()
{
	addRegisterClass(X86::EAX, &X86::GR32_RegClass);
	addRegisterClass(X86::EDX, &X86::GR32_RegClass);
	addRegisterClass(X86::ECX, &X86::GR32_RegClass);
	addRegisterClass(X86::EBX, &X86::GR32_RegClass);
	addRegisterClass(X86::ESI, &X86::GR32RegClass);
	addRegisterClass(X86::EDI, &X86::GR32RegClass);

	addRegisterClass(X86::FP0,  &X86::RFP80RegClass);
	addRegisterClass(X86::FP1,  &X86::RFP80RegClass);
	addRegisterClass(X86::FP2,  &X86::RFP80RegClass);
	addRegisterClass(X86::FP3,  &X86::RFP80RegClass);
	addRegisterClass(X86::FP4,  &X86::RFP80RegClass);
	addRegisterClass(X86::FP5,  &X86::RFP80RegClass);
	addRegisterClass(X86::FP6,  &X86::RFP80RegClass);

	addRegisterClass(X86::XMM0, &X86::FR64RegClass);
	addRegisterClass(X86::XMM1, &X86::FR64RegClass);
	addRegisterClass(X86::XMM2, &X86::FR64RegClass);
	addRegisterClass(X86::XMM3, &X86::FR64RegClass);
	addRegisterClass(X86::XMM4, &X86::FR64RegClass);
	addRegisterClass(X86::XMM5, &X86::FR64RegClass);
	addRegisterClass(X86::XMM6, &X86::FR64RegClass);
	addRegisterClass(X86::XMM7, &X86::FR64RegClass);

	addMemoryClass(-1, NearMem);
	addConstantClass(LargeConst);

}

X86AllocationClassInfo::X86AllocationClassInfo(const X86TargetMachine &tm):
	AllocationClassInfo(tm)
{
	if(WhatMetric == Speed)
	{
		loadCost = 8;
		storeCost = 8;
		moveCost = 1;
		intConstCost = 1;
		fpConstCost = 1;
		memConstCost = 8;
	}
	else
	{
		loadCost = 3;
		storeCost = 3;
		moveCost = 2;
		intConstCost = 5;
		fpConstCost = 6;
		memConstCost = 7;
	}
}

void X86AllocationClassInfo::initialize()
{
	setupClasses();
	setupCosts();
}

unsigned X86AllocationClassInfo::numCalleeSaves() const
{
	return 3;
}

unsigned X86AllocationClassInfo::numCallerSaves() const
{
	return 3;
}


//stack fp implementation currently doesn't support fp regs being live across blocks
bool X86AllocationClassInfo::usableAcrossBlocks(int c) const
{
	unsigned r = getRegister(c);
	if(r >= X86::FP0 && r <= X86::FP6)
	{
		return false;
	}
	return AllocationClassInfo::usableAcrossBlocks(c);
}

/* Return the cost of moving from class src to dst
 * Negative costs means move is not possible.
 *  */
int X86AllocationClassInfo::computeMoveCost(unsigned src, unsigned dst)
{
	int type1 = getType(src);
	int type2 = getType(dst);

	if(type1 == Register && type2 == Register)
	{
		if(src == dst)
			return 0;
		const TargetRegisterClass *sRC = getTargetClass(src);
		const TargetRegisterClass *dRC = getTargetClass(dst);
		if(sRC == dRC || sRC->hasSubClass(dRC) || dRC->hasSubClass(sRC))
			return moveCost;
	}
	else if(type1 == Memory)
	{
		if(type2 == Register)
			return loadCost;
		else if(type2 == AntiOnly)
			return storeCost;
	}
	else if(type1 == Constant)
	{
		if(type2 == Register) //load cost
		{
			if(getTargetClass(dst) == &X86::GR32_RegClass || getTargetClass(dst) == &X86::GR32RegClass)
				return intConstCost;
			else
				return fpConstCost;
		}
		else if(type2 == Memory)
			return memConstCost - storeCost; //store cost paid by anti
	}

	return AllocationClassInfo::computeMoveCost(src,dst);
}
int X86AllocationClassInfo::getInsnCost(MachineInstr *MI, int op_index, int c) const
{
	if(WhatMetric == Speed)
	{
		return getInsnSpeedCost(MI, op_index, c);
	}
	else
	{
		return getInsnSizeCost(MI, op_index, c);
	}
}

/* Handle X86 specific instruction costs here */
int X86AllocationClassInfo::getInsnSizeCost(MachineInstr *MI, int op_index, int c) const
{
	int ret = 0;
	int type = getType(c);
	int op = MI->getOpcode();

	if(type == Memory)
	{
		//one additional byte assuming short indexing off of ebp
		ret = 1;
		if(MI->hasSavedReg())
		{
			//a push/pop is actually a byte smaller than a move
			ret = -1;
		}
		//if this is redefining the op, just provide a cost for the first operand
		else if(op_index > 0 && MI->getOperand(op_index).isReg() &&
				MI->getOperand(0).isReg() &&
				MI->getOperand(op_index).getReg() == MI->getOperand(0).getReg())
			ret = 0;
		//a MOVr0 is implemented as 2-byte xor if the dst is a reg,
		//but a 7 byte move if the dest is memory
		else if(op == X86::MOV32r0 || op == X86::MOV16r0 || op == X86::MOV8r0)
			ret = 5;
		// INCr/DECL is a one byte insn if the op is a register, 3 bytes otherwise
		else if(op == X86::INC16r || op == X86::INC32r || op == X86::INC8r ||
				op == X86::DEC16r || op == X86::DEC32r || op == X86::DEC8r)
			ret = 2;
		//moving a constant to memory take an extra byte
		else if(op == X86::MOV32ri || op == X86::MOV16ri)
			ret = 2;
	}

	//eax sometimes saves a byte
	if(getCapacity(c) == 1 && getRegister(c) == X86::EAX)
	{
		if(op == X86::TEST32ri || op == X86::TEST8ri || op == X86::TEST16ri ||
				op == X86::CMP16ri || op == X86::CMP32ri || op == X86::CMP8ri)
			ret = -1;

	}

	return ret + AllocationClassInfo::getInsnCost(MI, op_index, c);
}


/* Handle X86 specific instruction costs here */
int X86AllocationClassInfo::getInsnSpeedCost(MachineInstr *MI, int op_index, int c) const
{
	int ret = 0;
	int type = getType(c);

	if(type == Memory)
	{
		ret = MI->getOperand(op_index).isDef() ? storeCost : loadCost;
		ret -= moveCost; //slightly prefer accessing from within the insn
		if(MI->getOperand(op_index).isRegister())
		{
			//check for a modifying operation since only the first cost gets counted
			unsigned r = MI->getOperand(op_index).getReg();
			for (int i = 0, e = MI->getNumOperands(); i != e; ++i)
			{
				//register operand?
				if(i == op_index)
					continue;
				MachineOperand &MO = MI->getOperand(i);
				if(MO.isRegister() && MO.getReg() == r)
				{
					ret += MI->getOperand(op_index).isDef() ? storeCost : loadCost;
					break;
				}
			}
		}
	}

	return ret + AllocationClassInfo::getInsnCost(MI, op_index, c);
}

}
