//===- X86InstrInfo.cpp - X86 Instruction Information -----------*- C++ -*-===//
//
//                     The LLVM Compiler Infrastructure
//
// This file was developed by the LLVM research group and is distributed under
// the University of Illinois Open Source License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
//
// This file contains the X86 implementation of the TargetInstrInfo class.
//
//===----------------------------------------------------------------------===//

#include "X86InstrInfo.h"
#include "X86.h"
#include "X86GenInstrInfo.inc"
#include "X86InstrBuilder.h"
#include "X86Subtarget.h"
#include "X86TargetMachine.h"
#include "llvm/ADT/STLExtras.h"
#include "llvm/CodeGen/MachineInstrBuilder.h"
#include "llvm/Target/TargetOptions.h"
#include "llvm/CodeGen/MachineFunction.h"
#include "llvm/CodeGen/MachineFrameInfo.h"
#include <iostream>
#include "llvm/CodeGen/LiveVariables.h"
#include "X86MachineFunctionInfo.h"
#include "llvm/Target/TargetOptions.h" 	
#include "llvm/CodeGen/MachineFunction.h" 	 
#include "llvm/CodeGen/MachineFrameInfo.h"
#include "llvm/CodeGen/SSARegMap.h"

using namespace llvm;


X86InstrInfo::X86InstrInfo(X86TargetMachine &tm)
  : TargetInstrInfo(X86Insts, array_lengthof(X86Insts)),
    TM(tm), RI(tm, *this) {
}

bool X86InstrInfo::isMoveInstr(const MachineInstr& MI,
                               unsigned& sourceReg,
                               unsigned& destReg) const {
  MachineOpCode oc = MI.getOpcode();
  if (oc == X86::MOV8rr || oc == X86::MOV16rr ||
      oc == X86::MOV32rr || oc == X86::MOV64rr ||
      oc == X86::MOV16to16_ || oc == X86::MOV32to32_ ||
      oc == X86::MOV_Fp3232  || oc == X86::MOVSSrr || oc == X86::MOVSDrr ||
      oc == X86::MOV_Fp3264 || oc == X86::MOV_Fp6432 || oc == X86::MOV_Fp6464 ||
      oc == X86::FsMOVAPSrr || oc == X86::FsMOVAPDrr ||
      oc == X86::MOVAPSrr || oc == X86::MOVAPDrr ||
      oc == X86::MOVSS2PSrr || oc == X86::MOVSD2PDrr ||
      oc == X86::MOVPS2SSrr || oc == X86::MOVPD2SDrr ||
      oc == X86::MMX_MOVD64rr || oc == X86::MMX_MOVQ64rr) {
      assert(MI.getNumOperands() >= 2 &&
             MI.getOperand(0).isRegister() &&
             MI.getOperand(1).isRegister() &&
             "invalid register-register move instruction");
      sourceReg = MI.getOperand(1).getReg();
      destReg = MI.getOperand(0).getReg();
      return true;
  }
  return false;
}

//dkoes - return true if MI is coalesable
bool X86InstrInfo::isCoalesceable(const MachineInstr& MI,
        unsigned& sourceReg,
        unsigned& destReg) const
{
	MachineOpCode oc = MI.getOpcode();
	if(oc == X86::EXTRACT_SUBREG)
	{
		 sourceReg = MI.getOperand(1).getReg();
		 destReg = MI.getOperand(0).getReg();
		return true; 
	}
	if(isMoveInstr(MI, sourceReg, destReg))
		return true;
	return false;
}
	
unsigned X86InstrInfo::isLoadFromStackSlot(MachineInstr *MI, 
                                           int &FrameIndex) const {
  switch (MI->getOpcode()) {
  default: break;
  case X86::MOV8rm:
  case X86::MOV16rm:
  case X86::MOV16_rm:
  case X86::MOV32rm:
  case X86::MOV32_rm:
  case X86::MOV64rm:
  case X86::LD_Fp64m:
  case X86::MOVSSrm:
  case X86::MOVSDrm:
  case X86::MOVAPSrm:
  case X86::MOVAPDrm:
  case X86::MMX_MOVD64rm:
  case X86::MMX_MOVQ64rm:
    if (MI->getOperand(1).isFrameIndex() && MI->getOperand(2).isImmediate() &&
        MI->getOperand(3).isRegister() && MI->getOperand(4).isImmediate() &&
        MI->getOperand(2).getImmedValue() == 1 &&
        MI->getOperand(3).getReg() == 0 &&
        MI->getOperand(4).getImmedValue() == 0) {
      FrameIndex = MI->getOperand(1).getFrameIndex();
      return MI->getOperand(0).getReg();
    }
    break;
  }
  return 0;
}

unsigned X86InstrInfo::isStoreToStackSlot(MachineInstr *MI,
                                          int &FrameIndex) const {
  switch (MI->getOpcode()) {
  default: break;
  case X86::MOV8mr:
  case X86::MOV16mr:
  case X86::MOV16_mr:
  case X86::MOV32mr:
  case X86::MOV32_mr:
  case X86::MOV64mr:
  case X86::ST_FpP64m:
  case X86::MOVSSmr:
  case X86::MOVSDmr:
  case X86::MOVAPSmr:
  case X86::MOVAPDmr:
  case X86::MMX_MOVD64mr:
  case X86::MMX_MOVQ64mr:
  case X86::MMX_MOVNTQmr:
    if (MI->getOperand(0).isFrameIndex() && MI->getOperand(1).isImmediate() &&
        MI->getOperand(2).isRegister() && MI->getOperand(3).isImmediate() &&
        MI->getOperand(1).getImmedValue() == 1 &&
        MI->getOperand(2).getReg() == 0 &&
        MI->getOperand(3).getImmedValue() == 0) {
      FrameIndex = MI->getOperand(0).getFrameIndex();
      return MI->getOperand(4).getReg();
    }
    break;
  }
  return 0;
}


bool X86InstrInfo::isReallyTriviallyReMaterializable(MachineInstr *MI) const {
  switch (MI->getOpcode()) {
  default: break;
  case X86::MOV8rm:
  case X86::MOV16rm:
  case X86::MOV16_rm:
  case X86::MOV32rm:
  case X86::MOV32_rm:
  case X86::MOV64rm:
  case X86::LD_Fp64m:
  case X86::MOVSSrm:
  case X86::MOVSDrm:
  case X86::MOVAPSrm:
  case X86::MOVAPDrm:
  case X86::MMX_MOVD64rm:
  case X86::MMX_MOVQ64rm:
    // Loads from constant pools are trivially rematerializable.
    return MI->getOperand(1).isRegister() && MI->getOperand(2).isImmediate() &&
           MI->getOperand(3).isRegister() && MI->getOperand(4).isConstantPoolIndex() &&
           MI->getOperand(1).getReg() == 0 &&
           MI->getOperand(2).getImmedValue() == 1 &&
           MI->getOperand(3).getReg() == 0;
  }
  // All other instructions marked M_REMATERIALIZABLE are always trivially
  // rematerializable.
  return true;
}

/// convertToThreeAddress - This method must be implemented by targets that
/// set the M_CONVERTIBLE_TO_3_ADDR flag.  When this flag is set, the target
/// may be able to convert a two-address instruction into a true
/// three-address instruction on demand.  This allows the X86 target (for
/// example) to convert ADD and SHL instructions into LEA instructions if they
/// would require register copies due to two-addressness.
///
/// This method returns a null pointer if the transformation cannot be
/// performed, otherwise it returns the new instruction.
///
MachineInstr *
X86InstrInfo::convertToThreeAddress(MachineFunction::iterator &MFI,
                                    MachineBasicBlock::iterator &MBBI,
                                    LiveVariables &LV) const {
  MachineInstr *MI = MBBI;
  // All instructions input are two-addr instructions.  Get the known operands.
  unsigned Dest = MI->getOperand(0).getReg();
  unsigned Src = MI->getOperand(1).getReg();

  MachineInstr *NewMI = NULL;
  // FIXME: 16-bit LEA's are really slow on Athlons, but not bad on P4's.  When
  // we have better subtarget support, enable the 16-bit LEA generation here.
  bool DisableLEA16 = true;

  switch (MI->getOpcode()) {
  default: return 0;
  case X86::SHUFPSrri: {
    assert(MI->getNumOperands() == 4 && "Unknown shufps instruction!");
    if (!TM.getSubtarget<X86Subtarget>().hasSSE2()) return 0;
    
    unsigned A = MI->getOperand(0).getReg();
    unsigned B = MI->getOperand(1).getReg();
    unsigned C = MI->getOperand(2).getReg();
    unsigned M = MI->getOperand(3).getImm();
    if (B != C) return 0;
    NewMI = BuildMI(get(X86::PSHUFDri), A).addReg(B).addImm(M);
    break;
  }
  case X86::SHL64ri: {
    assert(MI->getNumOperands() == 3 && "Unknown shift instruction!");
    // NOTE: LEA doesn't produce flags like shift does, but LLVM never uses
    // the flags produced by a shift yet, so this is safe.
    unsigned Dest = MI->getOperand(0).getReg();
    unsigned Src = MI->getOperand(1).getReg();
    unsigned ShAmt = MI->getOperand(2).getImm();
    if (ShAmt == 0 || ShAmt >= 4) return 0;
    
    NewMI = BuildMI(get(X86::LEA64r), Dest)
      .addReg(0).addImm(1 << ShAmt).addReg(Src).addImm(0);
    break;
  }
  case X86::SHL32ri: {
    assert(MI->getNumOperands() == 3 && "Unknown shift instruction!");
    // NOTE: LEA doesn't produce flags like shift does, but LLVM never uses
    // the flags produced by a shift yet, so this is safe.
    unsigned Dest = MI->getOperand(0).getReg();
    unsigned Src = MI->getOperand(1).getReg();
    unsigned ShAmt = MI->getOperand(2).getImm();
    if (ShAmt == 0 || ShAmt >= 4) return 0;
    
    unsigned Opc = TM.getSubtarget<X86Subtarget>().is64Bit() ?
      X86::LEA64_32r : X86::LEA32r;
    NewMI = BuildMI(get(Opc), Dest)
      .addReg(0).addImm(1 << ShAmt).addReg(Src).addImm(0);
    break;
  }
  case X86::SHL16ri: {
    assert(MI->getNumOperands() == 3 && "Unknown shift instruction!");
    // NOTE: LEA doesn't produce flags like shift does, but LLVM never uses
    // the flags produced by a shift yet, so this is safe.
    unsigned Dest = MI->getOperand(0).getReg();
    unsigned Src = MI->getOperand(1).getReg();
    unsigned ShAmt = MI->getOperand(2).getImm();
    if (ShAmt == 0 || ShAmt >= 4) return 0;
    
    if (DisableLEA16) {
      // If 16-bit LEA is disabled, use 32-bit LEA via subregisters.
      SSARegMap *RegMap = MFI->getParent()->getSSARegMap();
      unsigned Opc = TM.getSubtarget<X86Subtarget>().is64Bit()
        ? X86::LEA64_32r : X86::LEA32r;
      unsigned leaInReg = RegMap->createVirtualRegister(&X86::GR32RegClass);
      unsigned leaOutReg = RegMap->createVirtualRegister(&X86::GR32RegClass);
            
      MachineInstr *Ins =
        BuildMI(get(X86::INSERT_SUBREG), leaInReg).addReg(Src).addImm(2);
      Ins->copyKillDeadInfo(MI);
      
      NewMI = BuildMI(get(Opc), leaOutReg)
        .addReg(0).addImm(1 << ShAmt).addReg(leaInReg).addImm(0);
      
      MachineInstr *Ext =
        BuildMI(get(X86::EXTRACT_SUBREG), Dest).addReg(leaOutReg).addImm(2);
      Ext->copyKillDeadInfo(MI);
      
      MFI->insert(MBBI, Ins);            // Insert the insert_subreg
      LV.instructionChanged(MI, NewMI);  // Update live variables
      LV.addVirtualRegisterKilled(leaInReg, NewMI);
      MFI->insert(MBBI, NewMI);          // Insert the new inst
      LV.addVirtualRegisterKilled(leaOutReg, Ext);
      MFI->insert(MBBI, Ext);            // Insert the extract_subreg      
      return Ext;
    } else {
      NewMI = BuildMI(get(X86::LEA16r), Dest)
        .addReg(0).addImm(1 << ShAmt).addReg(Src).addImm(0);
    }
    break;
  }
  }

  // FIXME: None of these instructions are promotable to LEAs without
  // additional information.  In particular, LEA doesn't set the flags that
  // add and inc do.  :(
  if (0)
  switch (MI->getOpcode()) {
  case X86::INC32r:
  case X86::INC64_32r:
    assert(MI->getNumOperands() == 2 && "Unknown inc instruction!");
    NewMI = addRegOffset(BuildMI(get(X86::LEA32r), Dest), Src, 1);
    break;
  case X86::INC16r:
  case X86::INC64_16r:
    if (DisableLEA16) return 0;
    assert(MI->getNumOperands() == 2 && "Unknown inc instruction!");
    NewMI = addRegOffset(BuildMI(get(X86::LEA16r), Dest), Src, 1);
    break;
  case X86::DEC32r:
  case X86::DEC64_32r:
    assert(MI->getNumOperands() == 2 && "Unknown dec instruction!");
    NewMI = addRegOffset(BuildMI(get(X86::LEA32r), Dest), Src, -1);
    break;
  case X86::DEC16r:
  case X86::DEC64_16r:
    if (DisableLEA16) return 0;
    assert(MI->getNumOperands() == 2 && "Unknown dec instruction!");
    NewMI = addRegOffset(BuildMI(get(X86::LEA16r), Dest), Src, -1);
    break;
  case X86::ADD32rr:
    assert(MI->getNumOperands() == 3 && "Unknown add instruction!");
    NewMI = addRegReg(BuildMI(get(X86::LEA32r), Dest), Src,
                     MI->getOperand(2).getReg());
    break;
  case X86::ADD16rr:
    if (DisableLEA16) return 0;
    assert(MI->getNumOperands() == 3 && "Unknown add instruction!");
    NewMI = addRegReg(BuildMI(get(X86::LEA16r), Dest), Src,
                     MI->getOperand(2).getReg());
    break;
  case X86::ADD32ri:
  case X86::ADD32ri8:
    assert(MI->getNumOperands() == 3 && "Unknown add instruction!");
    if (MI->getOperand(2).isImmediate())
      NewMI = addRegOffset(BuildMI(get(X86::LEA32r), Dest), Src,
                          MI->getOperand(2).getImmedValue());
    break;
  case X86::ADD16ri:
  case X86::ADD16ri8:
    if (DisableLEA16) return 0;
    assert(MI->getNumOperands() == 3 && "Unknown add instruction!");
    if (MI->getOperand(2).isImmediate())
      NewMI = addRegOffset(BuildMI(get(X86::LEA16r), Dest), Src,
                          MI->getOperand(2).getImmedValue());
    break;
  case X86::SHL16ri:
    if (DisableLEA16) return 0;
  case X86::SHL32ri:
    assert(MI->getNumOperands() == 3 && MI->getOperand(2).isImmediate() &&
           "Unknown shl instruction!");
    unsigned ShAmt = MI->getOperand(2).getImmedValue();
    if (ShAmt == 1 || ShAmt == 2 || ShAmt == 3) {
      X86AddressMode AM;
      AM.Scale = 1 << ShAmt;
      AM.IndexReg = Src;
      unsigned Opc = MI->getOpcode() == X86::SHL32ri ? X86::LEA32r :X86::LEA16r;
      NewMI = addFullAddress(BuildMI(get(Opc), Dest), AM);
    }
    break;
  }

  if (NewMI) {
    NewMI->copyKillDeadInfo(MI);
    LV.instructionChanged(MI, NewMI);  // Update live variables
    MFI->insert(MBBI, NewMI);          // Insert the new inst    
  }
  return NewMI;
}

/// commuteInstruction - We have a few instructions that must be hacked on to
/// commute them.
///
MachineInstr *X86InstrInfo::commuteInstruction(MachineInstr *MI) const {
  // FIXME: Can commute cmoves by changing the condition!
  switch (MI->getOpcode()) {
  case X86::SHRD16rri8: // A = SHRD16rri8 B, C, I -> A = SHLD16rri8 C, B, (16-I)
  case X86::SHLD16rri8: // A = SHLD16rri8 B, C, I -> A = SHRD16rri8 C, B, (16-I)
  case X86::SHRD32rri8: // A = SHRD32rri8 B, C, I -> A = SHLD32rri8 C, B, (32-I)
  case X86::SHLD32rri8:{// A = SHLD32rri8 B, C, I -> A = SHRD32rri8 C, B, (32-I)
    unsigned Opc;
    unsigned Size;
    switch (MI->getOpcode()) {
    default: assert(0 && "Unreachable!");
    case X86::SHRD16rri8: Size = 16; Opc = X86::SHLD16rri8; break;
    case X86::SHLD16rri8: Size = 16; Opc = X86::SHRD16rri8; break;
    case X86::SHRD32rri8: Size = 32; Opc = X86::SHLD32rri8; break;
    case X86::SHLD32rri8: Size = 32; Opc = X86::SHRD32rri8; break;
    }
    unsigned Amt = MI->getOperand(3).getImmedValue();
    unsigned A = MI->getOperand(0).getReg();
    unsigned B = MI->getOperand(1).getReg();
    unsigned C = MI->getOperand(2).getReg();
    bool BisKill = MI->getOperand(1).isKill();
    bool CisKill = MI->getOperand(2).isKill();
    return BuildMI(get(Opc), A).addReg(C, false, false, CisKill)
      .addReg(B, false, false, BisKill).addImm(Size-Amt);
  }
  default:
    return TargetInstrInfo::commuteInstruction(MI);
  }
}

static X86::CondCode GetCondFromBranchOpc(unsigned BrOpc)
{
	switch (BrOpc)
	{
	default:
		return X86::COND_INVALID;
	case X86::JE:
		return X86::COND_E;
	case X86::JNE:
		return X86::COND_NE;
	case X86::JL:
		return X86::COND_L;
	case X86::JLE:
		return X86::COND_LE;
	case X86::JG:
		return X86::COND_G;
	case X86::JGE:
		return X86::COND_GE;
	case X86::JB:
		return X86::COND_B;
	case X86::JBE:
		return X86::COND_BE;
	case X86::JA:
		return X86::COND_A;
	case X86::JAE:
		return X86::COND_AE;
	case X86::JS:
		return X86::COND_S;
	case X86::JNS:
		return X86::COND_NS;
	case X86::JP:
		return X86::COND_P;
	case X86::JNP:
		return X86::COND_NP;
	case X86::JO:
		return X86::COND_O;
	case X86::JNO:
		return X86::COND_NO;
	}
}

unsigned X86::GetCondBranchFromCond(X86::CondCode CC)
{
	switch (CC)
	{
	default:
		assert(0 && "Illegal condition code!");
	case X86::COND_E:
		return X86::JE;
	case X86::COND_NE:
		return X86::JNE;
	case X86::COND_L:
		return X86::JL;
	case X86::COND_LE:
		return X86::JLE;
	case X86::COND_G:
		return X86::JG;
	case X86::COND_GE:
		return X86::JGE;
	case X86::COND_B:
		return X86::JB;
	case X86::COND_BE:
		return X86::JBE;
	case X86::COND_A:
		return X86::JA;
	case X86::COND_AE:
		return X86::JAE;
	case X86::COND_S:
		return X86::JS;
	case X86::COND_NS:
		return X86::JNS;
	case X86::COND_P:
		return X86::JP;
	case X86::COND_NP:
		return X86::JNP;
	case X86::COND_O:
		return X86::JO;
	case X86::COND_NO:
		return X86::JNO;
	}
}

/// GetOppositeBranchCondition - Return the inverse of the specified condition,
/// e.g. turning COND_E to COND_NE.
X86::CondCode X86::GetOppositeBranchCondition(X86::CondCode CC)
{
	switch (CC)
	{
	default:
		assert(0 && "Illegal condition code!");
	case X86::COND_E:
		return X86::COND_NE;
	case X86::COND_NE:
		return X86::COND_E;
	case X86::COND_L:
		return X86::COND_GE;
	case X86::COND_LE:
		return X86::COND_G;
	case X86::COND_G:
		return X86::COND_LE;
	case X86::COND_GE:
		return X86::COND_L;
	case X86::COND_B:
		return X86::COND_AE;
	case X86::COND_BE:
		return X86::COND_A;
	case X86::COND_A:
		return X86::COND_BE;
	case X86::COND_AE:
		return X86::COND_B;
	case X86::COND_S:
		return X86::COND_NS;
	case X86::COND_NS:
		return X86::COND_S;
	case X86::COND_P:
		return X86::COND_NP;
	case X86::COND_NP:
		return X86::COND_P;
	case X86::COND_O:
		return X86::COND_NO;
	case X86::COND_NO:
		return X86::COND_O;
	}
}

bool X86InstrInfo::isUnpredicatedTerminator(const MachineInstr *MI) const {
  const TargetInstrDescriptor *TID = MI->getInstrDescriptor();
  if (TID->Flags & M_TERMINATOR_FLAG) {
    // Conditional branch is a special case.
    if ((TID->Flags & M_BRANCH_FLAG) != 0 && (TID->Flags & M_BARRIER_FLAG) == 0)
      return true;
    if ((TID->Flags & M_PREDICABLE) == 0)
      return true;
    return !isPredicated(MI);
  }
  return false;
}

// For purposes of branch analysis do not count FP_REG_KILL as a terminator.
static bool isBrAnalysisUnpredicatedTerminator(const MachineInstr *MI,
                                               const X86InstrInfo &TII) {
  if (MI->getOpcode() == X86::FP_REG_KILL)
    return false;
  return TII.isUnpredicatedTerminator(MI);
}

bool X86InstrInfo::AnalyzeBranch(MachineBasicBlock &MBB, 
                                 MachineBasicBlock *&TBB,
                                 MachineBasicBlock *&FBB,
                                 std::vector<MachineOperand> &Cond) const {
  // If the block has no terminators, it just falls into the block after it.
  MachineBasicBlock::iterator I = MBB.end();
  if (I == MBB.begin() || !isBrAnalysisUnpredicatedTerminator(--I, *this))
    return false;

  // Get the last instruction in the block.
  MachineInstr *LastInst = I;
  
  // If there is only one terminator instruction, process it.
  if (I == MBB.begin() || !isBrAnalysisUnpredicatedTerminator(--I, *this)) {
    if (!isBranch(LastInst->getOpcode()))
      return true;
    
    // If the block ends with a branch there are 3 possibilities:
    // it's an unconditional, conditional, or indirect branch.
    
    if (LastInst->getOpcode() == X86::JMP) {
      TBB = LastInst->getOperand(0).getMachineBasicBlock();
      return false;
    }
    X86::CondCode BranchCode = GetCondFromBranchOpc(LastInst->getOpcode());
    if (BranchCode == X86::COND_INVALID)
      return true;  // Can't handle indirect branch.

    // Otherwise, block ends with fall-through condbranch.
    TBB = LastInst->getOperand(0).getMachineBasicBlock();
    Cond.push_back(MachineOperand::CreateImm(BranchCode));
    return false;
  }
  
  // Get the instruction before it if it's a terminator.
  MachineInstr *SecondLastInst = I;
  
  // If there are three terminators, we don't know what sort of block this is.
  if (SecondLastInst && I != MBB.begin() &&
      isBrAnalysisUnpredicatedTerminator(--I, *this))
    return true;

	// If the block ends with X86::JMP and a conditional branch, handle it.
	X86::CondCode
			BranchCode = GetCondFromBranchOpc(SecondLastInst->getOpcode());
	if (BranchCode != X86::COND_INVALID && LastInst->getOpcode() == X86::JMP)
	{
		TBB = SecondLastInst->getOperand(0).getMachineBasicBlock();
		Cond.push_back(MachineOperand::CreateImm(BranchCode));
		FBB = LastInst->getOperand(0).getMachineBasicBlock();
		return false;
	}

  // If the block ends with two X86::JMPs, handle it.  The second one is not
  // executed, so remove it.
  if (SecondLastInst->getOpcode() == X86::JMP && 
      LastInst->getOpcode() == X86::JMP) {
    TBB = SecondLastInst->getOperand(0).getMachineBasicBlock();
    I = LastInst;
    I->eraseFromParent();
    return false;
  }

  // Otherwise, can't handle this.
  return true;
}

unsigned X86InstrInfo::RemoveBranch(MachineBasicBlock &MBB) const {
  MachineBasicBlock::iterator I = MBB.end();
  if (I == MBB.begin()) return 0;
  --I;
  if (I->getOpcode() != X86::JMP && 
      GetCondFromBranchOpc(I->getOpcode()) == X86::COND_INVALID)
    return 0;
  
  // Remove the branch.
  I->eraseFromParent();
  
  I = MBB.end();
  
  if (I == MBB.begin()) return 1;
  --I;
  if (GetCondFromBranchOpc(I->getOpcode()) == X86::COND_INVALID)
    return 1;
  
  // Remove the branch.
  I->eraseFromParent();
  return 2;
}

unsigned
X86InstrInfo::InsertBranch(MachineBasicBlock &MBB, MachineBasicBlock *TBB,
                           MachineBasicBlock *FBB,
                           const std::vector<MachineOperand> &Cond) const {
  // Shouldn't be a fall through.
  assert(TBB && "InsertBranch must not be told to insert a fallthrough");
  assert((Cond.size() == 1 || Cond.size() == 0) &&
         "X86 branch conditions have one component!");

  if (FBB == 0) { // One way branch.
    if (Cond.empty()) {
      // Unconditional branch?
      BuildMI(&MBB, get(X86::JMP)).addMBB(TBB);
    } else {
      // Conditional branch.
      unsigned Opc = GetCondBranchFromCond((X86::CondCode)Cond[0].getImm());
      BuildMI(&MBB, get(Opc)).addMBB(TBB);
    }
    return 1;
  }
  
  // Two-way Conditional branch.
  unsigned Opc = GetCondBranchFromCond((X86::CondCode)Cond[0].getImm());
  BuildMI(&MBB, get(Opc)).addMBB(TBB);
  BuildMI(&MBB, get(X86::JMP)).addMBB(FBB);
  return 2;
}

bool X86InstrInfo::BlockHasNoFallThrough(MachineBasicBlock &MBB) const {
  if (MBB.empty()) return false;
  
  switch (MBB.back().getOpcode()) {
  case X86::RET:     // Return.
  case X86::RETI:
  case X86::TAILJMPd:
  case X86::TAILJMPr:
  case X86::TAILJMPm:
  case X86::JMP:     // Uncond branch.
  case X86::JMP32r:  // Indirect branch.
  case X86::JMP32m:  // Indirect branch through mem.
    return true;
  default: return false;
  }
}

bool X86InstrInfo::ReverseBranchCondition(std::vector<MachineOperand> &Cond) const
{
	assert(Cond.size() == 1 && "Invalid X86 branch condition!");
	Cond[0].setImm(GetOppositeBranchCondition((X86::CondCode)Cond[0].getImm()));
	return false;
}

const TargetRegisterClass *X86InstrInfo::getPointerRegClass() const
{
	const X86Subtarget *Subtarget = &TM.getSubtarget<X86Subtarget>();
	if (Subtarget->is64Bit())
		return &X86::GR64RegClass;
	else
		return &X86::GR32RegClass;
}

/// isDisp8 - Return true if this signed displacement fits in a 8-bit 
/// sign-extended field. 
static bool isDisp8(int Value)
{
	return Value == (signed char)Value;
}


/// isX86_64ExtendedReg - Is the MachineOperand a x86-64 extended register?
/// e.g. r8, xmm8, etc.
bool X86InstrInfo::isX86_64ExtendedReg(const MachineOperand &MO) const {
  if (!MO.isRegister()) return false;
  unsigned RegNo = MO.getReg();
  int DWNum = getRegisterInfo().getDwarfRegNum(RegNo);
  if (DWNum >= getRegisterInfo().getDwarfRegNum(X86::R8) &&
      DWNum <= getRegisterInfo().getDwarfRegNum(X86::R15))
    return true;
  if (DWNum >= getRegisterInfo().getDwarfRegNum(X86::XMM8) &&
      DWNum <= getRegisterInfo().getDwarfRegNum(X86::XMM15))
    return true;
  return false;
}

inline static bool isX86_64NonExtLowByteReg(unsigned reg) {
  return (reg == X86::SPL || reg == X86::BPL ||
          reg == X86::SIL || reg == X86::DIL);
}

/// determineREX - Determine if the MachineInstr has to be encoded with a X86-64
/// REX prefix which specifies 1) 64-bit instructions, 2) non-default operand
/// size, and 3) use of X86-64 extended registers.
unsigned X86InstrInfo::determineREX(const MachineInstr &MI) const {
  unsigned REX = 0;
  const TargetInstrDescriptor *Desc = MI.getInstrDescriptor();

  // Pseudo instructions do not need REX prefix byte.
  if ((Desc->TSFlags & X86II::FormMask) == X86II::Pseudo)
    return 0;
  if (Desc->TSFlags & X86II::REX_W)
    REX |= 1 << 3;

  unsigned NumOps = Desc->numOperands;
  if (NumOps) {
    bool isTwoAddr = NumOps > 1 &&
      Desc->getOperandConstraint(1, TOI::TIED_TO) != -1;

    // If it accesses SPL, BPL, SIL, or DIL, then it requires a 0x40 REX prefix.
    unsigned i = isTwoAddr ? 1 : 0;
    for (unsigned e = NumOps; i != e; ++i) {
      const MachineOperand& MO = MI.getOperand(i);
      if (MO.isRegister()) {
        unsigned Reg = MO.getReg();
        if (isX86_64NonExtLowByteReg(Reg))
          REX |= 0x40;
      }
    }

    switch (Desc->TSFlags & X86II::FormMask) {
    case X86II::MRMInitReg:
      if (isX86_64ExtendedReg(MI.getOperand(0)))
        REX |= (1 << 0) | (1 << 2);
      break;
    case X86II::MRMSrcReg: {
      if (isX86_64ExtendedReg(MI.getOperand(0)))
        REX |= 1 << 2;
      i = isTwoAddr ? 2 : 1;
      for (unsigned e = NumOps; i != e; ++i) {
        const MachineOperand& MO = MI.getOperand(i);
        if (isX86_64ExtendedReg(MO))
          REX |= 1 << 0;
      }
      break;
    }
    case X86II::MRMSrcMem: {
      if (isX86_64ExtendedReg(MI.getOperand(0)))
        REX |= 1 << 2;
      unsigned Bit = 0;
      i = isTwoAddr ? 2 : 1;
      for (; i != NumOps; ++i) {
        const MachineOperand& MO = MI.getOperand(i);
        if (MO.isRegister()) {
          if (isX86_64ExtendedReg(MO))
            REX |= 1 << Bit;
          Bit++;
        }
      }
      break;
    }
    case X86II::MRM0m: case X86II::MRM1m:
    case X86II::MRM2m: case X86II::MRM3m:
    case X86II::MRM4m: case X86II::MRM5m:
    case X86II::MRM6m: case X86II::MRM7m:
    case X86II::MRMDestMem: {
      unsigned e = isTwoAddr ? 5 : 4;
      i = isTwoAddr ? 1 : 0;
      if (NumOps > e && isX86_64ExtendedReg(MI.getOperand(e)))
        REX |= 1 << 2;
      unsigned Bit = 0;
      for (; i != e; ++i) {
        const MachineOperand& MO = MI.getOperand(i);
        if (MO.isRegister()) {
          if (isX86_64ExtendedReg(MO))
            REX |= 1 << Bit;
          Bit++;
        }
      }
      break;
    }
    default: {
      if (isX86_64ExtendedReg(MI.getOperand(0)))
        REX |= 1 << 0;
      i = isTwoAddr ? 2 : 1;
      for (unsigned e = NumOps; i != e; ++i) {
        const MachineOperand& MO = MI.getOperand(i);
        if (isX86_64ExtendedReg(MO))
          REX |= 1 << 2;
      }
      break;
    }
    }
  }
  return REX;
}


// getX86RegNum - This function maps LLVM register identifiers to their X86
// specific numbering, which is used in various places encoding instructions.
//
unsigned X86InstrInfo::getX86RegNum(unsigned RegNo) const
{
	switch (RegNo)
	{
	case X86::RAX:
	case X86::EAX:
	case X86::AX:
	case X86::AL:
		return N86::EAX;
	case X86::RCX:
	case X86::ECX:
	case X86::CX:
	case X86::CL:
		return N86::ECX;
	case X86::RDX:
	case X86::EDX:
	case X86::DX:
	case X86::DL:
		return N86::EDX;
	case X86::RBX:
	case X86::EBX:
	case X86::BX:
	case X86::BL:
		return N86::EBX;
	case X86::RSP:
	case X86::ESP:
	case X86::SP:
	case X86::SPL:
	case X86::AH:
		return N86::ESP;
	case X86::RBP:
	case X86::EBP:
	case X86::BP:
	case X86::BPL:
	case X86::CH:
		return N86::EBP;
	case X86::RSI:
	case X86::ESI:
	case X86::SI:
	case X86::SIL:
	case X86::DH:
		return N86::ESI;
	case X86::RDI:
	case X86::EDI:
	case X86::DI:
	case X86::DIL:
	case X86::BH:
		return N86::EDI;

	case X86::R8:
	case X86::R8D:
	case X86::R8W:
	case X86::R8B:
		return N86::EAX;
	case X86::R9:
	case X86::R9D:
	case X86::R9W:
	case X86::R9B:
		return N86::ECX;
	case X86::R10:
	case X86::R10D:
	case X86::R10W:
	case X86::R10B:
		return N86::EDX;
	case X86::R11:
	case X86::R11D:
	case X86::R11W:
	case X86::R11B:
		return N86::EBX;
	case X86::R12:
	case X86::R12D:
	case X86::R12W:
	case X86::R12B:
		return N86::ESP;
	case X86::R13:
	case X86::R13D:
	case X86::R13W:
	case X86::R13B:
		return N86::EBP;
	case X86::R14:
	case X86::R14D:
	case X86::R14W:
	case X86::R14B:
		return N86::ESI;
	case X86::R15:
	case X86::R15D:
	case X86::R15W:
	case X86::R15B:
		return N86::EDI;

	case X86::ST0:
	case X86::ST1:
	case X86::ST2:
	case X86::ST3:
	case X86::ST4:
	case X86::ST5:
	case X86::ST6:
	case X86::ST7:
		return RegNo-X86::ST0;

	case X86::XMM0:
	case X86::XMM1:
	case X86::XMM2:
	case X86::XMM3:
	case X86::XMM4:
	case X86::XMM5:
	case X86::XMM6:
	case X86::XMM7:
		return getRegisterInfo().getDwarfRegNum(RegNo) -getRegisterInfo().getDwarfRegNum(X86::XMM0);
	case X86::XMM8:
	case X86::XMM9:
	case X86::XMM10:
	case X86::XMM11:
	case X86::XMM12:
	case X86::XMM13:
	case X86::XMM14:
	case X86::XMM15:
		return getRegisterInfo().getDwarfRegNum(RegNo) -getRegisterInfo().getDwarfRegNum(X86::XMM8);

	default:
		assert(MRegisterInfo::isVirtualRegister(RegNo) &&
				"Unknown physical register!");
		assert(0 && "Register allocator hasn't allocated reg correctly yet!");
		return 0;
	}
}

// hasFP - Return true if the specified function should have a dedicated frame
// pointer register.  This is true if the function has variable sized allocas or
// if frame pointer elimination is disabled.
//
static bool hasFP(const MachineFunction &MF)
{
	return (NoFramePointerElim || 
			MF.getFrameInfo()->hasVarSizedObjects() ||
			MF.getInfo<X86MachineFunctionInfo>()->getForceFramePointer());
}

int X86InstrInfo::sizeMemModRMByte(const MachineInstr &MI, unsigned Op,
		unsigned PCAdj) const
{
	const MachineOperand &Op3 = MI.getOperand(Op+3);
	int DispSize = 0;
	int DispVal = 0;
	int size = 0;
	bool DispForReloc = false;
	const X86Subtarget *Subtarget = &TM.getSubtarget<X86Subtarget>();

	// Figure out what sort of displacement we have to handle here.
	if (Op3.isGlobalAddress() || Op3.isConstantPoolIndex()
			|| Op3.isJumpTableIndex() || Op3.isExternalSymbol())
	{
		DispSize = 4;
		DispVal = -1;

		if (Op3.isGlobalAddress() || Subtarget->is64Bit())
			DispForReloc = true;
	}
	else
	{
		DispVal = Op3.getImm();
	}

	const MachineOperand &Base = MI.getOperand(Op);
	const MachineOperand &IndexReg = MI.getOperand(Op+2);

	unsigned BaseReg;

	if (Base.isFrameIndex())
	{
		//unpack.
		const MachineFunction &MF = *MI.getParent()->getParent();
		int FrameIndex = Base.getFrameIndex();
		DispVal += MF.getFrameInfo()->getObjectOffset(FrameIndex);
		//slot size (saved ebp)
		DispVal += 4;
		if (Subtarget->is64Bit())
			DispVal += 4;

		if (hasFP(MF))
			BaseReg = X86::EBP;
		else
			BaseReg = X86::ESP;

		if (Subtarget->is64Bit())
		{
			if (hasFP(MF))
				BaseReg = X86::RBP;
			else
				BaseReg = X86::RSP;
		}
	}
	else
	{
		BaseReg = Base.getReg();
	}

	// Is a SIB byte needed?
	if (IndexReg.getReg() == 0&&(BaseReg == 0
			|| MRegisterInfo::isVirtualRegister(BaseReg)
			|| getX86RegNum(BaseReg) != N86::ESP))
	{
		if (BaseReg == 0)
		{ // Just a displacement?							 
			size++;
			DispSize = 4;
		}
		else
		{
			unsigned BaseRegNo = MRegisterInfo::isVirtualRegister(BaseReg) ? 0
					: getX86RegNum(BaseReg);
			if (!DispForReloc && DispVal == 0&& BaseRegNo != N86::EBP)
			{
				//  simple indirect register encoding... [EAX] f.e.
				size++;
			}
			else if (!DispForReloc && isDisp8(DispVal))
			{
				//  the disp8 encoding... [REG+disp8]
				size++;
				DispSize = 1;
			}
			else
			{
				// Emit the most general non-SIB encoding: [REG+disp32]
				size++;
				DispSize = 4;
			}
		}

	}
	else
	{ // We need a SIB byte, 
		assert(IndexReg.getReg() != X86::ESP &&
				IndexReg.getReg() != X86::RSP && "Cannot use ESP as index reg!");

		bool ForceDisp32 = false;
		bool ForceDisp8 = false;
		if (BaseReg == 0)
		{
			// If there is no base register,special case SIB byte with
			// MOD=0, BASE=5, to JUST get the index, scale, and displacement.
			size++;
			ForceDisp32 = true;
		}
		else if (DispForReloc)
		{
			// Emit the normal disp32 encoding.
			size++;
			ForceDisp32 = true;
		}
		else if (DispVal == 0&&(MRegisterInfo::isVirtualRegister(BaseReg)
				||getX86RegNum(BaseReg) != N86::EBP))
		{
			// Emit no displacement ModR/M byte
			size++;
		}
		else if (isDisp8(DispVal))
		{
			// Emit the disp8 encoding...
			size++;
			ForceDisp8 = true; // Make sure to force 8 bit disp if Base=EBP
		}
		else
		{
			// Emit the normal disp32 encoding...
			size++;
		}

		//the SIB byte
		size++;

		// Do we need to output a displacement?
		if (ForceDisp8)
		{
			DispSize = 1;
		}
		else if (DispVal != 0|| ForceDisp32)
		{
			DispSize = 4;
		}
	}

	return size+DispSize;
}

static unsigned sizeOfImm(const TargetInstrDescriptor &Desc)
{
	switch (Desc.TSFlags & X86II::ImmMask)
	{
	case X86II::Imm8:
		return 1;
	case X86II::Imm16:
		return 2;
	case X86II::Imm32:
		return 4;
	case X86II::Imm64:
		return 8;
	default:
		assert(0 && "Immediate size not set!");
		return 0;
	}
}

// instrSize - Return the size of the specified instruction in bytes.
// This implementation is adapted from the binary code emitter, so
// it should be quite accurate, at least for a fully specified insn.
int X86InstrInfo::instrSize(const MachineInstr& MI, bool ignore64 /* = false */) const
{

	int size = 0;
	unsigned Opcode = MI.getOpcode();
	const TargetInstrDescriptor &Desc = get(Opcode);
	const X86Subtarget *Subtarget = &TM.getSubtarget<X86Subtarget>();
	unsigned numOperands = MI.getNumFixedOperands();
	// check for repeat opcode prefix 
	if ((Desc.TSFlags & X86II::Op0Mask) == X86II::REP)
		size++;
	
	// check for operand size opcode prefix 
	if (Desc.TSFlags & X86II::OpSize)
		size++;

	// check the address size opcode 
	if (Desc.TSFlags & X86II::AdSize)
		size++;

	bool Need0FPrefix = false;
	switch (Desc.TSFlags & X86II::Op0Mask)
	{
	case X86II::TB:
		Need0FPrefix = true; // Two-byte opcode prefix
		break;
	case X86II::REP:
		break; // already handled.
	case X86II::XS: // F3 0F
		size++;
		Need0FPrefix = true;
		break;
	case X86II::XD: // F2 0F
		size++;
		Need0FPrefix = true;
		break;
	case X86II::D8:
	case X86II::D9:
	case X86II::DA:
	case X86II::DB:
	case X86II::DC:
	case X86II::DD:
	case X86II::DE:
	case X86II::DF:
		size++;
		break; // Two-byte opcode prefix
	default:
		assert(0 && "Invalid prefix!");
	case 0:
		break; // No prefix!
	}

	if (Subtarget->is64Bit() && !ignore64)
	{
		// REX prefix
		unsigned REX = determineREX(MI);
		if (REX)
			size++;
	}

	// 0x0F escape code
	if (Need0FPrefix)
		size++;

	// If this is a two-address instruction, skip one of the register operands.
	unsigned CurOp = 0;
	if(Desc.numOperands > 1)
		CurOp +=  (Desc.getOperandConstraint(1, TOI::TIED_TO) != -1);

	switch (Desc.TSFlags & X86II::FormMask)
	{
	default:
		assert(0 && "Unknown FormMask value in X86 instrSize!");
	case X86II::Pseudo:
		CurOp = numOperands;
		break;

	case X86II::RawFrm:
		size++;
		if (CurOp != numOperands)
		{
			const MachineOperand &MO = MI.getOperand(CurOp++);
			if (MO.isMachineBasicBlock())
			{
				size += 4;
			}
			else if (MO.isGlobalAddress())
			{
				size += 4;
			}
			else if (MO.isExternalSymbol())
			{
				size += 4;
			}
			else if (MO.isImmediate())
			{
				size += sizeOfImm(Desc);
			}
			else
			{
				assert(0 && "Unknown RawFrm operand!");
			}
		}
		break;

	case X86II::AddRegFrm:
		size++;
		CurOp++;
		if (CurOp != numOperands)
		{
			const MachineOperand &MO1 = MI.getOperand(CurOp++);
			if (MO1.isGlobalAddress())
			{
				size += 4;
			}
			else if (MO1.isExternalSymbol())
			{
				size += 4;
			}
			else if (MO1.isJumpTableIndex())
			{
				size += 4;
			}
			else
			{
				size += sizeOfImm(Desc);
			}
		}
		break;

	case X86II::MRMDestReg:
	{
		size++;
		size++;
		CurOp += 2;
		if (CurOp != numOperands)
		{
			CurOp++;
			size += sizeOfImm(Desc);
		}
		break;
	}
	case X86II::MRMDestMem:
	{
		size++;
		size += sizeMemModRMByte(MI, CurOp);
		CurOp += 5;
		if (CurOp != numOperands)
		{
			CurOp++;
			size += sizeOfImm(Desc);
		}
		break;
	}

	case X86II::MRMSrcReg:
		size++;
		size++;

		CurOp += 2;
		if (CurOp != numOperands)
		{
			CurOp++;
			size += sizeOfImm(Desc);
		}
		break;

	case X86II::MRMSrcMem:
	{
		unsigned PCAdj = (CurOp+5!= numOperands) ? sizeOfImm(Desc) : 0;

		size++;
		size += sizeMemModRMByte(MI, CurOp+1, PCAdj);
		CurOp += 5;
		if (CurOp != numOperands)
		{
			CurOp++;
			size += sizeOfImm(Desc);
		}
		break;
	}

	case X86II::MRM0r:
	case X86II::MRM1r:
	case X86II::MRM2r:
	case X86II::MRM3r:
	case X86II::MRM4r:
	case X86II::MRM5r:
	case X86II::MRM6r:
	case X86II::MRM7r:
		size++;
		size++;
		CurOp++;
		if (CurOp != numOperands )
		{
			CurOp++;
			size += sizeOfImm(Desc);
		}
		break;

	case X86II::MRM0m:
	case X86II::MRM1m:
	case X86II::MRM2m:
	case X86II::MRM3m:
	case X86II::MRM4m:
	case X86II::MRM5m:
	case X86II::MRM6m:
	case X86II::MRM7m:
	{
		unsigned PCAdj = (CurOp+4!= numOperands) ? (MI.getOperand(CurOp+4).isImmediate() ? sizeOfImm(Desc) : 4) : 0;

		size++;
		size += sizeMemModRMByte(MI, CurOp, PCAdj);
		CurOp += 4;

		if (CurOp != numOperands)
		{
			const MachineOperand &MO = MI.getOperand(CurOp++);
			if (MO.isImmediate())
				size += sizeOfImm(Desc);
			else if (MO.isGlobalAddress())
				size += 4;
			else if (MO.isJumpTableIndex())
				size += 4;
			else
				assert(0 && "Unknown operand!");
		}
		break;
	}

	case X86II::MRMInitReg:
		size++;
		// Duplicate register, used by things like MOV8r0 (aka xor reg,reg).
		size++;
		++CurOp;
		break;
	}

	if (!((Desc.Flags & M_VARIABLE_OPS) != 0||CurOp == numOperands
			&& "Unknown encoding!"))
	{
		std::cerr << "TSFlags "<< Desc.TSFlags<< "\n";
		std::cerr << "CurOp == "<< CurOp <<"\t"<< "numops = "
				<< numOperands << "\n";
		std::cerr << "sizeImmDesc "<< sizeOfImm(Desc) <<"\n";
		std::cerr << "MI "<< MI << "\n";
		assert(0 && "Unkown encoding!");
	}

	//perform fixup for specialized register dependent varients not
	//handled correctly above
	if(Opcode == X86::TEST32ri || Opcode == X86::TEST8ri || Opcode == X86::TEST16ri ||
			Opcode == X86::CMP16ri || Opcode == X86::CMP32ri || Opcode == X86::CMP8ri)
	{
		const MachineOperand &MO = MI.getOperand(0);
		if(MO.isRegister() && (MO.getReg() == X86::EAX || MO.getReg() == X86::AX || MO.getReg() == X86::AL))
			size--;
	}
	
	//check for fp pseudo instructions and guess as to their size
	if(size == 0 &&  (Desc.TSFlags & X86II::FormMask) == X86II::Pseudo
		&& (Desc.TSFlags & X86II::FPTypeMask) != X86II::NotFP)
	{
		size = 2; //base insn
		//look for an immediate and use its size to grow the predicted size
		//this is far from perfect
		for(unsigned op = 0; op < numOperands; op++)
		{
			const MachineOperand& MO = MI.getOperand(op);
			if(MO.isImmediate() && MO.getImmedValue() != 0)
			{
				size += 1;
				break;
			}
		}
	}
	else if(size == 0 && Opcode == X86::EXTRACT_SUBREG)
	{
		//it would be more accurate to leave this in but they get eliminated
		//post-ra so punt for now
		size = 2; //a movb insn
	}
	
	return size;
}

int X86InstrInfo::copyRegSize() const
{
	return 2;
}

